/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.web.rs;

import java.io.Serializable;
import java.util.Comparator;

import org.codehaus.jackson.annotate.JsonCreator;
import org.codehaus.jackson.annotate.JsonProperty;

/**
 * Represents repetition of a character.
 */
public class Repeat implements Serializable {

	/**
	 * Compares repeats based on their {@link #getPosition()} values.
	 */
	public static final Comparator<Repeat> POSITION_COMPARATOR = new Comparator<Repeat>() {

		@Override
		public int compare(Repeat o1, Repeat o2) {
			int a1 = o1.getPosition();
			int a2 = o2.getPosition();
			return (a1 < a2 ? -1 : (a1 == a2 ? 0 : 1));
		}
	};

	private static final long serialVersionUID = -6149798669166979439L;

	private final char ch;

	private transient int hashcode;

	private final int length;

	private final int position;

	/**
	 * 
	 * @param ch
	 *            Character which is repeated.
	 * @param position
	 *            Position where the sequence starts.
	 * @param length
	 *            The number of characters.
	 */
	@JsonCreator
	public Repeat(@JsonProperty("ch") final char ch,
			@JsonProperty("position") final int position,
			@JsonProperty("length") final int length) {
		this.ch = ch;
		this.length = length;
		this.position = position;
		this.hashcode = 0;
	}

	private int doHashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ch;
		result = prime * result + length;
		result = prime * result + position;
		if (result == 0)
			result = -1;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Repeat other = (Repeat) obj;
		if (ch != other.ch)
			return false;
		if (length != other.length)
			return false;
		if (position != other.position)
			return false;
		return true;
	}

	/**
	 * The character which this repetition represents.
	 * 
	 * @return
	 */
	public char getCh() {
		return ch;
	}

	/**
	 * The number of the repeated characters. e.g. "aaaa" has length of 4.
	 * 
	 * @return
	 */
	public int getLength() {
		return length;
	}

	/**
	 * Position in original character sequence where this repetition occurs.
	 * 
	 * @return
	 */
	public int getPosition() {
		return position;
	}

	@Override
	public int hashCode() {
		if (hashcode == 0) {
			hashcode = doHashCode();
		}
		return hashcode;
	}

	@Override
	public String toString() {
		return Repeat.class.getSimpleName() + " [ch=" + ch + ", length="
				+ length + ", position=" + position + "]";
	}

}
