/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.web.rs;

import java.io.Serializable;
import java.util.Comparator;

import org.codehaus.jackson.annotate.JsonCreator;
import org.codehaus.jackson.annotate.JsonProperty;

/**
 * Presents a part/fragment of a audio data.
 * <p>
 * Segment has a position and data. Position is a fingerprint sequence index
 * where it was taken from. Data is a subsequence of a complete fingerprint
 * sequence.
 */
public class Segment implements Serializable {

	/**
	 * Compares segments based on their {@link #getPosition()} values.
	 */
	public static final Comparator<Segment> POSITION_COMPARATOR = new Comparator<Segment>() {

		@Override
		public int compare(Segment o1, Segment o2) {
			int a1 = o1.getPosition();
			int a2 = o2.getPosition();
			return (a1 < a2 ? -1 : (a1 == a2 ? 0 : 1));
		}
	};

	private static final long serialVersionUID = 904967199291999494L;

	private final String data;

	private transient int hashcode;

	private final int position;

	@JsonCreator
	public Segment(@JsonProperty("data") final String data,
			@JsonProperty("position") final int position) {
		this.data = data;
		this.position = position;
		this.hashcode = 0;
	}

	private int doHashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((data == null) ? 0 : data.hashCode());
		result = prime * result + position;
		if (result == 0)
			result = -1;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Segment other = (Segment) obj;
		if (data == null) {
			if (other.data != null)
				return false;
		} else if (!data.equals(other.data))
			return false;
		if (position != other.position)
			return false;
		return true;
	}

	/**
	 * Data from original fingerprint sequence starting at position
	 * {@link #getPosition()}.
	 * 
	 * @return
	 */
	public String getData() {
		return data;
	}

	/**
	 * Position in complete fingerprint sequence where this segment was taken.
	 * 
	 * @return
	 */
	public int getPosition() {
		return position;
	}

	@Override
	public int hashCode() {
		if (hashcode == 0) {
			hashcode = doHashCode();
		}
		return hashcode;
	}

	@Override
	public String toString() {
		return Segment.class.getSimpleName() + " [data=" + data + ", position="
				+ position + "]";
	}

}
