/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.calibrator;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingArgumentException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.IOUtils;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.CnfManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.qos.logback.classic.util.ContextInitializer;

/**
 * Main class for calibrator. Parses command line arguments and executes
 * calibrator core if all ok.
 * 
 * @author Juha Heljoranta
 * 
 */
public class Calibrator {

	public static final String FILE_LIST = "file-list";

	private static final String HELP = "help";

	private static final String LICENSE = "license";

	private static final String SOURCE_DIR = "source-dir";

	private static final String VERSION = "version";

	private static final String WORK_DIR = "work-dir";

	public static void main(String[] args) {

		try {
			Calibrator main = new Calibrator();
			main.parseCommandline(args);
		} catch (Exception e) {
			Logger logger = LoggerFactory.getLogger(Calibrator.class);
			logger.error("Unhandled exception", e);
			e.printStackTrace();
		}

	}

	public static void toStderr(Object o) {
		System.err.println(o);
	}

	public static void toStdout(Object o) {
		System.out.println(o);
	}

	public Calibrator() throws IOException {
		if (System.getProperty(ContextInitializer.CONFIG_FILE_PROPERTY) == null) {
			File cnfDir = new CnfManager().getConfigDir();
			File loggerCnf = new File(cnfDir, "logback.xml");
			System.setProperty(ContextInitializer.CONFIG_FILE_PROPERTY,
					loggerCnf.getAbsolutePath());
		}
	}

	private String getLicenseInfo() throws IOException {
		InputStream is = getClass().getResourceAsStream("/license-info.txt");
		try {
			return IOUtils.toString(is, "UTF-8");
		} finally {
			is.close();
		}
	}

	private Options getOptions() {

		// submit and entity goes hand-to-hand
		Option sourceDir = new Option("s", SOURCE_DIR, true,
				"directory containing audio files");
		sourceDir.setArgName("dir");
		sourceDir.setOptionalArg(false);

		Option workDir = new Option("w", WORK_DIR, true,
				"directory for storing datafiles");
		workDir.setArgName("dir");
		workDir.setOptionalArg(false);

		Option help = new Option("h", HELP, false,
				"print this help text and exit");

		Option license = new Option(null, LICENSE, false,
				"print license information and exit");

		Option version = new Option("v", VERSION, false,
				"print version information and exit");

		OptionGroup group = new OptionGroup();
		group.addOption(help);
		group.addOption(version);
		group.addOption(license);

		Options options = new Options();
		options.addOptionGroup(group);
		options.addOption(workDir);
		options.addOption(sourceDir);

		return options;
	}

	private void parseCommandline(String[] args) throws ConfigurationException,
			ParseException, IOException {
		Options options = getOptions();
		CommandLineParser parser = new PosixParser();
		try {
			CommandLine cmdline = parser.parse(options, args);
			if (cmdline.hasOption(HELP)) {
				printHelp(options);
			} else if (cmdline.hasOption(VERSION)) {
				printVersion();
			} else if (cmdline.hasOption(LICENSE)) {
				toStdout(getLicenseInfo());
			} else if (cmdline.hasOption(SOURCE_DIR)
					&& cmdline.hasOption(WORK_DIR)) {
				CnfManager f = new CnfManager();
				Configuration c = f.getConfiguration(Cnf.calibrator,
						Cnf.framework, Cnf.fingerprint, Cnf.audioPrePros);
				String sourceDir = cmdline.getOptionValue(SOURCE_DIR);
				String workDir = cmdline.getOptionValue(WORK_DIR);
				File sd = new File(sourceDir);
				if (!sd.isDirectory()) {
					throw new IOException("No such directory: "
							+ sd.getAbsolutePath());
				}
				File wd = new File(workDir);
				if (!wd.isDirectory()) {
					if (!wd.mkdirs()) {
						throw new IOException("Cannot create directory: "
								+ wd.getAbsolutePath());
					}
				}

				new CalibratorCore(c, f).execute(sd, wd);
			} else if (cmdline.hasOption(SOURCE_DIR)
					|| cmdline.hasOption(WORK_DIR)) {
				toStderr("Both the " + WORK_DIR + " and the " + SOURCE_DIR
						+ " must be defined");
			} else {
				printHelp(options);
			}
		} catch (MissingArgumentException e) {
			toStderr(e.toString());
			printHelp(options);
		}

	}

	private void printHelp(Options options) {
		HelpFormatter formatter = new HelpFormatter();
		String cmdName = getClass().getSimpleName();
		String header = null;
		String footer = "If directory "
				+ SOURCE_DIR
				+ " doesn't have any audio files and the directory contains a file named '"
				+ FILE_LIST
				+ "' then contents of the file are regarded as a list of input files.";
		formatter.printHelp(cmdName, header, options, footer, true);
	}

	private void printVersion() throws IOException {
		String version = CnfManager.getVersion("jhears-calibrator");
		if (version == null) {
			version = "development";
		}
		toStdout("jHears calibrator, " + version);
	}

}
