/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.calibrator;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.DoubleBuffer;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.jhears.calibrator.lib.AudioPreprocessorFacade;
import org.jhears.calibrator.lib.AudioSampleFacade;
import org.jhears.calibrator.lib.AudioSampleFacade.FileCacheStatus;
import org.jhears.calibrator.lib.CodeBookGenerator;
import org.jhears.calibrator.lib.ConvertFactory;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.Cnf.CalibratorKey;
import org.jhears.cnf.Cnf.Framewrk;
import org.jhears.cnf.CnfManager;
import org.jhears.concurrent.Concurrent;
import org.jhears.io.StreamIterator;
import org.jhears.sample.InplaceBurnFunction;
import org.jhears.sample.RawToSmoothedFunction;
import org.jhears.sample.SampleFunctionInjector;
import org.jhears.sample.SpectrumToBarksFunction;
import org.jhears.security.Digester;
import org.jhears.util.Observer;
import org.picocontainer.MutablePicoContainer;
import org.picocontainer.PicoBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Function;
import com.google.common.base.Throwables;

public class CalibratorCore {

	private final Logger logger;

	private final Configuration configuration;

	private final CnfManager cnfFactory;

	private final boolean overwrite;

	public CalibratorCore(Configuration configuration, CnfManager cnfFactory) {
		this.cnfFactory = cnfFactory;
		logger = LoggerFactory.getLogger(CalibratorCore.class);
		this.configuration = configuration;
		overwrite = configuration.getBoolean(Framewrk.CACHE_OVERWRITE_EXISTING);
	}

	public void execute(File sourceDir, File workDir) {
		try {
			init();
			process(sourceDir, workDir);
		} catch (Exception e) {
			logger.error("Calibration failed. Uncaught exeption", e);
		}
	}

	private void init() {
		logger.info("Initializing.");
		long seed = configuration.getLong(CalibratorKey.SEED);
		logger.info("Using random seed " + seed);
	}

	private void process(File sourceDir, File workDir) throws IOException,
			InterruptedException, ConfigurationException,
			NoSuchAlgorithmException, CloneNotSupportedException {
		logger.info("Building loudness filter");
		Loudness loudness = new Loudness(configuration, cnfFactory);
		// TODO remove hardcoded values
		long timeout = 600;
		TimeUnit unit = TimeUnit.SECONDS;
		loudness.createLoudnessFilter(timeout, unit);
		logger.info("Loudness filter created");
		logger.info("Preprocessing audio files");
		AudioPreprocessorFacade preProsFacade = new AudioPreprocessorFacade(
				configuration, sourceDir, workDir);
		List<File> preprocessResults = preProsFacade.build();
		logger.info("Audio file preprocessing completed");
		logger.info("Applying loudness filter");
		FileCacheStatus loudnessFilteredSample = sampleWithLoudnessFilter(
				sourceDir, workDir, preprocessResults);
		logger.info("Loudness filtering completed");
		// note: burn doesn't need calibration if spectrum is normalized and
		// gain controlled.
		// logger.info("Creating burn filter");
		// createBurnFilter(loudnessFilteredSample);
		logger.info("Applying burn filter");
		File burnedBarks = applyBurnFilterAndConvertToBarks(loudnessFilteredSample
				.getFile());
		logger.info("Creating codebook.");
		CodeBookGenerator builder = new CodeBookGenerator(configuration,
				workDir);
		builder.generate(burnedBarks);
	}

	private boolean isAlreadyCached(File outFile) {
		return !overwrite && outFile.exists() && outFile.length() > 0;
	}

	private File applyBurnFilterAndConvertToBarks(File filteredSample)
			throws IOException {
		File outFile = new File(filteredSample.getParentFile(),
				filteredSample.getName() + ".burn");
		if (isAlreadyCached(outFile) && !overwrite) {
			return outFile;
		}
		final int sizeOfDouble = (Double.SIZE / Byte.SIZE);
		final int spectrumLen = configuration
				.getInt(Cnf.FingerprintKey.SAMPLE_WINDOW) / 2;
		int sampleSizeInBytes = spectrumLen * sizeOfDouble;
		BufferedInputStream bis = new BufferedInputStream(new FileInputStream(
				filteredSample));
		final BufferedOutputStream bos = new BufferedOutputStream(
				new FileOutputStream(outFile));
		StreamIterator si = new StreamIterator(bis, sampleSizeInBytes,
				sampleSizeInBytes, 1 << 20);
		final Iterator<byte[]> itr = si.iterator();
		MutablePicoContainer container = new PicoBuilder().withLifecycle()
				.withCaching().build();
		SampleFunctionInjector
				.addFunctionsToContainer(container, configuration);
		final SpectrumToBarksFunction spectToBarksFun = container
				.getComponent(SpectrumToBarksFunction.class);
		final InplaceBurnFunction burnFilterFun = container
				.getComponent(InplaceBurnFunction.class);
		final int barkBufSize = configuration
				.getInt(Cnf.FingerprintKey.BARK_BAND_STOP_COUNT) * sizeOfDouble;

		// Iterator<Callable<byte[]>> callables = new
		// Iterator<Callable<byte[]>>() {
		//
		// @Override
		// public void remove() {
		// throw new UnsupportedOperationException();
		// }
		//
		// @Override
		// public Callable<byte[]> next() {
		// return new Callable<byte[]>() {
		// private final byte[] data = itr.next();
		//
		// @Override
		// public byte[] call() throws Exception {
		// double[] spect = new double[spectrumLen];
		// DoubleBuffer.wrap(spect).put(
		// ByteBuffer.wrap(data).asDoubleBuffer());
		// burnFilterFun.apply(spect);
		// double[] barks = spectToBarksFun.apply(spect);
		// byte[] retval = new byte[barkBufSize];
		// ByteBuffer.wrap(retval).asDoubleBuffer()
		// .put(DoubleBuffer.wrap(barks));
		// return retval;
		// }
		// };
		// }
		//
		// @Override
		// public boolean hasNext() {
		// return itr.hasNext();
		// }
		//
		// };
		// ExecListener<byte[]> execListener = new ExecListener<byte[]>() {
		//
		// @Override
		// public void exception(Exception e) {
		// logger.error("Exception while executing", e);
		// }
		//
		// @Override
		// public void result(byte[] v) {
		// try {
		// bos.write(v);
		// } catch (IOException e) {
		// logger.error("Failed to write output", e);
		// throw Throwables.propagate(e);
		// }
		// }
		// };
		Function<byte[], byte[]> function = new Function<byte[], byte[]>() {

			@Override
			public byte[] apply(byte[] from) {
				double[] spect = new double[spectrumLen];
				DoubleBuffer.wrap(spect).put(
						ByteBuffer.wrap(from).asDoubleBuffer());
				burnFilterFun.apply(spect);
				double[] barks = spectToBarksFun.apply(spect);
				byte[] retval = new byte[barkBufSize];
				ByteBuffer.wrap(retval).asDoubleBuffer()
						.put(DoubleBuffer.wrap(barks));
				return retval;
			}
		};
		Observer<byte[]> observer = new Observer<byte[]>() {
			@Override
			public void observe(byte[] elt) {
				try {
					bos.write(elt);
				} catch (IOException e) {
					throw Throwables.propagate(e);
				}
			}
		};
		Concurrent.transform(itr, function, observer);
		bos.close();
		return outFile;
	}

	// private void createBurnFilter(FileCacheStatus loudnessFilteredSample)
	// throws IOException, ConfigurationException {
	// if (loudnessFilteredSample.isCached()) {
	// for (String key : configuration
	// .getStringArray(FingerprintKey.BURN_FILTER_VALUES)) {
	// if (!Cnf.DEFAULT_BURN_FILTER_VALUE.equals(key)) {
	// // loudness filter is already created
	// return;
	// }
	// }
	// }
	// BurnFilterCalibrator burnFilterCalibrator = new BurnFilterCalibrator(
	// configuration);
	// double[] burnFilterValues = burnFilterCalibrator
	// .calibrate(loudnessFilteredSample.getFile());
	// String[] saveValues = CnfUtil.toStringArray(burnFilterValues);
	// configuration
	// .setProperty(FingerprintKey.BURN_FILTER_VALUES, saveValues);
	// cnfFactory.saveProperty(Cnf.fingerprint,
	// FingerprintKey.BURN_FILTER_VALUES, saveValues);
	// }

	private FileCacheStatus sampleWithLoudnessFilter(File sourceDir,
			File workDir, List<File> files) throws IOException,
			InterruptedException, ConfigurationException,
			NoSuchAlgorithmException, CloneNotSupportedException {

		Digester digester = new Digester("MD5", 1 << 20);
		List<String> strings = new ArrayList<String>(files.size());
		for (File f : files) {
			strings.add(f.getName());
		}
		String outFileName = digester.getHexDigest(strings) + ".loudness";

		AudioSampleFacade asf = new AudioSampleFacade(configuration, workDir);
		MutablePicoContainer container = new PicoBuilder().withLifecycle()
				.withCaching().build();
		SampleFunctionInjector
				.addFunctionsToContainer(container, configuration);
		final RawToSmoothedFunction rawToSmoothed = container
				.getComponent(RawToSmoothedFunction.class);
		ConvertFactory convertFactory = new ConvertFactory() {

			@Override
			public Callable<double[]> createConvert(final String name,
					final byte[] sample) {
				return new Callable<double[]>() {

					@Override
					public double[] call() throws Exception {
						double[] rawSpectrum = rawToSmoothed.apply(sample);
						if (rawToSmoothed.isSilence(rawSpectrum)) {
							// ignore silence
							return null;
						}
						return rawSpectrum;
					}

					@Override
					public String toString() {
						return name;
					}
				};

			}
		};
		return asf.doSample(files, convertFactory, outFileName);

	}

}
