/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.calibrator;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Iterator;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.exec.DefaultExecutor;
import org.apache.commons.exec.ExecuteResultHandler;
import org.apache.commons.io.input.ClosedInputStream;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.Cnf.CalibratorKey;
import org.jhears.cnf.Cnf.FingerprintKey;
import org.jhears.cnf.CnfManager;
import org.jhears.cnf.CnfUtil;
import org.jhears.exec.DefaultExecuteResultHandler;
import org.jhears.exec.ExecUtil;
import org.jhears.io.ClosableIterator;
import org.jhears.io.StreamIterator;
import org.jhears.logging.OutputStreamLogger;
import org.jhears.logging.OutputStreamLogger.Level;
import org.jhears.pipe.Pipe;
import org.jhears.sample.RawToPowSpectFunction;
import org.jhears.sample.SampleFunctionInjector;
import org.jhears.sample.filter.LoudnessFilterCalibrator;
import org.jhears.util.Mth;
import org.picocontainer.MutablePicoContainer;
import org.picocontainer.PicoBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * Converts equal-loudness spectrum to flat spectrum. See Wikipedia article <a
 * href="http://en.wikipedia.org/wiki/Loudness">loudness</a>. Basic idea behind
 * this is to enable computer to perceive audio loudness equally across the
 * spectrum when using euclidean distance.
 * 
 * @author Juha Heljoranta
 * 
 */
public class Loudness {

	private final int sampleWindowInBytes;
	private final int sampleWindow;
	private final Configuration configuration;
	private final CnfManager cnfFactory;
	private RawToPowSpectFunction rawToSpectFun;

	public Loudness(Configuration configuration, CnfManager cnfFactory) {
		this.configuration = configuration;
		this.cnfFactory = cnfFactory;
		MutablePicoContainer container = new PicoBuilder().withLifecycle()
				.withCaching().build();
		SampleFunctionInjector.addFunctionsToContainer(container, configuration);
		rawToSpectFun = container.getComponent(RawToPowSpectFunction.class);
		int bytesPerSample = configuration
				.getInt(FingerprintKey.BYTES_PER_SAMPLE);
		sampleWindow = configuration.getInt(FingerprintKey.SAMPLE_WINDOW);
		sampleWindowInBytes = sampleWindow * bytesPerSample;
	}

	private ClosableIterator<double[]> iterator(final InputStream is) {
		final StreamIterator si = new StreamIterator(is, sampleWindowInBytes,
				sampleWindowInBytes / 4, sampleWindowInBytes * 4);
		final Iterator<byte[]> siter = si.iterator();
		return new ClosableIterator<double[]>() {

			@Override
			public void close() {
				try {
					si.close();
				} catch (IOException ignore) {
				}
			}

			@Override
			public boolean hasNext() {
				return siter.hasNext();
			}

			@Override
			public double[] next() {
				return rawToSpectFun.apply(siter.next());
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}

		};
	}

	public void createLoudnessFilter(long timeout, TimeUnit unit)
			throws IOException, InterruptedException, ConfigurationException {
//		LoudnessFilterLoader loundnessLoader = new LoudnessFilterLoader();

		double[] filterValues = CnfUtil.toDoubleArray(configuration
				.getStringArray(Cnf.FingerprintKey.LOUDNESS_FILTER_VALUES));
		// check if filter is already defined
		// by default the all filter values are 1 
		if (filterValues.length > 0
				&& Mth.sum(filterValues) != filterValues.length) {
			return;
		}

		String[] cmd = configuration.getStringArray(CalibratorKey.LOUDNESS_CMD);
		org.apache.commons.exec.CommandLine cmdline = ExecUtil
				.createCommandLine(cmd);
		AtomicBoolean retval = new AtomicBoolean(true);
		Semaphore semaphore = new Semaphore(0);
		ExecuteResultHandler handler = new DefaultExecuteResultHandler(retval,
				semaphore, cmd);
		PipedOutputStream src = new PipedOutputStream();
		PipedInputStream loudnessStream = new PipedInputStream(src);
		InputStream in = new ClosedInputStream();

		Logger stderrLog = LoggerFactory.getLogger(Loudness.class.getPackage()
				.getName() + ".loudness.sox.stdout");
		OutputStreamLogger osLogger = new OutputStreamLogger(stderrLog,
				Level.info);
		OutputStream err = osLogger.getOutputStream();
		osLogger.start(true);

		Pipe pipe = new Pipe(in, src, err);
		DefaultExecutor executor = ExecUtil.createExecutor(
				TimeUnit.MILLISECONDS.convert(timeout, unit), pipe);

		LoudnessFilterCalibrator loudnessCalibrator = new LoudnessFilterCalibrator(configuration);

		// start
		executor.execute(cmdline, handler);

		ClosableIterator<double[]> powSpectIterator = null;
		double[] values;
		try {
			powSpectIterator = iterator(loudnessStream);
			values = loudnessCalibrator.calibrate(powSpectIterator);
		} finally {
			if (powSpectIterator != null) {
				powSpectIterator.close();
			}
		}

		if (retval.compareAndSet(true, semaphore.tryAcquire(timeout, unit))) {
			String[] saveValues = CnfUtil.toStringArray(values);
			configuration.setProperty(FingerprintKey.LOUDNESS_FILTER_VALUES, saveValues);
			cnfFactory.saveProperty(Cnf.fingerprint,
					FingerprintKey.LOUDNESS_FILTER_VALUES, saveValues);
		}
		osLogger.close();

	}
}
