/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.calibrator.lib;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.filefilter.FileFilterUtils;
import org.apache.commons.io.output.CloseShieldOutputStream;
import org.jhears.calibrator.Calibrator;
import org.jhears.cnf.Cnf.Framewrk;
import org.jhears.concurrent.ThreadInterrupter;
import org.jhears.file.SimpleDirWalker;
import org.jhears.logging.OutputStreamLogger;
import org.jhears.logging.OutputStreamLogger.Level;
import org.jhears.sample.AudioPreprocessor;
import org.jhears.workpool.Work;
import org.jhears.workpool.WorkerPool;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Recursively scans audio files from source directory, runs audio files through
 * a audio preprocessor and place results into cache directory. In other words
 * this class converts audio files to raw format which jHears uses internally.
 * 
 */
public class AudioPreprocessorFacade {

	private class PreprocessWork implements Work {

		private final File file;
		private final long timeout;
		private final TimeUnit unit;

		public PreprocessWork(File file, long timeout, TimeUnit unit) {
			this.file = file;
			this.timeout = timeout;
			this.unit = unit;
		}

		@Override
		public void doWork() {
			ThreadInterrupter interrupter = new ThreadInterrupter(Thread
					.currentThread(), timeout, unit);
			try {
				boolean ok = false;
				interrupter.start();
				try {
					File cacheFile = audPrepro.process(file, overwrite);
					if (cacheFile != null) {
						if (cacheFile.isFile()) {
							results.add(cacheFile);
							ok = true;
						} else {
							logger
									.error("Audio pre-processor returned non-existing file. file="
											+ file.getAbsolutePath()
											+ " cacheFile="
											+ cacheFile.getAbsolutePath()
											+ " overwrite=" + overwrite);
						}
					}
				} finally {
					interrupter.cancel();
				}
				logResults(file, ok);
			} catch (Exception e) {
				logger.error("Unexpected error while processing file: "
						+ file.getAbsolutePath(), e);
			} finally {
				try {
					errorLogger.getOutputStream().flush();
				} catch (IOException e) {
					logger.error("Error logging failure", e);
				}
			}
		}
	}

	private final AudioPreprocessor audPrepro;

	private final OutputStreamLogger errorLogger;

	private int failCnt;

	private final Logger logger;

	private final Object mutex;

	private final boolean overwrite;

	private List<File> results;

	private final File sourceDirectory;

	private int successCnt;

	private final int threads;

	private final WorkerPool<PreprocessWork> workerPool;

	public AudioPreprocessorFacade(Configuration configuration,
			File sourceDirectory, File workDir) throws IOException,
			ConfigurationException {
		logger = LoggerFactory.getLogger(getClass());
		this.sourceDirectory = sourceDirectory;
		if (!workDir.isDirectory()) {
			throw new IOException("Invalid target directory "
					+ workDir.getAbsolutePath());
		}
		if (!sourceDirectory.isDirectory()) {
			throw new IOException("Invalid source directory "
					+ sourceDirectory.getAbsolutePath());
		}
		failCnt = 0;
		successCnt = 0;

		overwrite = configuration.getBoolean(Framewrk.CACHE_OVERWRITE_EXISTING);

		audPrepro = new AudioPreprocessor(configuration, workDir);
		Logger stdErrLog = LoggerFactory.getLogger("System.err");
		Level stdErrLevel = Level.info;
		errorLogger = new OutputStreamLogger(stdErrLog, stdErrLevel);
		errorLogger.start(true);
		CloseShieldOutputStream errStream = new CloseShieldOutputStream(
				errorLogger.getOutputStream());
		audPrepro.setErrorStream(errStream);
		mutex = new Object();
		threads = Runtime.getRuntime().availableProcessors();
		workerPool = new WorkerPool<PreprocessWork>(threads,
				new PreprocessWork(null, -1, null));
		// timeoutFactor = configuration
		// .getLong(Calibr.AUD_PREPROC_TIMEOUT_FACTOR);

	}

	/**
	 * 
	 * @return sorted list of files produced by the preprocessor
	 * @throws InterruptedException
	 * @throws IOException
	 */
	public List<File> build() throws InterruptedException, IOException {
		final Set<String> supportedFormats = audPrepro
				.listSupportedFileFormats();
		errorLogger.getOutputStream().flush();
		long maxtimeout = 1000L;
		SimpleDirWalker w = getDirWalker(supportedFormats);
		List<File> flist = w.getFiles(sourceDirectory);
		if (flist.isEmpty()) {
			flist = readFromFileList();
		}
		Collections.sort(flist);
		results = Collections
				.synchronizedList(new ArrayList<File>(flist.size()));
		for (File file : flist) {
			long timeout = file.length() + 1000;
			if (maxtimeout < timeout) {
				maxtimeout = timeout;
			}
			processFile(file, timeout, TimeUnit.MILLISECONDS);
		}
		close(maxtimeout * threads, TimeUnit.MILLISECONDS);
		errorLogger.getOutputStream().flush();
		Collections.sort(results);
		return results;
	}

	private List<File> readFromFileList() throws IOException {
		List<File> retval = new ArrayList<File>();
		File file = new File(sourceDirectory, Calibrator.FILE_LIST);
		if (file.isFile()) {
			BufferedReader br = new BufferedReader(new FileReader(file));
			try {
				String line;
				while ((line = br.readLine()) != null) {
					File f = new File(line);
					if (f.isFile()) {
						retval.add(f);
					} else {
						logger.warn("File not found: " + f.getAbsolutePath());
					}
				}
			} finally {
				br.close();
			}
		}
		return retval;
	}

	private void close(long timeout, TimeUnit unit) throws InterruptedException {
		if (!workerPool.stop(timeout, unit)) {
			logger.error("Workpool shutdown timed out");
		}
		logger.info("Work completed. Added " + successCnt + " entries. "
				+ failCnt + " entries skipped due errors.");
	}

	private SimpleDirWalker getDirWalker(final Set<String> supportedFormats) {
		return new SimpleDirWalker(FileFilterUtils
				.asFileFilter(new FileFilter() {

					@Override
					public boolean accept(File pathname) {
						String ext = FilenameUtils.getExtension(pathname
								.getName());
						return (ext != null && supportedFormats.contains(ext));
					}
				}));
	}

	private void logResults(File file, boolean ok) {
		synchronized (mutex) {
			if (ok) {
				successCnt++;
				logger.info("Processed file " + file.getName());
			} else {
				logger.error("Failed to process " + file.getName());
				failCnt++;
			}
		}
	}

	private void processFile(File file, long timeout, TimeUnit unit)
			throws InterruptedException {
		workerPool.addWork(new PreprocessWork(file, timeout, unit));
		if (logger.isDebugEnabled()) {
			logger.debug("Scheduled for preprocessing "
					+ file.getAbsolutePath());
		}
	}

}
