/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.calibrator.lib;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.DoubleBuffer;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.IOUtils;
import org.jhears.cnf.Cnf.AudioPreProsKey;
import org.jhears.cnf.Cnf.CalibratorKey;
import org.jhears.cnf.Cnf.FingerprintKey;
import org.jhears.cnf.Cnf.Framewrk;
import org.jhears.concurrent.Concurrent;
import org.jhears.util.BlockingQueueIteratorAdapter;
import org.jhears.util.Observer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Function;
import com.google.common.base.Throwables;

/**
 * Class will read contents of the audio preprocessing cache directory, sample
 * preprocessed audio data, perform audio analysis of the samples and stores
 * results for codebook generation.
 * <p>
 * Preprocessed audio data reading is configurable but typically sampling skips
 * lot of data. Skipping helps to keep data entropy high.
 * <p>
 * TODO: allow this class to create overlapping samples.
 * TODO: rewrite this class.
 */
public class AudioSampleFacade {

	private static final class NameSample {

		private final String name;

		private final byte[] sample;

		NameSample(final String name, final byte[] sample) {
			this.name = name;
			this.sample = sample;
		}
	}

	private class FacadeWorker {

		private int completedCounter = 0;;
		private final List<File> files;
		private final File outFile;

		private final int size;
		private final int nThreads;
		private final NameSample poison;
		private Iterator<NameSample> itr;
		private final ConvertFactory convertFactory;

		public FacadeWorker(File outFile, List<File> files,
				final ConvertFactory convertFactory) {

			this.convertFactory = convertFactory;
			nThreads = Runtime.getRuntime().availableProcessors();
			queue = new LinkedBlockingQueue<NameSample>(nThreads * 10);
			poison = new NameSample(null, null);
			BlockingQueueIteratorAdapter<NameSample> bqi = new BlockingQueueIteratorAdapter<NameSample>(
					queue, poison);
			itr = bqi.iterator();
			this.outFile = outFile;
			this.files = files;
			size = files.size();

		}

		public void sample() throws IOException, InterruptedException {
			logger.info("Sampling " + size + " files.");
			Thread samplerThread = new Thread(new Runnable() {

				@Override
				public void run() {
					Thread.currentThread().setName(FacadeWorker.class.getSimpleName());
					try {
						sampleFiles();
						queue.put(poison);
					} catch (InterruptedException e) {
						logger.error("Interrupted", e);
						Thread.currentThread().interrupt();
					} catch (Exception e) {
						logger.error("Fail", e);
					}
				}
			});
			try {
				bos = new BufferedOutputStream(new FileOutputStream(outFile));
				Function<NameSample, double[]> function = new Function<NameSample, double[]>() {

					@Override
					public double[] apply(NameSample from) {
						try {
							return convertFactory.createConvert(from.name, from.sample).call();
						} catch (Exception e) {
							logger.error("fail", e);
							throw Throwables.propagate(e);
						}
					}
				};
				Observer<double[]> observer = new Observer<double[]>() {
					@Override
					public void observe(double[] elt) {
						try {
							append(elt);
						} catch (IOException e) {
							throw Throwables.propagate(e);
						}
					}
				};
				samplerThread.start();
				Concurrent.transform(itr, function, observer );
			} finally {
				if (bos != null) {
					IOUtils.closeQuietly(bos);
				}
				samplerThread.interrupt();
			}
			logger.info("Sampled " + size + " files.");
		}

		private void sampleFiles() throws IOException, InterruptedException {
			int i = 0;
			for (File file : files) {
				i++;
				if (file.isFile()) {
					logger.info("Sampling " + Integer.toString(i) + "/" + size
							+ " (" + file.getName() + ").");
					sampleFile(file);
				} else {
					logger.warn("File not found: " + file.getAbsolutePath());
					completedCounter++;
				}
			}
		}
	}

	private static final int SIZEOF_DOUBLE = Double.SIZE / Byte.SIZE;

	/**
	 * Stream where the results are written.
	 */
	private BufferedOutputStream bos;

	private BlockingQueue<NameSample> queue;

	private final int bytesPerSample;

	private byte[] doubleBufData;

	private DoubleBuffer doubleBuffer;

	private final Logger logger;

	private final boolean overwrite;

	private final Random random;

	private final int sampleWindowInBytes;

	/**
	 * The number of bytes to skip after taking a sample.
	 */
	private final int skipBytesConstant;

	/**
	 * The maximum number of bytes to skip
	 */
	private final int skipSamplesMaxRnd;

	private final File sourceDirectory;

	private final File targetDirectory;

	public AudioSampleFacade(Configuration configuration, File workDir)
			throws IOException, ConfigurationException {
		logger = LoggerFactory.getLogger(AudioSampleFacade.class);
		sourceDirectory = new File(workDir,
				AudioPreProsKey.CNST_PRE_PROC_DIR_NAME);
		targetDirectory = new File(workDir,
				CalibratorKey.CNST_ANALYSIS_DIR_NAME);
		if (!targetDirectory.exists()) {
			if (!targetDirectory.mkdirs()) {
				throw new IOException("Failed to create directory: "
						+ targetDirectory.getAbsolutePath());
			}
		}
		if (!targetDirectory.isDirectory()) {
			throw new IOException("Invalid target directory " + targetDirectory);
		}
		if (!sourceDirectory.isDirectory()) {
			throw new IOException("Invalid source directory " + sourceDirectory);
		}

		overwrite = configuration.getBoolean(Framewrk.CACHE_OVERWRITE_EXISTING);
		bytesPerSample = configuration.getInt(FingerprintKey.BYTES_PER_SAMPLE);
		sampleWindowInBytes = configuration
				.getInt(FingerprintKey.SAMPLE_WINDOW)
				* bytesPerSample;
		skipBytesConstant = configuration
				.getInt(CalibratorKey.SKIP_SAMPLES_CONST)
				* bytesPerSample;
		skipSamplesMaxRnd = configuration
				.getInt(CalibratorKey.SKIP_SAMPLES_MAX_RND);
		this.random = new Random(configuration.getLong(CalibratorKey.SEED));
	}

	private void append(double[] data) throws IOException {
		if (doubleBuffer == null && data != null) {
			doubleBufData = new byte[SIZEOF_DOUBLE * data.length];
			doubleBuffer = ByteBuffer.wrap(doubleBufData).asDoubleBuffer();
		}
		if (data != null) {
			doubleBuffer.clear();
			doubleBuffer.put(data);
			bos.write(doubleBufData);
		} else {
			if (logger.isTraceEnabled()) {
				logger.trace("Skipping zero bark.");
			}
		}
	}

	public static class FileCacheStatus {

		private final File file;

		private final boolean cached;

		public FileCacheStatus(File file, boolean cached) {
			this.file = file;
			this.cached = cached;
		}

		public File getFile() {
			return file;
		}

		public boolean isCached() {
			return cached;
		}
		
	}

	public FileCacheStatus doSample(List<File> files,
			ConvertFactory convertFactory, String outFileName)
			throws IOException, InterruptedException {
		init();
		File outFile = getFile(outFileName);
		FileCacheStatus retval = new FileCacheStatus(outFile,
				isAlreadyCached(outFile));
		if (!retval.cached) {
			FacadeWorker worker = new FacadeWorker(outFile, files,
					convertFactory);
			worker.sample();
		}
		logger.info("Samping completed");
		return retval;
	}

	/**
	 * Calculates a checksum based on the names of the input files and uses the
	 * check sum as the name of the output file.
	 * 
	 * @param files
	 * @param filteringStage
	 * @return
	 * @throws IOException
	 */
	private File getFile(String name) throws IOException {
		return new File(targetDirectory, name);
	}

	private int getLength(File file) {
		long length = file.length();
		if (length > Integer.MAX_VALUE) {
			logger.warn("File '" + file.getName() + "' is too big (" + length
					+ "). Access limited to first " + Integer.MAX_VALUE
					+ " bytes.");
			length = Integer.MAX_VALUE;
		}
		length -= sampleWindowInBytes;
		return (int) length;
	}

	private int getRndSkip() {
		if (skipSamplesMaxRnd > 0) {
			return random.nextInt(skipSamplesMaxRnd) * bytesPerSample;
		} else {
			return 0;
		}
	}

	private void init() {
		doubleBuffer = null;
	}

	private boolean isAlreadyCached(File outFile) {
		return !overwrite && outFile.exists() && outFile.length() > 0;
	}

	private int sampleFile(File file) throws IOException, InterruptedException {
		int totalSamples = 0;
		String name = file.getName();
		DataInputStream dis = null;
		try {
			dis = new DataInputStream(new BufferedInputStream(
					new FileInputStream(file)));
			int length = getLength(file);
			int skip = skipBytesConstant + getRndSkip();
			long read = 0;
			while ((length - read) >= (skip + sampleWindowInBytes)) {
				byte[] sample = new byte[sampleWindowInBytes];
				try {
					dis.skip(skip);
					dis.readFully(sample);
				} catch (EOFException e) {
					throw new IOException("File size changed during sampling");
				}
				queue.put(new NameSample(name, sample));
				read += skip + sampleWindowInBytes;
				skip = skipBytesConstant + getRndSkip();
				totalSamples++;
			}
		} finally {
			if (dis != null) {
				dis.close();
			}
		}
		return totalSamples;
	}

}
