/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.calibrator.lib;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Random;
import java.util.UUID;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.Cnf.CalibratorKey;
import org.jhears.cnf.Cnf.FingerprintKey;
import org.jhears.cnf.CnfManager;
import org.jhears.io.BarkReader;
import org.jhears.io.ClosableIterator;
import org.jhears.math.ELBG;
import org.jhears.math.LBG.Result;
import org.jhears.math.Stat;
import org.jhears.util.Mth;
import org.joda.time.DateTimeZone;
import org.joda.time.format.DateTimeFormatter;
import org.joda.time.format.ISODateTimeFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.com.bytecode.opencsv.CSVWriter;

public class CodeBookGenerator {

	private final int barkCnt;

	private final Configuration configuration;

	private final double epsilon1;

	private final double epsilon2;
	private final Logger logger;
	private final double maxZeroValue;
	private final Random random;
	private BarkReader sampler;
	private final int symbols;
	private final File targetDirectory;

	private final int zeroSymbolIdx;

	public CodeBookGenerator(Configuration configuration, File workDir)
			throws IOException {
		logger = LoggerFactory.getLogger(getClass());
		this.configuration = configuration;
		this.random = new Random(configuration.getLong(CalibratorKey.SEED));
		barkCnt = configuration.getInt(FingerprintKey.BARK_BAND_STOP_COUNT);
		symbols = configuration.getInt(FingerprintKey.CODEBOOK_SYMBOL_COUNT);
		zeroSymbolIdx = configuration.getInt(FingerprintKey.ZERO_SYMBOL_IDX);
		maxZeroValue = configuration.getDouble(CalibratorKey.MAX_ZERO_VALUE);
		epsilon1 = configuration.getDouble(CalibratorKey.ELBG_EPSILON1);
		epsilon2 = configuration.getDouble(CalibratorKey.ELBG_EPSILON2);
		targetDirectory = new File(workDir,
				CalibratorKey.CNST_CODEBOOK_DIR_NAME);
		if (!targetDirectory.exists()) {
			if (!targetDirectory.mkdirs()) {
				throw new IOException("Failed to created directory: "
						+ targetDirectory.getAbsolutePath());
			}
		}
		if (!targetDirectory.isDirectory()) {
			throw new IOException("Invalid target directory "
					+ targetDirectory.getAbsolutePath());
		}
	}

	private void addSilence(double[][] matrix, int matrixRows, int zeros) {
		for (int i = matrixRows - zeros; i < matrixRows; i++) {
			matrix[i] = new double[barkCnt];
			for (int c = 0; c < barkCnt; c++) {
				matrix[i][c] = random.nextDouble() * maxZeroValue;
			}
		}
	}

	public void generate(File inputFile) throws IOException,
			ConfigurationException {
		// TODO: test if output already exist and skip if necessary
		sampler = new BarkReader(configuration, inputFile);
		int matrixRows = (int) sampler.getBarkCount();
		if (matrixRows < symbols * symbols) {
			logger.error("Not enough input data to generate a codebook.");
			return;
		}
		/*
		 * On average we'll have (matrixRows / symbols) rows of data
		 * representing a single symbol. To make sure that the resulting code
		 * book contains a symbol for zero we'll have to add artificial silence.
		 * This is required since the input data is usually stripped from noise
		 * below a threshold.
		 */
		int zeros = matrixRows / symbols;
		matrixRows += zeros;
		double[][] matrix = readSamples(matrixRows, zeros);

		logger.info("Injecting " + zeros + " rows of artificial silence.");
		addSilence(matrix, matrixRows, zeros);
		logger.info("Clustering...");
		ELBG elbg = new ELBG(matrix, symbols, epsilon1, epsilon2, random,
				Integer.MAX_VALUE);
		Result result = elbg.calculate();

		double[] trueZero = new double[barkCnt];
		if (isbookvalid(result, trueZero)) {
			logger.info("Clustering complete. Codebook ready.");
			// over write near-zero symbol with a zero value
			int zeroIdx = Mth.getClosestRow(result.y, trueZero);
			result.y[zeroIdx] = trueZero;
			// move zero to row zeroSymbolIdx
			if (zeroIdx != zeroSymbolIdx) {
				swapRows(zeroIdx, zeroSymbolIdx, result.y);
			}
			saveCodebook(result.y, inputFile.getName(), matrixRows - zeros,
					matrix);
		} else {
			logger.error("Clustering failed. The code book is not valid.");
		}

	}

	private boolean isbookvalid(Result result, double[] trueZero) {
		// distance between closest two rows divided by two
		double minDistance = minDistance(result.y) / 2.0;
		int zeroCnt = 0;
		/*
		 * Only one of the rows should be close to zero. That row represents the
		 * artificial silence generated before the codebook calculation.
		 */
		for (double[] y : result.y) {
			if (Mth.distance(trueZero, y) < minDistance) {
				zeroCnt++;
			}
		}
		if (zeroCnt != 1) {
			logger.error("Discarding current codebook. "
					+ "Expected codebook to have only one zero but it has "
					+ zeroCnt);
		}
		return zeroCnt == 1;
	}

	/**
	 * Returns a distance between two of the closest rows.
	 * 
	 * @param y
	 * @return squared distance
	 */
	private double minDistance(double[][] y) {
		double dist = Double.MAX_VALUE;
		for (int i = 0; i < y.length; i++) {
			for (int j = i + 1; j < y.length; j++) {
				double d = Mth.distance(y[i], y[j]);
				if (d < dist) {
					dist = d;
				}
			}
		}
		return dist;
	}

	private double[][] readSamples(int matrixRows, int zeros)
			throws IOException {
		logger.info("Loading " + (matrixRows - zeros) + " samples.");
		double[][] matrix = new double[matrixRows][];
		ClosableIterator<double[]> iter = sampler.iterator();
		for (int i = 0; i < matrixRows - zeros; i++) {
			matrix[i] = iter.next();
		}
		iter.close();
		return matrix;
	}

	private void saveCodebook(double[][] y, String fileName, int nonZeroRows,
			double[][] matrix) throws ConfigurationException, IOException {

		DateTimeFormatter dt = ISODateTimeFormat.dateTime().withZone(
				DateTimeZone.UTC);
		String date = dt.print(System.currentTimeMillis());
		String filename = fileName + "-codebook.properties";
		File file = new File(targetDirectory, filename);
		String uuid = UUID.randomUUID().toString();
		Configuration newCnf = new CnfManager().getCopy(Cnf.fingerprint,
				configuration);

		newCnf.setProperty(FingerprintKey.CREATION_DATE, date);
		newCnf.setProperty(FingerprintKey.CONFIGURATION_UUID, uuid);
		String version = CnfManager.getVersion("jhears-calibrator");
		if (version == null) {
			version = "development";
		}
		newCnf.setProperty(FingerprintKey.VERSION, version);
		newCnf.setProperty(FingerprintKey.CODEBOOK, y);
		// CnfUtil.copyCnfTo(filterConfiguration, newCnf);

		// String checksum = Clint.calculateChecksum(configuration);
		// newCnf.addProperty(Clint.CONFIGURATION_CHECKSUM_SHA1, checksum);
		PropertiesConfiguration pc = new PropertiesConfiguration();
		pc.copy(newCnf);
		pc.save(file);
		logger.info("Codebook saved: " + file.getAbsolutePath());

		Stat[] barkStatArr = new Stat[barkCnt];
		for (int column = 0; column < barkCnt; column++) {
			double[] barkColumn = new double[nonZeroRows];
			for (int i = 0; i < nonZeroRows; i++) {
				barkColumn[i] = matrix[i][column];
			}
			barkStatArr[column] = new Stat(barkColumn);
		}
		Stat[] cbStatArr = new Stat[symbols];
		for (int i = 0; i < symbols; i++) {
			cbStatArr[i] = new Stat(y[i]);
		}
		filename = fileName + "-codebook-statistics.csv";
		File statfile = new File(targetDirectory, filename);
		writeCsv(statfile, barkStatArr, cbStatArr, date, uuid, y);
		logger.info("Codebook statistical details saved: "
				+ statfile.getAbsolutePath());
		logger.info("If you are satified with the results, please rename "
				+ file.getAbsolutePath() + " to "
				+ Cnf.fingerprint.getFileName()
				+ " and distribute it to server and clients.");
	}

	private void swapRows(int rowA, int rowB, double[][] matrix) {
		double[] aux = matrix[rowA];
		matrix[rowA] = matrix[rowB];
		matrix[rowB] = aux;
	}

	private String[] toStrArray(int row, double[] v) {
		String[] retval = new String[v.length + 1];
		retval[0] = Integer.toString(row);
		for (int i = 0; i < v.length; i++) {
			retval[i + 1] = Double.toString(v[i]);
		}
		return retval;
	}

	private String[] toStrArray(int row, Stat s) {
		return new String[] { Integer.toString(row),
				Double.toString(s.getMin()), Double.toString(s.getMax()),
				Double.toString(s.getMean()), Double.toString(s.getMedian()),
				Double.toString(s.getQ1()), Double.toString(s.getQ3()),
				Double.toString(s.getVariance()), Integer.toString(s.getSize()) };
	}

	private void writeCsv(File file, Stat[] barkStatArr, Stat[] cbStatArr,
			String date, String uuid, double[][] y) throws IOException {

		CSVWriter cw = new CSVWriter(new BufferedWriter(new OutputStreamWriter(
				new FileOutputStream(file))));
		try {
			cw.writeNext(new String[] { "date" });
			cw.writeNext(new String[] { date });
			cw.writeNext(new String[] { "uuid" });
			cw.writeNext(new String[] { uuid });
			cw.writeNext(new String[] { "codebook-symbols" });
			cw.writeNext(new String[] { "symbol", "min", "max", "mean",
					"median", "q1", "q3", "variance", "size" });
			int r = 0;
			for (Stat s : cbStatArr) {
				cw.writeNext(toStrArray(r++, s));
			}
			cw.writeNext(new String[] { "input-per-bark" });
			cw.writeNext(new String[] { "bark", "min", "max", "mean", "median",
					"q1", "q3", "variance", "size" });
			r = 0;
			for (Stat s : barkStatArr) {
				cw.writeNext(toStrArray(r++, s));
			}
			cw.writeNext(new String[] { "codebook-vectors" });
			String[] headers = new String[barkCnt + 1];
			headers[0] = "vector";
			for (int i = 1; i < headers.length; i++) {
				headers[i] = "b" + Integer.toString(i);
			}
			cw.writeNext(headers);
			for (int i = 0; i < y.length; i++) {
				cw.writeNext(toStrArray(i + 1, y[i]));
			}
		} finally {
			cw.close();
		}
	}
}
