/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.math;

import java.util.List;
import java.util.Random;
import java.util.Set;

import org.jhears.util.IntList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The Enhanced {@link LBG} algorithm. By Giuseppe Patanè and Marco Russo. See
 * <a href="http://ai.unime.it/~gp/publications/abstracts/elbg.htm">ELBG</a>
 * [ai.unime.it] paper for details.
 * 
 * @author Juha Heljoranta
 * @see LBG
 * 
 */
public class ELBG extends LBG {

	/**
	 * List of high utilization cells. A.k.a pth cells.
	 */
	private List<WeightedInt> hiList;

	private int ithCell;

	private final Logger logger;

	/**
	 * List of low utilization cells. A.k.a ith cells.
	 */
	private List<WeightedInt> loUtilList;

	private int lthCell;

	private int pthCell;

	/**
	 * Utilization indexes
	 */
	private double[] uidxes;

	/**
	 * List of unused cell id for elbg calculation.
	 */
	private Set<Integer> unused;

	protected ELBG(final double[][] x, final double[][] y, double epsilon1,
			double epsilon2, Random random, int maxIterations,
			boolean doInitialize) {
		super(x, y, epsilon1, epsilon2, random, maxIterations, doInitialize);
		logger = LoggerFactory.getLogger(getClass());
	}

	public ELBG(final double[][] x, final double[][] y, double epsilon,
			Random random, int maxIterations) {
		this(x, y, epsilon, epsilon, random, maxIterations, false);
	}

	public ELBG(final double[][] x, final int codeWords) {
		this(x, new double[codeWords][], 0.01, 0.1, new Random(),
				Integer.MAX_VALUE, true);
	}

	/**
	 * 
	 * @param x
	 *            matrix containing input vectors
	 * @param codeWords
	 *            the number of codebook symbols to generate
	 * @param epsilon1
	 *            threshold for deciding when the codebook is good enough. The
	 *            smaller the value, the better the codebook and the longer the
	 *            computation. Typical value: 0.01.
	 * @param epsilon2
	 *            threshold for deciding when the cell split is good enough. The
	 *            smaller the value, the better the cell split and the longer
	 *            the computation. Typical value: 0.1.
	 * @param random
	 *            random instance to be used during the computation.
	 * @param maxIterations
	 *            The maximum number of allowed iterations. Depends a lot of
	 *            task at hand. A rough estimate could be the number of rows
	 *            multiplied by the number of code words.
	 */
	public ELBG(final double[][] x, final int codeWords, double epsilon1,
			double epsilon2, Random random, int maxIterations) {
		this(x, new double[codeWords][], epsilon1, epsilon2, random,
				maxIterations, true);
	}

	/**
	 * Prevent ith, pth and lth cell to be chosen again in remaining iterations.
	 */
	private void consume() {
		// lth cell cannot be any of these cells for remaining iterations
		unused.remove(Integer.valueOf(ithCell));
		unused.remove(Integer.valueOf(pthCell));
		unused.remove(Integer.valueOf(lthCell));
		// ith cell cannot be just consumed lth cell during remaining iterations
		removeValueFrom(lthCell, loUtilList);
		// pth cell cannot be chosen again for remaining iterations
		removeValueFrom(pthCell, hiList);
		// pth cell cannot be just consumed lth cell during remaining iterations
		removeValueFrom(lthCell, hiList);
	}

	@Override
	protected void elbg() {
		uidxes = calculateUtilityIndexes();
		initElgb();
		if (!executeElbg()) {
			// this extra block doesn't seem to have any effect so disable it
			// if (logger.isDebugEnabled()) {
			// logger.debug("m=" + getM()
			// + " no successful SoCAs executed. Retrying.");
			// }
			// initElgb();
			// executeElbg();
		}
		return;
	}

	private boolean executeElbg() {
		boolean success = false;
		while (next()) {
			if (soca(ithCell, pthCell, lthCell)) {
				logConfirmShift(ithCell, pthCell, lthCell);
				consume();
				success = true;
			}
		}
		return success;
	}

	private void initElgb() {
		unused = newUnusedList();
		hiList = utilListHi(uidxes);
		loUtilList = inWeightRandOrder(utilListLo(uidxes));
	}

	private void logConfirmShift(int ithCell, int pthCell, int lthCell) {
		if (logger.isTraceEnabled()) {
			logger.trace("m=" + getM() + " SoCA  executed. cell " + pthCell
					+ " split. cell " + ithCell + " merged into cell "
					+ lthCell + ".");

		}
	}

	private void logSoca(int ithCell, int pthCell, int lthCell) {
		if (logger.isTraceEnabled()) {
			logger.trace("m=" + getM() + " trying to execute SoCA. split cell "
					+ pthCell + " . merge cell " + ithCell + ". target cell "
					+ lthCell + ".");
		}
	}

	private boolean next() {
		if (!loUtilList.isEmpty() && !hiList.isEmpty()) {
			ithCell = loUtilList.remove(0).getValue();
			pthCell = inWeightRandOrder(hiList).get(0).getValue();
			return nextLth();
		}
		return false;
	}

	private boolean nextLth() {
		double dist = Double.POSITIVE_INFINITY;
		for (Integer u : unused) {
			int i = u.intValue();
			if (i == pthCell || i == ithCell) {
				continue;
			}
			double d = getCodeWordsDistance(ithCell, i);
			if (d <= dist) {
				dist = d;
				lthCell = i;
			}
		}
		return lthCell != -1;
	}

	private boolean removeValueFrom(int value, List<WeightedInt> list) {
		for (int i = 0; i < list.size(); i++) {
			if (value == list.get(i).getValue()) {
				list.remove(i);
				return true;
			}
		}
		return false;
	}

	private boolean soca(int ithCell, int pthCell, int lthCell) {
		logSoca(ithCell, pthCell, lthCell);
		Result res = splitCell(pthCell);
		IntList setL_ = combinedCell(ithCell, lthCell);
		double[] yl_ = calculateCentroid(setL_);
		double dold = getCellDist(ithCell) + getCellDist(pthCell)
				+ getCellDist(lthCell);
		double dnew = res.distortions[ITH] + res.distortions[PTH]
				+ cellDistortion(yl_, setL_);
		if (dnew <= dold) {
			mergeToCells(res, pthCell, ithCell);
			assignCellRowsToCell(setL_, lthCell);
			cellReAssign(lthCell, yl_, setL_);
			check();
			return true;
		}
		return false;
	}

}
