/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.math;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.jhears.util.IntList;
import org.jhears.util.Mth;
import org.jhears.util.WeightedRandom;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Linde-Buzo-Gray algorithm. From Wikipedia: "The Linde-Buzo-Gray algorithm is
 * a vector quantization algorithm to derive a good codebook. It is similar to
 * the k-means method in data clustering."
 * 
 * @author Juha Heljoranta
 * @see ELBG
 */
/*
 * This class contain some code which is only used by the ELBG.
 */
public class LBG {

	public static class Result {

		/**
		 * Voronoi partition /cell distortions. The distortion is a sum of cell
		 * vector distances from the centroid of the cell / voronoi partition.
		 * Array has as many elements as there are codebook symbols.
		 */
		public final double[] distortions;

		/**
		 * The number of iterations.
		 */
		public final int m;

		/**
		 * Voronoi partitions. Each {@link IntList} contains input matrix row
		 * numbers which belong cell / voronoi partition it presents. Array has
		 * as many elements as there are codebook symbols.
		 */
		public final IntList[] pmap;

		/**
		 * The codebook.
		 */
		public final double[][] y;

		public Result(final double[][] y, final IntList[] pmap,
				final double[] distortions, int m) {
			this.y = y;
			this.pmap = pmap;
			this.distortions = distortions;
			this.m = m;
		}
	}

	protected static final int ITH = 1;

	protected static final int PTH = 0;

	private static void addAtoB(final double[] a, final double[] b) {
		for (int i = 0; i < a.length; i++) {
			b[i] += a[i];
		}
	}

	private static double sumVectorValues(double[] v) {
		double sum = 0.0;
		for (int i = 0; i < v.length; i++) {
			sum += v[i];
		}
		return sum;
	}

	/**
	 * Voronoi partition /cell distortions. The distortion is a sum of cell
	 * vector distances from the centroid of the cell / voronoi partition. Array
	 * has as many elements as there are codebook symbols.
	 */
	private double[] cellDist;

	protected final int codeWords;

	private final boolean doInitialize;

	private int emptyCell;

	private final double epsilon1;

	private final double epsilon2;

	protected final WeightedRandom wr;

	private final Logger logger;

	/**
	 * The number of iterations.
	 */
	private int m;

	private final int maxIterations;

	private boolean paranoid;

	/**
	 * entry maps row in x to a codeword into y. so x[3]=2 is read: input vector
	 * 3 is assigned to codeword 2 in y. aka. voronoi partition.
	 */
	private final int partition[];

	/**
	 * Voronoi partitions. Each {@link IntList} contains input matrix row
	 * numbers which belong cell / voronoi partition it presents. Array has as
	 * many elements as there are codebook symbols.
	 */
	private final IntList[] partMap;

	private double prevDist;

	private final Random random;

	protected final int vectorLen;

	/**
	 * Input matrix.
	 */
	private final double[][] x;

	/**
	 * Codebook.
	 */
	private double[][] y;

	protected LBG(final double[][] x, final double[][] y, double epsilon1,
			double epsilon2, Random random, int maxIterations,
			boolean doInitialize) {
		logger = LoggerFactory.getLogger(getClass());
		this.x = x;
		this.y = y;
		this.epsilon1 = epsilon1;
		this.epsilon2 = epsilon2;
		this.random = random;
		this.maxIterations = maxIterations;
		this.doInitialize = doInitialize;
		this.codeWords = y.length;
		this.partition = new int[x.length];
		this.vectorLen = x[0].length;
		this.wr = new WeightedRandom(random);
		emptyCell = -1;
		partMap = new IntList[codeWords];
		prevDist = Double.MAX_VALUE;
		init();
	}

	public LBG(final double[][] x, final double[][] y, double epsilon,
			Random random, int maxIterations) {
		this(x, y, epsilon, epsilon, random, maxIterations, false);
	}

	public LBG(final double[][] x, final int codeWords) {
		this(x, new double[codeWords][], 0.01, 0.1, new Random(),
				Integer.MAX_VALUE, true);
	}

	public LBG(final double[][] x, final int codeWords, double epsilon1,
			double epsilon2, Random random, int maxIterations) {
		this(x, new double[codeWords][], epsilon1, epsilon2, random,
				maxIterations, true);
	}

	protected void assignCellRowsToCell(IntList cellRows, int cellId) {
		int size = cellRows.size();
		for (int r = 0; r < size; r++) {
			partition[cellRows.get(r)] = cellId;
		}
	}

	private void calcPartitionsAndDistortion() {
		logIterationStart();
		calculateVoronoiPartitions(); // NNC
		calculateCellDistortions();
	}

	public Result calculate() {
		if (doInitialize) {
			initializeRandom();
		}
		optimizeCodebook();
		check();
		return new Result(y, partMap, cellDist, m);
	}

	private void calculateBook() {
		if (emptyCell != -1) {
			handleEmptyCell();
		} else if (codeWords > 2) {
			elbg();
		}
		calculateNewCodebook(); // CC
	}

	private void calculateCellDistortions() {
		cellDist = new double[codeWords];
		for (int i = 0; i < codeWords; i++) {
			cellDist[i] = cellDistortion(i);
		}
		logDistortions();
	}

	protected double[] calculateCentroid(IntList set) {
		double[] c = new double[vectorLen];
		int size = set.size();
		for (int r = 0; r < size; r++) {
			addAtoB(x[set.get(r)], c);
		}
		divide(c, set.size());
		return c;
	}

	private void calculateNewCodebook() {
		y = new double[codeWords][vectorLen];
		for (int r = 0; r < x.length; r++) {
			addAtoB(x[r], y[partition[r]]);
		}
		for (int i = 0; i < codeWords; i++) {
			if (partMap[i].isEmpty()) {
				logger.error("Empty cell detected. cell id=" + i);
				emptyCell = i;
			} else {
				divide(y[i], partMap[i].size());
			}
		}
		logCodeBook();
	}

	protected double[] calculateUtilityIndexes() {
		double meandistortion = sumVectorValues(cellDist) / codeWords;
		double[] indexes = new double[codeWords];
		for (int i = 0; i < codeWords; i++) {
			indexes[i] = cellDist[i] / meandistortion;
		}
		return indexes;
	}

	private void calculateVoronoiPartitions() {
		clearPartitions();
		for (int r = 0; r < x.length; r++) {
			int cellId = getClosestCentroid(r);
			partition[r] = cellId;
			partMap[cellId].add(r);
		}
		emptyCellCheck();
		logPartSizes();
		check();
	}

	protected double cellDistortion(final double[] centroid, IntList list) {
		double d = 0.0;
		int size = list.size();
		for (int r = 0; r < size; r++) {
			d += Mth.distance(centroid, x[list.get(r)]);
		}
		return d;
	}

	private double cellDistortion(final int i) {
		double d = 0.0;
		IntList list = partMap[i];
		int size = list.size();
		for (int r = 0; r < size; r++) {
			d += Mth.distance(y[i], x[list.get(r)]);
		}
		return d;
	}

	protected void cellReAssign(int cellId, double[] centroid, IntList set) {
		partMap[cellId] = set;
		y[cellId] = centroid;
	}

	/**
	 * If {@link #paranoid} is true, the checks that the internal data
	 * structures are consistent.
	 */
	protected void check() {
		if (!paranoid) {
			return;
		}
		Set<Integer> ail = new HashSet<Integer>(x.length);
		boolean hasEmpty = checkEmptyAndConsistency(ail);
		if (hasEmpty != (emptyCell != -1)) {
			logger.error("Empty cell flag is wrong. Correct value is "
					+ hasEmpty);
		}
		checkEquals(ail);

	}

	private boolean checkEmptyAndConsistency(Set<Integer> ail) {
		boolean hasEmpty = false;
		for (int i = 0; i < codeWords; i++) {
			if (partMap[i].isEmpty()) {
				hasEmpty = true;
			}
			checkMappingConsistency(ail, i);
		}
		return hasEmpty;
	}

	private void checkEquals(Set<Integer> ail) {
		HashSet<Integer> all = new HashSet<Integer>(x.length);
		for (int r = 0; r < x.length; r++) {
			all.add(Integer.valueOf(r));
		}
		if (ail.size() != x.length) {
			logger.error("Metadata available for " + ail.size() + " rows "
					+ " but should be " + x.length);
		}
		all.removeAll(ail);
		if (!all.isEmpty()) {
			logger.error("Metadata missing for rows " + all.toString());
		}
	}

	private void checkLocateDuplicate(int cellId, int r) {
		for (int u = 0; u < cellId; u++) {
			for (int e = 0; e < partMap[u].size(); e++) {
				if (partMap[u].get(e) == r) {
					logger.error("Cell " + cellId + " contains a row " + r
							+ " which is alread in cell " + u);
					return;
				}
			}
		}
	}

	private void checkMappingConsistency(Set<Integer> ail, int cellId) {
		IntList cell = partMap[cellId];
		for (int j = 0; j < cell.size(); j++) {
			int r = cell.get(j);
			if (!ail.add(Integer.valueOf(r))) {
				checkLocateDuplicate(cellId, r);
			}
			if (partition[r] != cellId) {
				logger.error("Partition data corrupted. cell " + cellId
						+ " claims to own row " + r
						+ " but row is assigned to cell " + partition[r]);
			}
		}
	}

	private void clearPartitions() {
		for (int i = 0; i < codeWords; i++) {
			partMap[i].clear();
		}
	}

	/**
	 * Creates a new cell and adds rows from both cells into it.
	 * 
	 * @param cellA
	 * @param cellB
	 * @return
	 */
	protected IntList combinedCell(int cellA, int cellB) {
		IntList setL_ = new IntList(partMap[cellB].size()
				+ partMap[cellA].size());
		setL_.addAll(partMap[cellA]);
		setL_.addAll(partMap[cellB]);
		return setL_;
	}

	/**
	 * Creates a subset from source by using values in key list.
	 * 
	 * @param source
	 * @param keyList
	 * @return
	 */
	private IntList createSubset(IntList source, IntList keyList) {
		IntList newCellMap = new IntList(keyList.size());
		for (int i = 0; i < keyList.size(); i++) {
			int foreignIdx = keyList.get(i);
			int localIdx = source.get(foreignIdx);
			newCellMap.add(localIdx);
		}
		return newCellMap;
	}

	/**
	 * divide vector v with constant x
	 * 
	 * @param v
	 * @param x
	 */
	private void divide(final double[] v, final double x) {
		for (int i = 0; i < v.length; i++) {
			v[i] /= x;
		}
	}

	/**
	 * Super class overrides
	 */
	protected void elbg() {

	}

	private void emptyCellCheck() {
		for (int i = 0; i < codeWords; i++) {
			if (partMap[i].isEmpty()) {
				if (logger.isTraceEnabled()) {
					logger.trace("Empty cell found. Cell id=" + i);
				}
				emptyCell = i;
				return;
			}
		}
	}

	protected double getCellDist(int ithCell) {
		return cellDist[ithCell];
	}

	private double[][] getCellVectors(int cellId) {
		final double[][] px = new double[partMap[cellId].size()][];
		for (int i = 0; i < px.length; i++) {
			px[i] = x[partMap[cellId].get(i)];
		}
		return px;
	}

	private int getClosestCentroid(int row) {
		return Mth.getClosestRow(y, x[row]);
	}

	protected double getCodeWordsDistance(int cellA, int cellB) {
		return Mth.distance(y[cellA], y[cellB]);
	}

	/**
	 * Returns minimum and maximum values for the matrix column. retval[0] is
	 * the minimum and the retval[1] is the maximum;
	 * 
	 * @param col
	 * @param matrix
	 * @return
	 */
	private double[] getColumnMinMax(int col, double[][] matrix) {
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		for (int row = 0; row < matrix.length; row++) {
			double d = matrix[row][col];
			if (d < min) {
				min = d;
			}
			if (d > max) {
				max = d;
			}
		}
		return new double[] { min, max };
	}

	private int getEmptyCell() {
		for (int i = 0; i < codeWords; i++) {
			if (partMap[i].isEmpty()) {
				return i;
			}
		}
		return -1;
	}

	private double[][] getInitBook(double[][] cellPData) {
		double[][] initBook;
		if (m == 0 && codeWords == 2 && emptyCell != -1) {
			if (logger.isDebugEnabled()) {
				logger.debug("Empty cell and m=0. "
						+ "Cell split by hyperboxing has failed. "
						+ "Spliting cell from centroid.");
			}
			initBook = splitByCentroid(cellPData);
		} else {
			initBook = splitByHyperBoxing(cellPData);
		}
		return initBook;
	}

	protected int getM() {
		return m;
	}

	private List<Integer> getRowNumbers() {
		ArrayList<Integer> lList = new ArrayList<Integer>(x.length);
		for (int i = 0; i < x.length; i++) {
			lList.add(Integer.valueOf(i));
		}
		return lList;
	}

	protected List<WeightedInt> inWeightRandOrder(List<WeightedInt> list) {
		int len = list.size();
		double[] weights = new double[len];
		for (int i = 0; i < len; i++) {
			weights[i] = list.get(i).getWeight();
		}
		int[] indexes = wr.shuffle(weights);
		List<WeightedInt> retval = new ArrayList<WeightedInt>(len);
		for (int i : indexes) {
			retval.add(list.get(i));
		}
		return retval;
	}
	
	private void handleEmptyCell() {
		List<WeightedInt> hiUtilList = inWeightRandOrder(utilListHi(calculateUtilityIndexes()));
		if (!tryToSplitEmptyCell(hiUtilList)) {
			logger
					.warn("Empty cell detected and attempt to locate cell for spliting failed. Empty cell id="
							+ emptyCell + ".");
		}
		check();
	}

	private void init() {
		if (x.length < codeWords) {
			throw new IllegalArgumentException(
					"not enough input vectors to calculate " + codeWords
							+ " codewords");
		}
		m = 0;
		int size = x.length / codeWords * 2;
		for (int i = 0; i < codeWords; i++) {
			partMap[i] = new IntList(size);
		}

	}

	private void initializeRandom() {
		y = new double[codeWords][vectorLen];
		Set<Integer> iSet = randomRows();
		Iterator<Integer> ii = iSet.iterator();
		for (int i = 0; i < codeWords; i++) {
			int r = ii.next().intValue();
			y[i] = Arrays.copyOf(x[r], vectorLen);
		}
	}

	private void logCodeBook() {
		if (logger.isTraceEnabled()) {
			for (int i = 0; i < codeWords; i++) {
				logger.trace("m=" + m + " cell=" + i + " word="
						+ Arrays.toString(y[i]));
			}
		}
	}

	private void logDistortions() {
		if (logger.isTraceEnabled()) {
			for (int i = 0; i < codeWords; i++) {
				logger.trace("m=" + m + " cell=" + i + " distortion="
						+ cellDist[i]);
			}
		}
	}

	private void logIterationStart() {
		if (logger.isTraceEnabled()) {
			logger.trace("m=" + m + " starts");
		}
	}

	private void logPartSizes() {
		if (logger.isTraceEnabled()) {
			for (int i = 0; i < codeWords; i++) {
				logger.trace("m=" + m + " cell=" + i + " size="
						+ partMap[i].size());
			}
		}
	}

	protected void mergeToCells(Result res, int pthCell, int ithCell) {
		IntList pthPartition = partMap[pthCell];
		IntList i_partition = createSubset(pthPartition, res.pmap[ITH]);
		IntList p_partition = createSubset(pthPartition, res.pmap[PTH]);
		// overwrite part of the pthCell partition
		assignCellRowsToCell(i_partition, ithCell);
		// no need to re-assign partition pthCell to it self
		// assignCellRowsToCell(p_partition, pthCell, pthCell);
		partMap[ithCell] = i_partition;
		// some of the partition rows where overwritten with ithCell
		partMap[pthCell] = p_partition;
		y[ithCell] = res.y[ITH];
		y[pthCell] = res.y[PTH];
	}

	/**
	 * Calculate mean quantization error: sum(Di)/Np.
	 * 
	 * @return
	 */
	private double mqe() {
		double mqe = sumVectorValues(cellDist) / x.length;
		if (logger.isTraceEnabled()) {
			logger.trace("m=" + m + " mqe=" + mqe);
		}
		return mqe;
	}

	protected Set<Integer> newUnusedList() {
		Set<Integer> unused = new HashSet<Integer>(codeWords);
		for (int i = 0; i < codeWords; i++) {
			unused.add(Integer.valueOf(i));
		}
		return unused;
	}

	/**
	 * This is the (E)LGB main loop
	 */
	private void optimizeCodebook() {
		do {
			if (emptyCell != -1) {
				break; // failed to remove empty cell
			}
			calcPartitionsAndDistortion();
			// prevent division by zero
			double distortion = mqe() + Double.MIN_VALUE;
			double diff = (Math.abs(prevDist - distortion) / distortion);
			if (emptyCell == -1 && diff <= epsilon1 || m > maxIterations) {
				break;
			} else {
				calculateBook();
				prevDist = distortion;
				m++;
			}
		} while (true);
	}

	private Set<Integer> randomRows() {
		Set<Integer> iSet = new HashSet<Integer>(codeWords);
		if (x.length / 2 < codeWords) {
			List<Integer> lList = getRowNumbers();
			while (iSet.size() < codeWords) {
				iSet.add(lList.remove(random.nextInt(lList.size())));
			}
		} else {
			while (iSet.size() < codeWords) {
				iSet.add(Integer.valueOf(random.nextInt(x.length)));
			}
		}
		return iSet;
	}

	public void setParanoidMode(boolean enabled) {
		this.paranoid = enabled;
	}

	private double[][] splitByCentroid(final double[][] cellVectors) {
		double[] c = new double[vectorLen];
		int size = cellVectors.length;
		for (int r = 0; r < size; r++) {
			addAtoB(cellVectors[r], c);
		}
		divide(c, size);
		double[][] newc = new double[2][vectorLen];
		for (int i = 0; i < vectorLen; i++) {
			newc[0][i] = c[i] + epsilon1 * random.nextDouble();
			newc[1][i] = c[i] - epsilon1 * random.nextDouble();
		}
		return newc;
	}

	/**
	 * Approximate centroids for data. Hyper box is placed around the data two
	 * points dividing a diagonal of the hyper box into three segments are
	 * returned.
	 * 
	 * @param cellVectors
	 * @return
	 */
	private double[][] splitByHyperBoxing(final double[][] cellVectors) {
		final double[][] retval = new double[2][vectorLen];
		for (int col = 0; col < vectorLen; col++) {
			double[] minMax = getColumnMinMax(col, cellVectors);
			double delta = minMax[1] - minMax[0];
			retval[0][col] = minMax[0] + delta / 4.0;
			retval[1][col] = minMax[0] + delta * 3.0 / 4.0;
		}
		return retval;
	}

	protected Result splitCell(int cellId) {
		if (logger.isTraceEnabled()) {
			logger.trace("m=" + m + " splitting cell=" + cellId);
		}
		double[][] cellPData = getCellVectors(cellId);
		double[][] initBook = getInitBook(cellPData);
		LBG e2 = new LBG(cellPData, initBook, epsilon2, random, maxIterations);
		return e2.calculate();

	}

	private boolean tryToSplitEmptyCell(List<WeightedInt> hiUtilList) {
		for (WeightedInt pthCell : hiUtilList) {
			mergeToCells(splitCell(pthCell.getValue()), pthCell.getValue(),
					emptyCell);
			emptyCell = getEmptyCell();
			if (emptyCell == -1) {
				return true;
			}
		}
		return false;
	}

	protected List<WeightedInt> utilListHi(double[] uindexes) {
		ArrayList<WeightedInt> kwdiList = new ArrayList<WeightedInt>(codeWords);
		double sum = 0.0;
		for (int i = 0; i < codeWords; i++) {
			if (uindexes[i] > 1.0 && partMap[i].size() > 1) {
				sum += uindexes[i];
			}
		}
		for (int i = 0; i < codeWords; i++) {
			if (uindexes[i] > 1.0 && partMap[i].size() > 1) {
				kwdiList.add(new WeightedInt(uindexes[i] / sum, i));
			}
		}
		return kwdiList;
	}

	protected List<WeightedInt> utilListLo(double[] uindexes) {
		ArrayList<WeightedInt> kwdiList = new ArrayList<WeightedInt>(codeWords);
		for (int i = 0; i < codeWords; i++) {
			if (uindexes[i] < 1.0) {
				// prevent division by zero
				kwdiList.add(new WeightedInt(
						1.0 / (uindexes[i] + Double.MIN_NORMAL), i));
			}
		}
		return kwdiList;
	}
}
