/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.List;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.jhears.cnf.CnfManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Throwables;
import com.google.common.io.CharStreams;

public class ClientCommandLine {

	private static class HelpFormatter2 extends HelpFormatter {

		private boolean adjustPrintUsage = false;

		@Override
		public void printUsage(PrintWriter pw, int width, String app,
				Options options) {
			adjustPrintUsage = true;
			super.printUsage(pw, width, app, options);
			adjustPrintUsage = false;
		}

		@Override
		public void printWrapped(PrintWriter pw, int width,
				int nextLineTabStop, String text) {
			if (adjustPrintUsage) {
				super.printWrapped(pw, width, nextLineTabStop, text
						+ " [FILE...]");
			} else {
				super.printWrapped(pw, width, nextLineTabStop, text);
			}
		}

	}

	private static final String FETCH_ENTITY = "entity";

	private static final String FETCH_FINGERPRINT = "fingerprint";

	private static final String APPROVE_FINGERPRINT = "approve";

	private static final String FILE_FORMAT = "format";

	private static final String HELP = "help";

	private static final String LICENSE = "license";

	private static final String LIST_FILE_TYPES = "list-types";

	private static final String PRINT = "print";

	private static final String SEARCH = "search";

	private static final String UPLOAD = "upload";

	private static final String URL = "url";

	private static final String VERBOSE = "verbose";

	private static final String VERSION = "version";

	private final IJHearsClient client;

	private final Logger logger;

	public ClientCommandLine(IJHearsClient client) {
		this.logger = LoggerFactory.getLogger(ClientCommandLine.class);
		this.client = client;
	}

	private Options getOptions() {

		Option upload = new Option("u", UPLOAD, true,
				"Upload an audio fingerprint with given entity id");
		upload.setArgName("entityId");
		upload.setOptionalArg(false);

		Option format = new Option(null, FILE_FORMAT, true, "Input file format");
		format.setArgName("format");
		format.setOptionalArg(false);

		Option fetchEntity = new Option("e", FETCH_ENTITY, true,
				"Get all fingerprints associated to entityId");
		fetchEntity.setArgName("entityId");
		fetchEntity.setOptionalArg(false);

		Option search = new Option("s", SEARCH, false,
				"Search for matching audio fingerprint");

		Option url = new Option("r", URL, true, "Server URL");
		url.setArgName("url");
		url.setOptionalArg(false);

		Option fetchFingerprint = new Option("f", FETCH_FINGERPRINT, true,
				"Get fingerprint data associated to fingerprintId");
		fetchFingerprint.setArgName("fingerprintId");
		fetchFingerprint.setOptionalArg(false);

		Option approveFingerprint = new Option("a", APPROVE_FINGERPRINT, true,
				"Confirm that the fingerprint data matches the audio entity.");
		approveFingerprint.setArgName("fingerprintId");
		approveFingerprint.setOptionalArg(false);

		Option print = new Option("p", PRINT, false,
				"Print raw audio fingerprint");

		Option listSupportedFiles = new Option("l", LIST_FILE_TYPES, false,
				"List supported file extensions and exit");

		Option help = new Option("h", HELP, false,
				"Print this help text and exit");

		Option license = new Option(null, LICENSE, false,
				"Print license information and exit");

		Option version = new Option("v", VERSION, false,
				"Print version information and exit");

		Option verbose = new Option(null, VERBOSE, false, "Verbose output");

		OptionGroup group = new OptionGroup();
		group.addOption(upload);
		group.addOption(search);
		group.addOption(fetchFingerprint);
		group.addOption(approveFingerprint);
		group.addOption(fetchEntity);
		group.addOption(print);
		group.addOption(help);
		group.addOption(version);
		group.addOption(license);
		group.addOption(listSupportedFiles);

		Options options = new Options();
		options.addOptionGroup(group);
		options.addOption(url);
		options.addOption(format);
		options.addOption(verbose);

		return options;
	}

	private String getVersion() {
		String version = CnfManager.getVersion("jhears-client");
		if (version == null) {
			version = "development";
		}
		return version;
	}

	/**
	 * 
	 * @param args
	 * @return system exit status value
	 */
	@SuppressWarnings("unchecked")
	public int parseCommandLine(String[] args) {
		int retval = 0;
		CommandLineParser parser = new PosixParser();
		try {
			CommandLine cmdline = parser.parse(getOptions(), args);
			client.setOutput(System.out);
			client.setError(System.err);

			if (cmdline.hasOption(VERBOSE)) {
				client.setVerbose();
			}
			if (cmdline.hasOption(FILE_FORMAT)) {
				client.setInputFormat(cmdline.getOptionValue(FILE_FORMAT));
			}
			if (cmdline.hasOption(URL)) {
				client.setUrl(cmdline.getOptionValue(URL));
			}

			if (cmdline.hasOption(HELP)) {
				printHelp();
				return retval;
			} else if (cmdline.hasOption(VERSION)) {
				printVersion();
				return retval;
			} else if (cmdline.hasOption(LICENSE)) {
				printLicense();
				return retval;
			} else if (cmdline.hasOption(LIST_FILE_TYPES)) {
				client.listSupportedFiletypes();
				return retval;
			}

			List<String> l = cmdline.getArgList();
			for (String s : l) {
				if (!new File(s).isFile()) {
					throw new FileNotFoundException(new File(s).getPath());
				}
			}
			client.setInput(l);

			if (cmdline.hasOption(SEARCH)) {
				client.search();
			} else if (cmdline.hasOption(UPLOAD)) {
				client.upload(cmdline.getOptionValue(UPLOAD));
			} else if (cmdline.hasOption(PRINT)) {
				client.print();
			} else if (cmdline.hasOption(FETCH_FINGERPRINT)) {
				client.fetchFingerprint(cmdline
						.getOptionValue(FETCH_FINGERPRINT));
			} else if (cmdline.hasOption(APPROVE_FINGERPRINT)) {
				client.approveFingerprint(cmdline
						.getOptionValue(APPROVE_FINGERPRINT));
			} else if (cmdline.hasOption(FETCH_ENTITY)) {
				client.fetchEntity(cmdline.getOptionValue(FETCH_ENTITY));
			} else {
				printHelp();
			}
		} catch (Exception e) {
			logger.error("Execution failed", e);
			System.err.println("Fail:");
			for (Throwable t : Throwables.getCausalChain(e)) {
				System.err.println("  " + t.toString());
			}
			retval = 1;
		}
		return retval;
		//
		// if (cmdline.hasOption(PRINT)) {
		// init(cmdline.hasOption(URL) ? cmdline.getOptionValue(URL)
		// : null);
		// String[] remaining = cmdline.getArgs();
		// String fname = "-";
		// if (remaining.length > 0) {
		// fname = remaining[0];
		// }
		// retval = printToStdout(fname, cmdline.getOptionValue(WORK_DIR));
		// } else if (cmdline.hasOption(FINGERPRINT)) {
		// init(cmdline.hasOption(URL) ? cmdline.getOptionValue(URL)
		// : null);
		// retval = printFingerprint(cmdline.getOptionValue(FINGERPRINT));
		// } else if (cmdline.hasOption(AUDIO_ENTITY)) {
		// init(cmdline.hasOption(URL) ? cmdline.getOptionValue(URL)
		// : null);
		// retval = printAudioEntity(cmdline.getOptionValue(AUDIO_ENTITY));
		// } else if (cmdline.hasOption(SEARCH)) {
		// init(cmdline.hasOption(URL) ? cmdline.getOptionValue(URL)
		// : null);
		// String[] remaining = cmdline.getArgs();
		// String fname = "-";
		// if (remaining.length > 0) {
		// fname = remaining[0];
		// }
		// retval = search(fname, cmdline.getOptionValue(WORK_DIR));
		// } else if (cmdline.hasOption(UPLOAD)) {
		// if (cmdline.hasOption(ENTITY)) {
		// init(cmdline.hasOption(URL) ? cmdline.getOptionValue(URL)
		// : null);
		// String[] remaining = cmdline.getArgs();
		// String fname = "-";
		// if (remaining.length > 0) {
		// fname = remaining[0];
		// }
		// retval = upload(cmdline.getOptionValue(ENTITY), fname,
		// cmdline.getOptionValue(WORK_DIR));
		// } else {
		// toStdout("Upload requires " + entityId + " option.");
		// retval = 1;
		// }
		// } else {
		// printHelp(options);
		// }
	}

	private void printHelp() {
		HelpFormatter formatter = new HelpFormatter2();
		String cmdName = getClass().getSimpleName();
		String version = getVersion();
		String copyright = "jHears client version " + version
				+ ". Copyright (C) 2010  Juha Heljoranta.";
		System.out.println(copyright);
		System.out
				.println("License AGPLv3+: GNU Affero GPL version 3 or later <http://gnu.org/licenses/agpl.html>");
		System.out
				.println("This is free software: you are free to change and redistribute it.");
		formatter.printHelp(cmdName, null, getOptions(), null, true);
	}

	private void printLicense() throws IOException {
		InputStream is = getClass().getResourceAsStream("/license-info.txt");
		try {
			String license = CharStreams.toString(new InputStreamReader(is,
					"UTF-8"));
			System.out.println(license);
		} finally {
			is.close();
		}
	}

	private void printVersion() {
		System.out.println(getVersion());
	}

}
