/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIUtils;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.HttpParams;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.codehaus.jackson.JsonGenerationException;
import org.codehaus.jackson.map.JsonMappingException;
import org.codehaus.jackson.map.ObjectMapper;
import org.jhears.cnf.Cnf;
import org.jhears.web.rs.ApiPath;
import org.jhears.web.rs.AudioEntity;
import org.jhears.web.rs.Fingerprint;
import org.jhears.web.rs.InsertRequest;
import org.jhears.web.rs.SearchResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Charsets;
import com.google.common.io.ByteStreams;
import com.google.common.io.CharStreams;

/**
 * Class is used to communicate with JHears server.
 */
public class JHearsApi implements IJHearsApi {

	private final Configuration configuration;

	private final String host;

	private final HttpClient httpclient;

	private final HttpContext httpContext;

	private final Logger logger;

	private final ObjectMapper om;

	private String password;

	private final String path;

	private final int port;

	private final String scheme;

	private String username;

	public JHearsApi(Configuration configuration, URL url) {
		logger = LoggerFactory.getLogger(getClass());
		this.configuration = configuration;
		HttpParams params = new BasicHttpParams();
		params.setParameter("http.protocol.handle-authentication",
				Boolean.FALSE);
		httpclient = new DefaultHttpClient(params);
		httpContext = new BasicHttpContext();
		String tmpPath;
		if (url != null) {
			scheme = url.getProtocol();
			host = url.getHost();
			port = url.getPort();
			tmpPath = url.getPath();
		} else {
			scheme = configuration.getString(Cnf.ClientKey.SERVER_SCHEME);
			host = configuration.getString(Cnf.ClientKey.SERVER_HOST);
			port = configuration.getInt(Cnf.ClientKey.SERVER_PORT);
			tmpPath = configuration.getString(Cnf.ClientKey.SERVER_PATH);
		}
		// remove trailing slash
		while (tmpPath.endsWith("/")) {
			tmpPath = tmpPath.substring(0, tmpPath.length() - 1);
		}
		path = tmpPath;
		om = new ObjectMapper();

	}

	@Override
	public Boolean approveFingerprint(String fingerprintId) throws IOException,
			UnauthorizedException, URISyntaxException {
		HttpGet get = new HttpGet(getUri(ApiPath.GET_FINGERPRINT + "/"
				+ URLEncoder.encode(fingerprintId, Charsets.UTF_8.name())
				+ ApiPath.APPROVE_FINGERPRINT));
		HttpResponse response = httpclient.execute(get, httpContext);
		try {
			return handleResponse(response, Boolean.class);
		} finally {
			get.abort();
		}
	}

	protected boolean canTryLogin() {
		username = configuration.getString(Cnf.ClientKey.SERVER_USERNAME, null);
		password = configuration.getString(Cnf.ClientKey.SERVER_PASSWORD, null);
		return username != null && password != null;
	}

	@Override
	public AudioEntity getAudioEntity(String audioEntityId) throws IOException,
			UnauthorizedException, URISyntaxException {
		HttpGet get = new HttpGet(getUri(ApiPath.GET_AUDIO_ENTITY + "/"
				+ URLEncoder.encode(audioEntityId, Charsets.UTF_8.name())));
		HttpResponse response = httpclient.execute(get, httpContext);
		try {
			return handleResponse(response, AudioEntity.class);
		} finally {
			get.abort();
		}
	}

	@Override
	@SuppressWarnings("unchecked")
	public Configuration getConfiguration(String name) throws IOException,
			UnauthorizedException, URISyntaxException {
		HttpGet get = new HttpGet(
				getUri(ApiPath.GET_CONFIGURATION + "/" + name));
		HttpResponse response = httpclient.execute(get, httpContext);
		HashMap<String, Object> resp;
		try {
			resp = handleResponse(response, HashMap.class);
		} finally {
			get.abort();
		}
		if (resp != null) {
			Configuration c = new BaseConfiguration();
			for (Map.Entry<String, Object> entry : resp.entrySet()) {
				c.setProperty(entry.getKey(), entry.getValue());
			}
			return c;
		}
		return null;
	}

	@Override
	public Fingerprint getFingerprint(String fingerprintId) throws IOException,
			UnauthorizedException, URISyntaxException {
		HttpGet get = new HttpGet(getUri(ApiPath.GET_FINGERPRINT + "/"
				+ URLEncoder.encode(fingerprintId, Charsets.UTF_8.name())));
		HttpResponse response = httpclient.execute(get, httpContext);
		try {
			return handleResponse(response, Fingerprint.class);
		} finally {
			get.abort();
		}
	}

	private URI getLoginURI() throws URISyntaxException {
		String apiPath = this.path + "/login";
		return URIUtils.createURI(scheme, host, port, apiPath, null, null);
	}

	private URI getUri(String path) throws URISyntaxException {
		String apiPath = this.path + "/api" + path;
		String query = null;
		URI uri = URIUtils.createURI(scheme, host, port, apiPath, query, null);
		return uri;
	}

	private Boolean handleLogin(HttpResponse response) throws IOException,
			UnauthorizedException {
		HttpEntity entity;
		switch (response.getStatusLine().getStatusCode()) {
		case HttpStatus.SC_OK:
			entity = response.getEntity();
			// TODO detect content encoding
			byte[] out = ByteStreams.toByteArray(entity.getContent());
			String rval = new String(out, Charsets.UTF_8).trim();
			return Boolean.parseBoolean(rval);
		case HttpStatus.SC_UNAUTHORIZED:
			throw new UnauthorizedException();
		default:
			logger.warn("Server returned an unexpected status code: "
					+ response.getStatusLine());
			break;
		}
		return null;
	}

	private <T> T handleResponse(HttpResponse response, Class<T> clazz)
			throws IOException, UnauthorizedException {
		switch (response.getStatusLine().getStatusCode()) {
		case HttpStatus.SC_OK:
			return parseJson(response, clazz);
		case HttpStatus.SC_UNAUTHORIZED:
			throw new UnauthorizedException();
		default:
			logger.warn("Server returned an unexpected status code: "
					+ response.getStatusLine());
			break;
		}
		return null;
	}

	@Override
	public Fingerprint insert(InsertRequest insert)
			throws JsonGenerationException, JsonMappingException, IOException,
			URISyntaxException, UnauthorizedException {
		HttpPost post = new HttpPost(getUri(ApiPath.INSERT));
		int approximateJsonDataSize = insert.getFingerprint().length() * 2 + 32;
		StringEntity json = toJson(insert, approximateJsonDataSize);
		post.setEntity(json);
		HttpResponse response = httpclient.execute(post, httpContext);
		try {
			return handleResponse(response, Fingerprint.class);
		} finally {
			post.abort();
		}
	}

	protected Boolean login() throws URISyntaxException,
			ClientProtocolException, IOException, UnauthorizedException {
		HttpPost post = new HttpPost(getLoginURI());
		List<NameValuePair> formparams = new ArrayList<NameValuePair>();
		formparams.add(new BasicNameValuePair("username", username));
		formparams.add(new BasicNameValuePair("password", password));
		UrlEncodedFormEntity entity = new UrlEncodedFormEntity(formparams,
				Charsets.UTF_8.name());
		post.setEntity(entity);
		HttpResponse response = httpclient.execute(post, httpContext);
		try {
			return handleLogin(response);
		} finally {
			post.abort();
		}
	}

	private <T> T parseJson(HttpResponse response, Class<T> clazz)
			throws IOException {
		InputStream is = null;
		HttpEntity responseentity = response.getEntity();
		if (responseentity.getContentLength() == 0) {
			return null;
		}
		try {
			is = responseentity.getContent();
			return om.readValue(is, clazz);
		} finally {
			if (is != null) {
				is.close();
			}
		}
	}

	@Override
	public SearchResult[] search(String fingerprint) throws IOException,
			UnauthorizedException, URISyntaxException {
		HttpPost post = new HttpPost(getUri(ApiPath.SEARCH));
		int approximateJsonDataSize = fingerprint.length() + 16;
		StringEntity json = toJson(fingerprint, approximateJsonDataSize);
		post.setEntity(json);
		HttpResponse response = httpclient.execute(post, httpContext);
		try {
			return handleResponse(response, SearchResult[].class);
		} finally {
			post.abort();
		}
	}

	private StringEntity toJson(Object o, int initialBufferSize)
			throws JsonGenerationException, JsonMappingException, IOException {
		StringWriter sw;
		if (initialBufferSize < 0) {
			sw = new StringWriter();
		} else {
			sw = new StringWriter(initialBufferSize);
		}
		om.writeValue(sw, o);
		// conform to rfc4627. use utf-8.
		StringEntity entity = new StringEntity(sw.toString(),
				Charsets.UTF_8.name());
		entity.setContentType("application/json");
		return entity;
	}

}
