/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.output.CloseShieldOutputStream;
import org.apache.commons.lang.time.DurationFormatUtils;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.Cnf.FingerprintKey;
import org.jhears.cnf.Cnf.SearchKey;
import org.jhears.cnf.CnfManager;
import org.jhears.logging.OutputStreamLogger;
import org.jhears.logging.OutputStreamLogger.Level;
import org.jhears.sample.AudioPreprocessor;
import org.jhears.seq.FPSeqListener;
import org.jhears.seq.FPSeqProducer;
import org.jhears.seq.FSeqFactory;
import org.jhears.seq.MutableFSeq;
import org.jhears.seq.SeqMatch;
import org.jhears.web.rs.AudioEntity;
import org.jhears.web.rs.Fingerprint;
import org.jhears.web.rs.InsertRequest;
import org.jhears.web.rs.Match;
import org.jhears.web.rs.RepeatCluster;
import org.jhears.web.rs.SearchResult;
import org.jhears.web.rs.SegmentCluster;
import org.slf4j.LoggerFactory;

import com.google.common.base.Throwables;

public class JHearsClientNG implements IJHearsClient {

	private IJHearsApi client;

	private Configuration configuration;

	private PrintStream error;

	private final CnfManager factory;

	private String format;

	private List<String> inputFiles;

	private PrintStream out;

	private String url;

	private boolean verbose;

	public JHearsClientNG() {
		// this.logger = LoggerFactory.getLogger(JHearsClientNG.class);
		this.verbose = false;
		this.url = null;
		this.out = System.out;
		this.error = System.err;
		this.format = null;
		this.inputFiles = new ArrayList<String>(0);
		try {
			factory = new CnfManager();
		} catch (IOException e) {
			throw Throwables.propagate(e);
		}
	}

	@Override
	public void approveFingerprint(String fingerprintId) {
		try {
			init();
			Boolean ok = client.approveFingerprint(fingerprintId);
			if (ok != null) {
				if (ok.booleanValue()) {
					out.println("Fingerprint approval accepted.");
				} else {
					out.println("Fingerprint approval rejected. "
							+ "Perhaps you have alrady approved this fingerprint?");
				}
			} else {
				out.println("Fingerprint not found.");
			}
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	@Override
	public void fetchEntity(String entityId) {
		try {
			init();
			AudioEntity audioEntity = client.getAudioEntity(entityId);
			if (audioEntity != null) {
				if (verbose) {
					out.println(audioEntity.toString());
				} else {
					for (Fingerprint fp : audioEntity.getFingerprints()) {
						out.println(fp.getIdentifier());
					}
				}
			}
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	@Override
	public void fetchFingerprint(String fingerprintId) {
		try {
			init();
			Fingerprint fingerprint = client.getFingerprint(fingerprintId);
			if (fingerprint != null) {
				if (verbose) {
					out.println(fingerprint.toString());
				} else {
					out.println(fingerprint.toShortString());
				}
			}
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	private IJHearsApi getApi(java.net.URL url) throws ConfigurationException {
		IJHearsApi api = new JHearsApiLoginWrapper(new JHearsApi(
				factory.getConfiguration(Cnf.client), url));
		return api;
	}

	private void init() throws ConfigurationException, IOException,
			UnauthorizedException, URISyntaxException {
		URL url = null;
		if (this.url != null) {
			url = new URL(this.url);
		}
		client = getApi(url);
		loadConfiguration();
	}

	@Override
	public List<String> listSupportedFiletypes() {
		try {
			init();
			AudioPreprocessor p;
			p = new AudioPreprocessor(configuration, null);
			OutputStreamLogger errorLogger = new OutputStreamLogger(
					LoggerFactory.getLogger(getClass().getName()), Level.warn);
			errorLogger.start(false);
			CloseShieldOutputStream errStream = new CloseShieldOutputStream(
					errorLogger.getOutputStream());
			p.setErrorStream(errStream);
			Set<String> set = p.listSupportedFileFormats();
			errStream.flush();
			errorLogger.close();
			set.add("fingerprint");
			ArrayList<String> sortedList = new ArrayList<String>(set);
			Collections.sort(sortedList);
			return sortedList;
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	private void loadConfiguration() throws IOException, UnauthorizedException,
			URISyntaxException, ConfigurationException {
		Cnf[] cArr = new Cnf[] { Cnf.audioPrePros, Cnf.fingerprint,
				Cnf.framework, Cnf.search };
		for (Cnf cnf : cArr) {
			if (!factory.hasConfiguration(cnf)) {
				Configuration c = client.getConfiguration(cnf.toString());
				factory.saveConfiguration(cnf, c);
			}
		}
		configuration = factory.getConfiguration(Cnf.audioPrePros,
				Cnf.fingerprint, Cnf.framework, Cnf.search);
	}

	/**
	 * Prints to {@link #out}.
	 * 
	 * @return
	 */
	private FPSeqListener newFPSeqListener() {
		return new FPSeqListener() {

			private MutableFSeq fseq = new FSeqFactory(configuration)
					.getNewInstance();

			@Override
			public void append(char c) {
				fseq.append(c);
			}

			@Override
			public void append(char[] c) {
				fseq.append(c);
			}

			@Override
			public void eof() {
				if (verbose) {
					out.println(fseq.toString());
				} else {
					out.println(fseq.asString());
				}
				// out.println(fseq.flatten().toString());
			}
		};
	}

	/**
	 * Prints to {@code fseq}.
	 * 
	 * @param fseq
	 *            where to print.
	 * @return
	 */
	private FPSeqListener newFPSeqListener(final MutableFSeq fseq) {
		return new FPSeqListener() {

			@Override
			public void append(char c) {
				fseq.append(c);
			}

			@Override
			public void append(char[] c) {
				fseq.append(c);
			}

			@Override
			public void eof() {

			}
		};
	}

	private String positionToTime(int pos) {
		double symbolRate = configuration.getInt(FingerprintKey.SAMPLE_RATE)
				/ (double) configuration.getInt(FingerprintKey.SAMPLE_INTERVAL);
		long value = Math.round(Math.abs(pos) / symbolRate * 1000L);
		String retval = DurationFormatUtils.formatDurationHMS(value);
		if (pos < 0) {
			return "-" + retval;
		} else {
			return retval;
		}
	}

	@Override
	public void print() {
		try {
			init();
			FPSeqProducer producer;
			producer = new FPSeqProducer(configuration, null);
			if (inputFiles.isEmpty()) {
				FPSeqListener seqListener = newFPSeqListener();
				producer.addListener(seqListener);
				producer.produceFrom(System.in, format);
				producer.removeListener(seqListener);
			} else {
				for (String inputFile : inputFiles) {
					FPSeqListener seqListener = newFPSeqListener();
					producer.addListener(seqListener);
					producer.produceFrom(new File(inputFile));
					producer.removeListener(seqListener);
				}
			}
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	@Override
	public void search() {
		try {
			init();

			final MutableFSeq fseq = new FSeqFactory(configuration)
					.getNewInstance();
			FPSeqProducer producer;
			producer = new FPSeqProducer(configuration, null);
			if (inputFiles.isEmpty()) {
				FPSeqListener seqListener = newFPSeqListener(fseq);
				producer.addListener(seqListener);
				producer.produceFrom(System.in, format);
				producer.removeListener(seqListener);
			} else {
				for (String inputFile : inputFiles) {
					FPSeqListener seqListener = newFPSeqListener(fseq);
					producer.addListener(seqListener);
					producer.produceFrom(new File(inputFile));
					producer.removeListener(seqListener);
				}
			}

			SearchResult[] results = client.search(fseq.asString());
			// flatten so that the seqmatch works correctly
			fseq.flatten();
			SeqMatch match = new SeqMatch(
					configuration.getFloat(SearchKey.SEQ_MATCH_THRESHOLD));

			for (SearchResult result : results) {
				Set<Match> matches = result.getMatches();
				out.println("entityId: " + result.getAudioEntityId());
				for (Match m : matches) {
					if (verbose) {
						out.println("fingerprint: " + m.toString());
					} else {
						out.println("fingerprintId: " + m.getFingerprintId());
					}
					float maxConfidence = -1.0f;
					int maxconfpos = -1;
					int others = -1; // start with -1 to exclude self
					if (m.getSegmentClusters() != null) {
						for (SegmentCluster s : m.getSegmentClusters()) {
							others++;
							float confidence = match.segmentsInSeqConfidence(
									m.getSegments(), s.getRelativePosition(),
									fseq);
							int pos = s.getRelativePosition();
							if (confidence > maxConfidence) {
								maxConfidence = confidence;
								maxconfpos = pos;
							}
							if (verbose) {
								printConfidence(confidence, pos, -1);
							}
						}
					}
					if (m.getRepeatClusters() != null) {
						for (RepeatCluster r : m.getRepeatClusters()) {
							others++;
							float confidence = match.repeatsInSeqConfidence(
									m.getRepeats(), r.getRelativePosition(),
									fseq);
							int pos = r.getRelativePosition();
							if (confidence > maxConfidence) {
								maxConfidence = confidence;
								maxconfpos = pos;
							}
							if (verbose) {
								printConfidence(confidence, pos, -1);
							}
						}
					}
					if (!verbose && maxconfpos != -1) {
						printConfidence(maxConfidence, maxconfpos, others);
					}
				}
			}

		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	private void printConfidence(float maxConfidence, int maxconfpos,
			int matchCount) {
		String others = matchCount <= 0 ? "" : " + " + matchCount + " others";
		out.println("confidence: " + String.format("%.2f", maxConfidence)
				+ " @ " + maxconfpos + " (" + positionToTime(maxconfpos) + ")"
				+ others);
	}

	@Override
	public void setError(PrintStream error) {
		this.error = error;
	}

	@Override
	public void setInput(List<String> inputFiles) {
		this.inputFiles = inputFiles;
	}

	@Override
	public void setInputFormat(String format) {
		this.format = format;

	}

	@Override
	public void setOutput(PrintStream out) {
		this.out = out;
	}

	@Override
	public void setUrl(String url) {
		this.url = url;
	}

	@Override
	public void setVerbose() {
		this.verbose = true;

	}

	@Override
	public void upload(String entityId) {
		try {
			init();

			final MutableFSeq fseq = new FSeqFactory(configuration)
					.getNewInstance();
			FPSeqProducer producer = new FPSeqProducer(configuration, null);

			if (inputFiles.isEmpty()) {
				FPSeqListener seqListener = newFPSeqListener(fseq);
				producer.addListener(seqListener);
				producer.produceFrom(System.in, format);
				producer.removeListener(seqListener);
			} else {
				for (String inputFile : inputFiles) {
					FPSeqListener seqListener = newFPSeqListener(fseq);
					producer.addListener(seqListener);
					producer.produceFrom(new File(inputFile));
					producer.removeListener(seqListener);
				}
			}
			Fingerprint result = client.insert(new InsertRequest(entityId, fseq
					.asString()));
			if (result == null) {
				error.println("Server rejected the upload. "
						+ "The server probably already has the fingerprint, "
						+ "try to searching it.");
			} else {
				if (verbose) {
					out.println(result.toString());
				} else {
					out.println(result.getIdentifier());
				}
			}
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

}
