/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import java.io.File;
import java.io.FileDescriptor;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.Serializable;
import java.io.SyncFailedException;

import org.jhears.util.RuntimeIOException;

/**
 * Provides an infrastructure for a bit arrays stored into a file.
 * 
 * @author Juha Heljoranta
 * 
 */
public abstract class AbstractFileBitArray implements IBitArray, Serializable {

	/**
	 * The access mode in which the file is to be opened.
	 */
	public static enum Mode {

		/**
		 * Open for reading only. Invoking any of the <tt>write</tt> methods of
		 * the resulting object will cause an {@link java.io.IOException} to be
		 * thrown.
		 */
		R,

		/**
		 * Open for reading and writing. If the file does not already exist then
		 * an attempt will be made to create it.
		 */
		RW,

		/**
		 * Open for reading and writing, as with <tt>"rw"</tt>, and also require
		 * that every update to the file's content be written synchronously to
		 * the underlying storage device.
		 */
		RWD,

		/**
		 * Open for reading and writing, as with <tt>"rw"</tt>, and also require
		 * that every update to the file's content or metadata be written
		 * synchronously to the underlying storage device.
		 */
		RWS;
	}

	private static final long serialVersionUID = 8027645904383477604L;

	private final boolean autoClose;

	private final File file;

	private final long fileLength;

	/** The number of bits stored in this array. */
	private final int length;

	private final String mode;

	private RandomAccessFile raf;

	/**
	 * 
	 * @param length
	 *            The length of the bit array.
	 * @param file
	 *            The file which is used to store the bit array.
	 * @param fileLength
	 *            The length of the file.
	 * @param autoClose
	 *            true if the file should be closed automatically after each
	 *            operation (the file is automatically opened when needed).
	 * @param mode
	 *            The mode in which the file should be opened.
	 */
	public AbstractFileBitArray(final int length, File file, long fileLength,
			boolean autoClose, Mode mode) {
		this.length = length;
		this.file = file;
		this.fileLength = fileLength;
		this.autoClose = autoClose;
		this.mode = mode.toString().toLowerCase();
		truncate();
	}

	@Override
	public void clear() {
		try {
			openRaf();
			final byte[] buf = new byte[(1 << 13)];
			long remainingBytes = fileLength;
			while (remainingBytes > 0L) {
				long a = buf.length;
				if (remainingBytes < buf.length) {
					a = remainingBytes;
				}
				raf.write(buf, 0, (int) a);
				remainingBytes -= a;
			}
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		} finally {
			closeRaf();
		}
	}

	@Override
	public void clear(final int index) throws IndexOutOfBoundsException {
		if (index < length) {
			try {
				openRaf();
				clear(index, raf);
			} catch (IOException e) {
				throw new RuntimeIOException(e);
			} finally {
				closeRaf();
			}
		} else {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	protected abstract void clear(final int index, RandomAccessFile raf)
			throws IOException;

	@Override
	public void clearAll(final int[] indexes) throws IndexOutOfBoundsException {
		try {
			openRaf();
			for (final int i : indexes) {
				clear(i, raf);
			}
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		} finally {
			closeRaf();
		}
	}

	/**
	 * Close the file if its open.
	 */
	public void close() {
		try {
			if (raf != null) {
				raf.close();
			}
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		}
		raf = null;
	}

	private void closeRaf() {
		if (autoClose) {
			close();
		}

	}

	@Override
	public boolean get(final int index) throws IndexOutOfBoundsException {
		if (index < length) {
			try {
				openRaf();
				return get(index, raf);
			} catch (IOException e) {
				throw new RuntimeIOException(e);
			} finally {
				closeRaf();
			}
		} else {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	protected abstract boolean get(int index, RandomAccessFile raf)
			throws IOException;

	@Override
	public boolean[] getAll(final int[] indexes)
			throws IndexOutOfBoundsException {
		final boolean[] retval = new boolean[indexes.length];
		try {
			openRaf();
			for (int i = 0; i < indexes.length; i++) {
				retval[i] = get(indexes[i], raf);
			}
			return retval;
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		} finally {
			closeRaf();
		}
	}

	/**
	 * Return the length of this bit array.
	 * 
	 * @return The length of this bit array.
	 */
	public int length() {
		return length;
	}

	public void open() {
		try {
			openRaf();
		} catch (FileNotFoundException e) {
			throw new RuntimeIOException(e);
		}
	}

	private void openRaf() throws FileNotFoundException {
		if (raf == null) {
			raf = new RandomAccessFile(file, mode);
		}
	}

	@Override
	public void set(final int index) throws IndexOutOfBoundsException {
		if (index < length) {
			try {
				openRaf();
				set(index, raf);
			} catch (IOException e) {
				throw new RuntimeIOException(e);
			} finally {
				closeRaf();
			}
		} else {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	protected abstract void set(int index, RandomAccessFile raf)
			throws IOException;

	@Override
	public void setAll(final int[] indexes) throws IndexOutOfBoundsException {
		try {
			openRaf();
			for (final int i : indexes) {
				set(i, raf);
			}
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		} finally {
			closeRaf();
		}
	}

	@Override
	public int size() {
		return length;
	}

	/**
	 * Force all system buffers to synchronize with the underlying device.
	 * 
	 * @see FileDescriptor#sync()
	 */
	public void sync() {
		if (raf != null) {
			try {
				raf.getFD().sync();
			} catch (SyncFailedException e) {
				throw new RuntimeIOException(e);
			} catch (IOException e) {
				throw new RuntimeIOException(e);
			}
		}
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + " [length=" + length + ", file="
				+ file + "]";
	}

	private void truncate() {
		try {
			openRaf();
			raf.setLength(fileLength);
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		} finally {
			closeRaf();
		}
	}

}
