/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import java.util.Arrays;

import org.jhears.util.IntList;

public final class BitArrayUtil {

	public static interface LazyLoadingCompletionListener {
		public void loadingCompleted();
	}

	private static final int MAX = 1 << 18;

	/**
	 * Copy bit array to another bit-by-bit.
	 * <p>
	 * The number of copied bits is limited to smallest of the two arrays.
	 * 
	 * @param from
	 * @param to
	 * @return the number of bits copied.
	 */
	public static final int copy(IBitArray from, IBitArray to) {
		final int size = from.size() < to.size() ? from.size() : to.size();
		int remaining = size;
		int idx = 0;
		while (remaining > 0) {
			int copyLen = remaining;
			if (copyLen > MAX) {
				copyLen = MAX;
			}
			int[] indexes = new int[copyLen];
			for (int i = 0; i < copyLen; i++) {
				indexes[i] = idx++;
			}
			copy(from, to, indexes);
			remaining -= copyLen;
			// boolean[] setIndexes = from.getAll(indexes);
			// IntList setArr = new IntList(copyLen);
			// IntList clearArr = new IntList(copyLen);
			// for (int i = 0; i < copyLen; i++) {
			// if (setIndexes[i]) {
			// setArr.add(indexes[i]);
			// } else {
			// clearArr.add(indexes[i]);
			// }
			// }
			// to.setAll(setArr.toArray());
			// to.clearAll(clearArr.toArray());
		}
		return size;
	}

	public static void copy(final IBitArray from, final IBitArray to,
			final int[] indexes) {
		boolean[] setIndexes = from.getAll(indexes);
		IntList setArr = new IntList(indexes.length);
		IntList clearArr = new IntList(indexes.length);
		for (int i = 0; i < indexes.length; i++) {
			if (setIndexes[i]) {
				setArr.add(indexes[i]);
			} else {
				clearArr.add(indexes[i]);
			}
		}
		to.setAll(setArr.toArray());
		to.clearAll(clearArr.toArray());
	}

	/**
	 * Compares two bit arrays. Bit arrays are equal only if they have the same
	 * size and contents of the arrays are equal.
	 * 
	 * @param a
	 * @param b
	 * @return
	 */
	public static final boolean equals(IBitArray a, IBitArray b) {
		int size = a.size();
		if (b.size() != size) {
			return false;
		}
		int remaining = size;
		int idx = 0;
		while (remaining > 0) {
			int copyLen = remaining;
			if (copyLen > MAX) {
				copyLen = MAX;
			}
			int[] indexes = new int[copyLen];
			for (int i = 0; i < copyLen; i++) {
				indexes[i] = idx++;
			}
			boolean[] aIndexes = a.getAll(indexes);
			boolean[] bIndexes = b.getAll(indexes);
			if (!Arrays.equals(aIndexes, bIndexes)) {
				return false;
			}
			remaining -= copyLen;
		}
		return true;
	}

	/**
	 * Populates {@link CachedBitArray} on demand.
	 */
	public static ForwardingBitArray withLazyLoad(
			final CachedBitArray cachedBitArray) {
		return withLazyLoad(cachedBitArray, null);
	}

	/**
	 * Populates {@link CachedBitArray} on demand.
	 */
	public static ForwardingBitArray withLazyLoad(
			final CachedBitArray cachedBitArray,
			final LazyLoadingCompletionListener completionListner) {
		final ForwardingBitArray forwarding = new ForwardingBitArray();
		LazyListener<IBitArray> listener = new LazyListener<IBitArray>() {
			@Override
			public void setDelegate(IBitArray t) {
				forwarding.setDelegate(t);
				if (completionListner != null && t instanceof CachedBitArray) {
					completionListner.loadingCompleted();
				}
			}
		};
		// will be garbage collected after loading is done
		new LazyCachedBitArray(cachedBitArray, listener);
		return forwarding;
	}

	private BitArrayUtil() {
	}

}
