/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import java.io.Serializable;

/**
 * Wrapper class to build cached bit array.
 * <p>
 * Class encloses two {@link IBitArray} instances. The first is called slow and
 * the second is called fast.
 * <p>
 * Note that the bit arrays should be equal (
 * {@link BitArrayUtil#equals(IBitArray, IBitArray)} must return true).
 * <p>
 * Alternatively {@link BitArrayUtil#withLazyLoad(CachedBitArray)} can be used
 * for automatic lazy loading:
 * 
 * <pre>
 * BitArrayUtil.withLazyLoad(
 *   new CachedBitArray(persistentBitArray,
 *   new BitArray(persistentBitArray.size()))
 * </pre>
 * <p>
 * {@link #get(int)} and {@link #size()} methods are delegated to fast bit
 * array. Rest of methods ({@link #clear()}, {@link #clear(int)} and
 * {@link #set(int)}) are delegated to both of the bit arrays. When delegating
 * to both arrays the order is fist the fast bit array then the slow bit array.
 * <p>
 * If the slow bit array is instance of {@link ICountingBitArray} then the
 * {@link #clear(int)} will first clear the slow bit array bit and checks if the
 * slow bit array still has the bit set, if not then the fast bit array bit is
 * also cleared.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class CachedBitArray implements IBitArray, Serializable {

	private static final long serialVersionUID = 6845215528368835711L;

	private final IBitArray fastBitArray;

	private final IBitArray slowBitArray;

	private final boolean slowCanCount;

	/**
	 * Create a new instance.
	 * <p>
	 * Note that the bit arrays should be equal (
	 * {@link BitArrayUtil#equals(IBitArray, IBitArray)} must return {@code
	 * true}). {@link BitArrayUtil#withLazyLoad(CachedBitArray)} can be used for
	 * automatic lazy loading.
	 * 
	 * @param slowBitArray
	 * @param fastBitArray
	 */
	public CachedBitArray(IBitArray slowBitArray, IBitArray fastBitArray) {
		this.slowBitArray = slowBitArray;
		this.fastBitArray = fastBitArray;
		slowCanCount = slowBitArray instanceof ICountingBitArray;
	}

	@Override
	public void clear() {
		fastBitArray.clear();
		slowBitArray.clear();
	}

	@Override
	public void clear(int index) throws IndexOutOfBoundsException {
		if (slowCanCount) {
			slowBitArray.clear(index);
			if (!slowBitArray.get(index)) {
				fastBitArray.clear(index);
			}
		} else {
			fastBitArray.clear(index);
			slowBitArray.clear(index);
		}
	}

	@Override
	public void clearAll(int[] indexes) throws IndexOutOfBoundsException {
		if (slowCanCount) {
			slowBitArray.clearAll(indexes);
			boolean[] values = slowBitArray.getAll(indexes);
			for (int i = 0; i < indexes.length; i++) {
				if (!values[i]) {
					fastBitArray.clear(indexes[i]);
				}
			}
		} else {
			fastBitArray.clearAll(indexes);
			slowBitArray.clearAll(indexes);
		}
	}

	@Override
	public boolean get(int index) throws IndexOutOfBoundsException {
		return fastBitArray.get(index);
	}

	@Override
	public boolean[] getAll(int[] indexes) throws IndexOutOfBoundsException {
		return fastBitArray.getAll(indexes);
	}

	public IBitArray getFastBitArray() {
		return fastBitArray;
	}

	public IBitArray getSlowBitArray() {
		return slowBitArray;
	}

	@Override
	public void set(int index) throws IndexOutOfBoundsException {
		fastBitArray.set(index);
		slowBitArray.set(index);
	}

	@Override
	public void setAll(int[] indexes) throws IndexOutOfBoundsException {
		fastBitArray.setAll(indexes);
		slowBitArray.setAll(indexes);
	}

	@Override
	public int size() {
		return fastBitArray.size();
	}

	@Override
	public String toString() {
		return CachedBitArray.class.getSimpleName() + " [fastBitArray="
				+ fastBitArray + ", slowBitArray=" + slowBitArray + "]";
	}

}
