/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import java.io.Serializable;
import java.util.Arrays;

/**
 * Fixed length bit array.
 * <p>
 * Class is backed by array of byte values.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 */
public class CountingBitArray implements ICountingBitArray, Serializable {

	private static final long serialVersionUID = -7040404132522853603L;

	/** Array used to store values. */
	private final byte[] arr;

	/** The number of bits stored in this array. */
	private final int length;

	/**
	 * Creates a new instance.
	 * 
	 * @param length
	 *            The length of the bit array.
	 */
	public CountingBitArray(final int length) {
		this.length = length;
		arr = new byte[length];
	}

	/**
	 * Convenience constructor which allows to override internal data array.
	 * 
	 * @param length
	 * @param arr
	 */
	protected CountingBitArray(final int length, byte[] arr) {
		this.length = length;
		this.arr = arr;
	}

	@Override
	public void clear() {
		Arrays.fill(arr, (byte) 0);
	}

	@Override
	public void clear(final int index) throws IndexOutOfBoundsException {
		if ((arr[index] & 0xff) > 0) {
			arr[index] = (byte) ((arr[index] & 0xff) - 1);
		}
	}

	@Override
	public void clearAll(int[] indexes) throws IndexOutOfBoundsException {
		for (int i : indexes) {
			clear(i);
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CountingBitArray other = (CountingBitArray) obj;
		if (!Arrays.equals(arr, other.arr))
			return false;
		if (length != other.length)
			return false;
		return true;
	}

	@Override
	public boolean get(final int index) throws IndexOutOfBoundsException {
		return (arr[index] & 0xff) > 0;
	}

	@Override
	public boolean[] getAll(int[] indexes) throws IndexOutOfBoundsException {
		final boolean[] retval = new boolean[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			retval[i] = get(indexes[i]);
		}
		return retval;
	}

	/**
	 * Returns a copy of backing long array.
	 * 
	 * @return
	 */
	public byte[] getArray() {
		return Arrays.copyOf(arr, arr.length);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(arr);
		result = prime * result + length;
		return result;
	}

	/**
	 * Return the length of this bit array.
	 * 
	 * @return The length of this bit array.
	 */
	public int length() {
		return length;
	}

	@Override
	public void set(final int index) throws IndexOutOfBoundsException {
		if ((arr[index] & 0xff) < 0xff) {
			arr[index] = (byte) ((arr[index] & 0xff) + 1);
		} else {
			throw new CountingBitArrayOverflowException();
		}
	}

	@Override
	public void setAll(int[] indexes) throws IndexOutOfBoundsException {
		for (int i : indexes) {
			set(i);
		}
	}

	@Override
	public int size() {
		return length;
	}

	@Override
	public String toString() {
		return CountingBitArray.class.getSimpleName() + " [length=" + length
				+ ", arr.length=" + arr.length + ", arr.hashCode="
				+ Arrays.hashCode(arr) + "]";
	}

}
