/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import java.util.HashSet;

import org.jhears.util.IntList;

/**
 * Populates {@link CachedBitArray} in lazy fashion. This class is not intended
 * to be used after the loading is done for performance reasons. See
 * {@link BitArrayUtil#withLazyLoad(CachedBitArray)}.
 * <p>
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class LazyCachedBitArray implements IBitArray {

	public static final int MIN_LOAD_SIZE = 8192;

	private final CachedBitArray cachedBitArray;

	private boolean isPopulated;

	private final LazyListener<IBitArray> listener;

	private final Object mutex;

	private final BitArray populated;

	private int remainingToPopulate;

	public LazyCachedBitArray(CachedBitArray cachedBitArray,
			LazyListener<IBitArray> listener) throws IllegalArgumentException {
		if (cachedBitArray == null) {
			throw new IllegalArgumentException("CachedBitArray cannot be null");
		}
		if (listener == null) {
			throw new IllegalArgumentException("Listener cannot be null");
		}
		this.mutex = new Object();
		this.listener = listener;
		this.populated = new BitArray(cachedBitArray.size());
		this.remainingToPopulate = cachedBitArray.size();
		this.isPopulated = false;
		this.cachedBitArray = cachedBitArray;
		setPopulated(); // in case the array size is 0
		if (!isPopulated) {
			listener.setDelegate(this);
		}
	}

	@Override
	public void clear() {
		synchronized (mutex) {
			cachedBitArray.clear();
			remainingToPopulate = 0;
			setPopulated();
		}
	}

	@Override
	public void clear(int index) throws IndexOutOfBoundsException {
		synchronized (mutex) {
			doPopulate(nextRemainingIndexes(new int[] { index }));
			cachedBitArray.clear(index);
		}
	}

	@Override
	public void clearAll(int[] indexes) throws IndexOutOfBoundsException {
		synchronized (mutex) {
			doPopulate(nextRemainingIndexes(indexes));
			cachedBitArray.clearAll(indexes);
		}
	}

	private void doPopulate(int[] indexes) {
		int[] remaining = nextRemainingIndexes(indexes);
		BitArrayUtil.copy(cachedBitArray.getSlowBitArray(), cachedBitArray
				.getFastBitArray(), remaining);
		markPopulatedAll(remaining);
	}

	@Override
	public boolean get(int index) throws IndexOutOfBoundsException {
		synchronized (mutex) {
			doPopulate(nextRemainingIndexes(new int[] { index }));
			return cachedBitArray.get(index);
		}
	}

	@Override
	public boolean[] getAll(int[] indexes) throws IndexOutOfBoundsException {
		synchronized (mutex) {
			doPopulate(indexes);
			return cachedBitArray.getAll(indexes);
		}
	}

	public CachedBitArray getCachedBitArray() {
		return cachedBitArray;
	}

	public boolean isPopulated() {
		synchronized (mutex) {
			return isPopulated;
		}
	}

	private void markPopulated(int index) {
		if (!populated.get(index)) {
			populated.set(index);
			remainingToPopulate--;
			setPopulated();
		}
	}

	private void markPopulatedAll(int[] indexes) {
		for (int index : indexes) {
			markPopulated(index);
		}
	}

	/**
	 * 
	 * @param indexes
	 *            indexes to include
	 * @return
	 */
	private int[] nextRemainingIndexes(int[] indexes) {
		int max = Math.min(MIN_LOAD_SIZE, remainingToPopulate);
		max = Math.max(indexes.length, max);
		IntList list = new IntList(max);
		int size = cachedBitArray.size();
		HashSet<Integer> addedIndexes = new HashSet<Integer>(indexes.length);
		for (int i = 0; i < indexes.length && list.size() < max; i++) {
			if (!populated.get(indexes[i])) {
				addedIndexes.add(Integer.valueOf(indexes[i]));
				list.add(indexes[i]);
			}
		}
		for (int i = 0; i < size && list.size() < max; i++) {
			if (!addedIndexes.contains(Integer.valueOf(i)) && !populated.get(i)) {
				list.add(i);
			}
		}
		return list.toArray();
	}

	@Override
	public void set(int index) throws IndexOutOfBoundsException {
		synchronized (mutex) {
			doPopulate(nextRemainingIndexes(new int[] { index }));
			cachedBitArray.set(index);
		}
	}

	@Override
	public void setAll(int[] indexes) throws IndexOutOfBoundsException {
		synchronized (mutex) {
			doPopulate(indexes);
			cachedBitArray.setAll(indexes);
		}
	}

	/**
	 * If all bits are loaded mark the cache populated.
	 */
	private void setPopulated() {
		if (remainingToPopulate <= 0 && !isPopulated) {
			isPopulated = true;
			// populated = null; // release memory
			listener.setDelegate(cachedBitArray);
		}
	}

	@Override
	public int size() {
		return cachedBitArray.size();
	}

	@Override
	public String toString() {
		return LazyCachedBitArray.class.getSimpleName() + " [isPopulated="
				+ isPopulated + ", remainingToPopulate=" + remainingToPopulate
				+ "]";
	}

}
