/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.crypto;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Convenience class for producing cryptographically strong authentication
 * codes.
 * <p>
 * Pseudo code: {@code hmac(digest(password || salt), password)}.
 * </p>
 * If the salt and the authentication code are compromised the attacker is
 * forced to perform brute force key search in order to recover the password.
 * <p>
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class AuthCode {

	private final MessageDigest digest;

	private final HMac mac;

	private final Object mutex;

	/**
	 * Create an instance by using HMAC-SHA-512 and SHA-512 algorithms.
	 */
	public AuthCode() {
		this("SHA-512", new HMacImpl());
	}

	/**
	 * Create an instance by using custom message digest and hmac algorithms.
	 * 
	 * @param messageDigestAlgorithm
	 * @param mac
	 */
	private AuthCode(String messageDigestAlgorithm, HMac mac) {
		this.mutex = new Object();
		this.mac = mac;
		try {
			this.digest = MessageDigest.getInstance(messageDigestAlgorithm);
		} catch (NoSuchAlgorithmException e) {
			throw new IllegalArgumentException(e);
		}
	}

	/**
	 * Calculate an authentication code.
	 * <p>
	 * Pseudo code: {@code hmac(digest(password || salt), password)}.
	 * 
	 * @param salt
	 *            Preferably at least 8 bytes. 20 - 64 bytes should be quite
	 *            enough.
	 * @param password
	 *            secret used for authentication code.
	 * @return authentication code.
	 */
	public byte[] calculate(final byte[] salt, final byte[] password) {
		final byte[] hash;
		synchronized (mutex) {
			/*
			 * order is significant: first passwd then salt. see
			 * http://benlog.com/articles/2008/06/19/dont-hash-secrets/
			 */
			digest.reset();
			digest.update(password);
			hash = digest.digest(salt);
		}
		return mac.hmac(hash, password);
	}

}
