/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.crypto;

import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

/**
 * Convenience class for producing HMACs.
 * <p>
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class HMacImpl implements HMac {

	private final Mac mac;

	private final String macAlgorithm;

	private final Object mutex;

	/**
	 * Creates HMAC-SHA512 instance.
	 */
	public HMacImpl() {
		this("HmacSHA512");
	}

	/**
	 * Creates HMAC instance.
	 * 
	 * @param macAlgorithm
	 *            HMAC algorithm. e.g. "HmacSHA512".
	 */
	public HMacImpl(String macAlgorithm) {
		this.macAlgorithm = macAlgorithm;
		this.mutex = new Object();
		try {
			this.mac = Mac.getInstance(macAlgorithm);
		} catch (NoSuchAlgorithmException e) {
			throw new IllegalArgumentException(e);
		}
	}

	@Override
	public byte[] hmac(final byte[] key, final byte[] message) {
		final Key keySpec = new SecretKeySpec(key, macAlgorithm);
		try {
			synchronized (mutex) {
				mac.init(keySpec);
				mac.update(message);
				return mac.doFinal();
			}
		} catch (InvalidKeyException e) {
			throw new IllegalStateException(e);
		}
	}

}
