/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom;

import org.jhears.filter.bloom.util.BloomFilterBuilder;

/**
 * This class forms a bloom filter by combining a {@link HVBloomFilter} and
 * {@link IHasher}. Please, see {@link BloomFilterBuilder} for easy bloom filter
 * creation.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 * @param <T>
 *            Type of elements this filter holds.
 * @see BloomFilterBuilder
 */
public class BloomFilter<T> implements IBloomFilter<T> {

	/** Calculates hash values. */
	private final IHasher hasher;

	private final HVBloomFilter hvBf;

	private final int k;

	/**
	 * 
	 * @param hvBf
	 *            Hash value bloom filter providing the core logic.
	 * @param hasher
	 *            Hash utility used to create k hashes from objects when adding
	 *            and searching.
	 */
	public BloomFilter(HVBloomFilter hvBf, IHasher hasher) {
		this.hvBf = hvBf;
		this.hasher = hasher;
		this.k = hvBf.getK();
	}

	@Override
	public void add(T t) {
		hvBf.add(hasher.hash(t, k));
	}

	@Override
	public void clear() {
		hvBf.clear();
	}

	@Override
	public boolean contains(T t) {
		return hvBf.contains(hasher.hash(t, k));
	}

	@Override
	public long filterSize() {
		return hvBf.filterSize();
	}

	@Override
	public boolean isEmpty() {
		return hvBf.isEmpty();
	}

	@Override
	public boolean remove(T t) throws UnsupportedOperationException {
		return hvBf.remove(hasher.hash(t, k));
	}

	@Override
	public long size() {
		return hvBf.size();
	}

	@Override
	public boolean supportsRemove() {
		return hvBf.supportsRemove();
	}

}
