/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom;

import org.jhears.bitarray.IBitArray;

/**
 * Class holds an algorithm behind bloom filters. Modifies a separate state
 * object when objects are added/removed. This class by it self is immutable.
 * <p>
 * Filter consist from k slices of m bits. Filter expects k hash values.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class HVBloomFilter implements IHashValueBloomFilter {

	/** Bit array backing this filter. */
	private final IBitArray arr;

	/** The number of slices. */
	protected final int k;

	/** The the number of bits per bit array slice. */
	protected final int m;

	private final IBloomFilterState state;

	private final boolean supportsRemove;

	/**
	 * Create a new instance.
	 * 
	 * @param state
	 *            Stores bloom filter state.
	 * @throws IllegalArgumentException
	 *             If state m or k is less than one. If bitArray is null.
	 */
	public HVBloomFilter(IBloomFilterState state)
			throws IllegalArgumentException {
		if (state.getM() < 1) {
			throw new IllegalArgumentException(
					"Slice bit count must be greater than zero. m="
							+ state.getM());
		}
		if (state.getK() < 1) {
			throw new IllegalArgumentException(
					"The number of slices must be greater than zero. k="
							+ state.getK());
		}
		if (state.getBitArray() == null) {
			throw new IllegalArgumentException("Bit array is null");
		}
		this.state = state;
		this.k = state.getK();
		this.m = state.getM();
		this.arr = state.getBitArray();
		this.supportsRemove = state.supportsRemove();
	}

	/**
	 * Add element to this filter. Element is represented by array of hashes
	 * calculated from it.
	 * 
	 * @param hashValues
	 *            Hash values representing an element which is added to this
	 *            filter. Array must contain at least k elements.
	 */
	public void add(final int[] hashValues) {
		setAllBits(bitIndexes(hashValues));
		state.incSize();
	}

	/**
	 * Calculate bit array indexes from give hash values.
	 * 
	 * @param hashValues
	 *            Array of at least k hashes.
	 * @return Bit array indexes.
	 */
	protected int[] bitIndexes(final int[] hashValues) {
		final int[] indexes = new int[k];
		for (int i = 0; i < k; i++) {
			final int idx = (hashValues[i] % m) + (i * m);
			indexes[i] = idx;
		}
		return indexes;
	}

	@Override
	public void clear() {
		state.clear();
	}

	/**
	 * Clears all bits indicated by the given indexes.
	 * 
	 * @param bitIndexes
	 *            Bit array indexes which must be valid.
	 */
	protected void clearAllBits(final int[] bitIndexes) {
		arr.clearAll(bitIndexes);
	}

	/**
	 * Check if the element represented by array of hashes exists in this
	 * filter.
	 * 
	 * @param hashValues
	 *            Hash values representing an element which is searched.
	 * @return True of this filter might contain the element. Array must contain
	 *         at least k elements.
	 */
	public boolean contains(final int[] hashValues) {
		return isAllBitsSet(bitIndexes(hashValues));
	}

	@Override
	public long filterSize() {
		return state.filterSize();
	}

	public int getK() {
		return k;
	}

	/**
	 * Returns true if all of the given indexes are set.
	 * 
	 * @param bitIndexes
	 *            Bit array indexes which must be valid.
	 * @return Returns true if all of the given indexes are set.
	 */
	protected boolean isAllBitsSet(final int[] bitIndexes) {
		for (final boolean isSet : arr.getAll(bitIndexes)) {
			if (!isSet) {
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean isEmpty() {
		return state.isEmpty();
	}

	/**
	 * Remove element to this filter. Element is represented by array of hashes
	 * calculated from it.
	 * 
	 * @param hashValues
	 *            Hash values representing an element which is removed from this
	 *            filter. Array must contain at least k elements.
	 * @return true if this filter contained the element and element was
	 *         removed. false if this filter didn't contain the element.
	 * @throws UnsupportedOperationException
	 *             if this filter doesn't support removals.
	 */
	public boolean remove(final int[] hashValues)
			throws UnsupportedOperationException {
		if (!supportsRemove) {
			throw new UnsupportedOperationException();
		}
		int[] indexes = bitIndexes(hashValues);
		if (isAllBitsSet(indexes)) {
			clearAllBits(indexes);
			state.decSize();
			return true;
		}
		return false;
	}

	/**
	 * Sets all bits indicated by the given indexes.
	 * 
	 * @param bitIndexes
	 *            Bit array indexes which must be valid.
	 */
	protected void setAllBits(final int[] bitIndexes) {
		arr.setAll(bitIndexes);
	}

	@Override
	public long size() {
		return state.size();
	}

	public boolean supportsRemove() {
		return supportsRemove;
	}

	@Override
	public String toString() {
		return HVBloomFilter.class.getSimpleName() + " [state=" + state + "]";
	}

}
