/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.scalable;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Arrays;

import org.jhears.filter.bloom.HVBloomFilter;
import org.jhears.filter.bloom.IBloomFilterState;
import org.jhears.filter.bloom.IHashValueBloomFilter;

public abstract class AbstractSBFState implements IScalableBloomFilterState,
		Serializable {

	private static final long serialVersionUID = -120014745256485806L;

	private transient HVBloomFilter[] filters;

	private long filterSize;

	private int[] kArr;

	private final int m0;

	private int[] nArr;

	private final double p0;

	private final double r;

	private final int s;

	private long size;

	public AbstractSBFState(int s, double r, int m0, double p0) {
		this.filterSize = 0L;
		this.m0 = m0;
		this.p0 = p0;
		this.r = r;
		this.s = s;
		size = 0L;
		filters = new HVBloomFilter[0];
		nArr = new int[0];
		kArr = new int[0];
	}

	@Override
	public void appendNewFilter(int mi, int ki, int ni) {
		checkFilters();
		int i = filterCount();
		int newLength = i + 1;
		filters = Arrays.copyOf(filters, newLength);
		nArr = Arrays.copyOf(nArr, newLength);
		kArr = Arrays.copyOf(kArr, newLength);

		nArr[i] = ni;
		kArr[i] = ki;
		filters[i] = new HVBloomFilter(appendNewState(mi, ki));
		filterSize += filters[i].filterSize();
	}

	abstract protected IBloomFilterState appendNewState(int mi, int ki);

	private void checkFilters() {
		if (filters == null) {
			IBloomFilterState[] states = getBFStates();
			filters = new HVBloomFilter[states.length];
			for (int i = 0; i < states.length; i++) {
				filters[i] = new HVBloomFilter(states[i]);
			}
		}
	}

	@Override
	public void clear() {
		checkFilters();
		size = 0L;
		clearBFStates();
		filters = Arrays.copyOf(filters, 1);
		nArr = Arrays.copyOf(nArr, 1);
		kArr = Arrays.copyOf(kArr, 1);
		filterSize = filters[0].filterSize();
	}

	protected abstract void clearBFStates();

	@Override
	public void decSize() {
		size--;
	}

	@Override
	public void destroyLastFilter() {
		checkFilters();
		int newLength = filterCount() - 1;
		filterSize -= filters[newLength].filterSize();
		filters = Arrays.copyOf(filters, newLength);
		nArr = Arrays.copyOf(nArr, newLength);
		kArr = Arrays.copyOf(kArr, newLength);
		removeLastState();
	}

	@Override
	public long filterSize() {
		return filterSize;
	}

	protected abstract IBloomFilterState[] getBFStates();

	@Override
	public IHashValueBloomFilter getFilter(int i) {
		checkFilters();
		return filters[i];
	}

	@Override
	public int getK(int i) {
		return kArr[i];
	}

	@Override
	public int getM0() {
		return m0;
	}

	@Override
	public int getN(int i) {
		return nArr[i];
	}

	@Override
	public double getP0() {
		return p0;
	}

	@Override
	public double getR() {
		return r;
	}

	@Override
	public int getS() {
		return s;
	}

	@Override
	public void incSize() {
		size++;
	}

	@Override
	public boolean isEmpty() {
		return size == 0L;
	}

	private void readObject(ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		in.defaultReadObject();
		filters = null;
	}

	abstract protected void removeLastState();

	@Override
	public long size() {
		return size;
	}

	@Override
	public boolean supportsRemove() {
		checkFilters();
		return filters[0].supportsRemove();
	}

	@Override
	public String toString() {
		return AbstractSBFState.class.getSimpleName() + " [filterSize="
				+ filterSize + ", kArr=" + Arrays.toString(kArr) + ", m0=" + m0
				+ ", nArr=" + Arrays.toString(nArr) + ", p0=" + p0 + ", r=" + r
				+ ", s=" + s + ", size=" + size + ", states="
				+ Arrays.toString(getBFStates()) + "]";
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

}
