/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.scalable;

import org.jhears.filter.bloom.IHashValueBloomFilter;
import org.jhears.filter.bloom.util.BloomMath;

/**
 * Class holds an algorithm behind Scalable Bloom Filters. <blockquote>A SBF is
 * made up of a series of one or more (plain) Bloom Filters; when ﬁlters get
 * full due to the limit on the ﬁll ratio, a new one is added; querying is made
 * by testing for the presence in each ﬁlter.<br>
 * Each successive bloom ﬁlter is created with a tighter maximum error
 * probability on a geometric progression, so that the compounded probability
 * over the whole series converges to some wanted value, even accounting for an
 * inﬁnite series.</blockquote> <i>Almeida, Paulo; Baquero, Carlos; Preguiça,
 * Nuno; Hutchison, David (2007),
 * "<a href="http://gsd.di.uminho.pt/members/cbm/ps/dbloom.pdf">Scalable Bloom
 * Filters</a>", Information Processing Letters 101 (6): 255–261,
 * doi:10.1016/j.ipl.2006.10.007 </i>
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class HVScalableBloomFilter implements IHashValueBloomFilter {

	/** The number of slices in initial filter. */
	protected final int k0;

	/** The number of bits per slice in initial filter. */
	protected final int m0;

	/** False positive probability of initial filter. */
	protected final double p0;

	/** Tightening ratio for error probability. */
	protected final double r;

	/** Slice growth factor. Only sensible values are 2 and 4. */
	protected final int s;

	private final IScalableBloomFilterState state;

	/**
	 * Create a new instance.
	 * 
	 * @param state
	 *            Stores the scalable bloom filter state.
	 * @throws IllegalArgumentException
	 *             if state is invalid.
	 */
	public HVScalableBloomFilter(IScalableBloomFilterState state)
			throws IllegalArgumentException {
		this.r = state.getR();
		if (r >= 1.0 || r <= 0.0) {
			throw new IllegalArgumentException(
					"Argument r must be between zero and one. r=" + r);
		}
		this.s = state.getS();
		if (s < 1) {
			throw new IllegalArgumentException(
					"Argument s must be greater than zero. s=" + s);
		}
		this.p0 = state.getP0();
		if (p0 >= 1.0 || p0 <= 0.0) {
			throw new IllegalArgumentException(
					"Argument P0 must be between zero and one. P0=" + p0);
		}
		this.m0 = state.getM0();
		if (m0 < 1) {
			throw new IllegalArgumentException(
					"Argument m0 must be greater than zero. m0=" + m0);
		}
		this.state = state;
		// k0 = log2(P0^-1)
		this.k0 = (int) Math.round(Math.ceil(BloomMath.log2(1.0 / p0)));
		if (state.filterCount() == 0) {
			appendNewFilter();
		}
	}

	/**
	 * Adds an element to the first filter which maximum size is not yet
	 * reached. Method {@link #prepareAdd()} should be called before this method
	 * is invoked.
	 * <p>
	 * {@inheritDoc}
	 * 
	 * @throws IllegalStateException
	 *             If no filters where available.
	 */
	@Override
	public void add(int[] createHashValues) throws IllegalStateException {
		int fcnt = state.filterCount();
		for (int j = 0; j < fcnt; j++) {
			if (state.getFilter(j).size() < state.getN(j)) {
				state.getFilter(j).add(createHashValues);
				state.incSize();
				return;
			}
		}
		throw new IllegalStateException(
				"All filters full. Didn't call prepareAdd() before adding?");
	}

	/**
	 * Append a new filter.
	 */
	private void appendNewFilter() {
		int i = state.filterCount();

		// ki = i * log2(r^-1)
		int ki = k0
				+ (int) Math.round(Math.ceil((i * BloomMath.log2(1.0 / r))));
		// mi = m0 * s^i
		int mi = m0 * (int) Math.round(Math.pow(s, i));
		int ni = (int) (m0 * Math.pow(s, i) * BloomMath.LN2);
		state.appendNewFilter(mi, ki, ni);
	}

	/**
	 * Returns true if only one of the filters
	 * {@link IHashValueBloomFilter#contains(int[])} return true.
	 * 
	 * @param hashValues
	 * @return
	 */
	public boolean canRemove(int[] hashValues) {
		int cnt = 0;
		int fcnt = state.filterCount();
		for (int i = 0; i < fcnt; i++) {
			if (state.getFilter(i).contains(hashValues)) {
				cnt++;
			}
		}
		return cnt == 1;
	}

	@Override
	public void clear() {
		state.clear();
	}

	/**
	 * Removes unused filters, if any.
	 */
	public void compact() {
		while (destroyLast())
			;
	}

	/**
	 * Check if any of the filter array members contain the give value.
	 * 
	 * @param hashValues
	 *            Value to search for.
	 * @return True one of the filters contained the item.
	 */
	@Override
	public boolean contains(int[] hashValues) {
		int fcnt = state.filterCount();
		for (int j = 0; j < fcnt; j++) {
			if (state.getFilter(j).contains(hashValues)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Remove the last filter if it is empty.
	 * 
	 * @return true if the filter was removed.
	 */
	private boolean destroyLast() {
		int fcnt = state.filterCount();
		if (fcnt > 1) {
			if (state.getFilter(fcnt - 1).isEmpty()) {
				state.destroyLastFilter();
				return true;
			}
		}
		return false;
	}

	@Override
	public long filterSize() {
		return state.filterSize();
	}

	/**
	 * Return the number of slices in last filter.
	 * 
	 * @return
	 */
	public int getLastK() {
		return state.getK(state.filterCount() - 1);
	}

	@Override
	public boolean isEmpty() {
		return state.isEmpty();
	}

	/**
	 * Adds a new filter if the last filter is full.
	 */
	public void prepareAdd() {
		int lastIdx = state.filterCount() - 1;
		if (state.getFilter(lastIdx).size() >= state.getN(lastIdx)) {
			appendNewFilter();
		}
	}

	/**
	 * Scalable bloom filters support element removal if filters which the SBF
	 * is assembled ({@link IHashValueBloomFilter}) from supports removal. More
	 * over, the removal is possible if only one of the filters contain the
	 * element in question.
	 * <p>
	 * {@inheritDoc}
	 * 
	 * @see #canRemove(int[])
	 */
	@Override
	public boolean remove(int[] hashValues)
			throws UnsupportedOperationException {
		int cnt = 0;
		int idx = -1;
		int fcnt = state.filterCount();
		for (int i = 0; i < fcnt; i++) {
			if (state.getFilter(i).contains(hashValues)) {
				cnt++;
				idx = i;
			}
		}
		if (cnt == 1 && state.getFilter(idx).remove(hashValues)) {
			state.decSize();
			return true;
		}
		return false;
	}

	@Override
	public long size() {
		return state.size();
	}

	@Override
	public boolean supportsRemove() {
		return state.supportsRemove();
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + " [k0=" + k0 + ", state=" + state
				+ "]";
	}

}
