/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.scalable;

import org.jhears.filter.bloom.IHashValueBloomFilter;
import org.jhears.filter.bloom.util.BloomMath;

/**
 * Data object for storing the scalable bloom filter state.
 * 
 * @author Juha Heljoranta
 * 
 */
public interface IScalableBloomFilterState {

	/**
	 * Append a new filter.
	 * 
	 * @param mi
	 *            The number of bits in slice.
	 * @param ki
	 *            The number of slices in filter.
	 * @param ni
	 *            The maximum size of the filter.
	 */
	void appendNewFilter(int mi, int ki, int ni);

	/**
	 * Clear filter.
	 */
	void clear();

	/**
	 * Decrement filter size by one.
	 */
	void decSize();

	/**
	 * Remove the last filter.
	 */
	void destroyLastFilter();

	/**
	 * The number of filters.
	 * 
	 * @return
	 */
	int filterCount();

	/** Total size of this bloom filter in bits. */
	long filterSize();

	/**
	 * Get the ith filter.
	 * 
	 * @param i
	 * @return
	 */
	public IHashValueBloomFilter getFilter(int i);

	/**
	 * Get the slice count of the ith filter.
	 * 
	 * @param i
	 * @return
	 */
	int getK(int i);

	/**
	 * Initial filter slice size in bits. Preferably power of two.
	 * 
	 * @return
	 */
	int getM0();

	/**
	 * Get the maximum size of the ith filter.
	 * 
	 * @param i
	 * @return
	 */
	int getN(int i);

	/**
	 * Error probability for the first bloom filter. e.g. 0.25 value give 25 %
	 * of false positive rate in initial filter. Usually around 0.1-0.0001.
	 * Smaller values increase memory usage.
	 * <p>
	 * To use p0 as SBF compound error probability multiply p0 with
	 * <code>1/(1-r)</code> or use
	 * {@link BloomMath#optimalSBFP0(double, double)}.
	 * 
	 * @return
	 */
	double getP0();

	/**
	 * Error probability tightening ratio. Sane values are in range [0.5,1.0).
	 * Optimal value value is around 0.8 to 0.9. If set is expected to grow
	 * little values around 0.8 are good. If set is expected to grow more values
	 * around 0.9 are suitable.
	 * 
	 * @return
	 */
	double getR();

	/**
	 * Slice size growth factor. Typically 2 or 4. Preferably power of two. Use
	 * 2 if set is expected to grow little. Use 4 if set is expected to grow
	 * more. If more extreme growth is expected then values like 8 or 16 are
	 * suitable.
	 * 
	 * @return
	 */
	int getS();

	/**
	 * Increment filter size by one.
	 */
	void incSize();

	/**
	 * True if the filter is empty.
	 * 
	 * @return
	 */
	boolean isEmpty();

	/**
	 * The total number of elements in this filter.
	 * 
	 * @return
	 */
	long size();

	boolean supportsRemove();

}