/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.scalable;

import org.jhears.filter.bloom.IHasher;
import org.jhears.filter.bloom.util.BloomFilterBuilder;

/**
 * This class forms a scalable bloom filter by combining a
 * {@link HVScalableBloomFilter} and {@link IHasher}. Please, see
 * {@link BloomFilterBuilder} for easy scalable bloom filter creation.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 * @param <T>
 *            Type of elements this filter holds.
 * @see BloomFilterBuilder
 */
public class ScalableBloomFilter<T> implements IScalableBloomFilter<T> {

	/** Creates hash values. */
	protected final IHasher hasher;

	private final HVScalableBloomFilter hvSbf;

	/**
	 * 
	 * @param hvSbf
	 *            Hash value scalable bloom filter providing the core logic.
	 * @param hasher
	 *            Hash utility used to create k hashes from objects when adding
	 *            and searching.
	 */
	public ScalableBloomFilter(HVScalableBloomFilter hvSbf, IHasher hasher) {
		this.hvSbf = hvSbf;
		this.hasher = hasher;
	}

	@Override
	public void add(T t) {
		hvSbf.prepareAdd();
		hvSbf.add(hasher.hash(t, hvSbf.getLastK()));
	}

	@Override
	public void clear() {
		hvSbf.clear();
	}

	@Override
	public void compact() {
		hvSbf.compact();
	}

	@Override
	public boolean contains(T t) {
		return hvSbf.contains(hasher.hash(t, hvSbf.getLastK()));
	}

	@Override
	public long filterSize() {
		return hvSbf.filterSize();
	}

	@Override
	public boolean isEmpty() {
		return hvSbf.isEmpty();
	}

	@Override
	public boolean remove(T t) throws UnsupportedOperationException {
		return hvSbf.remove(hasher.hash(t, hvSbf.getLastK()));
	}

	@Override
	public long size() {
		return hvSbf.size();
	}

	@Override
	public boolean supportsRemove() {
		return hvSbf.supportsRemove();
	}

	@Override
	public String toString() {
		return ScalableBloomFilter.class.getSimpleName() + " [alg=" + hvSbf
				+ ", hasher=" + hasher + "]";
	}

}
