/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.util;

import org.jhears.filter.bloom.BitArrayBFState;
import org.jhears.filter.bloom.BloomFilter;
import org.jhears.filter.bloom.HVBloomFilter;
import org.jhears.filter.bloom.Hasher;
import org.jhears.filter.bloom.IBloomFilter;
import org.jhears.filter.bloom.IBloomFilterState;
import org.jhears.filter.bloom.IHasher;
import org.jhears.filter.bloom.scalable.BitArraySBFState;
import org.jhears.filter.bloom.scalable.IScalableBloomFilterState;
import org.jhears.filter.bloom.scalable.ScalableBloomFilter;
import org.jhears.filter.bloom.scalable.HVScalableBloomFilter;
import org.jhears.util.DefaultHashFunction;
import org.jhears.util.LCG;

/**
 * Convenience utility for creating fixed size and scalable bloom filters.
 * <p>
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class BloomFilterBuilder {

	/** Hash creation utility. */
	private final IHasher hasher;

	public BloomFilterBuilder() {
		this(new Hasher(new DefaultHashFunction<Object>(), new LCG()));
	}

	/**
	 * Create a new instance.
	 * 
	 * @param hasher
	 *            custom hash creator.
	 * @param factory
	 *            custom factory.
	 */
	public BloomFilterBuilder(IHasher hasher) {
		this.hasher = hasher;
	}

	protected <T> IBloomFilter<T> createBloomFilter(IBloomFilterState state) {
		return new BloomFilter<T>(new HVBloomFilter(state), hasher);
	}

	/**
	 * Returns bloom filter instance.
	 * 
	 * @param <T>
	 *            Type of elements handled by the filter.
	 * @param m
	 *            Optimal slice size.
	 * @param k
	 *            Optimal slice count.
	 * @return
	 */
	protected <T> IBloomFilter<T> createBloomFilter(int m, int k) {
		return createBloomFilter(new BitArrayBFState(m, k));
	}

	/**
	 * Returns scalable bloom filter instance.
	 * 
	 * @param <T>
	 *            Type of elements handled by the filter.
	 * @param growth
	 *            Estimated filter growth magnitude.
	 * @param m0
	 *            optimal slice size for initial filter.
	 * @param p0
	 *            Error probability for initial filter.
	 * @return
	 */
	protected <T> IBloomFilter<T> createScalableBloomFilter(Growth growth,
			int m0, double p0) {
		IScalableBloomFilterState state = new BitArraySBFState(growth.getS(),
				growth.getR(), m0, p0);
		return createScalableBloomFilter(state);
	}

	protected <T> IBloomFilter<T> createScalableBloomFilter(
			IScalableBloomFilterState state) {
		return new ScalableBloomFilter<T>(new HVScalableBloomFilter(state),
				hasher);
	}

	/**
	 * Creates a fixed size bloom filter.
	 * 
	 * @param <T>
	 *            Type of elements handled by the filter.
	 * @param initialCapacity
	 *            the initial capacity of the bloom filter.
	 * @param p
	 *            Error probability P. e.g. use 0.01 for 1% error probability.
	 * @return {@link BloomFilter} optimized to given arguments.
	 */
	public <T> IBloomFilter<T> fixedSizeBloomFilter(int initialCapacity,
			double p) {
		int m = BloomMath.optimalSliceSize(initialCapacity);
		int k = BloomMath.optimalSliceCount(p);
		return createBloomFilter(m, k);
	}

	/**
	 * Creates a scalable bloom filter.
	 * 
	 * @param <T>
	 *            Type of elements handled by the filter.
	 * @param initialCapacity
	 *            the initial capacity of the bloom filter.
	 * @param p
	 *            Error probability P. e.g. use 0.01 for 1% error probability.
	 * @param growth
	 *            Estimated filter growth magnitude.
	 * @return {@link ScalableBloomFilter} optimized to given arguments.
	 */
	public <T> IBloomFilter<T> scalableBloomFilter(int initialCapacity,
			double p, Growth growth) {
		int m0 = BloomMath.optimalSliceSize(initialCapacity);
		double p0 = BloomMath.optimalSBFP0(p, growth.getR());
		return createScalableBloomFilter(growth, m0, p0);
	}

}
