/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.util;

import org.jhears.util.MathUtil;

/**
 * Mathematical utilities and constants needed in bloom filters.
 * 
 * @author Juha Heljoranta
 * 
 */
public final class BloomMath {

	/** Value of ln(2). */
	public static final double LN2 = Math.log(2.0);

	/** Value of log(2)^2. */
	public static final double LN2POW2 = Math.pow(LN2, 2.0);

	/**
	 * Returns base 2 logarithm of a value d.
	 * 
	 * @param d
	 *            Value.
	 * @return Base 2 logarithm of value.
	 */
	public static double log2(final double d) {
		return Math.log(d) / LN2;
	}

	/**
	 * Calculate initial filter false positive probability P0 by using upper
	 * bound Scalable Bloom Filter error probability P.
	 * 
	 * @param p
	 *            Upper bound error probability for SBF.
	 * @param r
	 *            Tightening ration.
	 * @throws IllegalArgumentException
	 *             If either of the arguments is not in range (0,1).
	 * @return Optimal error probability P0 calculated from p and r.
	 */
	public static double optimalSBFP0(final double p, final double r)
			throws IllegalArgumentException {
		if (p <= 0.0 || p >= 1.0) {
			throw new IllegalArgumentException("Error probability p=" + p
					+ " must be in range (0,1).");
		}
		if (r <= 0.0 || r >= 1.0) {
			throw new IllegalArgumentException("Thightening ratio r=" + r
					+ " must be in range (0,1).");
		}
		// derived from P <= P0 * (1 / (1 -r))
		return (1.0 - r) * p;
	}

	/**
	 * Calculates the optimal number of slices for bloom filter.
	 * 
	 * @param p0
	 *            Error probability. Typically 0.01 - 0.0001. For example, value
	 *            0.1 means there is 10 % change of having a false positive when
	 *            checking if filter contains an element. Lower values mean
	 *            higher memory consumption.
	 * @throws IllegalArgumentException
	 *             If <code>initialSize</code> is less than one or if
	 *             <code>p0</code> is not in range (0,1).
	 * @return The optimal number of slices for the filter.
	 */
	public static int optimalSliceCount(final double p0)
			throws IllegalArgumentException {
		if (p0 >= 1.0 || p0 <= 0.0) {
			throw new IllegalArgumentException(
					"Error probability must be between zero an one. p0=" + p0);
		}
		return (int) Math.round(Math.ceil(BloomMath.log2(1.0 / p0)));
	}

	/**
	 * Calculates optimal slice size m.
	 * 
	 * @param filterSize
	 *            The number of elements bloom filter is expected to hold.
	 * @throws IllegalArgumentException
	 *             If filterSize is negative.
	 * @return The optimal number of bits per filter slice. Smallest returned
	 *         value is 2.
	 */
	public static int optimalSliceSize(int filterSize)
			throws IllegalArgumentException {
		if (filterSize < 0) {
			throw new IllegalArgumentException(
					"Initial size must be positive. initialSize=" + filterSize);
		}
		if (filterSize == 0) {
			filterSize = 1;
		}
		// derived from ni ~= m0 * s^i * ln(2)
		return (int) Math.round(Math.ceil(filterSize / BloomMath.LN2));
	}

	/**
	 * Returns absolute values of array elements. Method guarantees that all
	 * returned values are <code>>= 0</code>.
	 * <p>
	 * In case a value is <code>Integer.MIN_VALUE</code> the absolute is
	 * calculated from <code>(Integer.MIN_VALUE + 1)</code>.
	 * 
	 * @param values
	 * @return
	 */
	public static int[] positiveAbs(final int[] values) {
		final int len = values.length;
		final int[] retval = new int[len];
		for (int i = 0; i < len; i++) {
			retval[i] = MathUtil.positiveAbs(values[i]);
		}
		return retval;
	}

	private BloomMath() {
	}

}
