/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.bloom.util;

/**
 * Predefined scalable bloom filter growth magnitudes.
 * <p>
 * Correctly chosen growth magnitude keeps CPU and memory usage at minimum.
 * <p>
 * Each magnitude has a suitable slice growth and error probability tightening
 * factor known to work well after the filter has grown by that particular
 * magnitude. These values are tentative approximations and subject to change.
 * <p>
 * <i>Defined growth magnitudes and their associated values are subject to
 * change.</i>
 * 
 * @author Juha Heljoranta
 */
public enum Growth {

	/** Filter is expected to grow by 1 order of magnitude or less. */
	MAG_1(2, 0.85),
	/** Filter is expected to grow by 2 orders of magnitude. */
	MAG_2(2, 0.9),
	/** Filter is expected to grow by 3 orders of magnitude. */
	MAG_3(4, 0.9),
	/** Filter is expected to grow by 4 orders of magnitude. */
	MAG_4(4, 0.9),
	/** Filter is expected to grow by 5 orders of magnitude. */
	MAG_5(8, 0.9),
	/** Filter is expected to grow by 6 orders of magnitude or more. */
	MAG_6(16, 0.9);

	private final double r;

	private final int s;

	private Growth(int s, double r) {
		this.s = s;
		this.r = r;
	}

	/**
	 * Error probability tightening rate.
	 * 
	 * @return Error probability tightening rate.
	 */
	public double getR() {
		return r;
	}

	/**
	 * Filter growth rate.
	 * 
	 * @return Filter growth rate.
	 */
	public int getS() {
		return s;
	}

}
