/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.filter.newbloom.scalable;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Arrays;

import org.jhears.filter.bloom.IBloomFilterState;
import org.jhears.filter.bloom.scalable.AbstractSBFState;
import org.jhears.filter.newbloom.FcbaBFState;
import org.jhears.util.RuntimeIOException;

public class FcbaSBFState extends AbstractSBFState {

	private static final long serialVersionUID = 114945966628181229L;

	public static FcbaSBFState load(File dir) throws IOException {
		ObjectInputStream is = null;
		try {
			is = new ObjectInputStream(new BufferedInputStream(
					new FileInputStream(new File(dir, "state.dat"))));
			return (FcbaSBFState) is.readUnshared();
		} catch (ClassNotFoundException e) {
			throw new RuntimeException(e);
		} finally {
			if (is != null) {
				is.close();
			}
		}
	}

	private final File dir;

	private File[] files;

	private FcbaBFState[] states;

	public FcbaSBFState(int s, double r, int m0, double p0, File dir) {
		super(s, r, m0, p0);
		this.dir = dir;
		states = new FcbaBFState[0];
		files = new File[0];
		save();
	}

	@Override
	protected IBloomFilterState appendNewState(int mi, int ki) {
		int i = states.length;
		states = Arrays.copyOf(states, i + 1);
		files = Arrays.copyOf(files, i + 1);
		files[i] = new File(dir, FcbaBFState.class.getSimpleName() + "-" + i
				+ ".dat");
		try {
			files[i].createNewFile();
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		}
		states[i] = new FcbaBFState(mi, ki, files[i]);
		save();
		return states[i];
	}

	@Override
	protected void clearBFStates() {
		for (int i = 1; i < files.length; i++) {
			if (!files[i].delete()) {
				throw new RuntimeIOException(new IOException("Delete failed: "
						+ files[i].getAbsolutePath()));
			}
		}
		states = Arrays.copyOf(states, 1);
		states[0].clear();
		save();
	}

	@Override
	public int filterCount() {
		return states.length;
	}

	@Override
	protected IBloomFilterState[] getBFStates() {
		return states;
	}

	@Override
	protected void removeLastState() {
		int i = states.length - 1;
		if (!files[i].delete()) {
			throw new RuntimeIOException(new IOException("Delete failed: "
					+ files[i].getAbsolutePath()));
		}
		states = Arrays.copyOf(states, i);
		files = Arrays.copyOf(files, i);
		save();
	}

	public void save() {
		ObjectOutputStream os = null;
		try {
			os = new ObjectOutputStream(new BufferedOutputStream(
					new FileOutputStream(new File(dir, "state.dat"))));
			os.writeUnshared(this);
		} catch (IOException e) {
			throw new RuntimeIOException(e);
		} finally {
			if (os != null) {
				try {
					os.close();
				} catch (IOException ignore) {
				}
			}
		}
	}

	@Override
	public String toString() {
		return FcbaSBFState.class.getSimpleName() + " [states="
				+ Arrays.toString(states) + ", super=" + super.toString() + "]";
	}

}
