/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;

/**
 * Adaptation of {@link List} for {@code int} primitives.
 */
public class IntList implements Serializable, Cloneable {

	public static final int DEFAULT_GROWTH_PERCENTAGE = 67;

	private static final long serialVersionUID = -1515937802922294597L;

	private static boolean equals(int[] a, int[] a2, int len, int len2) {
		if (a == a2) {
			return true;
		}
		if (a == null || a2 == null) {
			return false;
		}
		if (len != len2) {
			return false;
		}
		for (int i = 0; i < len; i++) {
			if (a[i] != a2[i]) {
				return false;
			}
		}
		return true;
	}

	private int[] arr;

	private final int growthPercentage;

	private int size;

	public IntList() {
		this(16);
	}

	public IntList(int initialCapacity) {
		this(new int[initialCapacity], 0, false);
	}

	public IntList(int initialCapacity, int growthPercentage) {
		this(new int[initialCapacity], 0, false, growthPercentage);
	}

	public IntList(int[] arr) {
		this(arr, arr.length, true);
	}

	public IntList(int[] arr, int growthPercentage) {
		this(arr, arr.length, true, growthPercentage);
	}

	private IntList(int[] arr, int size, boolean copy) {
		this(arr, size, copy, DEFAULT_GROWTH_PERCENTAGE);
	}

	private IntList(int[] arr, int size, boolean copy, int growthPercentage) {
		if (growthPercentage < 0) {
			throw new IllegalArgumentException(
					"Growth percentage must be positive");
		}
		this.arr = copy ? Arrays.copyOf(arr, size) : arr;
		this.size = size;
		this.growthPercentage = growthPercentage;
	}

	public IntList(int[] arr, int size, int growthPercentage) {
		this(arr, size, true, growthPercentage);
	}

	public boolean add(int value) {
		checkSize();
		arr[size++] = value;
		return true;
	}

	public void add(int index, int value) {
		checkSize();
		if (index <= size) {
			System.arraycopy(arr, index, arr, index + 1, size - index);
			arr[index] = value;
			size++;
		} else {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	public boolean addAll(final int index, final int[] values) {
		return addAll(index, new IntList(values, values.length, false));
	}

	public boolean addAll(int index, IntList c) {
		int numNew = c.size;
		if (size + numNew > size) {
			arr = Arrays.copyOf(arr, size + numNew);
		}
		int numMoved = size - index;
		if (numMoved > 0) {
			System.arraycopy(arr, index, arr, index + numNew, numMoved);
		}
		System.arraycopy(c.arr, 0, arr, index, numNew);
		size += numNew;
		return numNew != 0;
	}

	public boolean addAll(int[] values) {
		if (values.length > 0) {
			if (size + values.length > size) {
				arr = Arrays.copyOf(arr, size + values.length);
			}
			System.arraycopy(values, 0, arr, size, values.length);
			size = size + values.length;
			return true;
		}
		return false;
	}

	public boolean addAll(IntList arrayIntList) {
		if (!arrayIntList.isEmpty()) {
			if (size + arrayIntList.size > size) {
				arr = Arrays.copyOf(arr, size + arrayIntList.size);
			}
			System.arraycopy(arrayIntList.arr, 0, arr, size, arrayIntList.size);
			size = size + arrayIntList.size;
			return true;
		}
		return false;
	}

	public int binarySearch(int key) {
		return Arrays.binarySearch(arr, 0, size, key);
	}

	private void checkSize() {
		if (arr.length == size) {
			int newSize = size + (size / 100 * growthPercentage);
			if (newSize == size) {
				newSize += 16;
			}
			arr = Arrays.copyOf(arr, newSize);
		}
	}

	public void clear() {
		size = 0;
	}

	@Override
	public Object clone() throws CloneNotSupportedException {
		IntList il = (IntList) super.clone();
		il.arr = arr.clone();
		return il;
	}

	public boolean contains(int value) {
		return indexOf(value) != -1;
	}

	public boolean containsAll(int[] values) {
		for (int value : values) {
			if (indexOf(value) == -1) {
				return false;
			}
		}
		return true;
	}

	public boolean containsAll(IntList c) {
		for (int i = 0; i < c.size; i++) {
			if (indexOf(c.arr[i]) == -1) {
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		IntList other = (IntList) obj;
		return equals(arr, other.arr, size, other.size);
	}

	public int get(int i) throws IndexOutOfBoundsException {
		if (i < size) {
			return arr[i];
		} else {
			throw new ArrayIndexOutOfBoundsException(i);
		}
	}

	@Override
	public int hashCode() {
		if (arr == null) {
			return 0;
		}
		int result = 1;
		for (int i = 0; i < size; i++) {
			result = 31 * result + arr[i];
		}
		return result;
	}

	public int indexOf(int value) {
		for (int i = 0; i < size; i++) {
			if (arr[i] == value) {
				return i;
			}
		}
		return -1;
	}

	public boolean isEmpty() {
		return size == 0;
	}

	public int lastIndexOf(int value) {
		for (int i = size - 1; i >= 0; i--) {
			if (arr[i] == value) {
				return i;
			}
		}
		return -1;
	}

	public int remove(int index) {
		if (index < size) {
			int retval = arr[index];
			System.arraycopy(arr, index + 1, arr, index, (size--) - index - 1);
			return retval;
		} else {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	private boolean remove(IntList c, boolean complement) {
		int ptr = 0;
		boolean modified = false;
		for (int i = 0; i < size; i++) {
			if (complement == c.contains(arr[i])) {
				arr[ptr++] = arr[i];
			}
		}
		if (ptr != size) {
			size = ptr;
			modified = true;
		}
		return modified;
	}

	public boolean removeAll(int[] values) {
		return removeAll(new IntList(values, values.length, false));
	}

	// public Iterator<Object> iterator() {
	// // TODO Auto-generated method stub
	// return null;
	// }

	public boolean removeAll(IntList c) {
		return remove(c, false);
	}

	// @Override
	// public ListIterator<Object> listIterator() {
	// // TODO Auto-generated method stub
	// return null;
	// }

	// @Override
	// public ListIterator<Object> listIterator(int index) {
	// // TODO Auto-generated method stub
	// return null;
	// }

	public boolean removeValue(int value) {
		int i = indexOf(value);
		if (i != -1) {
			remove(i);
			return true;
		}
		return false;
	}

	public boolean retainAll(int[] values) {
		return retainAll(new IntList(values, values.length, false));
	}

	public boolean retainAll(IntList c) {
		return remove(c, true);
	}

	public void reverse() {
		int i = 0;
		int j = size - 1;
		int tmp;
		while (j > i) {
			tmp = arr[j];
			arr[j] = arr[i];
			arr[i] = tmp;
			j--;
			i++;
		}
	}

	public int set(int index, int value) {
		if (index < size) {
			int retval = arr[index];
			arr[index] = value;
			return retval;
		} else {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	public int size() {
		return size;
	}

	public void sort() {
		Arrays.sort(arr, 0, size);
	}

	public IntList subList(int fromIndex, int toIndex) {
		int[] copy = Arrays.copyOfRange(arr, fromIndex, toIndex);
		return new IntList(copy, copy.length, false);
	}

	public int[] toArray() {
		return Arrays.copyOf(arr, size);
	}

	public int[] toArray(int[] a) {
		if (a == null || a.length < size) {
			a = new int[size];
		}
		System.arraycopy(arr, 0, a, 0, size);
		return a;
	}

	@Override
	public String toString() {
		return IntList.class.getSimpleName() + " [size=" + size + ", arr="
				+ Arrays.toString(arr) + "]";
	}

	public void trimToSize() {
		if (arr.length != size) {
			arr = Arrays.copyOf(arr, size);
		}
	}

}
