/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.io.Serializable;
import java.util.Random;

/**
 * Linear Congruential Generator. Pseudorandom number generator.
 * <p>
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class LCG implements Serializable {

	private static final long serialVersionUID = 1263224657175456559L;

	private final long increment;

	private final long modulus;

	private final long multiplier;

	/**
	 * Create instance compatible with {@link Random}.
	 * <p>
	 * In following example the values <code>a</code> and <code>b</code> are
	 * guaranteed to be identical.
	 * 
	 * <pre>
	 * int a = new LCG().nextInt(seed)
	 * int b = new Random(seed).nextInt()
	 * </pre>
	 * 
	 */
	public LCG() {
		this((1L << 48) - 1, 0x5DEECE66DL, 0xBL);
	}

	/**
	 * Create a new instance with custom parameters.
	 * <p>
	 * Please note that LCGs are extremely sensitive to the choice of the
	 * coefficients.
	 * <p>
	 * For details see {@link Random} source code and <a
	 * href="http://en.wikipedia.org/wiki/Linear_congruential_generator">Linear
	 * congruential generator</a> [wikipedia.org].
	 * 
	 * @param modulus
	 * @param multiplier
	 * @param increment
	 */
	public LCG(final long modulus, final long multiplier, final long increment) {
		this.modulus = modulus;
		this.multiplier = multiplier;
		this.increment = increment;
	}

	/**
	 * Convert seed value to a new seed value.
	 * 
	 * @param seed
	 *            Initial seed.
	 * @return Seed derived from initial seed.
	 */
	public long newSeedFrom(final long seed) {
		return (seed ^ multiplier) & modulus;
	}

	/**
	 * Returns a pseudo random integer by using given seed value.
	 * 
	 * @param seed
	 *            The seed value.
	 * @return Random value derived from seed.
	 */
	public int nextInt(final long seed) {
		return seedToInt(nextSeed(newSeedFrom(seed)));
	}

	/**
	 * Convenience method to generate several pseudo random integer values.
	 * 
	 * @param seed
	 *            The seed value.
	 * @param count
	 *            The number of numbers to return.
	 * @return
	 * @throws IllegalArgumentException
	 *             If count is negative.
	 */
	public int[] nextInts(final long seed, final int count)
			throws IllegalArgumentException {
		if (count < 0) {
			throw new IllegalArgumentException("Count cannot be negative");
		}
		long newSeed = newSeedFrom(seed);
		final int[] retval = new int[count];
		for (int i = 0; i < count; i++) {
			newSeed = nextSeed(newSeed);
			retval[i] = seedToInt(newSeed);
		}
		return retval;
	}

	/**
	 * Returns a pseudo random long by using given seed value.
	 * 
	 * @param seed
	 *            The seed value.
	 * @return Random value derived from seed.
	 */
	public long nextLong(long seed) {
		seed = nextSeed(newSeedFrom(seed));
		int a = seedToInt(seed);
		seed = nextSeed(seed);
		int b = seedToInt(seed);
		return ((long) a << 32) + b;
	}

	/**
	 * Create a seed from value.
	 * 
	 * @param seed
	 *            value used to create a new seed.
	 * @return new seed value
	 */
	public long nextSeed(final long seed) {
		return (seed * multiplier + increment) & modulus;
	}

	/**
	 * Extract an integer from seed.
	 * 
	 * @param seed
	 *            Value to extract from.
	 * @return Extracted value.
	 */
	public int seedToInt(final long seed) {
		return (int) (seed >>> 16);
	}

}
