/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import org.jhears.util.MathUtil;
import org.junit.Assert;
import org.junit.Test;

public abstract class IBitArrayTest {

	protected abstract IBitArray getBitArray(int capacity);

	@Test
	public void testClearOne() {
		int[] capacities = getCapacities();
		for (int capacity : capacities) {
			IBitArray arr = getBitArray(capacity);
			int[] indexes = MathUtil.seq(capacity);
			boolean[] expected = new boolean[capacity];
			Arrays.fill(expected, true);
			// set all bits
			arr.setAll(indexes);
			for (int i = 0; i < capacity; i++) {
				// toggle bit and verify
				Assert.assertTrue(arr.get(i));
				arr.clear(i);
				Assert.assertFalse(arr.get(i));
				arr.set(i);
				boolean[] actual = arr.getAll(indexes);
				Assert.assertTrue(Arrays.equals(expected, actual));
			}
		}
	}

	@Test
	public void testRandomBits() {
		int size = Long.SIZE * 16;
		Random r = new Random(3);
		Set<Integer> is = new HashSet<Integer>();
		while (is.size() < size / 2) {
			is.add(Integer.valueOf(r.nextInt(size)));
		}
		IBitArray arr = getBitArray(size);
		for (Integer integer : is) {
			int i = integer.intValue();
			arr.set(i);
		}
		for (int j = 0; j < size; j++) {
			Integer integer = Integer.valueOf(j);
			if (is.contains(integer)) {
				Assert.assertTrue(arr.get(j));
			} else {
				Assert.assertFalse(arr.get(j));
			}
		}
	}

	@Test
	public void testSetByOne() {
		int capacity = Long.SIZE * 2;

		IBitArray arr = getBitArray(capacity);

		for (int i = 0; i < capacity; i++) {
			arr.set(i);

			int[] indexesTrue = MathUtil.seq(i + 1);
			boolean[] expectedTrue = new boolean[i + 1];
			Arrays.fill(expectedTrue, true);
			int[] indexesFalse = MathUtil.seq(capacity - i - 1, i + 1);
			boolean[] expectedFalse = new boolean[capacity - i - 1];
			Arrays.fill(expectedFalse, false);

			boolean[] actualTrue = arr.getAll(indexesTrue);
			Assert.assertTrue(Arrays.equals(expectedTrue, actualTrue));

			boolean[] actualFalse = arr.getAll(indexesFalse);
			Assert.assertTrue(Arrays.equals(expectedFalse, actualFalse));
		}

	}

	@Test
	public void testSetGetClearAll() {
		int size = Long.SIZE * 16;
		Random r = new Random(0);
		Set<Integer> is = new HashSet<Integer>();
		while (is.size() < size / 2) {
			is.add(Integer.valueOf(r.nextInt(size)));
		}
		IBitArray arr = getBitArray(size);

		// setAll
		int[] indexes = new int[is.size()];
		ArrayList<Integer> ilist = new ArrayList<Integer>(is);
		for (int i = 0; i < indexes.length; i++) {
			indexes[i] = ilist.get(i).intValue();
		}
		arr.setAll(indexes);

		// fetch full array contents and verify
		int[] allIndexes = new int[size];
		for (int i = 0; i < size; i++) {
			allIndexes[i] = i;
		}
		boolean[] ss = arr.getAll(allIndexes);
		for (int i = 0; i < size; i++) {
			if (is.contains(Integer.valueOf(i))) {
				Assert.assertTrue(ss[i]);
			} else {
				Assert.assertFalse(ss[i]);
			}
		}

		// fetch only set indexes and verify
		ss = arr.getAll(indexes);
		boolean[] expecteds = new boolean[is.size()];
		Arrays.fill(expecteds, true);
		Assert.assertTrue(Arrays.equals(expecteds, ss));

		// clear arry
		arr.clearAll(indexes);
		ss = arr.getAll(allIndexes);
		expecteds = new boolean[size];
		Arrays.fill(expecteds, false);
		Assert.assertTrue(Arrays.equals(expecteds, ss));
	}

	protected abstract int[] getCapacities();

	@Test
	public void testSetOne() {
		int[] capacities = getCapacities();

		for (int capacity : capacities) {
			int[] indexes = MathUtil.seq(capacity);
			boolean[] expected = new boolean[capacity];
			Arrays.fill(expected, false);
			IBitArray arr = getBitArray(capacity);
			for (int i = 0; i < capacity; i++) {
				Assert.assertFalse(arr.get(i));
				arr.set(i);
				Assert.assertTrue(arr.get(i));
				arr.clear(i);
				boolean[] actual = arr.getAll(indexes);
				Assert.assertTrue(Arrays.equals(expected, actual));
				// for (int j = 0; j < capacity; j++) {
				// Assert.assertFalse(arr.get(i));
				// }
			}
		}
	}

}
