/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.bitarray;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.util.HashSet;
import java.util.Random;

import junit.framework.Assert;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.runners.MockitoJUnitRunner;

@RunWith(value = MockitoJUnitRunner.class)
public class LazyCachedBitArrayTest {

	@Test
	public void testDefaultLoadAll() {
		IBitArray slowBitArray = mock(IBitArray.class);
		IBitArray fastBitArray = mock(IBitArray.class);
		int size = 3;
		when(slowBitArray.size()).thenReturn(size);
		when(fastBitArray.size()).thenReturn(size);
		when(slowBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		when(fastBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		CachedBitArray cachedBitArray = new CachedBitArray(slowBitArray,
				fastBitArray);
		@SuppressWarnings("unchecked")
		LazyCachedBitArray lazy = new LazyCachedBitArray(cachedBitArray,
				mock(LazyListener.class));
		Assert.assertFalse(lazy.isPopulated());
		// will use default load size and fetch all
		lazy.get(0);
		Assert.assertTrue(lazy.isPopulated());
		verify(slowBitArray).getAll(any(int[].class));
		verify(fastBitArray).clearAll(any(int[].class));
		verify(fastBitArray).setAll(eq(new int[0]));
	}

	@SuppressWarnings("unchecked")
	@Test
	public void testSetValue() {
		IBitArray slowBitArray = mock(IBitArray.class);
		IBitArray fastBitArray = mock(IBitArray.class);
		int size = 3;
		when(slowBitArray.size()).thenReturn(size);
		when(fastBitArray.size()).thenReturn(size);
		when(slowBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		when(fastBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		CachedBitArray cachedBitArray = new CachedBitArray(slowBitArray,
				fastBitArray);
		LazyListener setter = mock(LazyListener.class);
		LazyCachedBitArray lazy = new LazyCachedBitArray(cachedBitArray, setter);
		Assert.assertFalse(lazy.isPopulated());
		// will use default load size and fetch all
		lazy.get(0);
		verify(setter).setDelegate(eq(cachedBitArray));
	}

	@Test
	public void testRandom() {
		int x = 100;
		int size = LazyCachedBitArray.MIN_LOAD_SIZE * x;
		IBitArray slowBitArray = new BitArray(size);
		IBitArray fastBitArray = new BitArray(size);

		HashSet<Integer> iset = new HashSet<Integer>(size / 2);
		Random r = new Random(0);
		while (iset.size() < size / 2) {
			iset.add(Integer.valueOf(r.nextInt(size)));
		}

		int[] setIndex = new int[size / 2];
		int l = 0;
		for (Integer i : iset) {
			setIndex[l++] = i.intValue();
		}
		slowBitArray.setAll(setIndex);

		CachedBitArray cachedBitArray = new CachedBitArray(slowBitArray,
				fastBitArray);
		LazyListener<IBitArray> setter = new LazyListener<IBitArray>() {
			@Override
			public void setDelegate(IBitArray t) {
				// nop
			}
		};
		LazyCachedBitArray lazy = new LazyCachedBitArray(cachedBitArray, setter);
		for (int i = 0; i < x * 2; i++) {
			iset.clear();
			while (iset.size() < x) {
				iset.add(Integer.valueOf(r.nextInt(size)));
			}
			int[] arr = new int[x];
			int j = 0;
			for (Integer k : iset) {
				arr[j++] = k.intValue();
			}
			lazy.getAll(arr);
			if (lazy.isPopulated()) {
				break;
			}
		}
		Assert.assertTrue(BitArrayUtil.equals(slowBitArray, fastBitArray));
		Assert.assertTrue(BitArrayUtil.equals(slowBitArray, lazy));
	}

	@Test
	public void testDefaultLoadInTwoParts() {
		IBitArray slowBitArray = mock(IBitArray.class);
		IBitArray fastBitArray = mock(IBitArray.class);
		int size = LazyCachedBitArray.MIN_LOAD_SIZE + 1;
		when(slowBitArray.size()).thenReturn(size);
		when(fastBitArray.size()).thenReturn(size);
		when(slowBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		when(fastBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		CachedBitArray cachedBitArray = new CachedBitArray(slowBitArray,
				fastBitArray);
		@SuppressWarnings("unchecked")
		LazyCachedBitArray lazy = new LazyCachedBitArray(cachedBitArray,
				mock(LazyListener.class));
		Assert.assertFalse(lazy.isPopulated());
		lazy.get(0);
		Assert.assertFalse(lazy.isPopulated());
		lazy.get(0);
		Assert.assertTrue(lazy.isPopulated());
		verify(slowBitArray, times(2)).getAll(any(int[].class));
		verify(fastBitArray, times(2)).clearAll(any(int[].class));
		verify(fastBitArray, times(2)).setAll(eq(new int[0]));
	}

	@Test
	public void testGetIndexes() {
		IBitArray slowBitArray = mock(IBitArray.class);
		IBitArray fastBitArray = mock(IBitArray.class);
		int size = 3;
		when(slowBitArray.size()).thenReturn(size);
		when(fastBitArray.size()).thenReturn(size);
		when(slowBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		when(fastBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		CachedBitArray cachedBitArray = new CachedBitArray(slowBitArray,
				fastBitArray);
		@SuppressWarnings("unchecked")
		LazyCachedBitArray lazy = new LazyCachedBitArray(cachedBitArray,
				mock(LazyListener.class));
		Assert.assertFalse(lazy.isPopulated());
		lazy.getAll(new int[] { 0 });
		Assert.assertTrue(lazy.isPopulated());
		lazy.getAll(new int[] { 1 });
		Assert.assertTrue(lazy.isPopulated());
		lazy.getAll(new int[] { 2 });
		Assert.assertTrue(lazy.isPopulated());
		verify(slowBitArray, times(3)).getAll(any(int[].class));
		verify(fastBitArray, times(3)).clearAll(any(int[].class));
		verify(fastBitArray, times(3)).setAll(eq(new int[0]));
	}

	@Test
	public void testGetIndexesNone() {
		IBitArray slowBitArray = mock(IBitArray.class);
		IBitArray fastBitArray = mock(IBitArray.class);
		int size = 3;
		when(slowBitArray.size()).thenReturn(size);
		when(fastBitArray.size()).thenReturn(size);
		when(slowBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		when(fastBitArray.getAll(any(int[].class))).thenReturn(
				new boolean[size]);
		CachedBitArray cachedBitArray = new CachedBitArray(slowBitArray,
				fastBitArray);
		@SuppressWarnings("unchecked")
		LazyCachedBitArray lazy = new LazyCachedBitArray(cachedBitArray,
				mock(LazyListener.class));
		Assert.assertFalse(lazy.isPopulated());
		lazy.getAll(new int[0]);
		Assert.assertTrue(lazy.isPopulated());
		verify(slowBitArray).getAll(any(int[].class));
		verify(fastBitArray).clearAll(any(int[].class));
		verify(fastBitArray).setAll(eq(new int[0]));
	}

}
