/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.cluster;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.jhears.util.IntList;

/**
 * Detects clusters of integer values in sorted arrays.
 * <p>
 * This class is thread safe.
 */
public class ClusterDetector {

	private final int maxDist;

	private final int minSize;

	/**
	 * 
	 * @param maxDist
	 *            Maximum distance between two tested values. That is, the
	 *            difference between two adjacent values must not exceed this.
	 *            If difference is less than or equal to this value then the
	 *            values potentially belong to same cluster.
	 * @param minSize
	 *            The minimum number of values which must satisfy
	 *            <code>maxDist</code> criteria before they are considered to
	 *            form a cluster.
	 * @throws IllegalArgumentException
	 *             If any of the arguments is negative.
	 */
	public ClusterDetector(int maxDist, int minSize) {
		if (maxDist < 0) {
			throw new IllegalArgumentException(
					"Maxumum distance must be positive.");
		}
		if (minSize < 0) {
			throw new IllegalArgumentException("Minimum size must be positive.");
		}
		this.maxDist = maxDist;
		this.minSize = minSize;
	}

	/**
	 * 
	 * @param positions
	 *            Values sorted with {@link Arrays#sort(int[])}. Method
	 *            calculates difference between consequent array elements and if
	 *            the value is less than or equal to {@link #getMaxDist()} the
	 *            values might form a cluster. If there are at east
	 *            {@link #getMinSize()} consequent elements satisfying
	 *            aforementioned criteria then the elements form a cluster.
	 * @return Cluster objects encapsulating the array indexes if the
	 *         <code>positions</code> argument.
	 */
	public List<Cluster> detectClusters(int[] positions) {
		List<Cluster> clusters = new ArrayList<Cluster>();
		if (positions.length == 0) {
			return clusters;
		}
		int[] distances = getDistances(positions);
		IntList clusterIndexes = new IntList();
		int i = 0;
		for (; i < distances.length; i++) {
			if (distances[i] <= maxDist) {
				clusterIndexes.add(i);
			} else {
				newCluster(clusters, clusterIndexes, i);
			}
		}
		newCluster(clusters, clusterIndexes, i);
		return clusters;
	}

	/**
	 * 
	 * @param positions
	 *            Array sorted in ascending order.
	 * @return
	 */
	private int[] getDistances(final int[] positions) {
		final int[] distances = new int[positions.length - 1];
		for (int i = 0; i < distances.length; i++) {
			distances[i] = positions[i + 1] - positions[i];
		}
		return distances;
	}

	public int getMaxDist() {
		return maxDist;
	}

	public int getMinSize() {
		return minSize;
	}

	private void newCluster(List<Cluster> clusters,
			IntList clusterIndexes, int currentIndex) {
		// the next index is not part of this cluster
		// currentIndex belongs to this cluster
		clusterIndexes.add(currentIndex);
		if (clusterIndexes.size() >= minSize) {
			clusters.add(new Cluster(clusterIndexes.toArray()));
		}
		clusterIndexes.clear();
	}

}
