/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.cnf;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.jar.Attributes;
import java.util.jar.Attributes.Name;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.CompositeConfiguration;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.apache.commons.configuration.SystemConfiguration;
import org.jhears.jar.ManifestUtil;
import org.slf4j.LoggerFactory;

/**
 * Provides configuration save/load functionality.
 * <p>
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class CnfManager {

	private static final String _FILE_SUFFIX = ".properties";

	private static final String _PNAME_SUFFIX = ".conf";

	private static final String _ETC = "etc";

	// private static final String FS = System.getProperty("file.separator");

	public static final String audioConfFname = Cnf.AudioPreProsKey._PREFIX
			+ _FILE_SUFFIX;

	public static final String audioConfPropName = Cnf.PREFIX + Cnf.SEPARATOR
			+ Cnf.AudioPreProsKey._PREFIX + _PNAME_SUFFIX;

	/**
	 * default calibrator configuration file name
	 */
	public static final String calibConfFname = Cnf.CalibratorKey._PREFIX
			+ _FILE_SUFFIX;

	/**
	 * name of a property which defines a calibrator configuration file
	 */
	public static final String calibConfPropName = Cnf.PREFIX + Cnf.SEPARATOR
			+ Cnf.CalibratorKey._PREFIX + _PNAME_SUFFIX;

	/**
	 * default client configuration file name
	 */
	public static final String clientConfFname = Cnf.ClientKey._PREFIX
			+ _FILE_SUFFIX;

	/**
	 * name of a property which defines a client configuration file
	 */
	public static final String clientConfPropName = Cnf.PREFIX + Cnf.SEPARATOR
			+ Cnf.ClientKey._PREFIX + _PNAME_SUFFIX;

	/** Property names which define configuration file locations. */
	private static final Set<String> cnfFileProperties = new HashSet<String>();

	/**
	 * Default directory name for jhears data files
	 */
	public static final String DEFAULT_HOME_DIR_NAME = "jhears-home";

	/**
	 * default fingerprint configuration file name
	 */
	public static final String fingerprintConfFname = Cnf.FingerprintKey._PREFIX
			+ _FILE_SUFFIX;

	/**
	 * name of a property which defines a fingerprint configuration file
	 */
	public static final String fingerprintConfPropName = Cnf.PREFIX
			+ Cnf.SEPARATOR + Cnf.FingerprintKey._PREFIX + _PNAME_SUFFIX;

	/**
	 * default framework configuration file name
	 */
	public static final String fwConfFname = Cnf.Framewrk._PREFIX
			+ _FILE_SUFFIX;

	/**
	 * name of a property which defines a client configuration file
	 */
	public static final String fwConfPropName = Cnf.PREFIX + Cnf.SEPARATOR
			+ Cnf.Framewrk._PREFIX + _PNAME_SUFFIX;

	/**
	 * nName of a property which defines a jhears home directory
	 */
	public static final String PROP_NAME_HOME = Cnf.PREFIX + ".home";

	public static final String searchConfFname = Cnf.SearchKey._PREFIX
			+ _FILE_SUFFIX;

	public static final String searchConfPropName = Cnf.PREFIX + Cnf.SEPARATOR
			+ Cnf.SearchKey._PREFIX + _PNAME_SUFFIX;

	/**
	 * default server configuration file name
	 */
	public static final String serverConfFname = Cnf.ServerKey._PREFIX
			+ _FILE_SUFFIX;

	/**
	 * name of a property which defines a client configuration file
	 */
	public static final String serverConfPropName = Cnf.PREFIX + Cnf.SEPARATOR
			+ Cnf.ServerKey._PREFIX + _PNAME_SUFFIX;

	static {
		cnfFileProperties.add(audioConfPropName);
		cnfFileProperties.add(calibConfPropName);
		cnfFileProperties.add(clientConfPropName);
		cnfFileProperties.add(fingerprintConfPropName);
		cnfFileProperties.add(fwConfPropName);
		cnfFileProperties.add(searchConfPropName);
		cnfFileProperties.add(serverConfPropName);
	}

	/**
	 * The home directory. Lookup order is: system property
	 * {@link #PROP_NAME_HOME}, environment property {@link #PROP_NAME_HOME}.
	 * The default is system property "user.dir"+{@link #DEFAULT_HOME_DIR_NAME}.
	 * 
	 * @return
	 * @throws IOException
	 *             if home directory doesn't exists or cannot be created.
	 */
	private static File homeDir() throws IOException {
		File homeDir;
		if (System.getProperty(PROP_NAME_HOME) != null) {
			homeDir = new File(System.getProperty(PROP_NAME_HOME));
		} else if (System.getenv(PROP_NAME_HOME) != null) {
			homeDir = new File(System.getenv(PROP_NAME_HOME));
		} else if (System.getProperty("user.dir") != null) {
			File cwd = new File(System.getProperty("user.dir"));
			homeDir = new File(cwd, DEFAULT_HOME_DIR_NAME);
		} else {
			throw new IOException("Cannot locate jHears home directory");
		}
		if (!homeDir.isDirectory()) {
			if (!homeDir.mkdirs()) {
				throw new IOException("Cannot create directory: "
						+ homeDir.getAbsolutePath());
			}
		}
		return homeDir;
	}

	private Configuration defaults;

	private final File home;

	private PropertiesConfiguration propConf;

	private final Object mutex = new Object();

	private static final Name manifestKey = new Attributes.Name(
			"Group-Id-Artifact-Id");

	public CnfManager() throws IOException {
		this.home = homeDir();
	}

	public File getHomeDir() {
		return home;
	}

	@SuppressWarnings("unchecked")
	private void addMissingKeys(PropertiesConfiguration conf,
			Configuration defaults) throws ConfigurationException {
		Set<String> confKeys = toHashSet(conf.getKeys());
		Set<String> defaultKeys = toHashSet(defaults.getKeys());
		defaultKeys.removeAll(confKeys);
		if (!defaultKeys.isEmpty()) {
			LoggerFactory.getLogger(getClass()).info(
					"Adding missing keys (" + defaultKeys
							+ ") to configuration file "
							+ conf.getFile().getAbsolutePath());
			for (String key : defaultKeys) {
				conf.setProperty(key, defaults.getProperty(key));
			}
			conf.save();
		}
	}

	private File getCnfFile(Cnf cnf) {
		File conf;
		if (System.getProperty(cnf.getPropertyName()) != null) {
			conf = new File(System.getProperty(cnf.getPropertyName()));
		} else {
			conf = new File(new File(home, _ETC), cnf.getFileName());
		}
		return conf;
	}

	public File getConfigDir() {
		return new File(home, _ETC);
	}

	/**
	 * Convenience method to fetch all configurations.
	 * 
	 * @return
	 * @throws ConfigurationException
	 */
	public Configuration getAllConfigurations() throws ConfigurationException {
		return getConfiguration(Cnf.values());
	}

	public Configuration getConfiguration(Cnf... configs)
			throws ConfigurationException {
		synchronized (mutex) {
			CompositeConfiguration compositeConfig = new CompositeConfiguration();
			compositeConfig.addConfiguration(getSystemConfiguration());

			for (Cnf c : configs) {

				defaults = null;
				propConf = null;

				loadConfig(c);

				if (defaults != null && propConf != null) {
					addMissingKeys(propConf, defaults);
					compositeConfig.addConfiguration(propConf);
				}
			}

			return compositeConfig;
		}
	}

	@SuppressWarnings("unchecked")
	public Configuration getCopy(Cnf confClass, Configuration configuration) {
		Configuration newCnf = new BaseConfiguration();
		for (String key : toHashSet(confClass.getDefaultConfig().getKeys())) {
			newCnf.setProperty(key, configuration.getProperty(key));
		}
		return newCnf;
	}

	@SuppressWarnings("unchecked")
	private Configuration getSystemConfiguration() {
		BaseConfiguration retval = new BaseConfiguration();
		SystemConfiguration system = new SystemConfiguration();
		Iterator<String> i = system.getKeys(Cnf.PREFIX);
		while (i.hasNext()) {
			String key = i.next();
			if (!cnfFileProperties.contains(key)) {
				// exclude configuration file info
				retval.setProperty(key, system.getProperty(key));
			}
		}
		return retval;
	}

	public boolean hasConfiguration(Cnf cnf) {
		return getCnfFile(cnf).isFile();
	}

	private void loadConfig(Cnf c) throws ConfigurationException {
		defaults = c.getDefaultConfig();
		propConf = new PropertiesConfiguration(getCnfFile(c));
	}

	public void saveConfiguration(Cnf cnf, Configuration configuration)
			throws ConfigurationException {
		synchronized (mutex) {
			File file = getCnfFile(cnf);
			PropertiesConfiguration pc = new PropertiesConfiguration();
			pc.append(configuration);
			pc.save(file);
		}
	}

	public void saveProperty(Cnf config, String key, Object value)
			throws ConfigurationException {
		synchronized (mutex) {
			PropertiesConfiguration pc = new PropertiesConfiguration();
			pc.append(getConfiguration(config));
			pc.setProperty(key, value);
			pc.save(getCnfFile(config));
		}
	}

	private Set<String> toHashSet(Iterator<String> keyIterator) {
		HashSet<String> set = new HashSet<String>();
		while (keyIterator.hasNext()) {
			set.add(keyIterator.next());
		}
		return set;
	}

	public static String getVersion(String artifactId) {
		return ManifestUtil.getImplementationVersion(manifestKey, "org.jhears:"
				+ artifactId);
	}

}
