/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.cnf;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;

import com.google.common.base.Throwables;

public final class CnfUtil {

	private CnfUtil() {
	}

	@SuppressWarnings("unchecked")
	public static final void copyCnfTo(Configuration from, Configuration to) {
		Iterator<String> ki = from.getKeys();
		while (ki.hasNext()) {
			String key = ki.next();
			to.setProperty(key, from.getProperty(key));
		}
	}

	public static final double[] toDoubleArray(String[] strArray) {
		double[] a = new double[strArray.length];
		for (int i = 0; i < strArray.length; i++) {
			a[i] = Double.parseDouble(strArray[i]);
		}
		return a;
	}

	public static final int[] toIntArray(String[] strArray) {
		int[] a = new int[strArray.length];
		for (int i = 0; i < strArray.length; i++) {
			a[i] = Integer.parseInt(strArray[i]);
		}
		return a;
	}

	public static List<String> getAllPublicProperties(Class<?> clazz) {
		ArrayList<String> list = new ArrayList<String>();
		for (Field field : clazz.getFields()) {
			if (String.class.equals(field.getType())
					&& !field.getName().startsWith("_")) {
				if (!hasAnnotation(field, InternalProperty.class)) {
					list.add(getField(field));
				}
			}
		}
		return list;
	}

	private static boolean hasAnnotation(Field field, Class<?> annotationType) {
		for (Annotation a : field.getAnnotations()) {
			if (annotationType.equals(a.annotationType())) {
				return true;
			}
		}
		return false;
	}

	private static String getField(Field field) {
		try {
			return (String) field.get(null);
		} catch (Exception e) {
			throw Throwables.propagate(e);
		}
	}

	public static List<String> getPrivateProperties(Class<?> clazz) {
		ArrayList<String> list = new ArrayList<String>();
		for (Field field : clazz.getFields()) {
			for (Annotation a : field.getAnnotations()) {
				if (InternalProperty.class.equals(a.annotationType())
						&& String.class.equals(field.getType())) {
					list.add(getField(field));
				}
			}
		}
		return list;
	}

	public static final void validateConfiguration(Cnf cnf,
			Configuration configuration) throws ConfigurationException {
		List<String> keys = CnfUtil.getAllPublicProperties(cnf.getKeyClass());
		// check for missing keys
		List<String> missingKeys = new ArrayList<String>();
		for (String key : keys) {
			if (configuration.getProperty(key) == null) {
				missingKeys.add(key);
			}
		}
		if (!missingKeys.isEmpty()) {
			throw new ConfigurationException(
					"Mandatory configuration entrie(s) missing: " + missingKeys);
		}
	}

	public static final String[] toStringArray(final double[] arr) {
		final String[] retval = new String[arr.length];
		for (int i = 0; i < arr.length; i++) {
			retval[i] = Double.toString(arr[i]);
		}
		return retval;
	}

	public static Map<String, Object> toMap(Configuration c) {
		Map<String, Object> cMap = new HashMap<String, Object>();
		@SuppressWarnings("unchecked")
		Iterator<String> ki = c.getKeys();
		while (ki.hasNext()) {
			String key = ki.next();
			Object value = c.getProperty(key);
			cMap.put(key, value);
		}
		return cMap;
	}

	// @SuppressWarnings("unchecked")
	// public static String checkSum(Configuration configuration) {
	// Iterator<String> ki = configuration.getKeys();
	// ArrayList<String> list = new ArrayList<String>();
	// while (ki.hasNext()) {
	// list.add(ki.next());
	// }
	// Collections.sort(list);
	// StringWriter sw = new StringWriter();
	// for (String key : list) {
	// sw.append(key).append(configuration.getProperty(key).toString());
	// }
	// return DigestUtils.shaHex(sw.toString());
	// }
}
