/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.io;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.DoubleBuffer;
import java.util.NoSuchElementException;

import org.apache.commons.configuration.Configuration;
import org.jhears.cnf.Cnf.FingerprintKey;

import com.google.common.base.Throwables;

public class BarkReader implements Iterable<double[]> {

	private final int barkCnt;

	private final int dataEntryInBytes;

	private final File file;

	public BarkReader(Configuration configuration, final File file) {
		this.file = file;
		barkCnt = configuration.getInt(FingerprintKey.BARK_BAND_STOP_COUNT);
		int SIZEOF_DOUBLE = Double.SIZE / Byte.SIZE;
		dataEntryInBytes = barkCnt * SIZEOF_DOUBLE;
	}

	public long getBarkCount() {
		return file.length() / dataEntryInBytes;
	}

	@Override
	public ClosableIterator<double[]> iterator() {
		return new ClosableIterator<double[]>() {

			private static final int LOADNEXT = -2;
			private final DoubleBuffer barkBuf;
			private final byte[] barkBytes;
			private final DataInputStream dis;
			private int eof;

			{
				try {
					dis = new DataInputStream(new BufferedInputStream(
							new FileInputStream(file)));
				} catch (FileNotFoundException e) {
					throw Throwables.propagate(e);
				}
				barkBytes = new byte[dataEntryInBytes];
				barkBuf = ByteBuffer.wrap(barkBytes).asDoubleBuffer();
				eof = LOADNEXT;
			}

			@Override
			public void close() {
				close(false);
			}

			private void close(boolean silent) {
				try {
					dis.close();
				} catch (IOException e) {
					if (!silent) {
						throw Throwables.propagate(e);
					}
				}
			}

			@Override
			public boolean hasNext() {
				if (eof == LOADNEXT) {
					try {
						dis.readFully(barkBytes);
					} catch (EOFException e) {
						eof = -1;
					} catch (IOException e) {
						close(true);
						throw Throwables.propagate(e);
					}
				}
				if (eof == -1) {
					close(false);
					return false;
				} else {
					return true;
				}
			}

			@Override
			public double[] next() {
				if (hasNext()) {
					eof = LOADNEXT;
					double[] bark = new double[barkCnt];
					barkBuf.get(bark);
					barkBuf.clear();
					return bark;
				} else {
					throw new NoSuchElementException();
				}
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
	}
}
