/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.io;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map.Entry;

/**
 * Accumulates write operations and flushes them when write buffer gets full.
 * Class cannot handle situations where a write would overwrite another pending
 * write in deterministic way.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class DelayedWriter {

	private static final class DWEntry implements Entry<Long, byte[]> {

		private final Long key;

		private byte[] value;

		public DWEntry(Long key, byte[] value) {
			this.key = key;
			this.value = value;
		}

		@Override
		public Long getKey() {
			return key;
		}

		@Override
		public byte[] getValue() {
			return value;
		}

		@Override
		public byte[] setValue(byte[] value) {
			byte[] retval = this.value;
			this.value = value;
			return retval;
		}

		@Override
		public String toString() {
			return DWEntry.class.getSimpleName() + " [key=" + key
					+ ", value.length="
					+ (value == null ? "null" : value.length) + "]";
		}

	}

	private final RandomAccessFile raf;

	private final List<DWEntry> pendingWrites;

	private long pendingBytes;

	private final int buffSize;

	private static final Comparator<DWEntry> comparator = new Comparator<DWEntry>() {

		@Override
		public int compare(DWEntry o1, DWEntry o2) {
			if (o1.getKey().longValue() < o2.getKey().longValue()) {
				return -1;
			} else if (o1.getKey().longValue() > o2.getKey().longValue()) {
				return 1;
			} else {
				return 0;
			}
		}
	};

	public DelayedWriter(RandomAccessFile raf, int buffSize) {
		this.raf = raf;
		this.buffSize = buffSize;
		pendingWrites = new ArrayList<DWEntry>();
		pendingBytes = 0;
	}

	/**
	 * Queue write operation. Write queue is flushed to disk in ascending write
	 * position order. Multiple writes with the same position are executed in
	 * arbitrary order. In general, no guarantees are given if a write would
	 * overwrite another pending write.
	 * 
	 * @param pos
	 * @param data
	 * @throws IOException
	 */
	public void write(long pos, byte[] data, int offset, int len)
			throws IOException {
		pendingWrites.add(new DWEntry(Long.valueOf(pos), Arrays.copyOfRange(
				data, offset, offset + len)));
		pendingBytes += len;
		if (pendingBytes >= buffSize) {
			flush();
		}
	}

	/**
	 * Flush pending operations.
	 * 
	 * @throws IOException
	 */
	public void flush() throws IOException {
		Collections.sort(pendingWrites, comparator);
		List<Entry<Long, byte[]>> writes = optimize(pendingWrites);
		for (Entry<Long, byte[]> entry : writes) {
			raf.seek(entry.getKey().longValue());
			raf.write(entry.getValue());
		}
		pendingWrites.clear();
		pendingBytes = 0;
	}

	/**
	 * Merge sequential write operations.
	 * 
	 * @param elist
	 *            list of write operations. Entry key is write position and
	 *            value is data to be written. List must be sorted in ascending
	 *            order by entry keys.
	 * @return
	 */
	static List<Entry<Long, byte[]>> optimize(
			final List<? extends Entry<Long, byte[]>> elist) {
		final int size = elist.size();
		List<Entry<Long, byte[]>> retval = new ArrayList<Entry<Long, byte[]>>(
				size);

		// list of byte arrays which are merged
		List<byte[]> merges = new ArrayList<byte[]>();

		for (int i = 0; i < size;) {
			final Entry<Long, byte[]> cur = elist.get(i);
			final long curPos = cur.getKey().longValue();
			if (cur.getValue() == null) {
				throw new IllegalStateException("i=" + i + ", curPos=" + curPos
						+ ", size=" + size + " elist=" + elist);
			}
			// total bytes for merging
			int mergeLen = cur.getValue().length;
			// next element considered for merging
			int j = i + 1;
			for (; j < size;) {
				final Entry<Long, byte[]> next = elist.get(j);
				if (next.getKey().longValue() == curPos + mergeLen) {
					mergeLen += next.getValue().length;
					merges.add(next.getValue());
					next.setValue(null);
					j++;
				} else {
					break;
				}
			}
			i = j;
			if (!merges.isEmpty()) {
				byte[] dest = new byte[mergeLen];
				int destPos = 0;
				byte[] cb = cur.getValue();
				System.arraycopy(cb, 0, dest, destPos, cb.length);
				destPos += cb.length;
				for (byte[] bs : merges) {
					System.arraycopy(bs, 0, dest, destPos, bs.length);
					destPos += bs.length;
				}
				cur.setValue(dest);
				merges.clear();
				mergeLen = -1;
			}
			if (cur.getValue() == null) {
				throw new IllegalStateException("i=" + i + ", curPos=" + curPos
						+ ", j=" + j + ", size=" + size + " elist=" + elist);
			}
			retval.add(cur);
		}
		return retval;
	}

	/**
	 * Flush and close.
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		try {
			flush();
		} finally {
			raf.close();
		}
	}

}
