/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.io;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Iterator;
import java.util.NoSuchElementException;

import com.google.common.base.Throwables;

/**
 * Allows to iterate over input stream contents using fixed step and chunk size.
 * Commonly used to chunk the input stream into overlapping byte arrays.
 */
public class StreamIterator implements Iterable<byte[]> {

	/**
	 * Contains bytes read from input stream.
	 */
	private final ByteBuffer bb;
	private final byte[] buf;
	private byte[] chunk;
	private final int chunkSize;
	private boolean eof;
	private InputStream is;

	/**
	 * Helps to divide the data into chunks properly.
	 */
	private final ByteBuffer stepb;
	private final int stepSize;

	public StreamIterator(InputStream is, int chunkSize, int stepSize,
			int bufSize) {
		if (stepSize < 1) {
			throw new IllegalArgumentException();
		}
		if (chunkSize < 1) {
			throw new IllegalArgumentException();
		}
		if (bufSize < chunkSize) {
			throw new IllegalArgumentException();
		}
		if (bufSize < stepSize) {
			throw new IllegalArgumentException();
		}
		if (is == null) {
			throw new NullPointerException();
		}
		this.is = is;
		this.chunkSize = chunkSize;
		this.stepSize = stepSize;
		this.eof = false;
		this.chunk = null;
		buf = new byte[bufSize];
		bb = ByteBuffer.wrap(buf);
		stepb = ByteBuffer.wrap(buf);
	}

	public void close() throws IOException {
		if (is != null) {
			is.close();
			is = null;
		}
	}

	private boolean hasNext() {
		if (chunk == null) {
			try {
				read2();
			} catch (IOException e) {
				throw Throwables.propagate(e);
			}
			if (chunk == null && eof) {
				return false;
			} else {
				return hasNext();
			}
		}
		return true;
	}

	@Override
	public Iterator<byte[]> iterator() {
		return new Iterator<byte[]>() {

			@Override
			public boolean hasNext() {
				return StreamIterator.this.hasNext();
			}

			@Override
			public byte[] next() {
				return StreamIterator.this.next();
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}

		};
	}

	private byte[] next() {
		if (hasNext()) {
			try {
				return chunk;
			} finally {
				chunk = null;
			}
		} else {
			throw new NoSuchElementException();
		}
	}

	private void read2() throws IOException {
		// read into bb
		if (!eof && bb.hasRemaining()) {
			int read = is.read(buf, bb.position(), bb.remaining());
			if (read == -1) {
				eof = true;
			} else {
				bb.position(bb.position() + read);
			}
		}
		stepb.limit(bb.position());
		if (stepb.remaining() >= chunkSize) {
			// extract chunk
			chunk = new byte[chunkSize];
			stepb.get(chunk);
			int p = stepb.position() - chunkSize + stepSize;
			if (p <= stepb.limit()) {
				stepb.position(p);
			}
		} else {
			// compact only when necessary
			stepb.compact();
			bb.position(stepb.position());
			stepb.clear();
			if (eof && chunk == null) {
				// fetch remaining
				stepb.limit(bb.position());
				if (stepb.remaining() >= chunkSize) {
					read2();
				}
			}
		}
	}

}
