/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.logging;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;

import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.Ostermiller.util.CircularByteBuffer;
import com.google.common.base.Preconditions;

/**
 * Passes stream contents to a logger. Uses {@link BufferedReader#readLine()} to
 * read what ever is written onto {@link OutputStream} accessed via
 * {@link #getOutputStream()}.
 * <p>
 * Note that the {@link OutputStream}s are not thread safe. Proper way to use
 * this class in multi threaded environment is to create a new
 * {@link OutputStreamLogger} instance for each thread which needs to log an
 * output stream.
 * <p>
 * Streams are buffered so calling {@link OutputStream#flush()} at least before
 * closing the stream is required in order to guarantee that all output is being
 * logged. For applications which don't want to buffer log entries should call
 * the flush method more ofthen.
 * <p>
 * Users of this class <b>must</b> either close the output stream or call the
 * {@link #close()}. Otherwise there is a possibility of resource leakage.
 * <p>
 * Invoking {@link #run()} blocks until streams are closed or an error occurs.
 * {@link #start(boolean)} basically calls
 * <code>new Thread(this).start();</code> with boolean flag controlling the
 * daemon status.
 */
// TODO this class needs some work
public class OutputStreamLogger implements Runnable {

	public static enum Level {
		debug, error, info, trace, warn;
	}

	private final BufferedReader br;

	private final Level level;

	private final Logger log;

	private final Logger logger;

	private final InputStream pis;

	private final OutputStream pos;

	public OutputStreamLogger(int bufferSize, Logger log, Level level)
			throws IOException {
		Preconditions.checkNotNull(level, "level");
		Preconditions.checkNotNull(log, "log");
		logger = LoggerFactory.getLogger(getClass());
		this.level = level;
		this.log = log;
		// TODO pass character encoding in constructor
		CircularByteBuffer cbb;
		if (bufferSize > 0) {
			cbb = new CircularByteBuffer(bufferSize);
			pis = cbb.getInputStream();
			pos = cbb.getOutputStream();
			br = new BufferedReader(new InputStreamReader(pis), bufferSize);
		} else {
			cbb = new CircularByteBuffer();
			pis = cbb.getInputStream();
			pos = cbb.getOutputStream();
			br = new BufferedReader(new InputStreamReader(pis));
		}
	}

	public OutputStreamLogger(Logger log, Level level) throws IOException {
		this(-1, log, level);
	}

	/**
	 * Closes associated streams.
	 * 
	 * @throws IOException
	 */
	public void close() {
		try {
			pos.flush();
		} catch (IOException ignore) {
		}
		IOUtils.closeQuietly(pos);
		IOUtils.closeQuietly(br);
	}

	public Level getLevel() {
		return level;
	}

	public Logger getLog() {
		return log;
	}

	/**
	 * Bytes written into OutputStream are logged by using
	 * {@link BufferedReader#readLine()} method. Remember to flush when
	 * appropriate. Closing the stream will stop the logger.
	 * 
	 * @return
	 */
	public OutputStream getOutputStream() {
		return pos;
	}

	private void log(String line) {
		switch (level) {
		case debug:
			log.debug(line);
			break;
		case error:
			log.error(line);
			break;
		case info:
			log.info(line);
			break;
		case trace:
			log.trace(line);
			break;
		case warn:
			log.warn(line);
			break;
		default:
			logger.warn("Undefined log level: " + level);
			log.error(line);
			break;
		}
	}

	@Override
	public void run() {
		try {
			String line;
			while ((line = br.readLine()) != null) {
				log(line);
			}
			if (logger.isDebugEnabled()) {
				logger.debug("End of stream for logger " + log.getName());
			}
		} catch (Exception e) {
			logger.error("Uncught exception. Thread "
					+ Thread.currentThread().getName() + "terminates", e);
		} finally {
			IOUtils.closeQuietly(pos);
			IOUtils.closeQuietly(br);
		}
	}

	public void start(boolean daemonThread) {
		Thread t = new Thread(this);
		t.setName(log.getName() + "-" + t.getId());
		t.setDaemon(daemonThread);
		t.start();
	}

}
