/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.pipe;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.io.IOUtils;

import com.Ostermiller.util.CircularByteBuffer;

/**
 * Mimics Unix pipelines where standard out of the first program is a standard
 * input of the next program.
 * <p>
 * Pipeline is constructed by connecting series of {@link Pipe}s with
 * {@link CircularByteBuffer}s.
 * <p>
 * After process has consumed input stream it should flush and close the output
 * stream. Failing to do causes {@link CircularByteBuffer} to block.
 */
public class Pipeline {

	/**
	 * Constructs a pipeline where output of a first pipe is input of a next
	 * pipe.
	 * 
	 * @param pipes
	 *            The number of pipes to create.
	 * @param is
	 *            Stream given to first pipe.
	 * @param os
	 *            Stream given to last pipe.
	 * @param error
	 *            Stream given to all pipes.
	 * @param bufSize
	 *            Size of the {@link CircularByteBuffer} buffer.
	 */
	public static List<Pipe> construct(int pipes, InputStream is,
			OutputStream os, OutputStream error, int bufSize)
			throws IOException {
		return new Pipeline(is, os, error, bufSize).getChain(pipes);
	}

	private final int bufSize;

	private CircularByteBuffer cbb;

	private final OutputStream error;

	private final InputStream is;

	private final OutputStream os;

	private InputStream pipeIn;

	private OutputStream pipeOut;

	private Pipeline(InputStream is, OutputStream os, OutputStream error,
			int bufSize) {
		this.is = is;
		this.os = os;
		this.error = error;
		this.bufSize = bufSize;
	}

	private void configureInputPipes(int i) throws IOException {
		if (i == 0) {
			pipeIn = is;
		} else {
			pipeIn = cbb.getInputStream();
		}
	}

	private void configureOutputPipes(int i, int size) {
		if (i == size - 1) {
			pipeOut = os;
		} else {
			cbb = new CircularByteBuffer(bufSize);
			pipeOut = cbb.getOutputStream();
		}
	}

	private List<Pipe> getChain(int size) throws IOException {
		initPipes();
		List<Pipe> plist = new ArrayList<Pipe>(size);
		for (int i = 0; i < size; i++) {
			configureInputPipes(i);
			configureOutputPipes(i, size);
			plist.add(new Pipe(pipeIn, pipeOut, error));
		}
		return plist;
	}

	private void initPipes() {
		IOUtils.closeQuietly(pipeIn);
		IOUtils.closeQuietly(pipeOut);
		pipeIn = null;
		pipeOut = null;
	}

}
