/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.sample;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.exec.CommandLine;
import org.apache.commons.exec.DefaultExecutor;
import org.apache.commons.exec.ExecuteResultHandler;
import org.apache.commons.exec.ExecuteWatchdog;
import org.apache.commons.exec.PumpStreamHandler;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.input.ClosedInputStream;
import org.apache.commons.io.output.ByteArrayOutputStream;
import org.apache.commons.io.output.TeeOutputStream;
import org.apache.commons.lang.text.StrSubstitutor;
import org.jhears.cnf.Cnf.AudioPreProsKey;
import org.jhears.cnf.Cnf.Framewrk;
import org.jhears.exec.DefaultExecuteResultHandler;
import org.jhears.exec.ExecUtil;
import org.jhears.file.FileContentMapper;
import org.jhears.pipe.Pipe;
import org.jhears.pipe.Pipeline;
import org.jhears.security.Digester;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Throwables;

/**
 * Class is responsible of audio data preprocessing. This includes translating
 * of various audio formats into raw format used in further audio analysis and
 * applying various filters to audio data to make it more suitable for analysis
 * algorithms.
 * 
 */
public class AudioPreprocessor {

	private final int bufSize;

	private final Configuration configuration;

	private OutputStream errorStream;

	private final Set<String> formatSet;

	private final Logger logger;

	private final FileContentMapper mapper;

	private final String[] pipeCmds;

	public AudioPreprocessor(Configuration configuration, File workDir)
			throws ConfigurationException, IOException {
		this.configuration = configuration;
		this.logger = LoggerFactory.getLogger(getClass());
		bufSize = configuration.getInt(AudioPreProsKey.PIPE_BUFFER_SIZE);
		pipeCmds = configuration
				.getStringArray(AudioPreProsKey.INPUT_PIPE_CHAIN_CMD_LIST);
		formatSet = new HashSet<String>();
		errorStream = System.err;
		if (workDir != null) {
			String algorithm = configuration
					.getString(Framewrk.CACHE_DIGEST_ALGORITHM);
			int bufferSize = configuration
					.getInt(Framewrk.CACHE_DIGEST_BUFFER_SIZE);
			File dir = new File(workDir, AudioPreProsKey.CNST_PRE_PROC_DIR_NAME);
			if (!dir.isDirectory()) {
				if (!dir.mkdirs()) {
					throw new IOException("Cannot create a directory: "
							+ dir.getAbsolutePath());
				}
			}
			try {
				mapper = new FileContentMapper(new Digester(algorithm,
						bufferSize), dir);
			} catch (Exception e) {
				throw Throwables.propagate(e);
			}
		} else {
			mapper = null;
		}
	}

	private boolean cache(File in, OutputStream out, File cacheFile)
			throws FileNotFoundException, IOException, InterruptedException {
		BufferedInputStream bis = new BufferedInputStream(new FileInputStream(
				in));
		BufferedOutputStream bos = new BufferedOutputStream(
				new FileOutputStream(cacheFile));
		OutputStream tee;
		if (out != null) {
			tee = new TeeOutputStream(out, bos);
		} else {
			tee = bos;
		}
		boolean retval = false;
		try {
			retval = process(FilenameUtils.getExtension(in.getName()), bis,
					tee, in.length() + 1000L, TimeUnit.MILLISECONDS);
			if (logger.isDebugEnabled()) {
				if (retval) {
					logger.debug("File " + in + " cached to " + cacheFile);
				} else {
					logger.debug("Failed to cache file " + in);
				}
			}
			return retval;
		} finally {
			IOUtils.closeQuietly(bis);
			IOUtils.closeQuietly(bos);
			IOUtils.closeQuietly(tee);
			if (!retval) {
				cacheFile.delete();
			}
		}
	}

	private boolean copy(File in, OutputStream os) throws FileNotFoundException {
		FileInputStream input = new FileInputStream(in);
		try {
			if (logger.isDebugEnabled()) {
				logger.debug("Using cached result for " + in);
			}
			IOUtils.copy(input, os);
			return true;
		} catch (IOException e) {
			logger.error("Copy failed", e);
			return false;
		} finally {
			IOUtils.closeQuietly(input);
		}
	}

	private StrSubstitutor createStrSubstitutor(String inputFormat) {
		HashMap<String, String> map = new HashMap<String, String>(1);
		map.put(AudioPreProsKey.STR_SUBST_INPUT_FMT, inputFormat);
		StrSubstitutor substitutor = new StrSubstitutor(map);
		return substitutor;
	}

	private String[] getAndConfigureCmd(String key, StrSubstitutor substitutor) {
		String[] cmdline = configuration.getStringArray(key);
		for (int i = 0; i < cmdline.length; i++) {
			cmdline[i] = substitutor.replace(cmdline[i]);
		}
		return cmdline;
	}

	public boolean isSupported(File file) throws IOException {
		return isSupported(file.getName());
	}

	public boolean isSupported(String filename) throws IOException {
		String ext = FilenameUtils.getExtension(filename);
		return isSupportedFormat(ext);
	}

	public boolean isSupportedFormat(String format) throws IOException {
		Set<String> formats = listSupportedFileFormats();
		return formats.contains(format);
	}

	public Set<String> listSupportedFileFormats() throws IOException {
		if (!formatSet.isEmpty()) {
			return formatSet;
		}

		String audioPreprocessorFormats = configuration
				.getString(AudioPreProsKey.FMT);
		String audioPreprocessorCmd = configuration
				.getString(AudioPreProsKey.CMD);

		CommandLine cmdline = new CommandLine(audioPreprocessorCmd);
		DefaultExecutor executor = new DefaultExecutor();
		ExecuteWatchdog watchdog = new ExecuteWatchdog(60000);

		ByteArrayOutputStream out = new ByteArrayOutputStream();
		ByteArrayOutputStream err = new ByteArrayOutputStream();
		InputStream in = new ClosedInputStream();

		PumpStreamHandler psh = new PumpStreamHandler(out, err, in);
		executor.setStreamHandler(psh);
		executor.setWatchdog(watchdog);
		executor.setExitValue(1);
		int exitValue = executor.execute(cmdline);

		if (exitValue != 1) {
			String errStr = new String(err.toByteArray());
			throw new IOException("Command " + audioPreprocessorCmd
					+ " exited with exit code " + exitValue + ". stderr="
					+ errStr);
		}
		BufferedReader br = new BufferedReader(new StringReader(new String(
				out.toByteArray())));
		String line = null;
		while ((line = br.readLine()) != null) {
			if (line.startsWith(audioPreprocessorFormats)
					&& formatSet.isEmpty()) {
				StringTokenizer t = new StringTokenizer(
						line.substring(audioPreprocessorFormats.length()), " ");
				while (t.hasMoreTokens()) {
					formatSet.add(t.nextToken());
				}
			}
		}
		if (logger.isDebugEnabled()) {
			logger.debug("Supported formats: " + formatSet);
		}
		return formatSet;
	}

	private boolean nonCaching(File in, OutputStream out)
			throws FileNotFoundException, IOException, InterruptedException {
		BufferedInputStream bis = new BufferedInputStream(new FileInputStream(
				in));
		boolean retval = false;
		try {
			retval = process(FilenameUtils.getExtension(in.getName()), bis,
					out, in.length() + 1000, TimeUnit.MILLISECONDS);
			return retval;
		} finally {
			IOUtils.closeQuietly(bis);
		}
	}

	/**
	 * 
	 * @param in
	 * @param overwrite
	 * @return cache file or null if caching failed.
	 * @throws IOException
	 * @throws InterruptedException
	 */
	public File process(File in, boolean overwrite) throws IOException,
			InterruptedException {
		if (!isSupported(in)) {
			throw new IOException("Unsupported file format");
		}
		File cacheFile = mapper.map(in);
		if (cacheFile.isFile() && !overwrite) {
			return cacheFile;
		} else {
			return cache(in, null, cacheFile) ? cacheFile : null;
		}
	}

	public boolean process(File in, OutputStream out) throws IOException,
			InterruptedException {
		if (!isSupported(in)) {
			throw new IOException("Unsupported file format");
		}
		File cacheFile = mapper != null ? mapper.map(in) : null;
		if (cacheFile != null && cacheFile.isFile()) {
			return copy(cacheFile, out);
		} else if (cacheFile != null) {
			return cache(in, out, cacheFile);
		} else {
			return nonCaching(in, out);
		}
	}

	public boolean process(String inputFormat, InputStream is, OutputStream os)
			throws IOException, InterruptedException {
		// timeout ~= 68 years...
		return process(inputFormat, is, os, Integer.MAX_VALUE, TimeUnit.SECONDS);
	}

	private boolean process(String inputFormat, InputStream is,
			OutputStream os, long timeout, TimeUnit unit) throws IOException,
			InterruptedException {
		AtomicBoolean retval = new AtomicBoolean(true);
		long timoutInMillis = TimeUnit.MILLISECONDS.convert(timeout, unit);
		StrSubstitutor substitutor = createStrSubstitutor(inputFormat);
		List<Pipe> pipes = Pipeline.construct(pipeCmds.length, is, os,
				errorStream, bufSize);
		Semaphore semaphore = new Semaphore(-pipeCmds.length + 1);
		int idx = 0;
		for (String key : pipeCmds) {
			String[] cmd = getAndConfigureCmd(key, substitutor);
			CommandLine cmdline = ExecUtil.createCommandLine(cmd);
			DefaultExecutor executor = ExecUtil.createExecutor(timoutInMillis,
					pipes.get(idx));
			ExecuteResultHandler handler = new DefaultExecuteResultHandler(
					retval, semaphore, cmd);
			// spawn
			executor.execute(cmdline, handler);
			idx++;
		}
		// wait until all processes have released the semaphore
		retval.compareAndSet(true, semaphore.tryAcquire(timeout, unit));
		return retval.get();
	}

	public void setErrorStream(OutputStream errorStream) {
		this.errorStream = errorStream;
	}

}
