/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.sample;

import com.google.common.base.Preconditions;

/**
 * The lower the value the more it gets attenuated. The higher the value less it
 * gets attenuated. Some what similar effect as GIMP "burn" layer filter has.
 * <p>
 * Algorithm: {@code (1-y^C/(x^C+y^C))*x} where {@code x} is a value to be
 * adjusted, {@code y} defines a 50 % cut point. {@code C} defines filter
 * sharpness.
 * 
 * @author Juha Heljoranta
 * 
 */
public class InplaceBurnFunction implements SampleFunction<double[], Void> {

	private final int[][] freqIdxs;
	private final double power;
	private final double[] values;

	/**
	 * 
	 * @param freqIdxs
	 * @param values
	 *            50% cut point values. Typically the mean or median of the
	 *            potential input values or some what less.
	 * @param power
	 *            Filter sharpness. Typically 2 < C < 12.
	 */
	public InplaceBurnFunction(final int[][] freqIdxs, double[] values,
			double power) {
		Preconditions.checkArgument(power > 0.0, "Power must be positive");
		Preconditions.checkNotNull(freqIdxs,
				"Frequency indexes must not be null");
		Preconditions.checkNotNull(values, "Filter values must not be null");
		this.freqIdxs = freqIdxs;
		this.values = values;
		this.power = power;
		// raise constant values beforehand
		for (int i = 0; i < values.length; i++) {
			Preconditions.checkArgument(values[i] >= 0.0,
					"Filter value cannot be negative");
			values[i] = Math.pow(values[i], power);
		}
	}

	@Override
	public Void apply(final double[] poweSpect) {
		for (final int[] bark : freqIdxs) {
			for (final int i : bark) {
				final double f = poweSpect[i];
				final double v = values[i];
				// (1-C/(x^Y+C))*x
				poweSpect[i] = (1.0 - v / (Math.pow(f, power) + v)) * f;
			}
		}
		return null;
	}

}