/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.sample;

/**
 * Converts raw audio samples to barks. This class is thread safe.
 */
public class RawToBarksFunction implements SampleFunction<byte[], double[]> {

	private final InplaceBurnFunction burnFilterFun;

	private final RawToSmoothedFunction rawToSmoothedFun;
	private final SpectrumToBarksFunction spectToBarksFun;

	public RawToBarksFunction(SpectrumToBarksFunction spectToBarksFun,
			InplaceBurnFunction burnFilterFun,
			RawToSmoothedFunction rawToSmoothFun) {
		this.spectToBarksFun = spectToBarksFun;
		this.burnFilterFun = burnFilterFun;
		this.rawToSmoothedFun = rawToSmoothFun;
	}

	/*
	 * Plan:
	 * 
	 * Input raw sample (int/short)
	 * 
	 * Convert sample to doubles.
	 * 
	 * Remove DC component.
	 * 
	 * Apply hannwindow.
	 * 
	 * Perform fft.
	 * 
	 * Apply loudness filter.
	 * 
	 * Calculate rms. if rms is too low regard data as silence (convert to
	 * zeroes).
	 * 
	 * Smooth spectrum (level down peaks).
	 * 
	 * Log spectrum.
	 * 
	 * Normalize spectrum (gain control).
	 * 
	 * Apply burn filter (the lower the value the more it is lowered). (Note to
	 * self: Burn filter works as expected if spectrum is normalized and gain
	 * controlled => no calibration needed.)
	 * 
	 * Convert to barks.
	 */
	@Override
	public double[] apply(byte[] data) {
		double[] powerspect = rawToSmoothedFun.apply(data);
		if (rawToSmoothedFun.isSilence(powerspect)) {
			return spectToBarksFun.getSilence();
		} else {
			burnFilterFun.apply(powerspect);
			return spectToBarksFun.apply(powerspect);
		}
	}
}
