/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.sample.filter;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.DoubleBuffer;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.io.IOUtils;
import org.jhears.cnf.Cnf;
import org.jhears.math.MatrixUtil;
import org.jhears.math.Stat;
import org.jhears.util.Mth;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Sets the filtering threshold to third quartile of input data.
 */
public class BurnFilterCalibrator {

	private final Logger logger;

	private final Configuration configuration;

	public BurnFilterCalibrator(Configuration configuration) {
		logger = LoggerFactory.getLogger(BurnFilterCalibrator.class);
		this.configuration = configuration;
	}

	public double[] calibrate(final File loudnessFilteredSample) throws IOException {
		final int sizeOfDouble = Double.SIZE / Byte.SIZE;
		int sampleSizeInBytes = configuration
				.getInt(Cnf.FingerprintKey.SAMPLE_WINDOW) / 2 * sizeOfDouble;
		double[] burnFilterMaximums = new double[configuration
				.getInt(Cnf.FingerprintKey.SAMPLE_WINDOW) / 2];
		final int rows = (int) (loudnessFilteredSample.length() / sampleSizeInBytes);

		File tmp = File.createTempFile(
				BurnFilterCalibrator.class.getSimpleName(), ".tmp",
				loudnessFilteredSample.getParentFile());

		int cols = configuration.getInt(Cnf.FingerprintKey.SAMPLE_WINDOW) / 2;
		logger.info("Transposing input matrix (rows=" + rows + ", cols=" + cols
				+ ")...");
		MatrixUtil.transposeMatrix(loudnessFilteredSample, tmp, rows, cols,
				sizeOfDouble, 1 << 20, 1 << 25);

		DataInputStream is = null;
		try {
			is = new DataInputStream(new BufferedInputStream(
					new FileInputStream(tmp), 1 << 20));
			byte[] buf = new byte[sizeOfDouble * rows];
			ByteBuffer bb = ByteBuffer.wrap(buf);
			DoubleBuffer db = DoubleBuffer.wrap(new double[rows]);
			logger.info("Calculating filter maximums...");
			for (int f = 0; f < cols; f++) {
				is.readFully(buf);
				db.clear();
				bb.clear();
				db.put(bb.asDoubleBuffer());
				Stat s = new Stat(db.array());
				// use median as a 50% cut point
				burnFilterMaximums[f] = s.getMedian();
			}

			// smooth the curve
			for (int i = 0; i < cols; i++) {
				burnFilterMaximums = Mth.simpleMovingAverage(
						burnFilterMaximums, 2);
			}
		} finally {
			if (is != null) {
				IOUtils.closeQuietly(is);
			}
			if (tmp.exists() && !tmp.delete()) {
				throw new IOException("Failed to delete file: "
						+ tmp.getAbsolutePath());
			}
		}

		return burnFilterMaximums;
	}
}
