/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.security;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.List;

import org.apache.commons.codec.binary.Hex;

/**
 * Calculates message digest from a contents of a file or a input stream.
 * <p>
 * This class is thread safe.
 */
public class Digester {

	private static final Charset UTF8 = Charset.forName("UTF-8");

	private final int bufferSize;

	private final MessageDigest md;

	private final Object mutex;

	public Digester(String algorithm, int bufferSize)
			throws NoSuchAlgorithmException, CloneNotSupportedException {
		this.bufferSize = bufferSize;
		this.mutex = new Object();
		md = MessageDigest.getInstance(algorithm);
		md.clone();
	}

	public byte[] digest(File file) throws IOException {
		BufferedInputStream bis = new BufferedInputStream(new FileInputStream(
				file));
		try {
			return digest(bis);
		} finally {
			bis.close();
		}
	}

	public byte[] digest(InputStream is) throws IOException {
		MessageDigest dig = getDigest();
		byte[] buf = new byte[bufferSize];
		int red;
		dig.reset();
		while ((red = is.read(buf)) != -1) {
			dig.update(buf, 0, red);
		}
		return dig.digest();
	}

	public String getAlgorithm() {
		return md.getAlgorithm();
	}

	private MessageDigest getDigest() {
		synchronized (mutex) {
			try {
				return (MessageDigest) md.clone();
			} catch (CloneNotSupportedException ignore) {
			}
		}
		return null;
	}

	/**
	 * Calculates a checksum from string list.
	 * 
	 * @param strings
	 * @return Hex encoded digest
	 */
	public String getHexDigest(List<String> strings) {
		MessageDigest md = getDigest();
		for (String str : strings) {
			md.update(str.getBytes(UTF8));
		}
		return new String(Hex.encodeHex(md.digest()));
	}
}
