/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

/**
 * Returns all possible combinations for the matrix rows. Input is a matrix
 * which each row contains all possible values values for that particular row.
 * See {@link #expand(int[][])} for an example.
 * <p>
 * This class is thread safe.
 */
public class Combinations {

	/**
	 * Expands matrix rows to all possible combinations. Each matrix row
	 * contains all possible values for that row.
	 * <p>
	 * e.g. for matrix [[1 2],[3,4],[5,6]] will expand to combinations: [1,3,5]
	 * [2,3,5] [1,4,5] [2,4,5] [1,3,6] [2,3,6] [1,4,6] [2,4,6].
	 * 
	 * @param matrix
	 *            Each row of the matrix contains one or more possible values
	 *            for that particular row.
	 * @return All possible combinations of the matrix rows.
	 */
	public int[][] expand(int[][] matrix) {
		int[] columns = new int[matrix.length];
		int[][] retval = new int[getCombinations(matrix)][];
		int i = 0;
		do {
			retval[i++] = rowValues(columns, matrix);
		} while (hasNextRow(0, matrix, columns));
		return retval;
	}

	/**
	 * Returns the number of possible combinations.
	 * 
	 * @param matrix
	 * @return
	 */
	private int getCombinations(int[][] matrix) {
		int combinations = 1;
		for (int[] row : matrix) {
			combinations *= row.length;
		}
		return combinations;
	}

	/**
	 * Recursively checks if the matrix has non-considered rows remaining.
	 * <p>
	 * Method increments value of current row of the <code>columns</code> on
	 * every call. If the value would overflow then the value is set to zero and
	 * next column is checked.
	 * 
	 * @param row
	 *            Current row to consider
	 * @param matrix
	 *            Rows contain all possible values for each respective row.
	 * @param columns
	 *            Indexes for tracking columns of the matrix.
	 * @return false If there is no more rows to consider. Otherwise returns
	 *         true.
	 */
	private boolean hasNextRow(final int row, int[][] matrix, int[] columns) {
		if (row >= columns.length) {
			return false;
		}
		columns[row]++;
		if (columns[row] < matrix[row].length) {
			return true;
		} else {
			columns[row] = 0;
			return hasNextRow(row + 1, matrix, columns);
		}
	}

	/**
	 * Selects row values from matrix columns. The column numbers are given in
	 * <code>columns</code> argument.
	 * 
	 * @param columns
	 *            column index for each matrix row.
	 * @param matrix
	 *            values in format matrix[row][column]
	 * @return row value from matrix at given columns.
	 */
	private int[] rowValues(int[] columns, int[][] matrix) {
		int[] retval = new int[matrix.length];
		for (int i = 0; i < matrix.length; i++) {
			retval[i] = matrix[i][columns[i]];
		}
		return retval;
	}

}
