/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.output.CloseShieldOutputStream;
import org.jhears.cnf.Cnf.FingerprintKey;
import org.jhears.cnf.Cnf.SearchKey;
import org.jhears.cnf.CnfUtil;
import org.jhears.io.StreamIterator;
import org.jhears.logging.OutputStreamLogger;
import org.jhears.logging.OutputStreamLogger.Level;
import org.jhears.sample.AudioPreprocessor;
import org.jhears.sample.RawToBarksFunction;
import org.jhears.sample.SampleFunctionInjector;
import org.jhears.util.Mth;
import org.picocontainer.MutablePicoContainer;
import org.picocontainer.PicoBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.Ostermiller.util.CircularByteBuffer;
import com.google.common.io.ByteStreams;
import com.google.common.io.CharStreams;
import com.google.common.io.Files;

/**
 * Produces symbols suitable for {@link FSeq}.
 */
public class FPSeqProducer {

	private static double distanceBetweenTwoClosestCodebookSymbols(
			final double[][] codebook) {
		double smallest = Double.POSITIVE_INFINITY;
		for (int i = 0; i < codebook.length; i++) {
			for (int j = i + 1; j < codebook.length; j++) {
				double d = Mth.distance(codebook[i], codebook[j]);
				if (d < smallest) {
					smallest = d;
				}
			}
		}
		return smallest;
	}

	private final AudioPreprocessor audioPreProcessor;

	private final int barkCnt;

	private final char[] chars;

	private final double[][] codebook;

	private final OutputStreamLogger errorLogger;

	private final List<FPSeqListener> listeners;

	private final Logger logger;

	// private final SampleProcessor processor;

	private final int sampleWindowInBytes;

	private final int stepSizeInBytes;

	private final int symbolCnt;

	private final double treshold;

	private final RawToBarksFunction rawToBarksFun;

	private final Configuration configuration;

	public FPSeqProducer(Configuration configuration, File workDir)
			throws IOException, ConfigurationException {
		this.configuration = configuration;
		logger = LoggerFactory.getLogger(getClass());
		chars = configuration.getString(FingerprintKey.SYMBOLS).toCharArray();
		sampleWindowInBytes = configuration
				.getInt(FingerprintKey.SAMPLE_WINDOW)
				* configuration.getInt(FingerprintKey.BYTES_PER_SAMPLE);
		// LoudnessFilterLoader loundnessLoader = new LoudnessFilterLoader();
		// Filter loudnessfilter = loundnessLoader.createFilter(configuration);
		// BurnFilterLoader burnFilterLoader = new BurnFilterLoader();
		// Filter burnfilter = burnFilterLoader.createFilter(configuration);
		// String filterLoaderClassName = configuration
		// .getString(AudioPreProsKey.SAMPLE_FILTER_LOADER_CLASS);
		// FilterLoader filterLoader = ClassUtil
		// .newInstance(filterLoaderClassName);
		// Filter filter = filterLoader.createFilter(configuration);
		// Filter filter = new FilterChain(new Filter[] { loudnessfilter,
		// burnfilter });
		// processor = new SampleProcessor(configuration, filter);
		MutablePicoContainer container = new PicoBuilder().withLifecycle()
				.withCaching().build();
		SampleFunctionInjector
				.addFunctionsToContainer(container, configuration);
		rawToBarksFun = container.getComponent(RawToBarksFunction.class);
		barkCnt = configuration.getInt(FingerprintKey.BARK_BAND_STOP_COUNT);
		symbolCnt = configuration.getInt(FingerprintKey.CODEBOOK_SYMBOL_COUNT);
		stepSizeInBytes = configuration.getInt(FingerprintKey.SAMPLE_INTERVAL)
				* configuration.getInt(FingerprintKey.BYTES_PER_SAMPLE);
		codebook = new double[symbolCnt][barkCnt];
		populateCodebook(configuration);
		audioPreProcessor = new AudioPreprocessor(configuration, workDir);
		Logger stdErrLog = LoggerFactory.getLogger("System.err");
		Level stdErrLevel = Level.info;
		errorLogger = new OutputStreamLogger(stdErrLog, stdErrLevel);
		errorLogger.start(true);
		CloseShieldOutputStream errStream = new CloseShieldOutputStream(
				errorLogger.getOutputStream());
		audioPreProcessor.setErrorStream(errStream);
		listeners = new ArrayList<FPSeqListener>();
		double dist = distanceBetweenTwoClosestCodebookSymbols(codebook);
		double margin = configuration
				.getDouble(SearchKey.SEQUENCE_ERROR_MARGIN);
		treshold = (dist / 2.0) * margin;
	}

	public void addListener(FPSeqListener listener) {
		listeners.add(listener);
	}

	/**
	 * get distances to codebook symbols. Returned array is sorted. The closest
	 * symbol is at index zero and the most distant symbol is the last.
	 * 
	 * @param barked
	 * @return
	 */
	private Distance[] getDistancesToSymbols(double[] barked) {
		Distance[] dia = new Distance[symbolCnt];
		for (int i = 0; i < symbolCnt; i++) {
			dia[i] = new Distance(i, Mth.distance(codebook[i], barked));
		}
		Arrays.sort(dia);
		return dia;
	}

	private void populateCodebook(Configuration configuration) {
		double[] v = CnfUtil.toDoubleArray(configuration
				.getStringArray(FingerprintKey.CODEBOOK));
		for (int r = 0; r < symbolCnt; r++) {
			for (int c = 0; c < barkCnt; c++) {
				codebook[r][c] = v[(r * barkCnt) + c];
			}
		}
	}

	private void process(byte[] sample) {
		double[] barked = rawToBarksFun.apply(sample);
		Distance[] distances = getDistancesToSymbols(barked);
		int symbolsBelowTreshold = 1;
		while (true) {
			double delta = distances[symbolsBelowTreshold].getDistance()
					- distances[0].getDistance();
			if (delta > treshold) {
				break;
			} else {
				symbolsBelowTreshold++;
			}
		}
		if (symbolsBelowTreshold == 1) {
			for (FPSeqListener l : listeners) {
				l.append(chars[distances[0].getIndex()]);
			}
		} else {
			char[] mutation = new char[symbolsBelowTreshold];
			for (int j = 0; j < symbolsBelowTreshold; j++) {
				mutation[j] = chars[distances[j].getIndex()];
			}
			for (FPSeqListener l : listeners) {
				l.append(mutation);
			}
		}
	}

	public void produceFrom(final InputStream in, final String format)
			throws IOException {
		if ("fingerprint".equals(format)) {
			final MutableFSeq fseq = new FSeqFactory(configuration)
					.getNewInstance();
			String fingerprint = CharStreams
					.toString(CharStreams.newReaderSupplier(
							ByteStreams.newInputStreamSupplier(ByteStreams
									.toByteArray(in)), Charset.defaultCharset()));
			fseq.append(fingerprint.trim());
			for (FPSeqListener l : listeners) {
				fseq.notifyListener(l);
			}
			return;
		}
		if (!audioPreProcessor.isSupportedFormat(format)) {
			throw new IOException("Unsupported file format: " + format);
		}
		errorLogger.getOutputStream().flush();
		// TODO remove hardcoded constant
		CircularByteBuffer cbb = new CircularByteBuffer(1 << 20, true);
		InputStream pis = cbb.getInputStream();
		final OutputStream pos = cbb.getOutputStream();
		new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					Thread.currentThread().setName(
							FPSeqProducer.class.getSimpleName()
									+ " audio preprocessor");
					audioPreProcessor.process(format, in, pos);
					errorLogger.getOutputStream().flush();
				} catch (Exception e) {
					logger.error("Uncaught exception", e);
				} finally {
					IOUtils.closeQuietly(pos);
				}
			}
		}).start();

		StreamIterator si = new StreamIterator(pis, sampleWindowInBytes,
				stepSizeInBytes, 1 << 20);
		try {
			for (byte[] sample : si) {
				process(sample);
			}
		} catch (RuntimeException e) {
			logger.error("Sample processing failed.", e);
		}
		for (FPSeqListener l : listeners) {
			l.eof();
		}
	}

	public void produceFrom(final File file) throws IOException {
		if ("fingerprint".equals(FilenameUtils.getExtension(file.getName()))) {
			final MutableFSeq fseq = new FSeqFactory(configuration)
					.getNewInstance();
			String fingerprint = CharStreams.toString(Files.newReaderSupplier(
					file, Charset.defaultCharset()));
			fseq.append(fingerprint.trim());
			for (FPSeqListener l : listeners) {
				fseq.notifyListener(l);
			}
			return;
		}
		if (!audioPreProcessor.isSupported(file)) {
			throw new IOException("Unsupported file format: ("
					+ FilenameUtils.getExtension(file.getName()) + ") file="
					+ file.getAbsolutePath());
		}
		errorLogger.getOutputStream().flush();
		// TODO remove hardcoded constant
		CircularByteBuffer cbb = new CircularByteBuffer(1 << 20, true);
		InputStream pis = cbb.getInputStream();
		final OutputStream pos = cbb.getOutputStream();
		new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					Thread.currentThread().setName(
							FPSeqProducer.class.getSimpleName()
									+ " audio preprocessor");
					audioPreProcessor.process(file, pos);
					errorLogger.getOutputStream().flush();
				} catch (Exception e) {
					logger.error("Uncaught exception", e);
				} finally {
					IOUtils.closeQuietly(pos);
				}
			}
		}).start();

		StreamIterator si = new StreamIterator(pis, sampleWindowInBytes,
				stepSizeInBytes, 1 << 20);
		try {
			for (byte[] sample : si) {
				process(sample);
			}
		} catch (RuntimeException e) {
			logger.error("Sample processing failed.", e);
		}
		for (FPSeqListener l : listeners) {
			l.eof();
		}
	}

	public void removeListener(FPSeqListener listener) {
		listeners.remove(listener);
	}

}
