/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.Iterator;

import org.jhears.web.rs.Segment;

/**
 * Fingerprint sequence is a string in which each character represents one
 * symbol in a codebook, for example: "G4DWOCc+e++NxpR7tkIubw". Moreover, the
 * fingerprint can have mutations. A mutation is a situation where two or more
 * symbols are roughly equally close to a codebook symbol. Mutations are marked
 * by using backslash (\) and a number telling how many following characters are
 * part of the mutation. For example, a sequence "ab\3cdefg" has two mutations
 * with interpretations: "abcfg", "abdfg" and "abefg", the first being the
 * non-mutated value. When the sequence is flattened the mutations are removed,
 * for example flattened version of previous example is "abcfg.
 */
public interface FSeq extends Iterable<Segment> {

	/**
	 * Return a String presentation of the fingerprint. e.g. "a\4qwerb".
	 * 
	 * @return
	 */
	public String asString();

	/**
	 * The number of full segments this fingerprint sequence has when flattened.
	 * Basically the value of {@link #getLength()}/{@link #getSegmentLength()}.
	 * 
	 * @return
	 */
	public int getFullSegmentCount();

	/**
	 * Length of a String returned by {@link #asString()} when flattened. That
	 * is, mutations do not affect the length.
	 * 
	 * @return
	 */
	public int getLength();

	/**
	 * The maximum number of mutations per symbol. e.g. if maxMutations is 2
	 * then a "a\3XYZ\3IJKb" is allowed but "a\4qwerb" is not.
	 * 
	 * @return
	 */
	public int getMaxMutations();

	/**
	 * The number of mutations this sequence has. e.g. "a\3XYZ\3IJKb" has 4
	 * mutations and "a\4qwerb" has 3 mutation.
	 * 
	 * @return
	 */
	public int getMutations();

	/**
	 * The length of a segment when this sequence is divided into
	 * {@link Segment}s.
	 * 
	 * @return
	 */
	public int getSegmentLength();

	/**
	 * Iterates over all possible {@link Segment}s of this sequence. e.g. if
	 * segment length is 3 and the sequence is "abcdef" then the iterator
	 * returns segments with data "abc", "bcd", "cde" and "def". Another example
	 * with a mutation: "ab\2cde" returns "abc", "abd", "bce" and "bde".
	 */
	@Override
	public Iterator<Segment> iterator();

}