/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

/**
 * Quantizes series of integer values to nearest quantization codebook values.
 * Alternatively the quantizer can take account an error margin and return a
 * list of quantized values near the given values.
 * <p>
 * This class is thread safe.
 */
public class IntQuantizer {

	/**
	 * Throws IllegalArgumentException if array has duplicates.
	 * 
	 * @param codebook
	 * @throws IllegalArgumentException
	 */
	private static void checkForDuplicates(int[] codebook)
			throws IllegalArgumentException {
		HashSet<Integer> hs = new HashSet<Integer>(codebook.length);
		for (int i : codebook) {
			hs.add(Integer.valueOf(i));
		}
		if (hs.size() != codebook.length) {
			throw new IllegalArgumentException(
					"Array contains duplicate elements.");
		}
	}

	/**
	 * Convert collection of integers to array of integers.
	 * 
	 * @param c
	 * @return
	 */
	private static int[] convert(Collection<Integer> c) {
		int[] arr = new int[c.size()];
		Iterator<Integer> itr = c.iterator();
		int i = 0;
		while (itr.hasNext()) {
			arr[i++] = itr.next().intValue();
		}
		return arr;
	}

	private final int[] codebook;

	private final Combinations combinations;

	private final double maxMult;

	private final double minMult;

	private final int offset;

	/**
	 * 
	 * @param codebook
	 *            Values sorted in ascending order used for quantization. e.g.
	 *            4, 8, 16, ...
	 * 
	 * @param offset
	 *            constant value which is added to upper bound of the search and
	 *            subtracted from lower bound of the search. i.e. constant error
	 *            margin. e.g. 1.
	 * @param minMultiplier
	 *            search value is multiplied with this constant and rounded
	 *            downwards for lower search limit. i.e. search error margin
	 *            relative to searched value. e.g. 0.9 for 10 % margin.
	 * 
	 * @param maxMultiplier
	 *            search value is multiplied with this constant and rounded
	 *            upwards for upper search limit.i.e. search error margin
	 *            relative to searched value. e.g. 1.1 for 10 % margin.
	 * @throws IllegalArgumentException
	 *             If codebook is empty or contains duplicates or is not sorted
	 *             properly. If <code>minMultiplier > maxMultiplier</code> or if
	 *             offset is negative.
	 */
	public IntQuantizer(Collection<Integer> codebook, int offset,
			double minMultiplier, double maxMultiplier)
			throws IllegalArgumentException {
		this(convert(codebook), offset, minMultiplier, maxMultiplier);
	}

	/**
	 * @param codebook
	 *            Values sorted in ascending order used for quantization. e.g.
	 *            4, 8, 16, ...
	 * 
	 * @param offset
	 *            constant value which is added to upper bound of the search and
	 *            subtracted from lower bound of the search. i.e. constant error
	 *            margin. e.g. 1.
	 * @param minMultiplier
	 *            search value is multiplied with this constant and rounded
	 *            downwards for lower search limit. i.e. search error margin
	 *            relative to searched value. e.g. 0.9 for 10 % margin.
	 * 
	 * @param maxMultiplier
	 *            search value is multiplied with this constant and rounded
	 *            upwards for upper search limit.i.e. search error margin
	 *            relative to searched value. e.g. 1.1 for 10 % margin.
	 * @throws IllegalArgumentException
	 *             If codebook is empty or contains duplicates or is not sorted
	 *             properly. If <code>minMultiplier > maxMultiplier</code> or if
	 *             offset is negative.
	 */
	public IntQuantizer(int[] codebook, int offset, double minMultiplier,
			double maxMultiplier) throws IllegalArgumentException {
		checkForDuplicates(codebook);
		if (minMultiplier > maxMultiplier) {
			throw new IllegalArgumentException("minMultiplier > maxMultiplier");
		}
		if (offset < 0) {
			throw new IllegalArgumentException("offset cannot be negative");
		}
		if (codebook.length == 0) {
			throw new IllegalArgumentException(
					"The codebook must have at least one entry");
		}
		this.minMult = minMultiplier;
		this.offset = offset;
		this.maxMult = maxMultiplier;
		this.codebook = Arrays.copyOf(codebook, codebook.length);
		Arrays.sort(this.codebook);
		if (!Arrays.equals(this.codebook, codebook)) {
			throw new IllegalArgumentException("The codebook is not sorted.");
		}
		combinations = new Combinations();
	}

	/**
	 * Returns a copy of the codebook.
	 * 
	 * @return
	 */
	public int[] getCodebook() {
		return Arrays.copyOf(codebook, codebook.length);
	}

	private int[] getFilledArray(int from, int to) {
		int[] arr = new int[to - from + 1];
		for (int i = 0; i < arr.length; i++) {
			arr[i] = from + i;
		}
		return arr;
	}

	/**
	 * Returns index of the nearest bucket e.g. If buckets is [10, 20]: for
	 * values -1, 14, 15, 21 the method returns values 0, 0, 1, 1.
	 * 
	 * @param value
	 * @return
	 */
	private int getIdx(int value) {
		int idx = Arrays.binarySearch(codebook, value);
		if (idx < 0) {
			idx = -idx - 1;
			if (idx > 0 && idx < codebook.length) {
				int a = codebook[idx - 1]; // less than value
				int b = codebook[idx]; // greater than value
				if ((value - a) < (b - value)) {
					// value is closer to a than b.
					idx--;
				}
			} else if (idx == codebook.length) {
				idx--;
			}
		}
		return idx;
	}

	/**
	 * Returns all possible indexes of the codebook for the value which are
	 * within the error range.
	 * 
	 * @param a
	 * @return
	 */
	private int[] indexesFor(int a) {
		int min = (int) Math.round(Math.floor(a * minMult)) - offset;
		int max = (int) Math.round(Math.ceil(a * maxMult)) + offset;
		int minidx = getIdx(min);
		int maxidx = getIdx(max);
		int[] arr = getFilledArray(minidx, maxidx);
		return arr;
	}

	/**
	 * Exactly the same as the {@link #nearestIndexes(int[])} except that the
	 * all possible indexes within the error margins, given in constructor, are
	 * returned.
	 * <p>
	 * e.g. if codebook is 4, 8, 16, offset = 1, minMultiplier = 0.9,
	 * maxMultiplier = 1.1 and the input values are 6, 12, 24 the method return
	 * values [0, 1, 2 ], [ 1, 1, 2 ], [ 0, 2, 2 ] and [ 1, 2, 2 ].
	 * 
	 * @return
	 */
	public int[][] nearbyIndexes(int[] values) {
		int[][] key = new int[values.length][];
		for (int i = 0; i < values.length; i++) {
			key[i] = indexesFor(values[i]);
		}
		return combinations.expand(key);
	}

	/**
	 * Array element is rounded to its nearest codebook value and the index of
	 * that codebook symbol is returned. This is done for each array element. In
	 * other words the values are quantized to nearest codebook values and the
	 * indexes of the codebook values are returned.
	 * <p>
	 * e.g. if codebook is 4, 8, 12, 16 and the input values are 9, 3, 14 the
	 * method returns values 1, 0 and 3 (the indexes of the codebook values).
	 * 
	 * @param values
	 *            Input values to be quantized.
	 * @return Codebook indexes nearest to the quantized values.
	 */
	public int[] nearestIndexes(int[] values) {
		int[] key = new int[values.length];
		for (int i = 0; i < values.length; i++) {
			key[i] = getIdx(values[i]);
		}
		return key;
	}

}
