/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.List;

import org.jhears.web.rs.Segment;

/**
 * Modifiable {@link FSeq}.
 */
public interface MutableFSeq extends FSeq {

	public MutableFSeq append(final char c);

	public MutableFSeq append(final char[] c);

	public MutableFSeq append(String str);

	public ImmutableFSeq asImmutableFSeq();

	public MutableFSeq copy();

	/**
	 * Returns the number of characters which are not equal to given character.
	 * 
	 * @param censorChar
	 * @return
	 */
	public int countNonCensoredChars(char censorChar);

	public MutableFSeq flatten();

	public int getIgnoredMutations();

	/**
	 * Returns non-overlapping segments of the sequence which contain less than
	 * <code>threshold</code> of <code>censorChar</code> characters.
	 * 
	 * @param threshold
	 *            the maximum allowed number of censorChar per segment.
	 * @param censorChar
	 *            the character which is used for discriminating segments.
	 * @throws IllegalStateException
	 *             If sequence has mutations.
	 * @return Segments which have less than threshold characters.
	 */
	public List<Segment> getNonLowComplexitySegments(final int threshold,
			final char censorChar) throws IllegalStateException;

	/**
	 * Get segments from given positions if possible. Positions out of the valid
	 * range are ignored.
	 * 
	 * @param positions
	 * @return
	 * @throws IllegalStateException
	 *             If sequence has mutations.
	 */
	public List<Segment> getSegmentsFromPositions(final List<Integer> positions)
			throws IllegalStateException;

	/**
	 * Replaces areas which have low complexity with <code>censorChar</code>.
	 * Parts of a sequence with long stretches of single or only few different
	 * characters are considered to have low complexity and long stretches with
	 * lot of different characters without repetitions are high complexity.
	 * <p>
	 * Example of low complexity: "aaaaaabaaaabaaaababaaaabbbaaaaa"</br> Example
	 * of high complexity: "5ugWeWTKk0F4FVg+rlpXwX5zm1AI/XB"</br> Example of
	 * long stretch of 'a's censored seq with '0':
	 * "Is4oBaP9daa/jaa0000000000000aaIac0a1a8vgDaTQ"</br>
	 * <p>
	 * <p>
	 * Note that this will flatten the sequence.
	 * <p>
	 * Examples:
	 * 
	 * <pre>
	 * Original seq: &quot;12345678aaaaaaaa12345678&amp;quot
	 * replaceLowComplexity(4, 3, 2, '0')
	 * Resulting seq: &quot;12345678aa0000aa12345678&quot;
	 * 
	 * Original seq: &quot;12345678axaxaxaxa12345678&quot;
	 * replaceLowComplexity(5, 2, 2, '0')
	 * Resulting seq: &quot;12345678a0000000a12345678&quot;
	 * </pre>
	 * 
	 * @param complexitySegLen
	 *            The number of characters subjected for censoring at time.
	 * @param duplicatesPerSegment
	 *            the number of duplicate characters the segment must at least
	 *            to have for its being subjected for censoring.
	 * @param margin
	 *            the number of characters left untouched around the censored
	 *            area. Tip: This value should be at least (
	 *            <code>complexitySegLen -
	 *            duplicatesPerSegment - 1</code>) to avoid censoring too much.
	 * @param censorChar
	 *            areas which are censored are replaced with this character.
	 * @return self with low complexity ares replaced with the censorChar.
	 */
	public MutableFSeq replaceLowComplexity(final int complexitySegLen,
			final int duplicatesPerSegment, final int margin,
			final char censorChar);

	public MutableFSeq subseq(int beginIndex, int endIndex);

	public List<Segment> toSegments();

	public String toString();

	void notifyListener(FPSeqListener listener);

}
