/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.jhears.web.rs.Segment;

public class MutableFSeqImpl implements MutableFSeq {

	public static final int MAX_MUT_LEN = 16;

	/**
	 * Returns an array containing the number of duplicate characters within
	 * each segment in given array.
	 * <p>
	 * Examples:
	 * 
	 * <ul>
	 * <li>&quot;abc&quot; has zero duplicates
	 * <li>&quot;cbc&quot; has 1 duplicate
	 * <li>&quot;ccc&quot; has three duplicates
	 * <li>&quot;cxcc&quot; has three duplicates
	 * </ul>
	 * 
	 * @param ca
	 *            array to inspect for duplicate characters within given segment
	 *            length
	 * @param complexitySegLen
	 *            the segment length
	 * @param ign
	 *            "anti-joker" character. isn't duplicate of any other character
	 *            not even it self. In other words, sequences of these
	 *            characters turn into sequences of zeros.
	 * @return
	 */
	static int[] duplicatesPerSegment(final char[] ca,
			final int complexitySegLen, final char ign) {
		final int[] refPosArr = makePositionArray(complexitySegLen);
		final int[] freqArr = new int[ca.length - complexitySegLen + 1];
		for (int i = 0; i < freqArr.length; i++) {
			int[] freqarr = new int[complexitySegLen];
			// remaining segment positions to consider
			int[] posarr = Arrays.copyOf(refPosArr, refPosArr.length);
			for (int idx = 0; idx < complexitySegLen; idx++) {
				int pos = posarr[idx];
				if (pos == -1) {
					continue; // position was already checked
				}
				int caPos = i + pos;
				for (int cursor = pos + 1; cursor < complexitySegLen; cursor++) {
					if (ca[caPos] == ca[i + cursor]) {
						posarr[cursor] = -1; // mark position checked
						if (ca[caPos] != ign) {
							freqarr[idx]++;
						}
					}
				}
			}
			for (int f : freqarr) {
				freqArr[i] += f;
			}
		}
		return freqArr;
	}

	private static int[] makePositionArray(final int seglen) {
		final int[] poa = new int[seglen];
		for (int i = 0; i < seglen; i++) {
			poa[i] = i;
		}
		return poa;
	}

	private final char[] chars;

	private int ignoredMutations;

	private int length;

	private final int maxMutations;

	private int mutations;

	private final int[] rev;

	private StringBuilder sb;

	private final int segmentLength;

	private final int[] seqMutations;

	private final char smallestChar;

	public MutableFSeqImpl(final int segmentLength, final int maxMutations,
			char smallestChar, char[] chars, int[] rev) {
		this.segmentLength = segmentLength;
		this.maxMutations = maxMutations;
		this.chars = chars;
		this.rev = rev;
		this.sb = new StringBuilder();
		this.seqMutations = new int[segmentLength];
		this.length = 0;
		this.mutations = 0;
		this.smallestChar = smallestChar;
	}

	public MutableFSeqImpl(MutableFSeqImpl seq) {
		this.smallestChar = seq.smallestChar;
		this.seqMutations = seq.seqMutations;
		this.segmentLength = seq.segmentLength;
		this.rev = seq.rev;
		this.maxMutations = seq.maxMutations;
		this.chars = seq.chars;
		this.ignoredMutations = seq.ignoredMutations;
		this.length = seq.length;
		this.mutations = seq.mutations;
		this.sb = new StringBuilder(seq.sb);
	}

	private MutableFSeqImpl(String seq, int length, int mutations,
			int segmentLength, int maxMutations, int[] seqMutations,
			char smallestChar, char[] chars, int[] rev) {
		this.smallestChar = smallestChar;
		this.chars = chars;
		this.rev = rev;
		this.sb = new StringBuilder(seq);
		this.length = length;
		this.mutations = mutations;
		this.segmentLength = segmentLength;
		this.maxMutations = maxMutations;
		this.seqMutations = seqMutations;
	}

	public MutableFSeq append(final char c) {
		try {
			return append(rev[c - smallestChar]);
		} catch (ArrayIndexOutOfBoundsException e) {
			e.printStackTrace();
			throw e;
		}
	}

	public MutableFSeq append(final char[] c) {
		final int[] ia = new int[c.length];
		for (int j = 0; j < c.length; j++) {
			ia[j] = rev[c[j] - smallestChar];
		}
		return append(ia);
	}

	public MutableFSeq append(final int i) {
		sb.append(chars[i]);
		seqMutations[length % segmentLength] = 0;
		length++;
		return this;
	}

	public MutableFSeq append(final int[] i) {
		final int len = i.length;
		if (len == 0) {
			return this;
		} else if (len == 1) {
			return append(i[0]);
		} else if (len - 1 > maxMutations) {
			ignoredMutations += len - 1 - maxMutations;
			return append(Arrays.copyOf(i, maxMutations + 1));
		}
		seqMutations[length % segmentLength] = len - 1;
		int sum = 0;
		for (int j = 0; j < segmentLength; j++) {
			sum += seqMutations[j];
		}
		if (sum > maxMutations) {
			int newlen = maxMutations - (sum - len);
			ignoredMutations += len - newlen;
			return append(Arrays.copyOf(i, newlen));
		}
		final char[] ca = new char[len];
		for (int j = 0; j < len; j++) {
			ca[j] = chars[i[j]];
		}
		length++;
		mutations += len - 1;
		char d = Character.forDigit(len, MAX_MUT_LEN);
		if (d == '\0') {
			throw new IllegalArgumentException(
					"Cannot handle mutations longer than " + MAX_MUT_LEN);
		}
		sb.append('\\').append(d).append(ca);
		return this;
	}

	public MutableFSeq append(String str) {
		final char[] ca = str.toCharArray();
		final int size = ca.length;

		boolean esc = false;
		int ptr = 0;
		int i = 0;
		for (; i < size; i++) {
			if (esc) {
				esc = false;
				if (ptr < i - 1) {
					bulkAppend(Arrays.copyOfRange(ca, ptr, i - 1));
				}
				int len = Character.digit(ca[i], MAX_MUT_LEN);
				if (len < 2) {
					throw new IllegalArgumentException("" + len);
				} else if (len + i >= size) {
					throw new IllegalArgumentException();
				} else {
					append(Arrays.copyOfRange(ca, i + 1, i + len + 1));
					i += len;
					ptr = i + 1;
				}
			} else {
				esc = ca[i] == '\\';
			}
		}
		if (esc) {
			throw new IllegalArgumentException();
		}
		if (ptr < i) {
			bulkAppend(Arrays.copyOfRange(ca, ptr, i));
		}
		return this;
	}
	
	@Override
	public void notifyListener(FPSeqListener listener) {
		final char[] ca = sb.toString().toCharArray();
		final int size = ca.length;

		boolean esc = false;
		int ptr = 0;
		int i = 0;
		for (; i < size; i++) {
			if (esc) {
				esc = false;
				if (ptr < i - 1) {
					for (int x = ptr; x < i - 1; x++) {
						listener.append(ca[x]);
					}
				}
				int len = Character.digit(ca[i], MAX_MUT_LEN);
					listener.append(Arrays.copyOfRange(ca, i + 1, i + len + 1));
					i += len;
					ptr = i + 1;
			} else {
				esc = ca[i] == '\\';
			}
		}
		if (ptr < i) {
			for (int x = ptr; x < i; x++) {
				listener.append(ca[x]);
			}
		}
		listener.eof();
	}

	public ImmutableFSeq asImmutableFSeq() {
		return new ImmutableFSeq(asString(), length, mutations, segmentLength,
				maxMutations);
	}

	@Override
	public String asString() {
		return sb.toString();
	}

	private MutableFSeq bulkAppend(final char[] str) {
		for (char c : str) {
			append(c);
		}
		return this;
	}

	private void censor(final char[] ca, final int[] duplicatesPerSegArr,
			final int duplicatesPerSegment, final int margin,
			final int complexitySegLen, final char censorChar) {
		for (int i = 0; i < duplicatesPerSegArr.length; i++) {
			if (duplicatesPerSegArr[i] >= duplicatesPerSegment) {
				int j = i;
				for (; j < duplicatesPerSegArr.length; j++) {
					if (duplicatesPerSegArr[j] < duplicatesPerSegment) {
						break;
					}
				}
				int idelta = i + (i == 0 ? 0 : margin);
				int jdelta = j + complexitySegLen
						- (j == duplicatesPerSegArr.length ? 0 : margin) - 1;
				Arrays.fill(ca, idelta, jdelta, censorChar);
				i = j;
			}
		}
	}

	public MutableFSeq copy() {
		return new MutableFSeqImpl(this);
	}

	@Override
	public int countNonCensoredChars(char censorChar) {
		if (mutations != 0) {
			throw new IllegalStateException("Sequence has mutations");
		}
		int size = 0;
		for (char c : sb.toString().toCharArray()) {
			if (c != censorChar) {
				size++;
			}
		}
		return size;
	}

	public MutableFSeqImpl flatten() {
		if (mutations == 0) {
			return this;
		}
		final char[] ca = sb.toString().toCharArray();
		StringBuilder newSb = new StringBuilder(length);
		final int size = ca.length;
		boolean esc = false;
		int i = 0;
		for (; i < size; i++) {
			if (esc) {
				esc = false;
				int len = Character.digit(ca[i], MAX_MUT_LEN);
				if (len == -1) {
					throw new IllegalArgumentException("" + len);
				}
				newSb.append(ca[i + 1]);
				i += len;
			} else {
				esc = ca[i] == '\\';
				if (!esc) {
					newSb.append(ca[i]);
				}
			}
		}
		sb = newSb;
		mutations = 0;
		return this;
	}

	@Override
	public int getFullSegmentCount() {
		return length / segmentLength;
	}

	public int getIgnoredMutations() {
		return ignoredMutations;
	}

	@Override
	public int getLength() {
		return length;
	}

	@Override
	public int getMaxMutations() {
		return maxMutations;
	}

	@Override
	public int getMutations() {
		return mutations;
	}

	public List<Segment> getNonLowComplexitySegments(final int threshold,
			final char censorChar) {
		if (mutations != 0) {
			throw new IllegalStateException("The sequence has mutations.");
		}
		final char[] ca = sb.toString().toCharArray();
		final int size = ca.length / segmentLength;
		ArrayList<Segment> segList = new ArrayList<Segment>(size);
		for (int i = 0; i < size; i++) {
			int count = 0;
			final int begin = i * segmentLength;
			final int end = begin + segmentLength;
			for (int j = begin; j < end; j++) {
				if (ca[j] == censorChar) {
					count++;
				}
			}
			if (count <= threshold) {
				segList.add(new Segment(new String(Arrays.copyOfRange(ca,
						begin, end)), begin));
			}
		}
		return segList;
	}

	@Override
	public int getSegmentLength() {
		return segmentLength;
	}

	public List<Segment> getSegmentsFromPositions(final List<Integer> positions)
			throws IllegalStateException {
		if (mutations != 0) {
			throw new IllegalStateException(
					"Cannot get segments if sequence has mutations.");
		}
		final ArrayList<Segment> slist = new ArrayList<Segment>(positions
				.size());
		final char[] ca = sb.toString().toCharArray();
		final int max = ca.length - segmentLength;
		for (final Integer i : positions) {
			final int pos = i.intValue();
			if (pos >= 0 && pos <= max) {
				final String s = new String(Arrays.copyOfRange(ca, pos, pos
						+ segmentLength));
				slist.add(new Segment(s, pos));
			}
		}
		return slist;
	}

	@Override
	public Iterator<Segment> iterator() {
		return new SegmentIterator(segmentLength, sb.toString().toCharArray());
	}

	public MutableFSeq replaceLowComplexity(final int complexitySegLen,
			final int maxDuplicatesPerSegment, final int margin,
			final char censorChar) {
		if (maxDuplicatesPerSegment >= complexitySegLen) {
			throw new IllegalArgumentException("treshold >= complexitySegLen");
		}
		if (maxDuplicatesPerSegment < 0) {
			throw new IllegalArgumentException("maxDuplicatesPerSegment < 0");
		}
		if (complexitySegLen < 0) {
			throw new IllegalArgumentException("complexitySegLen < 0");
		}
		if (margin < 0) {
			throw new IllegalArgumentException("margin < 0");
		}
		char c = chars[rev[censorChar - smallestChar]];
		final char[] str = flatten().sb.toString().toCharArray();
		censor(str, duplicatesPerSegment(str, complexitySegLen, c),
				maxDuplicatesPerSegment, margin, complexitySegLen, c);
		sb = new StringBuilder(new String(str));
		return this;
	}

	public MutableFSeq subseq(int beginIndex, int endIndex) {
		if (beginIndex < 0) {
			throw new IllegalArgumentException();
		}
		if (beginIndex > endIndex) {
			throw new IllegalArgumentException();
		}
		if (endIndex > length) {
			throw new IllegalArgumentException();
		}
		if (mutations == 0) {
			return new MutableFSeqImpl(sb.toString().substring(beginIndex,
					endIndex), endIndex - beginIndex, mutations, segmentLength,
					maxMutations, new int[segmentLength], smallestChar, chars,
					rev);
		}
		final char[] ca = sb.toString().toCharArray();
		final int size = ca.length;
		boolean esc = false;
		int i = 0;
		int pos = 0;
		int newLength = 0;
		int begIdx = -1;
		int endIdx = -1;
		int newMutations = 0;
		for (; i < size; i++) {
			if (esc) {
				esc = false;
				int len = Character.digit(ca[i], MAX_MUT_LEN);
				if (len == -1) {
					throw new IllegalArgumentException("" + len);
				}
				i += len;
				if (begIdx != -1) {
					newMutations += len - 1;
				}
				if (endIdx != -1) {
					endIdx = i;
					break;
				}
			} else {
				if (pos == beginIndex) {
					begIdx = i;
				}
				if (pos == endIndex) {
					endIdx = i;
				}
				pos++;
				esc = ca[i] == '\\';
				if (begIdx != -1 && endIdx == -1) {
					newLength++;
				}
			}
			if (endIdx != -1) {
				break;
			}
		}
		if (pos == endIndex) {
			endIdx = i;
		}
		/*
		 * TODO: optimize.
		 * 
		 * Take last segmentLength elements and reconstruct newSeqMutations from
		 * them.
		 * 
		 * int[] newSeqMutations = new int[segmentLength]; return new
		 * MutableFSeqImpl(new String(ca, begIdx, endIdx - begIdx), newLength,
		 * newMutations, segmentLength, maxMutations, newSeqMutations,
		 * smallestChar, chars, rev);
		 */
		return new MutableFSeqImpl(segmentLength, maxMutations, smallestChar,
				chars, rev).append(new String(ca, begIdx, endIdx - begIdx));
	}

	@Override
	public List<Segment> toSegments() {
		ArrayList<Segment> slist = new ArrayList<Segment>(length);
		for (Segment s : this) {
			slist.add(s);
		}
		return slist;
	}

	@Override
	public String toString() {
		return MutableFSeqImpl.class.getSimpleName() + " [length=" + length
				+ ", mutations=" + mutations + ", sb=" + sb + "]";
	}

}
