/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Class iterates over String mutations. Strings are fixed length and they have
 * maximum number of mutations per row.
 * <p>
 * This class doesn't perform any input validation so the user of this class is
 * responsible of correct input values.
 * <p>
 * The iterator becomes usable after user has called {@link #insertRow(char[])}
 * method as many times as the string length given in constructor. At any time a
 * new row can be inserted but the iterator loses its current state when doing
 * so and starts from beginning.
 */
public class MutatingStringIterator implements Iterator<String> {

	private boolean hasNext;

	/**
	 * Current column progress indexes for each matrix row.
	 */
	private final int[] key;

	/**
	 * Segment mutation information. Each row contains mutations for segment
	 * elements. e.g. if the segment length is 8 then the matrix has 8 rows.
	 * Columns contain alternative characters for the row.
	 */
	private final char[][] matrix;

	private final int rows;

	/**
	 * Constructs a new iterator for fixed sized string mutations.
	 * 
	 * @param rows
	 *            mutated string length
	 */
	public MutatingStringIterator(int rows) {
		this.rows = rows;
		this.matrix = new char[rows][1];
		key = new int[rows];
		hasNext = false;
	}

	/**
	 * Returns true if column first value is non-zero for all rows.
	 * 
	 * @return
	 */
	private boolean checkNext() {
		for (int row = 0; row < rows; row++) {
			if (matrix[row][0] == 0) {
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean hasNext() {
		return hasNext;
	}

	/**
	 * Recursively checks if there are non-considered columns in the matrix.
	 * Method returns immediately when a non-considered entry is found. The
	 * value for the {@link #key} is incremented for each inspected row. Method
	 * returns false only if there are no more rows to consider.
	 * 
	 * @param row
	 * @return
	 */
	private boolean hasNextKey(final int row) {
		if (row >= rows) {
			return false;
		}
		key[row]++;
		char c = 0;
		if (key[row] < matrix[row].length) {
			c = matrix[row][key[row]];
		}
		if (c == 0) {
			key[row] = 0;
			return hasNextKey(row + 1);
		} else {
			return true;
		}
	}

	/**
	 * Inserts a new row at the end. Shifts other rows towards beginning
	 * throwing first row away.
	 * 
	 * @param row
	 */
	public void insertRow(char[] row) {
		int i = 0;
		for (; i < rows - 1; i++) {
			matrix[i] = matrix[i + 1];
		}
		matrix[i] = row;
		hasNext = checkNext();
	}

	@Override
	public String next() {
		if (hasNext) {
			StringBuilder word = new StringBuilder(rows);
			for (int row = 0; row < rows; row++) {
				final char c = matrix[row][key[row]];
				word.append(c);
			}
			hasNext = hasNextKey(0);
			return word.toString();
		}
		throw new NoSuchElementException();
	}

	@Override
	public void remove() {
		throw new UnsupportedOperationException();
	}

}
