/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.Arrays;

import org.jhears.web.rs.Repeat;

/**
 * Represents multiple {@link Repeat} objects.
 */
public class RepeatGroup {

	private final char[] chars;

	private final int position;

	private final int[] values;

	/**
	 * 
	 * @param position
	 *            The position of the first {@link Repeat} object.
	 * @param chars
	 *            Characters of the all repeat objects.
	 * @param values
	 *            The first <code>chars.length</code> values of the array are
	 *            {@link Repeat#getLength()}s and the next
	 *            <code>chars.length - 1</code> values are positions relative to
	 *            first entry position (the value of the first entry is excluded
	 *            since its already given in <code>position</code>).
	 * @throws NullPointerException
	 *             If chars or values is null.
	 * @throws IllegalArgumentException
	 *             If <code>chars.length &lt; 2</code> or
	 *             <code>(values.length * 2 - 1) != chars.length</code>
	 */
	public RepeatGroup(int position, char[] chars, int[] values)
			throws NullPointerException, IllegalArgumentException {
		if (chars.length < 2) {
			throw new IllegalArgumentException(
					"Must have at least two characters");
		}
		if (values.length != (chars.length * 2 - 1)) {
			throw new IllegalArgumentException("Values length must be "
					+ (chars.length * 2 - 1) + " because chars.length="
					+ chars.length + " but was " + values.length);
		}
		this.position = position;
		this.chars = chars;
		this.values = values;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		RepeatGroup other = (RepeatGroup) obj;
		if (!Arrays.equals(chars, other.chars))
			return false;
		if (position != other.position)
			return false;
		if (!Arrays.equals(values, other.values))
			return false;
		return true;
	}

	/**
	 * Characters from the {@link Repeat#getCh()}.
	 * 
	 * @return
	 */
	public char[] getChars() {
		return chars;
	}

	/**
	 * Absolute position of the first character.
	 * 
	 * @return
	 */
	public int getPosition() {
		return position;
	}

	/**
	 * The first <code>chars.length</code> values of the array are
	 * {@link Repeat#getLength()}s and the next <code>chars.length - 1</code>
	 * values are positions relative to first entry position (the value of the
	 * first entry is excluded since its already available via
	 * {@link #getPosition()}).
	 * 
	 * @return
	 */
	public int[] getValues() {
		return values;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(chars);
		result = prime * result + position;
		result = prime * result + Arrays.hashCode(values);
		return result;
	}

	@Override
	public String toString() {
		return RepeatGroup.class.getSimpleName() + " [chars="
				+ Arrays.toString(chars) + ", position=" + position
				+ ", values=" + Arrays.toString(values) + "]";
	}

}
