/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.jhears.web.rs.Repeat;

/**
 * Performs grouping of {@link Repeat} objects into {@link RepeatGroup} objects
 * and vice versa.
 * <p>
 * This class is thread safe.
 */
public class RepeatGrouper {

	private final int repeatsPerGroup;

	/**
	 * 
	 * @param repeatsPerGroup
	 *            The number of {@link Repeat} objects per entry.
	 */
	public RepeatGrouper(int repeatsPerGroup) {
		this.repeatsPerGroup = repeatsPerGroup;
	}

	public int getRepeatsPerGroup() {
		return repeatsPerGroup;
	}

	/**
	 * Expands {@link RepeatGroup}s back to {@link Repeat}s.
	 * 
	 * @param groups
	 *            If collection contains overlapping {@link RepeatGroup}s then
	 *            the returned list will contain duplicates.
	 * @return
	 */
	public List<Repeat> groupToRepeat(Collection<RepeatGroup> groups) {
		int size = groups.size() * repeatsPerGroup;
		List<Repeat> retval = new ArrayList<Repeat>(size);
		for (RepeatGroup rg : groups) {
			retval.addAll(groupToRepeat(rg));
		}
		return retval;
	}

	/**
	 * Expands single {@link RepeatGroup} back to original list of repeats.
	 * 
	 * @param group
	 * @return
	 */
	public List<Repeat> groupToRepeat(RepeatGroup group) {
		List<Repeat> retval = new ArrayList<Repeat>(repeatsPerGroup);
		int position = group.getPosition();
		int[] values = group.getValues();
		char[] chars = group.getChars();
		int len = chars.length;
		for (int i = 0; i < len; i++) {
			int p = position + (i > 0 ? values[len + i - 1] : 0);
			retval.add(new Repeat(chars[i], p, values[i]));
		}
		return retval;
	}

	/**
	 * Extract positions of each {@link Repeat} which give {@link RepeatGroup}
	 * represents.
	 * 
	 * @param group
	 * @return
	 * @see Repeat#getPosition()
	 */
	public int[] positions(RepeatGroup group) {
		int firstPos = group.getPosition();
		int[] retval = new int[repeatsPerGroup];
		retval[0] = firstPos;
		int[] values = group.getValues();
		for (int j = 1; j < repeatsPerGroup; j++) {
			retval[j] = values[j + repeatsPerGroup - 1] + firstPos;
		}
		return retval;
	}

	/**
	 * Transforms list of {@link Repeat} objects to {@link RepeatGroup} objects.
	 * 
	 * @param list
	 * @param overlap
	 *            If true then <code>list.size() - repeatsPerGroup + 1</code>
	 *            entries are returned else <code>list.size()
				/ {@link #getRepeatsPerGroup()}</code> entries are returned.
	 * @return
	 */
	public List<RepeatGroup> repeatToGroup(List<Repeat> list, boolean overlap) {
		if (list.size() < repeatsPerGroup) {
			return new ArrayList<RepeatGroup>(0);
		}
		int size = overlap ? list.size() - repeatsPerGroup + 1 : list.size()
				/ repeatsPerGroup;
		List<RepeatGroup> retval = new ArrayList<RepeatGroup>(size);
		for (int i = 0; i < size; i++) {
			// minus one since we exclude position of the first
			// which is always zero

			// array of lengths and positions
			int[] lenPosArr = new int[repeatsPerGroup * 2 - 1];
			char[] chars = new char[repeatsPerGroup];
			int offset = overlap ? i : i * repeatsPerGroup;
			for (int j = 0; j < repeatsPerGroup; j++) {
				chars[j] = list.get(offset + j).getCh();
				lenPosArr[j] = list.get(offset + j).getLength();
			}
			int firstPos = list.get(offset + 0).getPosition();
			// extract positions relative to first
			for (int j = 0; j < repeatsPerGroup - 1; j++) {
				lenPosArr[j + repeatsPerGroup] = list.get(offset + j + 1)
						.getPosition()
						- firstPos;
			}
			retval.add(new RepeatGroup(firstPos, chars, lenPosArr));
		}
		return retval;
	}

	/*
	 * smooth the sequence.
	 * 
	 * encode to repeats.
	 * 
	 * convert to groups.
	 */
	// public List<RepeatGroup2> repeatToGroup2(List<Repeat> list, boolean
	// overlap) {
	// if (list.size() < repeatsPerGroup) {
	// return new ArrayList<RepeatGroup2>(0);
	// }
	// int size = overlap ? list.size() - repeatsPerGroup + 1 : list.size()
	// / repeatsPerGroup;
	// List<RepeatGroup2> retval = new ArrayList<RepeatGroup2>(size);
	// for (int i = 0; i < size; i++) {
	// int[] lengths = new int[repeatsPerGroup];
	// char[] chars = new char[repeatsPerGroup];
	// int offset = overlap ? i : i * repeatsPerGroup;
	// for (int j = 0; j < repeatsPerGroup; j++) {
	// chars[j] = list.get(offset + j).getCh();
	// lengths[j] = list.get(offset + j).getLength();
	// }
	// retval.add(new RepeatGroup2(chars, lengths));
	// }
	// return retval;
	// }

}
