/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.Segment;

import uk.ac.shef.wit.simmetrics.similaritymetrics.NeedlemanWunch;

/**
 * Checks if the sequence matches given segments with certain confidence.
 */
public final class SeqMatch {

	private static class SubString {
		private final int position;

		private final String str;

		public SubString(String str, int position) {
			this.str = str;
			this.position = position;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			SubString other = (SubString) obj;
			if (position != other.position)
				return false;
			if (str == null) {
				if (other.str != null)
					return false;
			} else if (!str.equals(other.str))
				return false;
			return true;
		}

		public int getPosition() {
			return position;
		}

		public String getStr() {
			return str;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + position;
			result = prime * result + ((str == null) ? 0 : str.hashCode());
			return result;
		}

		@Override
		public String toString() {
			return SubString.class.getSimpleName() + " [position=" + position
					+ ", str=" + str + "]";
		}
	}

	private final NeedlemanWunch nw;

	private final float threshold;

	public SeqMatch(float threshold) {
		this.threshold = threshold;
		nw = new NeedlemanWunch();
	}

	private List<SubString> extractSubStrings(List<Repeat> repeats,
			int relativePosition, MutableFSeq seq) {
		final ArrayList<SubString> slist = new ArrayList<SubString>(repeats
				.size());
		final char[] ca = seq.asString().toCharArray();
		for (final Repeat r : repeats) {
			int start = r.getPosition() + relativePosition;
			int end = start + r.getLength();
			if (end <= 0 || start >= ca.length) {
				// fully out of range
				continue;
			}
			if (start < 0) {
				start = 0; // trim off the missing head
			}
			if (end > ca.length) {
				end = ca.length; // trim off the missing tail
			}
			final String s = new String(Arrays.copyOfRange(ca, start, end));
			slist.add(new SubString(s, start));
		}
		return slist;
	}

	public float repeatsInSeqConfidence(List<Repeat> repeats,
			int relativePosition, MutableFSeq seq) {
		int size = repeats.size();
		if (size == 0) {
			return 0.0f;
		}
		List<Integer> positions = new ArrayList<Integer>(size);
		for (Repeat r : repeats) {
			positions.add(Integer.valueOf(r.getPosition() + relativePosition));
		}
		float multiplier = 1.0f;
		List<SubString> seqSubStrings = extractSubStrings(repeats,
				relativePosition, seq);
		if (seqSubStrings.size() != size) {
			if (seqSubStrings.isEmpty()) {
				return 0.0f;
			}
			if (seqSubStrings.size() < Math.floor(size * threshold)) {
				// not enough matching segments
				// the resulting match cannot never be higher than this.
				multiplier = (float) seqSubStrings.size() / (float) size;
			}
			ArrayList<Repeat> newRepeats = new ArrayList<Repeat>(seqSubStrings
					.size());
			/*
			 * Perform something like: segments.retainAll(seqSegments).
			 * 
			 * Both collections the seqSegments and segments should contain only
			 * equivalent segments in correct order.
			 */
			// TODO optimize, this uses brute force
			for (SubString s : seqSubStrings) {
				for (Repeat o : repeats) {
					if (o.getPosition() + relativePosition == s.getPosition()) {
						newRepeats.add(o);
					}
				}
			}
			repeats = newRepeats;
			size = repeats.size();
		}
		float match = 0.0f;
		for (int i = 0; i < size; i++) {
			char[] charSeq = new char[repeats.get(i).getLength()];
			Arrays.fill(charSeq, repeats.get(i).getCh());
			String string1 = new String(charSeq);
			String string2 = seqSubStrings.get(i).getStr();
			match += nw.getSimilarity(string1, string2);
		}
		match /= size;
		return match * multiplier;
	}

	/**
	 * Returns true if the segments are in the sequence with high enough
	 * confidence.
	 * 
	 * @param segments
	 *            Segments to check for. Typically from database/cache.
	 * @param relativePosition
	 *            Relative match position. Calculated as
	 *            <code>querySegment.getPosition() - cacheSegment.getPosition()</code>
	 *            .
	 * @param seq
	 *            sequence which should contain containing the segments.
	 *            Typically the sequence used in query.
	 * @return
	 */
	public boolean segmentsInSeq(List<Segment> segments, int relativePosition,
			MutableFSeq seq) {
		return segmentsInSeqConfidence(segments, relativePosition, seq) >= threshold;
	}

	public boolean repeatsInSeq(List<Repeat> repeats, int relativePosition,
			MutableFSeq seq) {
		return repeatsInSeqConfidence(repeats, relativePosition, seq) >= threshold;
	}

	/**
	 * Returns approximated confidence that the sequence contains the segments
	 * or most of them.
	 * 
	 * @param segments
	 *            Segments to check for. Typically from database/cache.
	 * @param relativePosition
	 *            Relative match position. Calculated as
	 *            <code>querySegment.getPosition() - cacheSegment.getPosition()</code>
	 *            .
	 * @param seq
	 *            sequence which should contain containing the segments.
	 *            Typically the sequence used in query.
	 * @return
	 */
	public float segmentsInSeqConfidence(List<Segment> segments,
			int relativePosition, MutableFSeq seq) {
		int size = segments.size();
		if (size == 0) {
			return 0.0f;
		}
		List<Integer> positions = new ArrayList<Integer>(size);
		for (Segment s : segments) {
			positions.add(Integer.valueOf(s.getPosition() + relativePosition));
		}
		float multiplier = 1.0f;
		List<Segment> seqSegments = seq.getSegmentsFromPositions(positions);
		if (seqSegments.size() != size) {
			if (seqSegments.isEmpty()) {
				return 0.0f;
			}
			if (seqSegments.size() < Math.floor(size * threshold)) {
				// not enough matching segments
				// the resulting match cannot never be higher than this.
				multiplier = (float) seqSegments.size() / (float) size;
			}
			ArrayList<Segment> newSegments = new ArrayList<Segment>(seqSegments
					.size());
			/*
			 * Perform something like: segments.retainAll(seqSegments).
			 * 
			 * Both collections the seqSegments and segments should contain only
			 * equivalent segments in correct order.
			 */
			// TODO optimize, this uses brute force
			for (Segment s : seqSegments) {
				for (Segment o : segments) {
					if (o.getPosition() + relativePosition == s.getPosition()) {
						newSegments.add(o);
					}
				}
			}
			segments = newSegments;
			size = segments.size();
		}
		float match = 0.0f;
		for (int i = 0; i < size; i++) {
			String string1 = segments.get(i).getData();
			String string2 = seqSegments.get(i).getData();
			match += nw.getSimilarity(string1, string2);
		}
		match /= size;
		return match * multiplier;
	}

}
