/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.seq;

import java.util.Arrays;

import org.jhears.web.rs.Repeat;

/**
 * Fills "bumps" in otherwise repetitive sequence of characters. e.g.
 * "123aaaaaXaaaaa123" -> "123aaaaaaaaaaa123".
 * <p>
 * This class is thread safe as long as the character sequence which is
 * processed is not shared with multiple threads.
 */
public final class SequenceSmoother implements ISmoother {

	/**
	 * Performs the actual smoothing and provides thread safety by wrapping
	 * state tracking variables.
	 */
	private class Smoother {

		private final char[] ca;

		private char previous;

		private Repeat prevRepeat;

		private int sameCnt;

		public Smoother(char[] ca) {
			this.ca = ca;
		}

		public void process() {
			final int len = ca.length;
			if (len == 0) {
				return; // nothing to do
			}

			init();

			int i = 1;
			for (; i < len; i++) {
				if (ca[i] == previous) {
					sameCnt++;
				} else {
					processGap(i - sameCnt);
					previous = ca[i];
					sameCnt = 1;
				}
			}
			processGap(i - sameCnt);
		}

		public void init() {
			previous = ca[0];
			sameCnt = 1;
			prevRepeat = null;
		}

		/**
		 * Checks if there is long enough repetition and tries to fill the gap
		 * between previous repetition.
		 * 
		 * @param minSame
		 * @param maxGap
		 * @param ca
		 * @param repeatStartPos
		 */
		private void processGap(int repeatStartPos) {
			if (sameCnt >= minSame) {
				Repeat repeat = new Repeat(previous, repeatStartPos, sameCnt);
				if (prevRepeat != null) {
					tryToFillGap(prevRepeat, repeat);
				}
				prevRepeat = repeat;
			}
		}

		/**
		 * Fills a gap between two repetitions if the gap is less than the
		 * maxGap or if the gap minus repeated characters within the gap less
		 * than the maxGap.
		 * 
		 * @param maxGap
		 *            maximum allowed gap length. if the gap is longer than this
		 *            then the gap will not be filled.
		 * @param ca
		 *            array which contains the repetitions
		 * @param prevRepeat
		 *            previously found repeat
		 * @param repeat
		 *            current repeat
		 */
		private void tryToFillGap(Repeat prevRepeat, Repeat repeat) {
			int prevEndIdx = prevRepeat.getPosition() + prevRepeat.getLength();
			int gapLength = repeat.getPosition() - prevEndIdx;
			// gap can be longer than the maxGap if it contains repeated
			// characters
			int charsInGap = SequenceSmoother.countChars(ca, repeat.getCh(),
					prevEndIdx, repeat.getPosition());
			if (repeat.getCh() == prevRepeat.getCh()
					&& gapLength - charsInGap <= maxGap) {
				Arrays.fill(ca, prevEndIdx, prevEndIdx + gapLength, prevRepeat
						.getCh());
			}
		}

	}

	/**
	 * Count occurrences of a character in array.
	 * 
	 * @param ca
	 *            array to search for
	 * @param ch
	 *            character to search for
	 * @param beginIndex
	 *            start index (inclusive).
	 * @param endIndex
	 *            end index (exclusive).
	 * @throws NullPointerException
	 *             if ca is null.
	 * @throws ArrayIndexOutOfBoundsException
	 *             if beginIndex < ca.length or endIndex > ca.length.
	 * @return
	 */
	public static int countChars(final char[] ca, char ch, int beginIndex,
			int endIndex) throws NullPointerException,
			ArrayIndexOutOfBoundsException {
		int cnt = 0;
		for (int i = beginIndex; i < endIndex; i++) {
			if (ca[i] == ch) {
				cnt++;
			}
		}
		return cnt;
	}

	private final int maxGap;

	private final int minSame;

	/**
	 * 
	 * @param minSame
	 *            The minimum number of the character sequence surrounded by the
	 *            gap on both sides.
	 * @param maxGap
	 *            Maximum allowed length between two repeating sequences. If the
	 *            gap contains one or more character which are the same as in
	 *            the surrounding character sequences then the maximum gap
	 *            length is temporarily increased by the number of the same
	 *            characters.
	 * @throws IllegalArgumentException
	 *             if minSame or maxGap is less than one.
	 */
	public SequenceSmoother(final int minSame, final int maxGap)
			throws IllegalArgumentException {
		if (minSame < 1) {
			throw new IllegalArgumentException("minSame must greater than zero");
		}
		if (maxGap < 1) {
			throw new IllegalArgumentException("maxGap must greater than zero");
		}
		this.minSame = minSame;
		this.maxGap = maxGap;
	}

	/**
	 * Replaces character occurrences which differ from are surrounding
	 * repeating character sequences.
	 * 
	 * Examples where the minSame = 4 and maxGap = 2:
	 * 
	 * <pre>
	 * "123aaaaXaaaa456" -&gt; "123aaaaaaaaa456"
	 * "123aaaaXXaaaa456" -&gt; "123aaaaaaaaaa456"
	 * "123aaaaXaXaaaa456" -&gt; "123aaaaaaaaaaa456"
	 * "123aaaaXaaXaaaa456" -&gt; "123aaaaaaaaaaaa456"
	 * "123aaaaXaaaXaaaa456" -&gt; "123aaaaaaaaaaaaa456"
	 * 
	 * "123aaaXaaaa456" -&gt; "123aaaXaaaa456"
	 * "123aaaaXaaa456" -&gt; "123aaaaXaaa456"
	 * "123aaaaXXXaaaa456" -&gt; "123aaaaXXXaaaa456"
	 * "123aaaaXXaXaaaa456" -&gt; "123aaaaXXaXaaaa456"
	 * "123aaaaXXaaXaaaa456" -&gt; "123aaaaXXaaXaaaa456"
	 * "123aaaaXaXaXaaaa456" -&gt; "123aaaaXaXaXaaaa456"
	 * "123aaaaXaaaXXaaaa456" -&gt; "123aaaaXaaaXXaaaa456"
	 * 
	 * </pre>
	 * 
	 * @param ca
	 *            Array to be smoothed.
	 * @throws NullPointerException
	 *             If ca is null.
	 */
	@Override
	public void smooth(final char[] ca) throws NullPointerException {
		new Smoother(ca).process();
	}

}
