/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.workpool;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.jhears.concurrent.Executors;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class has a fixed number of workers. Workers take work from a queue and
 * execute it. Queue size is the same as number of workers. That is, the class
 * can hold n*2 units of work at most (n units in execution and another n in
 * queue).
 * <p>
 * TODO this class needs some cleanup
 */
public class WorkerPool<T extends Work> {

	private class Worker implements Runnable {

		@Override
		public void run() {
			T work = null;
			while (true) {
				try {
					work = queue.take();
					if (work == poison) {
						if (logger.isDebugEnabled()) {
							logger.debug("Worker is poisoned");
						}
						return;
					}
					if (logger.isTraceEnabled()) {
						logger.trace("Worker starts processing " + work);
					}
					work.doWork();
				} catch (Exception e) {
					logger.error("Unhandled Exception while executing " + work,
							e);
				}
			}
		}
	}

	private final ExecutorService executor;

	private final Logger logger;

	private final T poison;

	private final LinkedBlockingQueue<T> queue;

	private final int threads;

	public WorkerPool(final int threads, final T poison) {
		this.logger = LoggerFactory.getLogger(getClass());
		if (threads < 1) {
			throw new IllegalArgumentException(
					"number of threads must be at least one");
		}
		if (poison == null) {
			throw new NullPointerException("Poison cannot be null");
		}
		this.threads = threads;
		queue = new LinkedBlockingQueue<T>(threads);
		executor = Executors.newFixedBoundedThreadPool(threads);
		this.poison = poison;
		for (int i = 0; i < threads; i++) {
			executor.execute(new Worker());
		}
	}

	/**
	 * Adds work into queue. Blocks if there are no threads available.
	 * 
	 * @param work
	 * @throws NullPointerException
	 *             if the work is null
	 * @throws InterruptedException
	 * @throws IllegalArgumentException
	 *             if the work is the same as the poison given in constructor
	 */
	public void addWork(final T work) throws InterruptedException,
			NullPointerException, IllegalArgumentException {
		if (work == poison) {
			throw new IllegalArgumentException("Cannot add poison as a work");
		}
		queue.put(work);
	}

	/**
	 * Waits until all the work is done.
	 * <p>
	 * Calls {@link #addWork(Work, long, TimeUnit)} for each thread with the
	 * poison given in the constructor.
	 * <p>
	 * Total maximum timeout value is: timeout * (threads + 1). Note that the
	 * the timeout should be the same that the user would user for the
	 * {@link #addWork(Work, long, TimeUnit)} method.
	 * 
	 * @param timeout
	 * @param unit
	 * @return true if didn't time out during the stop procedure.
	 * @throws InterruptedException
	 */
	public boolean stop(final long timeout, final TimeUnit unit)
			throws InterruptedException {
		if (logger.isDebugEnabled()) {
			logger.debug("pool is stopping");
		}
		executor.shutdown();
		boolean retval = true;
		for (int i = 0; i < threads; i++) {
			if (!queue.offer(poison, timeout, unit)) {
				if (logger.isDebugEnabled()) {
					logger.debug("Queue poisoning timed out");
				}
				retval = false;
			}
		}
		if (!executor.awaitTermination(timeout, unit)) {
			if (logger.isDebugEnabled()) {
				logger
						.debug("Timed out while waiting for executor service terminate");
			}
			executor.shutdownNow();
			retval = false;
		}
		return retval;
	}

}
