/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.math;

import java.util.Arrays;

import org.jhears.util.Mth;
import org.junit.Assert;
import org.junit.Test;

public class SmoothTest {

	private static final int X = 11;

	@Test
	public void emptyNull() {
		double[] data = null;
		for (int i = 0; i < X; i++) {
			Assert.assertNull(Mth.simpleMovingAverage(data, i));
		}
	}

	@Test
	public void emptyEmpty() {
		double[] data = new double[0];
		for (int i = 0; i < X; i++) {
			Assert.assertArrayEquals(new double[0], Mth.simpleMovingAverage(
					data, i), 0.0);
		}
	}

	@Test
	public void emptySingleElement() {
		double[] data = new double[] { 7 };
		double[] expected = Arrays.copyOf(data, data.length);
		for (int i = 0; i < X; i++) {
			Assert.assertArrayEquals(expected,
					Mth.simpleMovingAverage(data, i), 0.0);
		}
	}

	@Test
	public void emptyTwoElements() {
		double[] data = new double[] { 7, 11 };
		double[] expected = Arrays.copyOf(data, data.length);
		for (int i = 0; i < X; i++) {
			Assert.assertArrayEquals(expected,
					Mth.simpleMovingAverage(data, i), 0.0);
		}
	}

	@Test
	public void threeElementsNeq0() {
		double[] data = new double[] { 7, 11, 13 };
		double[] expected = Arrays.copyOf(data, data.length);
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 0),
				0.0);
	}

	@Test
	public void threeElementsNgt0() {
		double[] data = new double[] { 7, 11, 13 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		for (int i = 1; i < X; i++) {
			Assert.assertArrayEquals("i=" + i, expected, Mth
					.simpleMovingAverage(data, i), eps);
		}
	}

	private static final double eps = 0.0001;

	@Test
	public void fourElementsNeq0() {
		double[] data = new double[] { 7, 11, 13, 17 };
		double[] expected = Arrays.copyOf(data, data.length);
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 0),
				0.0);
	}

	@Test
	public void fourElementsNgt0() {
		double[] data = new double[] { 7, 11, 13, 17 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (11 + 13 + 17) / 3.0;
		for (int i = 1; i < X; i++) {
			Assert.assertArrayEquals("i=" + i, expected, Mth
					.simpleMovingAverage(data, i), eps);
		}
	}

	@Test
	public void fiveElementsNeq0() {
		double[] data = new double[] { 7, 11, 13, 17, 19 };
		double[] expected = Arrays.copyOf(data, data.length);
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 0),
				0.0);
	}

	@Test
	public void fiveElementsNeq1() {
		double[] data = new double[] { 7, 11, 13, 17, 19 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (11 + 13 + 17) / 3.0;
		expected[3] = (13 + 17 + 19) / 3.0;
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 1),
				eps);
	}

	@Test
	public void fiveElementsNgt1() {
		double[] data = new double[] { 7, 11, 13, 17, 19 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (7 + 11 + 13 + 17 + 19) / 5.0;
		expected[3] = (13 + 17 + 19) / 3.0;
		for (int i = 2; i < X; i++) {
			Assert.assertArrayEquals("i=" + i, expected, Mth
					.simpleMovingAverage(data, i), eps);
		}
	}

	@Test
	public void sixElementsNeq0() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23 };
		double[] expected = Arrays.copyOf(data, data.length);
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 0),
				0.0);
	}

	@Test
	public void sixElementsNeq1() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (11 + 13 + 17) / 3.0;
		expected[3] = (13 + 17 + 19) / 3.0;
		expected[4] = (17 + 19 + 23) / 3.0;
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 1),
				eps);
	}

	@Test
	public void sixElementsNgt1() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (7 + 11 + 13 + 17 + 19) / 5.0;
		expected[3] = (11 + 13 + 17 + 19 + 23) / 5.0;
		expected[4] = (17 + 19 + 23) / 3.0;
		for (int i = 2; i < X; i++) {
			Assert.assertArrayEquals("i=" + i, expected, Mth
					.simpleMovingAverage(data, i), eps);
		}
	}

	@Test
	public void sevenElementsNeq0() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23, 29 };
		double[] expected = Arrays.copyOf(data, data.length);
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 0),
				0.0);
	}

	@Test
	public void sevenElementsNeq1() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23, 29 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (11 + 13 + 17) / 3.0;
		expected[3] = (13 + 17 + 19) / 3.0;
		expected[4] = (17 + 19 + 23) / 3.0;
		expected[5] = (19 + 23 + 29) / 3.0;
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 1),
				eps);
	}

	@Test
	public void sevenElementsNeq2() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23, 29 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (7 + 11 + 13 + 17 + 19) / 5.0;
		expected[3] = (11 + 13 + 17 + 19 + 23) / 5.0;
		expected[4] = (13 + 17 + 19 + 23 + 29) / 5.0;
		expected[5] = (19 + 23 + 29) / 3.0;
		Assert.assertArrayEquals(expected, Mth.simpleMovingAverage(data, 2),
				eps);
	}

	@Test
	public void sevenElementsNgt2() {
		double[] data = new double[] { 7, 11, 13, 17, 19, 23, 29 };
		double[] expected = Arrays.copyOf(data, data.length);
		expected[1] = (7 + 11 + 13) / 3.0;
		expected[2] = (7 + 11 + 13 + 17 + 19) / 5.0;
		expected[3] = (7 + 11 + 13 + 17 + 19 + 23 + 29) / 7.0;
		expected[4] = (13 + 17 + 19 + 23 + 29) / 5.0;
		expected[5] = (19 + 23 + 29) / 3.0;
		for (int i = 3; i < X; i++) {
			Assert.assertArrayEquals("i=" + i, expected, Mth
					.simpleMovingAverage(data, i), eps);
		}
	}

}
