/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.Iterator;
import java.util.Map;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.Map.Entry;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.apache.commons.lang.ArrayUtils;
import org.jhears.util.HashFunction;
import org.jhears.util.IterationObserver;
import org.jhears.util.Serialization;

import com.sleepycat.je.Cursor;
import com.sleepycat.je.CursorConfig;
import com.sleepycat.je.Database;
import com.sleepycat.je.DatabaseConfig;
import com.sleepycat.je.DatabaseEntry;
import com.sleepycat.je.LockMode;
import com.sleepycat.je.OperationStatus;
import com.sleepycat.je.Transaction;

public class BerkeleyMap<K, V> implements DatabaseMap<K, V> {

	private static interface CursorOp {

		void with(Cursor cursor);

	}

	private static final String ISE_MSG = "Map cannot be accessed multiple times within same thread.";

	/** Protects the {@link #clear()} method. */
	private final ReentrantReadWriteLock clearLock;

	private Database database;

	private final String databaseName;

	private final HashFunction<K> function;

	/** BDB supports only one active transaction in your thread at a time. */
	private final ThreadLocal<Boolean> guard;

	private final BerkeleyMapManager manager;

	private final Serialization serialization;

	public BerkeleyMap(BerkeleyMapManager manager, String databaseName,
			DatabaseConfig config, Serialization serialization,
			HashFunction<K> hashFunction) {
		this.manager = manager;
		this.databaseName = databaseName;
		this.clearLock = new ReentrantReadWriteLock();
		this.database = manager.openDatabase(databaseName,
				config == null ? BerkeleyMapManager.getDefaultDatabaseConfig()
						: config);
		this.serialization = serialization;
		function = hashFunction;
		guard = new ThreadLocal<Boolean>() {
			@Override
			protected Boolean initialValue() {
				return Boolean.FALSE;
			}
		};
	}

	/**
	 * Closes the database, truncates it and finally opens it again.
	 * <p>
	 * {@inheritDoc}
	 */
	@Override
	public void clear() {
		clearLock.writeLock().lock();
		try {
			manager.closeDatabase(database);
			manager.truncateDatabase(databaseName);
			database = manager.openDatabase(databaseName);
		} finally {
			clearLock.writeLock().unlock();
		}
	}

	private void cursorOperation(CursorOp cOp) {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			clearLock.readLock().lock();
			guard.set(Boolean.TRUE);
			Cursor cursor = null;
			CursorConfig cursorConfig = new CursorConfig();
			cursorConfig.setReadUncommitted(true);
			Transaction txn = database.getEnvironment().beginTransaction(null,
					null);
			try {
				cursor = database.openCursor(txn, cursorConfig);
				if (cursor != null) {
					cOp.with(cursor);
					cursor.close();
				}
				txn.commit();
			} catch (RuntimeException e) {
				if (cursor != null) {
					cursor.close();
				}
				txn.abort();
				throw e;
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public V get(K key) {
		DatabaseEntry keyEntry = toDatabaseEntry(key);
		DatabaseEntry dataEntry = new DatabaseEntry();
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			if (OperationStatus.SUCCESS == database.get(null, keyEntry,
					dataEntry, null)) {
				return serialization.<V>deserialize(dataEntry.getData());
			} else {
				return null;
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public void getAll(final Iterator<K> iterator,
			final IterationObserver<V> observer) {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			Transaction txn = database.getEnvironment().beginTransaction(null,
					null);
			try {
				while (iterator.hasNext() && observer.observeNext()) {
					K key = iterator.next();
					DatabaseEntry keyEntry = toDatabaseEntry(key);
					DatabaseEntry dataEntry = new DatabaseEntry();
					if (database.get(txn, keyEntry, dataEntry, null) == OperationStatus.SUCCESS) {
						final V v = serialization.<V>deserialize(dataEntry
								.getData());
						observer.observe(v);
					} else {
						observer.observe(null);
					}
				}
				txn.commit();
			} catch (RuntimeException e) {
				txn.abort();
				throw e;
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public void iterateKeys(final IterationObserver<K> observer) {
		cursorOperation(new CursorOp() {

			@Override
			public void with(Cursor cursor) {
				DatabaseEntry foundKey = new DatabaseEntry();
				DatabaseEntry foundData = new DatabaseEntry();
				// prevent data from being fetched
				foundData.setPartial(0, 0, true);
				while (cursor.getNext(foundKey, foundData,
						LockMode.READ_UNCOMMITTED) == OperationStatus.SUCCESS
						&& observer.observeNext()) {
					observer.observe(toKey(foundKey));
				}
			}
		});
	}

	@Override
	public void iterateMapEntries(final IterationObserver<Entry<K, V>> observer) {
		cursorOperation(new CursorOp() {

			@Override
			public void with(Cursor cursor) {
				DatabaseEntry foundKey = new DatabaseEntry();
				DatabaseEntry foundData = new DatabaseEntry();
				while (cursor.getNext(foundKey, foundData,
						LockMode.READ_UNCOMMITTED) == OperationStatus.SUCCESS
						&& observer.observeNext()) {
					observer.observe(toMapEntry(foundKey, foundData));
				}
			}

		});
	}

	@Override
	public void put(K key, V value) {
		DatabaseEntry keyEntry = toDatabaseEntry(key);
		DatabaseEntry dataEntry = new DatabaseEntry(serialization
				.serialize(value));
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			if (database.put(null, keyEntry, dataEntry) != OperationStatus.SUCCESS) {
				throw new TransactionException();
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public void putAll(Iterator<Map.Entry<K, V>> iterator) {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			Transaction txn = database.getEnvironment().beginTransaction(null,
					null);
			try {
				while (iterator.hasNext()) {
					Map.Entry<K, V> entry = iterator.next();
					DatabaseEntry keyEntry = toDatabaseEntry(entry.getKey());
					DatabaseEntry dataEntry = new DatabaseEntry(serialization
							.serialize(entry.getValue()));
					if (database.put(txn, keyEntry, dataEntry) != OperationStatus.SUCCESS) {
						txn.abort();
						throw new TransactionException();
					}
				}
				txn.commit();
			} catch (RuntimeException e) {
				txn.abort();
				throw e;
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public boolean remove(K key) {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			DatabaseEntry keyEntry = toDatabaseEntry(key);
			return OperationStatus.SUCCESS == database.delete(null, keyEntry);
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public void removeAll(Iterator<K> iterator) {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			Transaction txn = database.getEnvironment().beginTransaction(null,
					null);
			try {
				while (iterator.hasNext()) {
					DatabaseEntry keyEntry = toDatabaseEntry(iterator.next());
					database.delete(txn, keyEntry);
				}
				txn.commit();
			} catch (RuntimeException e) {
				txn.abort();
				throw e;
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	@Override
	public long size() {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		try {
			guard.set(Boolean.TRUE);
			clearLock.readLock().lock();
			return database.count();
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	/**
	 * Same as {@link #clear()} but with following exceptions:
	 * <ul>
	 * <li>Database stays open during the clear operation.
	 * <li>put and get operations are allowed during this method.
	 * <li>Fast if database is very small. Performs in O(n).
	 * </ul>
	 */
	public void slowClear() {
		if (guard.get().booleanValue()) {
			throw new IllegalStateException(ISE_MSG);
		}
		clearLock.readLock().lock();
		try {
			guard.set(Boolean.TRUE);
			Cursor cursor = null;
			CursorConfig cursorConfig = new CursorConfig();
			cursorConfig.setReadUncommitted(true);
			Transaction txn = database.getEnvironment().beginTransaction(null,
					null);
			try {
				cursor = database.openCursor(txn, cursorConfig);
				DatabaseEntry foundKey = new DatabaseEntry();
				DatabaseEntry foundData = new DatabaseEntry();
				// prevent data from being fetched
				foundData.setPartial(0, 0, true);
				if (cursor != null) {
					while (cursor.getNext(foundKey, foundData,
							LockMode.READ_UNCOMMITTED) == OperationStatus.SUCCESS) {
						cursor.delete();
					}
					cursor.close();
				}
				txn.commit();
			} catch (RuntimeException e) {
				if (cursor != null) {
					cursor.close();
				}
				txn.abort();
				throw e;
			}
		} finally {
			guard.set(Boolean.FALSE);
			clearLock.readLock().unlock();
		}
	}

	private DatabaseEntry toDatabaseEntry(K key) {
		byte[] keyHashCodeBytes = LongUtil.makeBytes(function.hash(key));
		byte[] keySerialized = serialization.serialize(key);
		return new DatabaseEntry(ArrayUtils.addAll(keyHashCodeBytes,
				keySerialized));
	}

	private K toKey(DatabaseEntry databaseEntry) {
		byte[] data = databaseEntry.getData();
		return serialization.<K>deserialize(ArrayUtils.subarray(data, 8,
				data.length));
	}

	private Map.Entry<K, V> toMapEntry(DatabaseEntry foundKey,
			DatabaseEntry foundData) {
		final K k = toKey(foundKey);
		final V v = serialization.<V>deserialize(foundData.getData());
		return new SimpleImmutableEntry<K, V>(k, v);
	}

	@Override
	public String toString() {
		return BerkeleyMap.class.getSimpleName() + " [databaseName="
				+ databaseName + "]";
	}
}
