/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import com.sleepycat.je.Database;
import com.sleepycat.je.DatabaseConfig;
import com.sleepycat.je.Durability;
import com.sleepycat.je.Environment;
import com.sleepycat.je.EnvironmentConfig;

public class BerkeleyMapManager {

	public static DatabaseConfig getDefaultDatabaseConfig() {
		DatabaseConfig cnf = new DatabaseConfig();
		cnf.setAllowCreate(true);
		cnf.setTransactional(true);
		cnf.setBtreeComparator(new ByteArrayComparator());
		cnf.setOverrideBtreeComparator(true);
		return cnf;
	}

	public static EnvironmentConfig getDefaultEnvironmentConfig() {
		EnvironmentConfig cnf = new EnvironmentConfig();
		cnf.setAllowCreate(true);
		cnf.setTransactional(true);
		cnf.setDurability(Durability.COMMIT_WRITE_NO_SYNC);
		return cnf;
	}

	private final Environment dbEnv;

	private final Map<String, Database> openDatabases;

	public BerkeleyMapManager(File envHome) {
		this(envHome, null);
	}

	public BerkeleyMapManager(File envHome, EnvironmentConfig envConfig) {
		EnvironmentConfig cnf;
		if (envConfig == null) {
			cnf = getDefaultEnvironmentConfig();
		} else {
			cnf = envConfig.clone();
		}
		dbEnv = new Environment(envHome, cnf);
		openDatabases = new HashMap<String, Database>();
	}

	public void closeDatabase(String databaseName) {
		synchronized (openDatabases) {
			Database database = openDatabases.remove(databaseName);
			if (database != null) {
				database.close();
			}
		}
	}

	public void closeDatabase(Database database) {
		synchronized (openDatabases) {
			// calling close() will set name to null...
			String name = database.getDatabaseName();
			database.close();
			openDatabases.remove(name);
		}
	}

	public Environment getEnv() {
		return dbEnv;
	}

	private Database open(String databaseName, DatabaseConfig dbConfig) {
		DatabaseConfig cnf;
		if (dbConfig == null) {
			cnf = getDefaultDatabaseConfig();
		} else {
			cnf = dbConfig.clone();
		}
		return dbEnv.openDatabase(null, databaseName, cnf);
	}

	public Database openDatabase(String databaseName) {
		return openDatabase(databaseName, null);
	}

	public Database openDatabase(String databaseName, DatabaseConfig dbConfig) {
		Database database;
		synchronized (openDatabases) {
			database = openDatabases.get(databaseName);
			if (database == null) {
				database = open(databaseName, dbConfig);
				openDatabases.put(databaseName, database);
			}
		}
		return database;
	}

	public void removeDatabase(String databaseName) {
		dbEnv.removeDatabase(null, databaseName);
	}

	public void shutDown() {
		if (dbEnv != null) {
			synchronized (openDatabases) {
				for (Database database : openDatabases.values()) {
					database.close();
				}
				openDatabases.clear();
			}
			dbEnv.cleanLog(); // Clean the log before closing
			dbEnv.close();
		}
	}

	public void truncateDatabase(String databaseName) {
		synchronized (openDatabases) {
			dbEnv.truncateDatabase(null, databaseName, false);
		}
	}

}
