/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.Serializable;
import java.util.Comparator;

/**
 * Compares byte arrays which length must be at least eight bytes. The first
 * eight bytes are treated as big endian encoded long values and the primary
 * comparison method is to use the long values.
 * <p>
 * If long values are equal then the remaining bytes are compared one by one as
 * long as both byte arrays have enough bytes.
 * <p>
 * If previous comparisons didn't resolve the issue then lengths of the arrays
 * are compared.
 * <p>
 * This comparator can return zero only if array contents are identical.
 * 
 * @author Juha Heljoranta
 * 
 */
public class ByteArrayComparator implements Comparator<byte[]>, Serializable {

	private static final long serialVersionUID = -1504913081459803788L;

	@Override
	public int compare(final byte[] o1, final byte[] o2) {
		final int len1 = o1.length;
		final int len2 = o2.length;
		final int len = len1 < len2 ? len1 : len2;
		final long v1 = LongUtil.makeLong(o1);
		final long v2 = LongUtil.makeLong(o2);
		if (v1 < v2) {
			return -1;
		} else if (v1 > v2) {
			return 1;
		}
		for (int i = 8; i < len; i++) {
			if (o1[i] < o2[i]) {
				return -1;
			} else if (o1[i] > o2[i]) {
				return 1;
			}
		}
		if (len1 < len2) {
			return -1;
		} else if (len1 > len2) {
			return 1;
		}
		return 0;
	}

}
