/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;

/**
 * Implements transaction idiom:
 * 
 * <pre>
 * Session s = provider.openSession();
 * Transaction tx = null;
 * try {
 * 	tx = s.beginTransaction();
 * 	value = withOpen(s); // do some work
 * 	tx.commit();
 * } catch (RuntimeException e) {
 * 	if (tx != null)
 * 		tx.rollback();
 * 	throw new RuntimeException(e);
 * } finally {
 * 	s.close();
 * }
 * </pre>
 * 
 * @see GeneralDao
 * @param <E>
 */
public class DAO {

	private final SessionProvider provider;

	public DAO(SessionProvider provider) {
		this.provider = provider;
	}

	/**
	 * Calls {@link Transactional#with(Session, Object)}.
	 * 
	 * @param transactional
	 * @param arg
	 *            Argument passed through to
	 *            {@link Transactional#with(Session, Object)}.
	 * 
	 * @throws TransactionException
	 *             If a RuntimeException exception occurs during transaction and
	 *             the exception is is not instance of
	 *             {@link HibernateException}. Use
	 *             {@link TransactionException#getCause()} to access the
	 *             original exception.
	 * @throws HibernateException
	 *             Exceptions thrown by hibernate are re-thrown after rollback.
	 * @return return value of the {@link Transactional#with(Session, Object)}.
	 */
	public <F, T> T execute(Transactional<F, T> transactional, F arg)
			throws TransactionException, HibernateException {
		Session s = provider.openSession();
		try {
			Transaction tx = null;
			try {
				tx = s.beginTransaction();
				T value = transactional.with(s, arg); // do some work
				tx.commit();
				return value;
			} catch (HibernateException e) {
				if (tx != null) {
					tx.rollback();
				}
				throw e;
			} catch (RuntimeException e) {
				if (tx != null) {
					tx.rollback();
				}
				throw new TransactionException(e);
			}
		} finally {
			if (s != null) {
				s.close();
			}
		}
	}

	public SessionProvider getSessionProvider() {
		return provider;
	}

}
