/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.Iterator;
import java.util.Map;

import org.jhears.util.IterationObserver;

/**
 * Similar to {@link java.util.Map} but uses iterators for bulk operations.
 * 
 * @author Juha Heljoranta
 * 
 * @param <K>
 * @param <V>
 */
public interface DatabaseMap<K, V> {

	/**
	 * Clears the map.
	 */
	public void clear();

	/**
	 * Return value associated to a key or null if map doesn't contain the key.
	 * 
	 * @param key
	 *            Key to lookup.
	 * @return Value associated to a key or null if map doesn't contain the key.
	 */
	public V get(final K key);

	/**
	 * Iterate over map keys using the given observer.
	 * 
	 * @param observer
	 *            Observer used during the iteration.
	 */
	public void iterateKeys(IterationObserver<K> observer);

	public void iterateMapEntries(IterationObserver<Map.Entry<K, V>> observer);

	// /**
	// * Iterate over values between from and to.
	// *
	// * @param observer
	// * @param from
	// * @param to
	// */
	// public void iterateKeys(IterationObserver<K> observer, Object from,
	// Object to);

	/**
	 * Add give key value pair to map. If map already has a value associated to
	 * the key then the value will be replaced.
	 * 
	 * @param key
	 * @param value
	 */
	public void put(final K key, final V value);

	/**
	 * Puts all values returned by the iterator;
	 * 
	 * @param iterator
	 */
	public void putAll(final Iterator<Map.Entry<K, V>> iterator);

	/**
	 * Gets all values returned by the key iterator. Return value associated to
	 * a key or null if map doesn't contain the key.
	 * 
	 * @param iterator
	 * @param observer
	 */
	public void getAll(final Iterator<K> iterator, IterationObserver<V> observer);

	/**
	 * Removes key and associated value from the map.
	 * 
	 * @param key
	 * @return if a key was removed.
	 */
	public boolean remove(final K key);

	/**
	 * Returns the number of keys in this map.
	 * 
	 * @return the number of keys in this map.
	 */
	public long size();

	/**
	 * Removes all values returned by the iterator.
	 * 
	 * @param iterator
	 */
	public void removeAll(Iterator<K> iterator);

}
