/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.hibernate.CacheMode;
import org.hibernate.Criteria;
import org.hibernate.FetchMode;
import org.hibernate.LockMode;
import org.hibernate.Query;
import org.hibernate.ScrollMode;
import org.hibernate.ScrollableResults;
import org.hibernate.Session;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.jhears.tuple.Tuple2;
import org.jhears.tuple.Tuple3;
import org.jhears.tuple.Tuple4;
import org.jhears.tuple.Tuple5;
import org.jhears.tuple.Tuple6;
import org.jhears.util.IterationObserver;

/**
 * Provides convenience methods to perform common tasks with {@link DAO}.
 */
public class GeneralDao implements IGenericDao {

	private static final String count_star = "select count(*) from ";

	private static final int COUNT_STAR_SB_SIZE = 48;

	private static final String count_star2 = " where ";
	private static final String count_star3 = " = :value";
	private static final String delete_post = " o where o.id = :id";
	private static final String delete_pre = "delete ";

	private static void setFetchModes(
			final Map<String, FetchMode> associationPathFetchModes, Criteria c) {
		if (associationPathFetchModes != null) {
			for (Map.Entry<String, FetchMode> entry : associationPathFetchModes
					.entrySet()) {
				c.setFetchMode(entry.getKey(), entry.getValue());
			}
		}
	}

	private final int batchSize;

	private final DAO dao;

	private final Transactional<Tuple3<Class<?>, String, Object>, Long> tCount = new Transactional<Tuple3<Class<?>, String, Object>, Long>() {

		@Override
		public Long with(Session s, Tuple3<Class<?>, String, Object> arg) {
			if (arg.get2() == null) {
				StringBuilder sb = new StringBuilder(COUNT_STAR_SB_SIZE);
				sb.append(count_star).append(arg.get1().getSimpleName());
				return (Long) s.createQuery(sb.toString()).uniqueResult();
			} else {
				StringBuilder sb = new StringBuilder(COUNT_STAR_SB_SIZE * 2);
				sb.append(count_star).append(arg.get1().getSimpleName());
				sb.append(count_star2);
				sb.append(arg.get2());
				sb.append(count_star3);
				return (Long) s.createQuery(sb.toString()).setParameter(
						"value", arg.get3()).uniqueResult();
			}
		}

	};

	private final Transactional<Object, Void> tDelete = new Transactional<Object, Void>() {

		@Override
		public Void with(Session s, Object object) {
			s.delete(object);
			return null;
		}

	};

	private final Transactional<Class<?>, Void> tDeleteAll = new Transactional<Class<?>, Void>() {

		@Override
		public Void with(Session s, Class<?> clazz) {
			s.createQuery("DELETE " + clazz.getSimpleName()).executeUpdate();
			return null;
		}

	};
	private final Transactional<Class<?>, Void> tDeleteAllSafe = new Transactional<Class<?>, Void>() {

		@Override
		public Void with(Session s, Class<?> clazz) {
			String select = "select distinct elt.id from "
					+ clazz.getSimpleName() + " elt";
			ScrollableResults results = null;
			try {
				results = s.createQuery(select).scroll(ScrollMode.FORWARD_ONLY);
				int i = 0;
				while (results.next()) {
					Serializable id = (Serializable) results.get(0);
					s.delete(s.get(clazz, id));
					if (batchSize > 1) {
						if (++i % batchSize != 0) {
							continue;
						}
						i = 0;
					}
					s.flush();
					s.clear();
				}
			} finally {
				if (results != null) {
					results.close();
				}
			}
			return null;
		}

	};

	private final Transactional<Tuple3<Class<?>, String, Object>, Boolean> tDeleteFwp = new Transactional<Tuple3<Class<?>, String, Object>, Boolean>() {

		@Override
		public Boolean with(Session s, Tuple3<Class<?>, String, Object> tuple) {
			Criteria c = s.createCriteria(tuple.get1()).add(
					Restrictions.eq(tuple.get2(), tuple.get3()));
			Object o = c.uniqueResult();
			if (o != null) {
				s.delete(o);
			}
			return Boolean.valueOf(o != null);
		}

	};

	private final Transactional<Tuple2<Class<?>, Serializable>, Boolean> tDeleteHql = new Transactional<Tuple2<Class<?>, Serializable>, Boolean>() {

		@Override
		public Boolean with(Session s, Tuple2<Class<?>, Serializable> tuple) {
			StringBuilder sb = new StringBuilder(COUNT_STAR_SB_SIZE);
			sb.append(delete_pre).append(tuple.get1().getSimpleName()).append(
					delete_post);
			Query q = s.createQuery(sb.toString());
			q.setParameter("id", tuple.get2());
			int deletedEntities = q.executeUpdate();
			return Boolean.valueOf(deletedEntities > 0);
		}

	};

	private final Transactional<Tuple2<Class<?>, Serializable>, Boolean> tDeleteT2 = new Transactional<Tuple2<Class<?>, Serializable>, Boolean>() {

		@Override
		public Boolean with(Session s, Tuple2<Class<?>, Serializable> tuple) {
			Object o = s.get(tuple.get1(), tuple.get2());
			if (o != null) {
				s.delete(o);
			}
			return Boolean.valueOf(o != null);
		}

	};

	private final Transactional<Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>>, Object> tGet = new Transactional<Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>>, Object>() {

		@Override
		public Object with(
				Session s,
				Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>> arg) {
			if (arg.get3() == null) {
				return s.get(arg.get1(), arg.get2());
			} else {
				Criteria c = s.createCriteria(arg.get1()).add(
						Restrictions.idEq(arg.get2())).add(
						Restrictions.eq(arg.get3(), arg.get4()));
				setFetchModes(arg.get5(), c);
				return c.uniqueResult();
			}
		}

	};

	private final Transactional<Class<?>, List<Object>> tGetAll = new Transactional<Class<?>, List<Object>>() {

		@SuppressWarnings("unchecked")
		@Override
		public List<Object> with(Session s, Class<?> clazz) {
			return s.createCriteria(clazz).setResultTransformer(
					Criteria.DISTINCT_ROOT_ENTITY).list();
		}

	};

	private final Transactional<Tuple2<Class<?>, Object[]>, List<Object>> tGetAllIds = new Transactional<Tuple2<Class<?>, Object[]>, List<Object>>() {

		@SuppressWarnings("unchecked")
		@Override
		public List<Object> with(Session s, Tuple2<Class<?>, Object[]> arg) {
			return s.createCriteria(arg.get1()).add(
					Restrictions.in("id", arg.get2())).list();
		}

	};

	private final Transactional<Tuple3<Class<?>, String, Object[]>, List<Object>> tGetAllWithPropertyIn = new Transactional<Tuple3<Class<?>, String, Object[]>, List<Object>>() {

		@SuppressWarnings("unchecked")
		@Override
		public List<Object> with(Session s,
				Tuple3<Class<?>, String, Object[]> arg) {
			return s.createCriteria(arg.get1()).add(
					Restrictions.in(arg.get2(), arg.get3())).list();
		}

	};

	private final Transactional<Tuple4<String, String, Object, Map<String, FetchMode>>, Object> tGetEntityWithProperty = new Transactional<Tuple4<String, String, Object, Map<String, FetchMode>>, Object>() {

		@Override
		public Object with(Session s,
				Tuple4<String, String, Object, Map<String, FetchMode>> arg) {
			Criteria c = s.createCriteria(arg.get1()).add(
					Restrictions.eq(arg.get2(), arg.get3()));
			setFetchModes(arg.get4(), c);
			return c.uniqueResult();
		}

	};

	private final Transactional<Tuple6<Class<?>, String, Object, Integer, Integer, Map<String, FetchMode>>, Object> tGetWithProperty = new Transactional<Tuple6<Class<?>, String, Object, Integer, Integer, Map<String, FetchMode>>, Object>() {

		@Override
		public Object with(
				Session s,
				Tuple6<Class<?>, String, Object, Integer, Integer, Map<String, FetchMode>> arg) {
			Criteria c = s.createCriteria(arg.get1()).add(
					Restrictions.eq(arg.get2(), arg.get3()));
			if (arg.get4() != null && arg.get4().intValue() >= 0) {
				c.setFirstResult(arg.get4().intValue());
			}
			if (arg.get5() != null && arg.get5().intValue() >= 0) {
				c.setMaxResults(arg.get5().intValue());
			}
			setFetchModes(arg.get6(), c);
			return c.uniqueResult();
		}
	};

	private final Transactional<Tuple6<Class<?>, IterationObserver<Object>, String, Object, Integer, Integer>, Void> tIterate = new Transactional<Tuple6<Class<?>, IterationObserver<Object>, String, Object, Integer, Integer>, Void>() {

		@Override
		public Void with(
				Session s,
				Tuple6<Class<?>, IterationObserver<Object>, String, Object, Integer, Integer> arg) {
			IterationObserver<Object> observer = arg.get2();
			ScrollableResults results = null;
			try {
				Criteria criteria = s.createCriteria(arg.get1());
				if (arg.get3() != null) {
					criteria = criteria.add(Restrictions.eq(arg.get3(), arg
							.get4()));
				}
				if (arg.get5() != null && arg.get5().intValue() >= 0) {
					criteria.setFirstResult(arg.get5().intValue());
				}
				if (arg.get6() != null && arg.get6().intValue() >= 0) {
					criteria.setMaxResults(arg.get6().intValue());
				}
				results = criteria.setLockMode(LockMode.READ).setCacheMode(
						CacheMode.IGNORE).scroll(ScrollMode.FORWARD_ONLY);
				while (results.next() && observer.observeNext()) {
					Object elt = results.get(0);
					observer.observe(elt);
					s.evict(elt);
				}
			} finally {
				if (results != null) {
					results.close();
				}
			}
			return null;
		}

	};

	private final Transactional<Object, Object> tMerge = new Transactional<Object, Object>() {

		@Override
		public Object with(Session s, Object object) {
			return s.merge(object);
		}

	};

	private final Transactional<Object, Void> tPersist = new Transactional<Object, Void>() {

		@Override
		public Void with(Session s, Object object) {
			s.persist(object);
			return null;
		}

	};

	private final Transactional<Object, Void> tRefresh = new Transactional<Object, Void>() {

		@Override
		public Void with(Session s, Object object) {
			s.refresh(object);
			return null;
		}

	};

	private final Transactional<Object, Serializable> tSave = new Transactional<Object, Serializable>() {

		@Override
		public Serializable with(Session s, Object object) {
			return s.save(object);
		}

	};

	private final Transactional<Object, Void> tSaveOrUpdate = new Transactional<Object, Void>() {

		@Override
		public Void with(Session s, Object object) {
			s.saveOrUpdate(object);
			return null;
		}

	};

	private final Transactional<Object, Void> tUpdate = new Transactional<Object, Void>() {

		@Override
		public Void with(Session s, Object object) {
			s.update(object);
			return null;
		}

	};

	public GeneralDao(DAO dao) {
		this.dao = dao;
		batchSize = dao.getSessionProvider().getJdbcBatchSize();
	}

	public long count(final Class<?> clazz) {
		return count(clazz, null, null);
	}

	/**
	 * Calls {@link Session#createQuery(String)} with
	 * <code>"select count(*) from "
	 * +clazz.getSimpleName()</code>.
	 * 
	 * @param clazz
	 * @param propertyName
	 * @param propertyValue
	 * @return
	 */
	public long count(final Class<?> clazz, String propertyName,
			Object propertyValue) {
		return dao.execute(tCount, new Tuple3<Class<?>, String, Object>(clazz,
				propertyName, propertyName));
	}

	/**
	 * Calls {@link Session#get(Object)} followed by
	 * {@link Session#delete(Object)}.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param id
	 * @return
	 */
	public <E> boolean delete(final Class<E> clazz, final Serializable id) {
		return dao.execute(tDeleteT2, new Tuple2<Class<?>, Serializable>(clazz,
				id));
	}

	/**
	 * Calls {@link Session#delete(Object)}.
	 * 
	 * @param object
	 */
	public void delete(final Object object) {
		dao.execute(tDelete, object);
	}

	public <E> void deleteAll(final Class<E> clazz) {
		dao.execute(tDeleteAll, clazz);
	}

	/**
	 * 
	 * @param <E>
	 * @param clazz
	 */
	public <E> void deleteAllSafe(final Class<E> clazz) {
		dao.execute(tDeleteAllSafe, clazz);
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} with property name
	 * restriction. Finally calls {@link Session#delete(Object)}.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param propertyName
	 * @param propertyValue
	 * @return true of object with give property was found and deleted.
	 */
	public <E> boolean deleteFirstWithProperty(final Class<E> clazz,
			final String propertyName, final Object propertyValue) {
		return dao.execute(tDeleteFwp, new Tuple3<Class<?>, String, Object>(
				clazz, propertyName, propertyValue));
	}

	/**
	 * Calls {@link Session#createQuery(String)} with
	 * <code>"delete "+clazz.getSimpleName()+" o where o.id = :id"</code>.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param id
	 * @return
	 */
	public <E> boolean deleteHql(final Class<E> clazz, final Serializable id) {
		return dao.execute(tDeleteHql, new Tuple2<Class<?>, Serializable>(
				clazz, id));
	}

	/**
	 * Calls {@link Session#get(Class, Serializable)}.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param id
	 * @return
	 */
	public <E> E get(final Class<E> clazz, final Serializable id) {
		return get(clazz, id, null, null);
	}

	public <E> E get(final Class<E> clazz, final Serializable id,
			final Map<String, FetchMode> associationPathFetchModes) {
		return get(clazz, id, null, associationPathFetchModes);
	}

	/**
	 * Same as {@link #get(Class, Serializable)} but adds additional property
	 * restriction.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param id
	 * @param propertyName
	 * @param propertyValue
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public <E> E get(final Class<E> clazz, final Serializable id,
			final String propertyName, final Object propertyValue) {
		Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>> tuple = new Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>>(
				clazz, id, propertyName, propertyValue, null);
		return (E) dao.execute(tGet, tuple);
	}

	@SuppressWarnings("unchecked")
	public <E> E get(final Class<E> clazz, final Serializable id,
			final String propertyName, final Object propertyValue,
			final Map<String, FetchMode> associationPathFetchModes) {
		Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>> tuple = new Tuple5<Class<?>, Serializable, String, Object, Map<String, FetchMode>>(
				clazz, id, propertyName, propertyValue,
				associationPathFetchModes);
		return (E) dao.execute(tGet, tuple);
	}

	@SuppressWarnings("unchecked")
	public <E> List<E> getAll(final Class<E> clazz) {
		return (List<E>) dao.execute(tGetAll, clazz);
	}

	/**
	 * Delegates to {@link #getAll(Class, Object[])}.
	 * 
	 * @see #getAll(Class, Object[])
	 */
	public <E> List<E> getAll(final Class<E> clazz,
			final Collection<? extends Serializable> ids) {
		return getAll(clazz, ids.toArray());
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} with id restriction.
	 * 
	 * @param <E>
	 * @param class1
	 * @param id
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public <E> List<E> getAll(final Class<E> clazz, final Object[] ids) {
		return (List<E>) dao.execute(tGetAllIds,
				new Tuple2<Class<?>, Object[]>(clazz, ids));
	}

	@SuppressWarnings("unchecked")
	public <E> List<E> getAllWithPropertyIn(final Class<E> clazz,
			final String propertyName, final Object[] ids) {
		return (List<E>) dao
				.execute(tGetAllWithPropertyIn,
						new Tuple3<Class<?>, String, Object[]>(clazz,
								propertyName, ids));
	}

	public DAO getDao() {
		return dao;
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} with property name
	 * restriction.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param propertyName
	 * @param propertyValue
	 * @return
	 */
	public <E> E getWithProperty(final Class<E> clazz,
			final String propertyName, final Object propertyValue) {
		return getWithProperty(clazz, propertyName, propertyValue, null);
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} with property name
	 * restriction.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param propertyName
	 * @param propertyValue
	 * @param associationPathFetchModes
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public <E> E getWithProperty(final Class<E> clazz,
			final String propertyName, final Object propertyValue,
			final Map<String, FetchMode> associationPathFetchModes) {
		return (E) dao
				.execute(
						tGetWithProperty,
						new Tuple6<Class<?>, String, Object, Integer, Integer, Map<String, FetchMode>>(
								clazz, propertyName, propertyValue, null, null,
								associationPathFetchModes));

	}

	/**
	 * Calls {@link Session#createCriteria(Class)} with property name
	 * restriction.
	 * 
	 * @param <E>
	 * @param entityName
	 * @param propertyName
	 * @param propertyValue
	 * @return
	 */
	public <E> E getWithProperty(final String entityName,
			final String propertyName, final Object propertyValue) {
		return this.<E>getWithProperty(entityName, propertyName, propertyValue, null);
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} with property name
	 * restriction.
	 * 
	 * @param <E>
	 * @param entityName
	 * @param propertyName
	 * @param propertyValue
	 * @param associationPathFetchModes
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public <E> E getWithProperty(final String entityName,
			final String propertyName, final Object propertyValue,
			final Map<String, FetchMode> associationPathFetchModes) {
		return (E) dao.execute(tGetEntityWithProperty,
				new Tuple4<String, String, Object, Map<String, FetchMode>>(
						entityName, propertyName, propertyValue,
						associationPathFetchModes));
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} and scrolls through the
	 * results.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param observer
	 *            Observer which is notified for each result element.
	 */
	public <E> void iterate(final Class<E> clazz,
			final IterationObserver<E> observer) {
		iterateWithProperty(clazz, observer, null, null);
	}

	/**
	 * Same as {@link #iterate(Class, IterationObserver)} but supports result
	 * paging.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param observer
	 * @param offset
	 *            Offset from where the iterator starts. Negative value is
	 *            treated as 0.
	 * @param maxResults
	 *            The maximum number of results to return. Negative values are
	 *            is ignored.
	 */
	public <E> void iterate(final Class<E> clazz,
			final IterationObserver<E> observer, int offset, int maxResults) {
		iterateWithProperty(clazz, observer, null, null, offset, maxResults);
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} and scrolls through the
	 * results.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param observer
	 *            Observer which is notified for each result element.
	 * @param propertyName
	 * @param propertyValue
	 */
	public <E> void iterateWithProperty(final Class<E> clazz,
			final IterationObserver<E> observer, final String propertyName,
			final Object propertyValue) {
		iterateWithProperty(clazz, observer, null, null, -1, -1);
	}

	/**
	 * Calls {@link Session#createCriteria(Class)} and scrolls through the
	 * results.
	 * 
	 * @param <E>
	 * @param clazz
	 * @param observer
	 *            Observer which is notified for each result element.
	 * @param propertyName
	 * @param propertyValue
	 * @param offset
	 *            Offset from where the iterator starts. Negative value is
	 *            treated as 0.
	 * @param maxResults
	 *            The maximum number of results to return. Negative values are
	 *            is ignored.
	 */
	@SuppressWarnings("unchecked")
	public <E> void iterateWithProperty(final Class<E> clazz,
			final IterationObserver<E> observer, final String propertyName,
			final Object propertyValue, int offet, int maxResults) {
		Tuple6<Class<?>, IterationObserver<Object>, String, Object, Integer, Integer> tuple = new Tuple6<Class<?>, IterationObserver<Object>, String, Object, Integer, Integer>(
				clazz, (IterationObserver<Object>) observer, propertyName,
				propertyValue, Integer.valueOf(offet), Integer
						.valueOf(maxResults));
		dao.execute(tIterate, tuple);
	}

	/**
	 * Calls {@link Session#merge(Object)}.
	 * 
	 * @param object
	 * @return
	 */
	public Object merge(final Object object) {
		return dao.execute(tMerge, object);
	}

	/**
	 * Calls {@link Session#persist(Object)}.
	 * 
	 * @param object
	 */
	public void persist(final Object object) {
		dao.execute(tPersist, object);
	}

	/**
	 * Calls {@link Session#refresh(Object)}.
	 * 
	 * @param object
	 */
	public void refresh(final Object object) {
		dao.execute(tRefresh, object);
	}

	/**
	 * Calls {@link Session#save(Object)}.
	 * 
	 * @param object
	 * @return
	 */
	public Serializable save(final Object object) {
		return dao.execute(tSave, object);
	}

	/**
	 * Calls {@link Session#saveOrUpdate(Object)}.
	 * 
	 * @param object
	 */
	public void saveOrUpdate(final Object object) {
		dao.execute(tSaveOrUpdate, object);
	}

	/**
	 * Calls {@link Session#update(Object)}.
	 * 
	 * @param object
	 */
	public void update(final Object object) {
		dao.execute(tUpdate, object);
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> T get(Class<T> clazz, String propertyName, Object propertyValue) {
		Tuple6<Class<?>, String, Object, Integer, Integer, Map<String, FetchMode>> tuple = new Tuple6<Class<?>, String, Object, Integer, Integer, Map<String, FetchMode>>(
				clazz, propertyName, propertyValue, null, null, null);
		return (T) dao.execute(tGetWithProperty, tuple);
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> List<T> list(Class<T> clazz, int first, int count,
			String propertyName, boolean ascending) {
		Tuple5<Class<?>, Integer, Integer, Boolean, String> tuple = new Tuple5<Class<?>, Integer, Integer, Boolean, String>(
				clazz, Integer.valueOf(first), Integer.valueOf(count), Boolean
						.valueOf(ascending), propertyName);
		return (List<T>) dao.execute(tGetWithPro, tuple);

	}

	private final Transactional<Tuple5<Class<?>, Integer, Integer, Boolean, String>, List<Object>> tGetWithPro = new Transactional<Tuple5<Class<?>, Integer, Integer, Boolean, String>, List<Object>>() {

		@SuppressWarnings("unchecked")
		@Override
		public List<Object> with(Session s,
				Tuple5<Class<?>, Integer, Integer, Boolean, String> arg) {
			Criteria c = s.createCriteria(arg.get1());
			if (arg.get2() != null && arg.get2().intValue() >= 0) {
				c.setFirstResult(arg.get2().intValue());
			}
			if (arg.get3() != null && arg.get3().intValue() >= 0) {
				c.setMaxResults(arg.get3().intValue());
			}
			if (arg.get5() != null && arg.get4() != null) {
				if (arg.get4().booleanValue()) {
					c.addOrder(Order.asc(arg.get5()));
				} else {
					c.addOrder(Order.desc(arg.get5()));
				}
			}
			return c.list();
		}
	};

	@SuppressWarnings("unchecked")
	@Override
	public <T> List<T> list(Class<T> clazz, int first, int count) {
		Tuple5<Class<?>, Integer, Integer, Boolean, String> tuple = new Tuple5<Class<?>, Integer, Integer, Boolean, String>(
				clazz, Integer.valueOf(first), Integer.valueOf(count), null,
				null);
		return (List<T>) dao.execute(tGetWithPro, tuple);
	}
}
