/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.Iterator;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.Map.Entry;

import org.hibernate.LockMode;
import org.hibernate.ScrollMode;
import org.hibernate.ScrollableResults;
import org.hibernate.Session;
import org.jhears.tuple.Tuple2;
import org.jhears.util.HashFunction;
import org.jhears.util.IterationObserver;
import org.jhears.util.Serialization;

public class HibernateMap<K, V> implements DatabaseMap<K, V> {

	private final int batchSize;

	private final DAO dao;

	private final HashFunction<K> keyHashFunction;

	private final Integer mapId;

	private final Serialization serialization;

	private final Transactional<Void, Void> tClear = new Transactional<Void, Void>() {

		@Override
		public Void with(Session s, Void arg) {
			s.getNamedQuery(HBlob.HQL_CLEAR).setInteger(HBlob.pNameMapId,
					mapId.intValue()).executeUpdate();
			return null;
		}
	};

	private final Transactional<K, V> tGet = new Transactional<K, V>() {

		@Override
		public V with(Session s, K key) {
			Long hashCode = Long.valueOf(keyHashFunction.hash(key));
			ScrollableResults sr = null;
			try {
				sr = s.getNamedQuery(HBlob.HQL_GET).setInteger(
						HBlob.pNameMapId, mapId.intValue()).setLong(
						HBlob.pNameHashCode, hashCode.longValue()).scroll(
						ScrollMode.FORWARD_ONLY);
				while (sr.next()) {
					HBlob blob = (HBlob) sr.get(0);
					K k = serialization.<K>deserialize(blob.getKeyBytes());
					if (key.equals(k)) {
						return serialization.<V>deserialize(blob.getValueBytes());
					}
				}
			} finally {
				if (sr != null) {
					sr.close();
				}
			}
			return null;
		}
	};
	
	private final Transactional<Tuple2<Iterator<K>, IterationObserver<V>>, Void> tGetAll = new Transactional<Tuple2<Iterator<K>, IterationObserver<V>>, Void>() {

		@Override
		public Void with(Session s,
				Tuple2<Iterator<K>, IterationObserver<V>> arg) {
			final Iterator<K> itr = arg.get1();
			final IterationObserver<V> o = arg.get2();
			int i = 0;
			while (itr.hasNext() && o.observeNext()) {
				o.observe(tGet.with(s, itr.next()));
				if (batchSize > 1) {
					if (++i % batchSize != 0) {
						continue;
					}
					i = 0;
				}
				s.flush();
				s.clear();
			}
			return null;
		}
	};

	private final Transactional<IterationObserver<K>, Void> tIterateKeys = new Transactional<IterationObserver<K>, Void>() {

		@Override
		public Void with(Session s, IterationObserver<K> observer) {
			ScrollableResults sr = null;
			try {
				sr = s.getNamedQuery(HBlob.HQL_GET_KEYS).setInteger(
						HBlob.pNameMapId, mapId.intValue()).scroll(
						ScrollMode.FORWARD_ONLY);
				while (observer.observeNext() && sr.next()) {
					HBlob blob = (HBlob) sr.get(0);
					K k = serialization.<K>deserialize(blob.getKeyBytes());
					observer.observe(k);
				}
			} finally {
				if (sr != null) {
					sr.close();
				}
			}
			return null;
		}
	};

	private final Transactional<IterationObserver<Entry<K, V>>, Void> tMapEntries = new Transactional<IterationObserver<Entry<K, V>>, Void>() {

		@Override
		public Void with(Session s, IterationObserver<Entry<K, V>> observer) {
			ScrollableResults sr = null;
			try {
				sr = s.getNamedQuery(HBlob.HQL_GET_KEYS).setInteger(
						HBlob.pNameMapId, mapId.intValue()).scroll(
						ScrollMode.FORWARD_ONLY);
				while (observer.observeNext() && sr.next()) {
					HBlob blob = (HBlob) sr.get(0);
					final K k = serialization.<K>deserialize(blob.getKeyBytes());
					final V v = serialization.<V>deserialize(blob.getValueBytes());
					observer.observe(new SimpleImmutableEntry<K, V>(k, v));
				}
			} finally {
				if (sr != null) {
					sr.close();
				}
			}
			return null;
		}

	};

	final Transactional<Entry<K, V>, Void> tPut = new Transactional<Entry<K, V>, Void>() {

		@Override
		public Void with(Session s, Entry<K, V> tupple) {
			K key = tupple.getKey();
			V value = tupple.getValue();
			Long hashCode = Long.valueOf(keyHashFunction.hash(key));
			ScrollableResults sr = null;
			boolean updated = false;
			try {
				sr = s.getNamedQuery(HBlob.HQL_GET).setInteger(
						HBlob.pNameMapId, mapId.intValue()).setLong(
						HBlob.pNameHashCode, hashCode.longValue()).setLockMode(
						HBlob.alias, LockMode.PESSIMISTIC_WRITE).scroll(
						ScrollMode.FORWARD_ONLY);
				while (sr.next()) {
					HBlob blob = (HBlob) sr.get(0);
					K k = serialization.<K>deserialize(blob.getKeyBytes());
					if (key.equals(k)) {
						blob.setValueBytes(serialization.serialize(value));
						updated = true;
						break;
					}
				}
			} finally {
				if (sr != null) {
					sr.close();
				}
			}
			if (!updated) {
				HBlob blob = new HBlob();
				blob.setHashCode(hashCode);
				blob.setKeyBytes(serialization.serialize(key));
				blob.setValueBytes(serialization.serialize(value));
				blob.setMapId(mapId);
				s.save(blob);
			}
			return null;
		}
	};

	private final Transactional<Iterator<Entry<K, V>>, Void> tPutAll = new Transactional<Iterator<Entry<K, V>>, Void>() {

		@Override
		public Void with(Session s, Iterator<Entry<K, V>> iterator) {
			int i = 0;
			while (iterator.hasNext()) {
				Entry<K, V> tuple = iterator.next();
				tPut.with(s, tuple);
				if (batchSize > 1) {
					if (++i % batchSize != 0) {
						continue;
					}
					i = 0;
				}
				s.flush();
				s.clear();
			}
			return null;
		}
	};

	final Transactional<K, Boolean> tRemove = new Transactional<K, Boolean>() {

		@Override
		public Boolean with(Session s, K key) {
			Long hashCode = Long.valueOf(keyHashFunction.hash(key));

			ScrollableResults sr = null;
			try {
				sr = s.getNamedQuery(HBlob.HQL_GET).setInteger(
						HBlob.pNameMapId, mapId.intValue()).setLong(
						HBlob.pNameHashCode, hashCode.longValue()).setLockMode(
						HBlob.alias, LockMode.PESSIMISTIC_WRITE).scroll(
						ScrollMode.FORWARD_ONLY);
				while (sr.next()) {
					HBlob blob = (HBlob) sr.get(0);
					K k = serialization.<K>deserialize(blob.getKeyBytes());
					if (key.equals(k)) {
						s.delete(blob);
						return Boolean.TRUE;
					}
				}
			} finally {
				if (sr != null) {
					sr.close();
				}
			}

			return Boolean.FALSE;
		}
	};

	private final Transactional<Iterator<K>, Void> tRemoveAll = new Transactional<Iterator<K>, Void>() {

		@Override
		public Void with(Session s, Iterator<K> iterator) {
			int i = 0;
			final int batchSize = dao.getSessionProvider().getJdbcBatchSize();
			while (iterator.hasNext()) {
				K key = iterator.next();
				tRemove.with(s, key);
				if (batchSize > 1) {
					if (++i % batchSize != 0) {
						continue;
					}
					i = 0;
				}
				s.flush();
				s.clear();
			}
			return null;
		}
	};

	private final Transactional<Void, Long> tSize = new Transactional<Void, Long>() {

		@Override
		public Long with(Session s, Void v) {
			return (Long) s.getNamedQuery(HBlob.HQL_GET_SIZE).setInteger(
					"mapId", mapId.intValue()).uniqueResult();
		}
	};

	public HibernateMap(Serialization serialization, DAO dao,
			HashFunction<K> keyHashFunction, Integer mapId) {
		this.serialization = serialization;
		this.dao = dao;
		this.keyHashFunction = keyHashFunction;
		this.mapId = mapId;
		batchSize = dao.getSessionProvider().getJdbcBatchSize();
	}

	@Override
	public void clear() {
		dao.execute(tClear, null);
	}

	@Override
	public V get(final K key) {
		return dao.execute(tGet, key);
	}

	@Override
	public void getAll(Iterator<K> iterator, IterationObserver<V> observer) {
		dao.execute(tGetAll, new Tuple2<Iterator<K>, IterationObserver<V>>(
				iterator, observer));
	}

	Transactional<Void, Void> gettClear() {
		return tClear;
	}

	Transactional<K, V> gettGet() {
		return tGet;
	}

	@Override
	public void iterateKeys(IterationObserver<K> observer) {
		dao.execute(tIterateKeys, observer);
	}

	@Override
	public void iterateMapEntries(IterationObserver<Entry<K, V>> observer) {
		dao.execute(tMapEntries, observer);
	}

	@Override
	public void put(final K key, final V value) {
		dao.execute(tPut, new SimpleImmutableEntry<K, V>(key, value));
	}

	@Override
	public void putAll(Iterator<Entry<K, V>> iterator) {
		dao.execute(tPutAll, iterator);
	}

	@Override
	public boolean remove(final K key) {
		return dao.execute(tRemove, key).booleanValue();
	}

	@Override
	public void removeAll(Iterator<K> iterator) {
		dao.execute(tRemoveAll, iterator);
	}

	@Override
	public long size() {
		return dao.execute(tSize, null).longValue();
	}

}
