/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.AbstractMap.SimpleImmutableEntry;

import org.hibernate.Session;
import org.jhears.util.DefaultHashFunction;
import org.jhears.util.HashFunction;
import org.jhears.util.Serialization;
import org.jhears.util.Serializer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Convenience class for map management. Using of this class is subject to
 * following restrictions:
 * <ul>
 * <li>{@link #deleteMap(String)} - user must make sure that map is not used no
 * more.
 * </ul>
 * 
 * @author Juha Heljoranta
 * 
 */
public class HibernateMapManager {

	private final DAO dao;

	private final HashFunction<String> tableHashFunction;

	private final Serialization defaultSerialization;

	private final GeneralDao generalDao;

	private final Logger logger;

	private final HibernateMap<String, Integer> mapCatalog;

	public HibernateMapManager(DAO dao, Serialization serialization) {
		this.logger = LoggerFactory.getLogger(HibernateMapManager.class);
		this.dao = dao;
		this.generalDao = new GeneralDao(dao);
		defaultSerialization = serialization;
		tableHashFunction = new DefaultHashFunction<String>();
		Integer mapCatalogId = Integer.valueOf(0);
		BlobIdentity catalogIdentity = new BlobIdentity();
		// override generated id
		catalogIdentity.setId(mapCatalogId);
		// make sure that Identity with id=0 exists
		generalDao.saveOrUpdate(catalogIdentity);
		mapCatalog = new HibernateMap<String, Integer>(new Serializer(), dao,
				tableHashFunction, mapCatalogId);
	}

	public int countMaps() {
		return (int) mapCatalog.size();
	}

	// public <K, V> DatabaseMap<K, V> createMap(final String name) {
	// return createMap(name, defaultSerialization, defaultHashFunction);
	// }

	public <K, V> DatabaseMap<K, V> createMap(final String name,
			HashFunction<K> hashFunction) {
		return createMap(name, defaultSerialization, hashFunction);
	}

	// public <K, V> DatabaseMap<K, V> createMap(final String name,
	// Serialization serialization) {
	// return createMap(name, serialization, defaultHashFunction);
	// }

	public <K, V> DatabaseMap<K, V> createMap(final String name,
			Serialization serialization, HashFunction<K> hashFunction) {
		final BlobIdentity identity = new BlobIdentity();
		dao.execute(new Transactional<Void, Void>() {

			@Override
			public Void with(Session s, Void arg) {
				s.save(identity);
				SimpleImmutableEntry<String, Integer> tuple = new SimpleImmutableEntry<String, Integer>(
						name, identity.getId());
				mapCatalog.tPut.with(s, tuple);
				return null;
			}

		}, null);
		Integer mapId = identity.getId();
		return new HibernateMap<K, V>(serialization, dao, hashFunction, mapId);
	}

	public boolean deleteMap(final String name) {
		return dao.execute(new Transactional<Void, Boolean>() {

			@Override
			public Boolean with(Session s, Void arg) {
				Integer id = mapCatalog.gettGet().with(s, name);
				if (id != null) {
					// remove all map contents
					new HibernateMap<String, Integer>(defaultSerialization,
							dao, tableHashFunction, id).gettClear().with(s, null);

					mapCatalog.tRemove.with(s, name);
					// clean up identity table
					BlobIdentity identity = (BlobIdentity) s.get(
							BlobIdentity.class, id);
					if (identity != null) {
						s.delete(identity);
					} else {
						logger.error("Inconsistent data. Deleted table name="
								+ name + " with non-existing identity id=" + id
								+ ". This is bug.");
					}
					return Boolean.TRUE;
				}
				return Boolean.FALSE;
			}
		}, null).booleanValue();
	}

	// public <K, V> DatabaseMap<K, V> getMap(final String name) {
	// return getMap(name, defaultSerialization, defaultHashFunction);
	// }

	public <K, V> DatabaseMap<K, V> getMap(final String name,
			HashFunction<K> hashFunction) {
		return getMap(name, defaultSerialization, hashFunction);
	}

	// public <K, V> DatabaseMap<K, V> getMap(final String name,
	// Serialization serialization) {
	// return getMap(name, serialization, defaultHashFunction);
	// }

	public <K, V> DatabaseMap<K, V> getMap(final String name,
			Serialization serialization, HashFunction<K> hashFunction) {
		Integer mapId = mapCatalog.get(name);
		if (mapId != null) {
			return new HibernateMap<K, V>(serialization, dao, hashFunction,
					mapId);
		} else {
			return null;
		}
	}

}
