/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.nio.charset.Charset;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;

import org.jhears.util.AbstractIterationObserver;
import org.jhears.util.IterationObserver;
import org.junit.Assert;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class DatabaseMapTest {

	public static class ThreadId {
		// Atomic integer containing the next thread ID to be assigned
		private static final AtomicInteger nextId = new AtomicInteger(0);

		// Thread local variable containing each thread's ID
		private static final ThreadLocal<Integer> threadId = new ThreadLocal<Integer>() {
			@Override
			protected Integer initialValue() {
				return nextId.getAndIncrement();
			}
		};

		// Returns the current thread's unique ID, assigning it if necessary
		public static int get() {
			return threadId.get();
		}
	}

	public static class IdGn {
		// Atomic integer containing the next thread ID to be assigned
		private static final AtomicInteger nextId = new AtomicInteger(0);

		// Returns the current thread's unique ID, assigning it if necessary
		public static int get() {
			return nextId.getAndIncrement();
		}
	}

	private static final Logger logger = LoggerFactory
			.getLogger(DatabaseMapTest.class);

	private static final Charset ASCII = Charset.forName("US-ASCII");

	private static String rndStr(int len, Random r) {
		byte[] ba = new byte[len];
		r.nextBytes(ba);
		for (int i = 0; i < len; i++) {
			int x = ba[i] & 0xff;
			x = (x % (0x7f - 0x20)) + 0x20;
			ba[i] = (byte) x;
		}
		return new String(ba, ASCII);
	}

	protected abstract DatabaseMap<String, String> createMap();

	@Test
	public void testBigValue() {

		DatabaseMap<String, String> m = createMap();
		m.clear();
		byte[] ba;
		// put 2k key and 2MiB value

		String key;
		String value;

		ba = new byte[1 << 11];
		Arrays.fill(ba, (byte) 'k');
		key = new String(ba, ASCII);
		ba = new byte[1 << 21];
		Arrays.fill(ba, (byte) 'v');
		value = new String(ba, ASCII);

		m.put(key, value);

		// verify
		ba = new byte[1 << 11];
		Arrays.fill(ba, (byte) 'k');
		key = new String(ba, ASCII);
		ba = new byte[1 << 21];
		Arrays.fill(ba, (byte) 'v');
		value = new String(ba, ASCII);

		Assert.assertEquals(value, m.get(key));

	}

	@Test
	public void testIterator() {
		DatabaseMap<String, String> m = createMap();
		m.clear();

		final int count = 1 << 6;
		final Random r = new Random(0);
		m.putAll(new Iterator<Map.Entry<String, String>>() {

			int i = 0;

			@Override
			public boolean hasNext() {
				return i < count;
			}

			@Override
			public Map.Entry<String, String> next() {
				i++;
				final String key = rndStr(1 << 6, r);
				final String value = rndStr(1 << 10, r);
				return new SimpleImmutableEntry<String, String>(key, value);
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}
		});
		final AtomicInteger ai = new AtomicInteger(0);
		m.iterateKeys(new AbstractIterationObserver<String>() {

			@Override
			public void observe(String elt) {
				ai.incrementAndGet();
			}

		});
		Assert.assertEquals(count, ai.get());
		m.clear();
	}

	@Test
	public void testPutGetRemoveAll() {
		Random r = new Random(0);
		DatabaseMap<String, String> m = createMap();
		HashMap<String, String> reference = new HashMap<String, String>();
		int count = 64;
		for (int i = 0; i < count; i++) {
			String key = "key-" + Integer.toHexString(r.nextInt());
			String value = "value-" + Integer.toHexString(r.nextInt());
			reference.put(key, value);
		}
		m.putAll(new HashMap<String, String>(reference).entrySet().iterator());
		final ArrayList<String> valueList = new ArrayList<String>();
		IterationObserver<String> observer = new AbstractIterationObserver<String>() {
			@Override
			public void observe(String elt) {
				valueList.add(elt);
			}
		};
		m.getAll(new HashMap<String, String>(reference).keySet().iterator(),
				observer);
		Assert.assertEquals(reference.size(), valueList.size());
		Assert.assertTrue(new HashSet<String>(reference.values())
				.equals(new HashSet<String>(valueList)));
		Assert.assertEquals(reference.size(), m.size());
		m.removeAll(new HashMap<String, String>(reference).keySet().iterator());
		Assert.assertEquals(0, m.size());
		valueList.clear();
		observer = new AbstractIterationObserver<String>() {
			@Override
			public void observe(String elt) {
				valueList.add(elt);
			}
		};
		m.getAll(new HashMap<String, String>(reference).keySet().iterator(),
				observer);
		Assert.assertEquals(reference.size(), valueList.size());
		for (String value : valueList) {
			Assert.assertNull(value);
		}
	}

	@Test
	public void testRandom() {
		DatabaseMap<String, String> m = createMap();
		m.clear();
		HashMap<String, String> reference = new HashMap<String, String>();
		reference.clear();
		Random r = new Random(0);
		int size = 100;
		for (int i = 0; i < size; i++) {
			String key;
			String value;
			boolean expect;
			boolean actual;
			switch (r.nextInt(7)) {
			case 0:
			case 1:
				key = "key-" + Integer.toHexString(r.nextInt());
				value = "value-" + Integer.toHexString(r.nextInt());
				expect = reference.put(key, value) != null;
				m.put(key, value);
				break;
			case 2:
				if (!reference.isEmpty()) {
					ArrayList<String> keys = new ArrayList<String>(reference
							.keySet());
					key = keys.get(r.nextInt(keys.size()));
					expect = reference.remove(key) != null;
					actual = m.remove(key);
					if (expect != actual) {
						throw new RuntimeException();
					}
				}
				break;
			case 3:
				if (!reference.isEmpty()) {
					ArrayList<String> keys = new ArrayList<String>(reference
							.keySet());
					key = keys.get(r.nextInt(keys.size()));
					m.get(key);
					if (!reference.get(key).equals(m.get(key))) {
						throw new RuntimeException();
					}
				}
				break;
			case 4:
				m = createMap();
				break;
			case 5:
				final HashSet<String> itrActual = new HashSet<String>();
				m.iterateKeys(new AbstractIterationObserver<String>() {

					@Override
					public void observe(String elt) {
						itrActual.add(elt);
					}

				});
				if (!reference.keySet().equals(itrActual)) {
					throw new RuntimeException();
				}
				break;
			case 7:
				if (reference.size() != m.size()) {
					throw new RuntimeException();
				}
				break;
			default:
				break;
			}
		}
	}

	// @Test
	// public void testMultithreading() throws InterruptedException,
	// ExecutionException {
	// final DatabaseMap<String, String> map = createMap();
	// map.clear();
	// final int referenceSize = 1 << 10;
	// final Map<String, String> referenceMap = new HashMap<String, String>();
	// final Random r = new Random(0);
	// for (int i = 0; i < referenceSize; i++) {
	// String key = rndStr(1 << 6, r);
	// String value = rndStr(1 << 10, r);
	// referenceMap.put(key, value);
	// }
	// final List<String> keys = new ArrayList<String>(referenceMap.keySet());
	// ArrayList<Callable<String>> tasks = new ArrayList<Callable<String>>();
	//
	// // populate to 50%
	// for (int i = 0; i < referenceSize / 2; i++) {
	// String key = keys.get(r.nextInt(referenceSize));
	// map.put(key, referenceMap.get(key));
	// }
	//
	// final int ops = 1 << 14;
	// tasks.add(new Callable<String>() {
	//
	// @Override
	// public String call() throws Exception {
	// try {
	// for (int i = 0; i < ops; i++) {
	// String key = keys.get(r.nextInt(referenceSize));
	// map.put(key, referenceMap.get(key));
	// }
	// } catch (Exception e) {
	// logger.error("", e);
	// throw e;
	// }
	// return "";
	// }
	// });
	//
	// tasks.add(new Callable<String>() {
	//
	// @Override
	// public String call() throws Exception {
	// try {
	// for (int i = 0; i < ops; i++) {
	// String key = keys.get(r.nextInt(referenceSize));
	// map.get(key);
	// }
	// } catch (Exception e) {
	// logger.error("", e);
	// throw e;
	// }
	// return "";
	// }
	// });
	//
	// tasks.add(new Callable<String>() {
	//
	// @Override
	// public String call() throws Exception {
	// try {
	// for (int i = 0; i < ops; i++) {
	// String key = keys.get(r.nextInt(referenceSize));
	// map.remove(key);
	// }
	// } catch (Exception e) {
	// logger.error("", e);
	// throw e;
	// }
	// return "";
	// }
	// });
	//
	// exec(tasks);
	// }

//	private void exec(List<Callable<String>> tasks)
//			throws InterruptedException, ExecutionException {
//		int threads = tasks.size();
//		ExecutorService executorService = Executors.newFixedThreadPool(threads);
//		ExecutorCompletionService<String> ecs = new ExecutorCompletionService<String>(
//				executorService);
//
//		for (Callable<String> task : tasks) {
//			ecs.submit(task);
//		}
//
//		int cnt = 0;
//		while (cnt < tasks.size()) {
//			Future<String> future = ecs.take();
//			future.get();
//			cnt++;
//		}
//	}

	@Test
	public void parallelTestRandom() throws InterruptedException,
			ExecutionException {
		int threads = 3;
		ExecutorService executorService = Executors.newFixedThreadPool(threads);
		ExecutorCompletionService<String> ecs = new ExecutorCompletionService<String>(
				executorService);
		ArrayList<Callable<String>> tasks = new ArrayList<Callable<String>>();
		for (int i = 0; i < threads * 4; i++) {
			tasks.add(new Callable<String>() {

				@Override
				public String call() throws Exception {
					try {
						testRandom();
					} catch (Exception e) {
						logger.error("", e);
						throw e;
					}
					return "";
				}
			});
		}

		for (Callable<String> task : tasks) {
			ecs.submit(task);
		}

		int cnt = 0;
		while (cnt < tasks.size()) {
			Future<String> future = ecs.take();
			future.get();
			cnt++;
		}
	}
}
