/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.auth;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.hibernate.Session;
import org.hibernate.criterion.Restrictions;
import org.jhears.crypto.AuthCode;
import org.jhears.data.DAO;
import org.jhears.data.RoleDataObj;
import org.jhears.data.Transactional;
import org.jhears.data.UserDataObj;
import org.jhears.user.UserInformation;
import org.jhears.user.UserInformationCentral;

import com.google.common.base.Charsets;
import com.google.common.base.Function;
import com.google.common.base.Throwables;
import com.google.common.collect.Collections2;

public class UserInformationCentralImpl implements UserInformationCentral {

	private final AuthCode authCode;

	private final DAO dao;

	public UserInformationCentralImpl(DAO dao) {
		authCode = new AuthCode();
		this.dao = dao;
	}

	@Override
	public boolean authenticate(UserInformation info, String plaintextPassword) {
		byte[] salt;
		byte[] pass;
		try {
			salt = Hex.decodeHex(info.getPasswordSalt().toCharArray());
			pass = Hex.decodeHex(info.getPassword().toCharArray());
		} catch (DecoderException e) {
			throw Throwables.propagate(e);
		}
		byte[] code = authCode.calculate(salt, plaintextPassword
				.getBytes(Charsets.UTF_8));
		boolean retval = pass != null && Arrays.equals(pass, code);
		return retval;
	}

	private UserDataObj findUserByName(final String username) {
		Transactional<Void, UserDataObj> transactional = new Transactional<Void, UserDataObj>() {

			@Override
			public UserDataObj with(Session s, Void arg) {
				UserDataObj user = (UserDataObj) s.createCriteria(
						UserDataObj.class).add(
						Restrictions.eq("name", username)).uniqueResult();
				if (user != null) {
					// load lazy collection
					user.getRoles().size();
				}
				return user;
			}

		};
		return dao.execute(transactional, null);
	}

	@Override
	public UserInformation getUserInformation(final String username) {
		UserDataObj user = findUserByName(username);
		if (user == null) {
			return null;
		}
		final List<String> roles = getUserRoleNames(user);
		final String passwd = user.getProperties().get(UserDataObj.PASSWORD);
		final String salt = user.getProperties().get(UserDataObj.PASSWORD_SALT);
		return new UserInformation() {

			private static final long serialVersionUID = -2702549142517913965L;

			@Override
			public String getPassword() {
				return passwd;
			}

			@Override
			public String getPasswordSalt() {
				return salt;
			}

			@Override
			public List<String> getRoles() {
				return roles;
			}

			@Override
			public String getUsername() {
				return username;
			}
		};
	}

	private List<String> getUserRoleNames(final UserDataObj user) {
		Function<RoleDataObj, String> function = new Function<RoleDataObj, String>() {

			@Override
			public String apply(RoleDataObj from) {
				return from.getName();
			}

		};
		return new ArrayList<String>(Collections2
				.transform(user.getRoles(), function));
	}

}
