/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.Serializable;
import java.util.Set;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.OneToMany;

/**
 * Audio entity represents some uniquely identifiable audio clip.
 * <p>
 * For example, 'Metallica (Founded 1981, group), Ride the Lightning (2000-02-22
 * release), The Call of Ktulu' defines an audio entity in precision which is
 * nearly impossible to get confused with some other audio entity.
 * <p>
 * Moreover, the entity should be described by using some pre-existing indexing
 * mechanism. For example, MusicBrainz url like,
 * http://musicbrainz.org/track/0d08fdff-335a-4448-8632-76c720b2497c.html or
 * some similar system.
 * <p>
 * Audio entity can have several audio {@link FingerprintDataObj} associated
 * into it. Each audio fingerprint should describe the audio entity as good as
 * they can. In perfect world there would be only single audio fingerprint for
 * each audio entity. However, due constraints set by the reality there is a
 * possibilities such as misuse, human error, software error and so on.
 * Therefore, having multiple fingerprints per audio entity allows the system to
 * cope with reality with bit higher confidence.
 */
@Entity
public class AudioEntityDataObj implements Serializable {

	private static final long serialVersionUID = -206916593747222061L;

	private Set<FingerprintDataObj> fingerprints;

	private Long id;

	private String identifier;

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AudioEntityDataObj other = (AudioEntityDataObj) obj;
		if (identifier == null) {
			if (other.identifier != null)
				return false;
		} else if (!identifier.equals(other.identifier))
			return false;
		return true;
	}

	/**
	 * Set of fingerprints associated to this entity.
	 * 
	 * @return
	 */
	@OneToMany(fetch = FetchType.LAZY, cascade = { CascadeType.DETACH,
			CascadeType.MERGE, CascadeType.PERSIST, CascadeType.REFRESH,
			CascadeType.REMOVE }, mappedBy = "audioEntity")
	public Set<FingerprintDataObj> getFingerprints() {
		return fingerprints;
	}

	/**
	 * Internal identifier, should not be exposed to clients.
	 * 
	 * @return
	 */
	@Id
	@GeneratedValue
	public Long getId() {
		return id;
	}

	/**
	 * Uniquely identifies the entity. Maximum length is
	 * {@value #MAX_IDENTIFIER_LENGTH}.
	 * 
	 * @return
	 */
	@Column(nullable = false, unique = true, updatable = false)
	public String getIdentifier() {
		return identifier;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((identifier == null) ? 0 : identifier.hashCode());
		return result;
	}

	public void setFingerprints(Set<FingerprintDataObj> fingerprints) {
		this.fingerprints = fingerprints;
	}

	@SuppressWarnings("unused")
	private void setId(Long id) {
		this.id = id;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	@Override
	public String toString() {
		// Set<FingerprintDataObj> fpSet = getFingerprints();
		// ArrayList<String> fpIds = new ArrayList<String>();
		// if (fpSet != null) {
		// for (FingerprintDataObj fp : getFingerprints()) {
		// fpIds.add(fp.getIdentifier());
		// }
		// }
		return AudioEntityDataObj.class.getSimpleName() + "identifier="
				+ identifier + "]";
		// + " [fingerprints.identifier=" + fpIds + ", id=" + id
		// + ", identifier=" + identifier + "]";
	}
}
