/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.Serializable;
import java.util.Arrays;

import org.jhears.filter.bloom.HVBloomFilter;
import org.jhears.filter.bloom.IHashValueBloomFilter;
import org.jhears.filter.bloom.scalable.IScalableBloomFilterState;
import org.jhears.util.HashFunction;
import org.jhears.util.Serialization;

public class CountingDatabaseSBFState implements IScalableBloomFilterState {

	private static class Constants implements Serializable {

		private static final long serialVersionUID = -8768618342857084051L;
		private final int m0;
		private final double p0;
		private final double r;
		private final int s;

		public Constants(int s, double r, int m0, double p0) {
			this.s = s;
			this.r = r;
			this.m0 = m0;
			this.p0 = p0;
		}

		@Override
		public String toString() {
			return Constants.class.getSimpleName() + " [m0=" + m0 + ", p0="
					+ p0 + ", r=" + r + ", s=" + s + "]";
		}

	}

	private final Constants constants;

	private final String databaseName;

	private HVBloomFilter[] filters;

	private long filterSize;

	private final HashFunction<?> hashFunction;

	private int[] kArr;

	private final BerkeleyMapManager manager;

	private int[] nArr;

	private final Serialization serialization;

	private long size;

	private final BerkeleyMap<String, Serializable> stateMap;

	private final String stateName;

	private String[] stateNames;

	private DatabaseBFState[] states;

	@SuppressWarnings("unchecked")
	public CountingDatabaseSBFState(int s, double r, int m0, double p0,
			String databaseName, BerkeleyMapManager manager,
			HashFunction<?> hashFunction, Serialization serialization) {
		this.databaseName = databaseName;
		this.manager = manager;
		this.hashFunction = hashFunction;
		this.serialization = serialization;
		stateName = stateName(databaseName);
		stateMap = new BerkeleyMap<String, Serializable>(manager, stateName,
				BerkeleyMapManager.getDefaultDatabaseConfig(), serialization,
				(HashFunction<String>) hashFunction);

		Constants tmpConstants = (Constants) stateMap.get("constants");
		if (tmpConstants == null) {
			this.constants = new Constants(s, r, m0, p0);
			if (s == -1 || r == -1.0 || m0 == -1 || p0 == -1.0) {
				throw new IllegalArgumentException("Invalid values found: "
						+ constants);
			}
			stateMap.put("constants", constants);
			this.size = 0L;
			states = new DatabaseBFState[0];
			filters = new HVBloomFilter[0];
			stateNames = new String[0];
			nArr = new int[0];
			kArr = new int[0];
			stateMap.put("size", Long.valueOf(size));
			stateMap.put("nArr", nArr);
			stateMap.put("kArr", kArr);
			stateMap.put("stateNames", stateNames);
		} else {
			this.constants = tmpConstants;
			this.size = Long.valueOf((Long) stateMap.get("size"));
			stateNames = (String[]) stateMap.get("stateNames");
			nArr = (int[]) stateMap.get("nArr");
			kArr = (int[]) stateMap.get("kArr");

			int len = stateNames.length;
			states = new DatabaseBFState[len];
			filters = new HVBloomFilter[len];
			for (int i = 0; i < len; i++) {
				states[i] = load(i);
				filters[i] = new HVBloomFilter(states[i]);
			}
		}
	}

	public CountingDatabaseSBFState(String databaseName,
			BerkeleyMapManager manager, HashFunction<?> hashFunction,
			Serialization serialization) {
		this(-1, -1.0, -1, -1.0, databaseName, manager, hashFunction,
				serialization);
	}

	@Override
	public void appendNewFilter(int mi, int ki, int ni) {
		int i = filterCount();
		int newLength = i + 1;
		filters = Arrays.copyOf(filters, newLength);
		nArr = Arrays.copyOf(nArr, newLength);
		kArr = Arrays.copyOf(kArr, newLength);
		states = Arrays.copyOf(states, newLength);
		stateNames = Arrays.copyOf(stateNames, newLength);
		nArr[i] = ni;
		kArr[i] = ki;
		stateNames[i] = stateName(i);

		DatabaseBFState state = create(i, mi, ki);
		state.clear();
		states[i] = state;

		filters[i] = new HVBloomFilter(state);
		filterSize += filters[i].filterSize();

		stateMap.put("nArr", nArr);
		stateMap.put("kArr", kArr);
		stateMap.put("stateNames", stateNames);
	}

	@Override
	public void clear() {
		int cnt = filterCount();
		if (cnt == 0) {
			return;
		}
		size = 0L;
		for (int i = 1; i < cnt; i++) {
			states[i].destroy();
		}
		filters = Arrays.copyOf(filters, 1);
		nArr = Arrays.copyOf(nArr, 1);
		kArr = Arrays.copyOf(kArr, 1);
		states = Arrays.copyOf(states, 1);
		stateNames = Arrays.copyOf(stateNames, 1);
		stateMap.put("nArr", nArr);
		stateMap.put("kArr", kArr);
		stateMap.put("stateNames", stateNames);
		stateMap.put("size", Long.valueOf(size));
		filters[0].clear();
		filterSize = filters[0].filterSize();
	}

	public void close() {
		for (DatabaseBFState state : states) {
			state.close();
		}
		manager.closeDatabase(stateName);
	}

	private DatabaseBFState create(int i, int mi, int ki) {
		String name = stateNames[i];
		return new DatabaseBFState(mi, ki, manager, name, serialization,
				hashFunction);
	}

	@Override
	public void decSize() {
		stateMap.put("size", Long.valueOf(--size));
	}

	public void destroy() {
		close();
		for (DatabaseBFState state : states) {
			state.destroy();
		}
		manager.removeDatabase(stateName);
	}

	@Override
	public void destroyLastFilter() {
		int newLength = filterCount() - 1;
		filterSize -= filters[newLength].filterSize();
		states[newLength].destroy();
		filters = Arrays.copyOf(filters, newLength);
		nArr = Arrays.copyOf(nArr, newLength);
		kArr = Arrays.copyOf(kArr, newLength);
		states = Arrays.copyOf(states, newLength);
		stateNames = Arrays.copyOf(stateNames, newLength);
		stateMap.put("nArr", nArr);
		stateMap.put("kArr", kArr);
		stateMap.put("stateNames", stateNames);
	}

	@Override
	public int filterCount() {
		return states.length;
	}

	@Override
	public long filterSize() {
		return filterSize;
	}

	@Override
	public IHashValueBloomFilter getFilter(int i) {
		return filters[i];
	}

	@Override
	public int getK(int i) {
		return kArr[i];
	}

	@Override
	public int getM0() {
		return constants.m0;
	}

	@Override
	public int getN(int i) {
		return nArr[i];
	}

	@Override
	public double getP0() {
		return constants.p0;
	}

	@Override
	public double getR() {
		return constants.r;
	}

	@Override
	public int getS() {
		return constants.s;
	}

	@Override
	public void incSize() {
		stateMap.put("size", Long.valueOf(++size));
	}

	@Override
	public boolean isEmpty() {
		return size == 0L;
	}

	private DatabaseBFState load(int i) {
		String name = stateNames[i];
		return new DatabaseBFState(manager, name, serialization, hashFunction);
	}

	@Override
	public long size() {
		return size;
	}

	private String stateName(int i) {
		return databaseName + "-" + i;
	}

	private String stateName(String databaseName2) {
		return databaseName + "-"
				+ IScalableBloomFilterState.class.getSimpleName();
	}

	@Override
	public boolean supportsRemove() {
		return true;
	}

	@Override
	public String toString() {
		return CountingDatabaseSBFState.class.getSimpleName() + " [constants="
				+ constants + ", databaseName=" + databaseName
				+ ", filterSize=" + filterSize + ", filters="
				+ Arrays.toString(filters) + ", kArr=" + Arrays.toString(kArr)
				+ ", nArr=" + Arrays.toString(nArr) + ", size=" + size
				+ ", stateName=" + stateName + "]";
	}

}
