/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.Iterator;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.Map.Entry;

import org.jhears.bitarray.BitArray;
import org.jhears.bitarray.BitArrayUtil;
import org.jhears.util.IterationObserver;
import org.jhears.util.AbstractIterationObserver;

public final class DBBitArrayUtil {

	private static final class Itr implements Iterator<Entry<Long, Long>> {

		private final long[] data;

		private int idx;

		private Itr(long[] data) {
			this.data = data;
			idx = 0;
		}

		@Override
		public boolean hasNext() {
			return idx < data.length;
		}

		@Override
		public Entry<Long, Long> next() {
			final Long value = Long.valueOf(data[idx]);
			final Long key = Long.valueOf(idx++);
			return new SimpleImmutableEntry<Long, Long>(key, value);
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}

	/**
	 * Returns a copy of {@link DatabaseBitArray} converted to {@link BitArray}.
	 * Implementation is considerably faster than using
	 * {@link BitArrayUtil#copy(org.jhears.util.IBitArray, org.jhears.util.IBitArray)}
	 * .
	 * 
	 * @param database
	 *            bit array to copy.
	 * @return {@link BitArray} which contents equal to {@code database}.
	 */
	public static final BitArray bitArrayCopy(DatabaseBitArray database) {
		int dbSize = database.size();
		final long[] bitArrayData = new long[dbSize / Long.SIZE + 1];
		IterationObserver<Entry<Long, Long>> observer = new AbstractIterationObserver<Entry<Long, Long>>() {

			@Override
			public void observe(Entry<Long, Long> elt) {
				int arrIdx = elt.getKey().intValue();
				long data = elt.getValue().longValue();
				bitArrayData[arrIdx] = data;
			}
		};
		database.getMap().iterateMapEntries(observer);
		return new BitArray(dbSize, bitArrayData) {
			private static final long serialVersionUID = -4939251356870060668L;
		};
	}

	/**
	 * Returns a copy of {@link DatabaseCountingBitArray} converted to {@link BitArray}.
	 * Implementation is considerably faster than using
	 * {@link BitArrayUtil#copy(org.jhears.util.IBitArray, org.jhears.util.IBitArray)}
	 * 
	 * @param database
	 *            bit array to copy.
	 * @return {@link BitArray} which contents equal to {@code database}.
	 */
	public static final BitArray bitArrayCopy(DatabaseCountingBitArray database) {
		int dbSize = database.size();
		final BitArray ba = new BitArray(dbSize);
		IterationObserver<Entry<Long, Byte>> observer = new AbstractIterationObserver<Entry<Long, Byte>>() {

			@Override
			public void observe(Entry<Long, Byte> elt) {
				int arrIdx = elt.getKey().intValue();
				if (elt.getValue().byteValue() > 0) {
					ba.set(arrIdx);
				}
			}
		};
		database.getMap().iterateMapEntries(observer);
		return ba;
	}

	/**
	 * Copy contents of a {@link BitArray} to {@link DatabaseBitArray}.
	 * <p>
	 * Implementation is considerably faster than using
	 * {@link BitArrayUtil#copy(org.jhears.util.IBitArray, org.jhears.util.IBitArray)}
	 * 
	 * @param fromBitArray
	 *            source.
	 * @param toDatabase
	 *            destination.
	 */
	public static final void copy(BitArray fromBitArray,
			DatabaseBitArray toDatabase) {
		long[] data = fromBitArray.getArray();
		toDatabase.getMap().putAll(new Itr(data));
	}

}
