/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.Serializable;
import java.util.concurrent.atomic.AtomicBoolean;

import org.jhears.bitarray.BitArray;
import org.jhears.bitarray.BitArrayUtil;
import org.jhears.bitarray.CachedBitArray;
import org.jhears.bitarray.IBitArray;
import org.jhears.bitarray.BitArrayUtil.LazyLoadingCompletionListener;
import org.jhears.filter.bloom.IBloomFilterState;
import org.jhears.util.HashFunction;
import org.jhears.util.Serialization;
import org.slf4j.LoggerFactory;

import com.google.common.util.concurrent.AbstractExecutionThreadService;

public class DatabaseBFState implements IBloomFilterState {

	private static class Constants implements Serializable {

		private static final long serialVersionUID = -2932779286558763177L;

		private final int k;
		private final int m;
		private final boolean supportsRemove;

		public Constants(int m, int k, boolean supportsRemove) {
			this.m = m;
			this.k = k;
			this.supportsRemove = supportsRemove;
		}

		@Override
		public String toString() {
			return Constants.class.getSimpleName() + " [k=" + k + ", m=" + m
					+ ", supportsRemove=" + supportsRemove + "]";
		}

	}

	private static class LazyLoader extends AbstractExecutionThreadService {

		private final AtomicBoolean isPopulated;
		private final IBitArray lazy;
		private final int sleepIntervalInMillis;
		private final String threadName;

		public LazyLoader(IBitArray lazy, String threadName,
				final int sleepIntervalInMillis, AtomicBoolean isPopulated) {
			super();
			this.lazy = lazy;
			this.threadName = threadName;
			this.sleepIntervalInMillis = sleepIntervalInMillis;
			this.isPopulated = isPopulated;
		}

		@Override
		protected void run() {
			Thread.currentThread().setName(
					Thread.currentThread().getName() + " lazy loading "
							+ threadName);
			while (State.RUNNING.equals(state()) && !isPopulated.get()
					&& !Thread.currentThread().isInterrupted()) {
				lazy.get((int) (Math.random() * lazy.size()));
				try {
					Thread.sleep(sleepIntervalInMillis);
				} catch (InterruptedException e) {
					Thread.currentThread().interrupt();
					return; // obey
				}
			}
		}
	}

	private final IBitArray arr;

	private final Constants constants;

	private final String dbName;

	private AbstractExecutionThreadService lazyLoaderService;

	private final BerkeleyMapManager manager;

	private long size;

	private final DatabaseMap<String, Serializable> stateMap;

	private final String stateName;

	public DatabaseBFState(BerkeleyMapManager manager, String databaseName,
			Serialization serialization, HashFunction<?> hashFunction) {
		this(-1, -1, manager, databaseName, serialization, hashFunction);
	}

	@SuppressWarnings("unchecked")
	public DatabaseBFState(int m, int k, BerkeleyMapManager manager,
			String databaseName, Serialization serialization,
			HashFunction<?> hashFunction) {
		this.manager = manager;
		this.stateName = stName(databaseName);
		this.stateMap = createStateMap(manager, serialization, hashFunction);

		Constants tmpConstants = (Constants) stateMap.get("constants");
		if (tmpConstants == null) {
			this.constants = new Constants(m, k, true);
			if (m == -1 || k == -1) {
				throw new IllegalArgumentException("Invalid values found: "
						+ constants);
			}
			stateMap.put("constants", constants);
			this.size = 0L;
			stateMap.put("size", Long.valueOf(size));
			dbName = dbName(databaseName);
		} else {
			this.constants = tmpConstants;
			this.size = Long.valueOf((Long) stateMap.get("size"));
			this.dbName = dbName(databaseName);
		}
		DatabaseMap<Long, Byte> baMap = new BerkeleyMap<Long, Byte>(manager,
				dbName, BerkeleyMapManager.getDefaultDatabaseConfig(),
				serialization, (HashFunction<Long>) hashFunction);
		IBitArray slowBitArray = new DatabaseCountingBitArray(baMap,
				constants.m * constants.k);
		IBitArray fastBitArray = new BitArray(constants.m * constants.k);

		this.arr = withLazyLoading(slowBitArray, fastBitArray);

	}

	@Override
	public void clear() {
		size = 0L;
		stateMap.put("size", Long.valueOf(size));
		this.arr.clear();
	}

	public void close() {
		lazyLoaderService.stopAndWait();
		manager.closeDatabase(dbName);
		manager.closeDatabase(stateName);
	}

	@SuppressWarnings("unchecked")
	private BerkeleyMap<String, Serializable> createStateMap(
			BerkeleyMapManager manager, Serialization serialization,
			HashFunction<?> hashFunction) {
		return new BerkeleyMap<String, Serializable>(manager, stateName,
				BerkeleyMapManager.getDefaultDatabaseConfig(), serialization,
				(HashFunction<String>) hashFunction);
	}

	private String dbName(String databaseName) {
		return databaseName + "-" + IBitArray.class.getSimpleName();
	}

	@Override
	public void decSize() {
		stateMap.put("size", Long.valueOf(--size));
	}

	public void destroy() {
		close();
		manager.removeDatabase(dbName);
		manager.removeDatabase(stateName);
	}

	@Override
	public long filterSize() {
		return arr.size();
	}

	@Override
	public IBitArray getBitArray() {
		return arr;
	}

	@Override
	public int getK() {
		return constants.k;
	}

	@Override
	public int getM() {
		return constants.m;
	}

	@Override
	public void incSize() {
		stateMap.put("size", Long.valueOf(++size));
	}

	@Override
	public boolean isEmpty() {
		return size == 0L;
	}

	@Override
	public long size() {
		return size;
	}

	private String stName(String databaseName) {
		return databaseName + "-" + IBloomFilterState.class.getSimpleName();
	}

	@Override
	public boolean supportsRemove() {
		return constants.supportsRemove;
	}

	@Override
	public String toString() {
		return DatabaseBFState.class.getSimpleName() + " [arr=" + arr
				+ ", constants=" + constants + ", size=" + size
				+ ", stateName=" + stateName + "]";
	}

	private IBitArray withLazyLoading(IBitArray slowBitArray,
			IBitArray fastBitArray) {
		final AtomicBoolean isPopulated = new AtomicBoolean(false);
		final CachedBitArray cached = new CachedBitArray(slowBitArray,
				fastBitArray);
		LazyLoadingCompletionListener completionListner = new LazyLoadingCompletionListener() {
			@Override
			public void loadingCompleted() {
				isPopulated.set(true);
				LoggerFactory.getLogger(DatabaseBFState.class).info(
						"Loading completed: " + cached);
			}
		};
		IBitArray bitArray = BitArrayUtil.withLazyLoad(cached,
				completionListner);
		lazyLoaderService = new LazyLoader(bitArray, CachedBitArray.class
				.getSimpleName(), 100, isPopulated);
		lazyLoaderService.start();
		return bitArray;
	}
}
