/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.LinkedHashMap;

import org.jhears.bitarray.IBitArray;
import org.jhears.util.AbstractIterationObserver;
import org.jhears.util.ArrayIterator;
import org.jhears.util.IterationObserver;

/**
 * Bit array backed by {@link DatabaseMap}.
 * 
 * @author Juha Heljoranta
 * 
 */
public class DatabaseBitArray implements IBitArray {

	private static interface Op {
		public Long eval(Long old, int i);
	}

	private static final Op CLEAR = new Op() {
		@Override
		public Long eval(Long old, int i) {
			return Long.valueOf(old.longValue() & ~(1L << (i % Long.SIZE)));
		}
	};

	private static final Op SET = new Op() {
		@Override
		public Long eval(Long old, int i) {
			return Long.valueOf(old.longValue() | 1L << (i % Long.SIZE));
		}
	};

	private final DatabaseMap<Long, Long> map;

	private final int size;

	public DatabaseBitArray(DatabaseMap<Long, Long> map, int size) {
		this.map = map;
		this.size = size;
	}

	@Override
	public void clear() {
		map.clear();
	}

	@Override
	public void clear(final int index) throws IndexOutOfBoundsException {
		Long key = Long.valueOf(index / Long.SIZE);
		Long oldValue = map.get(key);
		if (oldValue == null) {
			oldValue = Long.valueOf(0);
		}
		Long newValue = Long.valueOf(oldValue.longValue()
				& ~(1L << (index % Long.SIZE)));
		map.put(key, newValue);
	}

	@Override
	public void clearAll(int[] indexes) throws IndexOutOfBoundsException {
		withIndexes(indexes, CLEAR);
	}

	@Override
	public boolean get(int index) throws IndexOutOfBoundsException {
		Long key = Long.valueOf(index / Long.SIZE);
		Long value = map.get(key);
		return value != null ? (value.longValue() & (1L << (index % Long.SIZE))) != 0L
				: false;
	}

	@Override
	public boolean[] getAll(int[] indexes) throws IndexOutOfBoundsException {
		Long[] values = getAllValues(toKeys(indexes));
		boolean[] retval = new boolean[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			retval[i] = values[i] != null ? (values[i].longValue() & (1L << (indexes[i] % Long.SIZE))) != 0L
					: false;
		}
		return retval;
	}

	private Long[] getAllValues(Long[] keys) {
		final Long[] values = new Long[keys.length];
		IterationObserver<Long> observer = new AbstractIterationObserver<Long>() {

			private int i = 0;

			@Override
			public void observe(Long elt) {
				values[i++] = elt;

			}

		};
		map.getAll(new ArrayIterator<Long>(keys), observer);
		return values;
	}

	public DatabaseMap<Long, Long> getMap() {
		return map;
	}

	@Override
	public void set(int index) throws IndexOutOfBoundsException {
		Long key = Long.valueOf(index / Long.SIZE);
		Long oldValue = map.get(key);
		if (oldValue == null) {
			oldValue = Long.valueOf(0);
		}
		Long newValue = Long.valueOf(oldValue.longValue()
				| 1L << (index % Long.SIZE));
		map.put(key, newValue);
	}

	@Override
	public void setAll(int[] indexes) throws IndexOutOfBoundsException {
		withIndexes(indexes, SET);
	}

	@Override
	public int size() {
		return size;
	}

	private Long[] toKeys(int[] indexes) {
		Long[] retval = new Long[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			retval[i] = Long.valueOf(indexes[i] / Long.SIZE);
		}
		return retval;
	}

	@Override
	public String toString() {
		return DatabaseBitArray.class.getSimpleName() + " [map=" + map
				+ ", size=" + size + "]";
	}

	private void withIndexes(int[] indexes, Op op) {

		Long[] keys = toKeys(indexes);
		Long[] oldValues = getAllValues(keys);

		LinkedHashMap<Long, Long> hmap = new LinkedHashMap<Long, Long>(
				indexes.length);
		for (int i = 0; i < indexes.length; i++) {
			hmap.put(keys[i], oldValues[i]);
		}

		for (int i = 0; i < indexes.length; i++) {
			Long key = keys[i];
			Long oldValue = hmap.get(key);
			if (oldValue == null) {
				oldValue = Long.valueOf(0);
			}
			Long newValue = op.eval(oldValue, indexes[i]);
			hmap.put(key, newValue);
		}

		map.putAll(hmap.entrySet().iterator());
	}

}
