/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.LinkedHashMap;

import org.jhears.bitarray.CountingBitArrayOverflowException;
import org.jhears.bitarray.ICountingBitArray;
import org.jhears.util.AbstractIterationObserver;
import org.jhears.util.ArrayIterator;
import org.jhears.util.IterationObserver;

/**
 * Bit array backed by {@link DatabaseMap}.
 * 
 * @author Juha Heljoranta
 * 
 */
public class DatabaseCountingBitArray implements ICountingBitArray {

	private static interface Op {
		public Byte eval(Byte old);
	}

	private static final Op CLEAR = new Op() {
		@Override
		public Byte eval(Byte old) {
			if (old != null) {
				int b = old.byteValue() & 0xff;
				if (b > 0) {
					b--;
				}
				return Byte.valueOf((byte) b);
			}
			return null;
		}
	};

	private static final Op SET = new Op() {
		@Override
		public Byte eval(Byte old) {
			Byte newValue;
			if (old == null) {
				newValue = Byte.valueOf((byte) 1);
			} else {
				int b = old.byteValue() & 0xff;
				if (b < 0xff) {
					b++;
				} else {
					throw new CountingBitArrayOverflowException();
				}
				newValue = Byte.valueOf((byte) b);
			}
			return newValue;
		}
	};

	private final DatabaseMap<Long, Byte> map;

	private final int size;

	public DatabaseCountingBitArray(DatabaseMap<Long, Byte> map, int size) {
		this.map = map;
		this.size = size;
	}

	@Override
	public void clear() {
		map.clear();
	}

	@Override
	public void clear(final int index) throws IndexOutOfBoundsException {
		Long key = Long.valueOf(index);
		Byte oldValue = map.get(key);
		if (oldValue != null) {
			int b = oldValue.byteValue() & 0xff;
			if (b > 0) {
				b--;
			}
			map.put(key, Byte.valueOf((byte) b));
		}
	}

	@Override
	public void clearAll(int[] indexes) throws IndexOutOfBoundsException {
		withIndexes(indexes, CLEAR);
	}

	@Override
	public boolean get(int index) throws IndexOutOfBoundsException {
		Long key = Long.valueOf(index);
		Byte value = map.get(key);
		return value != null ? (value.byteValue() & 0xff) > 0 : false;
	}

	@Override
	public boolean[] getAll(int[] indexes) throws IndexOutOfBoundsException {
		Byte[] values = getAllValues(toKeys(indexes));
		boolean[] retval = new boolean[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			retval[i] = values[i] != null ? (values[i].byteValue() & 0xff) > 0
					: false;
		}
		return retval;
	}

	private Byte[] getAllValues(Long[] keys) {
		final Byte[] values = new Byte[keys.length];
		IterationObserver<Byte> observer = new AbstractIterationObserver<Byte>() {

			private int i = 0;

			@Override
			public void observe(Byte elt) {
				values[i++] = elt;

			}

		};
		map.getAll(new ArrayIterator<Long>(keys), observer);
		return values;
	}

	public DatabaseMap<Long, Byte> getMap() {
		return map;
	}

	@Override
	public void set(int index) throws IndexOutOfBoundsException,
			CountingBitArrayOverflowException {
		Long key = Long.valueOf(index);
		Byte oldValue = map.get(key);
		Byte newValue;
		if (oldValue == null) {
			newValue = Byte.valueOf((byte) 1);
		} else {
			int b = oldValue.byteValue() & 0xff;
			if (b < 0xff) {
				b++;
			} else {
				throw new CountingBitArrayOverflowException();
			}
			newValue = Byte.valueOf((byte) b);
		}
		map.put(key, newValue);
	}

	@Override
	public void setAll(int[] indexes) throws IndexOutOfBoundsException {
		withIndexes(indexes, SET);
	}

	@Override
	public int size() {
		return size;
	}

	private Long[] toKeys(int[] indexes) {
		Long[] retval = new Long[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			retval[i] = Long.valueOf(indexes[i]);
		}
		return retval;
	}

	@Override
	public String toString() {
		return DatabaseCountingBitArray.class.getSimpleName() + " [map=" + map
				+ ", size=" + size + "]";
	}

	private void withIndexes(int[] indexes, Op op) {

		Long[] keys = toKeys(indexes);
		Byte[] oldValues = getAllValues(keys);

		LinkedHashMap<Long, Byte> hmap = new LinkedHashMap<Long, Byte>(
				indexes.length);
		for (int i = 0; i < indexes.length; i++) {
			hmap.put(keys[i], oldValues[i]);
		}

		for (int i = 0; i < indexes.length; i++) {
			Long key = keys[i];
			Byte oldValue = hmap.get(key);
			Byte newValue = op.eval(oldValue);
			if (newValue != null) {
				hmap.put(key, newValue);
			}
		}

		map.putAll(hmap.entrySet().iterator());
	}
}
