/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.criterion.Restrictions;
import org.jhears.server.IUser;
import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.Segment;

/*
 * Note: It seems that hibernate 3.5.1-Final has some issues with 
 * eager fetching. Workaround is to use lazy fetching and 
 * manually invoke methods to initialize collections. 
 */
public class FPDAO {

	private static final String SELECT_ENTITIES_BY_FP_ID = "select ae from "
			+ AudioEntityDataObj.class.getSimpleName()
			+ " as ae inner join ae.fingerprints as fps where fps.id in (:ids)";

	private final DAO dao;

	private final GeneralDao generalDao;

	public FPDAO(GeneralDao generalDao) {
		this.generalDao = generalDao;
		this.dao = generalDao.getDao();
	}

	private FingerprintDataObj append(Session s,
			AudioEntityDataObj audioEntity, List<Segment> segments,
			List<Repeat> repeats) {
		FingerprintDataObj fp = new FingerprintDataObj();
		fp.setIdentifier(UUID.randomUUID().toString());
		fp.setApprovals(Long.valueOf(0));
		s.save(fp);
		audioEntity.getFingerprints().add(fp);
		fp.setAudioEntity(audioEntity);
		fp.setSegments(segments);
		fp.setRepeats(repeats);
		return fp;
	}

	public void deleteAll() {
		generalDao.deleteAllSafe(AudioEntityDataObj.class);
	}

	public boolean deleteEntityWithId(Long id) {
		return generalDao.delete(AudioEntityDataObj.class, id);
	}

	public boolean deleteEntityWithIdentifier(final String identifier,
			final IUser user) {
		Transactional<Void, Boolean> transactional = new Transactional<Void, Boolean>() {

			@Override
			public Boolean with(Session s, Void arg) {
				AudioEntityDataObj o = (AudioEntityDataObj) s.createCriteria(
						AudioEntityDataObj.class).add(
						Restrictions.eq("identifier", identifier))
						.uniqueResult();
				if (o != null) {
					s.delete(o);
					log(s, AudioEntityDataObj.class.getSimpleName(), o.getId(),
							user == null ? null : user.getId(), "DELETE",
							identifier);
					return Boolean.TRUE;
				}
				return Boolean.FALSE;
			}
		};
		return dao.execute(transactional, null).booleanValue();
		// boolean retval =
		// generalDao.deleteFirstWithProperty(AudioEntityDataObj.class,
		// );
		// if (retval) {
		// f(s)
		// }
		// return retval;
	}

	public boolean deleteFingerprintWithId(Long id) {
		return generalDao.delete(FingerprintDataObj.class, id);
	}

	public boolean deleteFingerprintWithIdentifier(final String identifier,
			final IUser user) {
		Transactional<Void, Boolean> transactional = new Transactional<Void, Boolean>() {

			@Override
			public Boolean with(Session s, Void arg) {
				FingerprintDataObj o = (FingerprintDataObj) s.createCriteria(
						FingerprintDataObj.class).add(
						Restrictions.eq("identifier", identifier))
						.uniqueResult();
				if (o != null) {
					s.delete(o);
					log(s, AudioEntityDataObj.class.getSimpleName(), o.getId(),
							user == null ? null : user.getId(), "DELETE",
							identifier);
					return Boolean.TRUE;
				}
				return Boolean.FALSE;
			}
		};
		return dao.execute(transactional, null).booleanValue();
	}

	@SuppressWarnings("unchecked")
	public List<AudioEntityDataObj> getAudioEntitiesByFingerprintId(
			final Set<Long> fingerprintIds) {
		if (fingerprintIds.isEmpty()) {
			return new ArrayList<AudioEntityDataObj>(0);
		}
		Transactional<Void, List<AudioEntityDataObj>> transactional = new Transactional<Void, List<AudioEntityDataObj>>() {

			@Override
			public List<AudioEntityDataObj> with(Session s, Void arg) {
				Query q = s.createQuery(SELECT_ENTITIES_BY_FP_ID);
				q.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
				List<AudioEntityDataObj> retval = q.setParameterList("ids",
						fingerprintIds).list();
				if (retval != null) {
					// fetch lazy collections
					for (AudioEntityDataObj ae : retval) {
						for (FingerprintDataObj fp : ae.getFingerprints()) {
							fp.getRepeats();
							fp.getSegments();
						}
					}
				}
				return retval;
			}
		};
		return dao.execute(transactional, null);
	}

	public AudioEntityDataObj getEntityWithId(final Long id) {
		return generalDao.getDao().execute(
				new Transactional<Void, AudioEntityDataObj>() {

					@Override
					public AudioEntityDataObj with(Session s, Void arg) {
						AudioEntityDataObj retval = (AudioEntityDataObj) s.get(
								AudioEntityDataObj.class, id);
						if (retval != null) {
							// fetch lazy collections
							retval.getFingerprints().size();
							for (FingerprintDataObj fp : retval
									.getFingerprints()) {
								fp.getRepeats();
								fp.getSegments();
							}
						}
						return retval;
					}
				}, null);
		// return generalDao.get(AudioEntityDataObj.class, id);
	}

	public AudioEntityDataObj getEntityWithIdentifier(final String identifier) {
		return generalDao.getDao().execute(
				new Transactional<Void, AudioEntityDataObj>() {

					@Override
					public AudioEntityDataObj with(Session s, Void arg) {
						Criteria c = s.createCriteria(AudioEntityDataObj.class)
								.add(Restrictions.eq("identifier", identifier));
						AudioEntityDataObj retval = (AudioEntityDataObj) c
								.uniqueResult();
						if (retval != null) {
							// fetch lazy collections
							for (FingerprintDataObj fp : retval
									.getFingerprints()) {
								fp.getRepeats();
								fp.getSegments();
							}
						}
						return retval;
					}
				}, null);
		// return generalDao.getWithProperty(AudioEntityDataObj.class,
		// "identifier", identifier);
	}

	public FingerprintDataObj getFingerprintWithIdentifierPlusEntity(
			final String identifier) {
		return generalDao.getDao().execute(
				new Transactional<Void, FingerprintDataObj>() {

					@Override
					public FingerprintDataObj with(Session s, Void arg) {
						Criteria c = s.createCriteria(FingerprintDataObj.class)
								.add(Restrictions.eq("identifier", identifier));
						FingerprintDataObj o = (FingerprintDataObj) c
								.uniqueResult();
						if (o != null) {
							// fetch lazy collections
							o.getAudioEntity().getIdentifier();
							o.getRepeats();
							o.getSegments();
							return o;
						}
						return null;
					}
				}, null);
		// return generalDao.getWithProperty(FingerprintDataObj.class,
		// "identifier", identifier);
	}

	public Boolean inApp(final String fingerprintId, final IUser user) {
		return generalDao.getDao().execute(new Transactional<Void, Boolean>() {

			@Override
			public Boolean with(Session s, Void arg) {
				Criteria c = s.createCriteria(FingerprintDataObj.class).add(
						Restrictions.eq("identifier", fingerprintId));
				FingerprintDataObj o = (FingerprintDataObj) c.uniqueResult();
				if (o != null) {
					boolean notApprovedYet = true;
					if (user != null) {
						// check if user has already approved the fingerprint
						QB qb = new QB();
						qb.eq(LogEntry.OBJECT_ID, o.getId());
						qb.eq(LogEntry.OBJECT_ENTITY, FingerprintDataObj.class
								.getSimpleName());
						qb.eq(LogEntry.USER_ID, user.getId());
						qb.eq(LogEntry.KEY, "APPROVE");
						notApprovedYet = qb.build(s.createCriteria(LogEntry.class))
								.uniqueResult() == null;
					}
					if (notApprovedYet) {
						log(s, FingerprintDataObj.class.getSimpleName(), o
								.getId(), user == null ? null : user.getId(),
								"APPROVE", o.getIdentifier());
						o.setApprovals(o.getApprovals().longValue() + 1);
					}
					return Boolean.valueOf(notApprovedYet);
				}
				return null;
			}
		}, null);
	}

	private void log(Session s, String objectEntity, Long objectId, Long userId,
			String key, String value) {
		LogEntry le = new LogEntry();
		le.setObjectEntity(objectEntity);
		le.setObjectId(objectId);
		le.setUserId(userId);
		le.setKey(key);
		le.setValue(value);
		le.setTimeInMillis(System.currentTimeMillis());
		s.save(le);
	}

	public FingerprintDataObj getFingerprintWithIdentifier(
			final String identifier) {
		return generalDao.getDao().execute(
				new Transactional<Void, FingerprintDataObj>() {

					@Override
					public FingerprintDataObj with(Session s, Void arg) {
						Criteria c = s.createCriteria(FingerprintDataObj.class)
								.add(Restrictions.eq("identifier", identifier));
						FingerprintDataObj o = (FingerprintDataObj) c
								.uniqueResult();
						if (o != null) {
							// fetch lazy collections
							o.getRepeats();
							o.getSegments();
							return o;
						}
						return null;
					}
				}, null);
	}

	public FingerprintDataObj getFingerpritWithIdPlusEntity(final Long id) {
		return generalDao.getDao().execute(
				new Transactional<Void, FingerprintDataObj>() {

					@Override
					public FingerprintDataObj with(Session s, Void arg) {
						FingerprintDataObj o = (FingerprintDataObj) s.get(
								FingerprintDataObj.class, id);
						if (o != null) {
							// fetch lazy collections
							o.getRepeats();
							o.getSegments();
							o.getAudioEntity().getIdentifier();
							return o;
						}
						return null;
					}
				}, null);
	}

	public FingerprintDataObj getFingerpritWithId(final Long id) {
		return generalDao.getDao().execute(
				new Transactional<Void, FingerprintDataObj>() {

					@Override
					public FingerprintDataObj with(Session s, Void arg) {
						FingerprintDataObj o = (FingerprintDataObj) s.get(
								FingerprintDataObj.class, id);
						if (o != null) {
							// fetch lazy collections
							o.getRepeats();
							o.getSegments();
							return o;
						}
						return null;
					}
				}, null);
	}

	private FingerprintDataObj insert(Session s, String entityId,
			List<Segment> segments, List<Repeat> repeats, IUser user) {
		FingerprintDataObj fp = new FingerprintDataObj();
		fp.setIdentifier(UUID.randomUUID().toString());
		fp.setSegments(segments);
		fp.setRepeats(repeats);
		fp.setApprovals(Long.valueOf(0));
		s.save(fp);
		AudioEntityDataObj audioEntity = new AudioEntityDataObj();
		audioEntity.setIdentifier(entityId);
		s.save(audioEntity);
		log(s, AudioEntityDataObj.class.getSimpleName(), audioEntity.getId(),
				user == null ? null : user.getId(), "CREATE", audioEntity
						.getIdentifier());
		HashSet<FingerprintDataObj> fplist = new HashSet<FingerprintDataObj>(1);
		fplist.add(fp);
		audioEntity.setFingerprints(fplist);
		fp.setAudioEntity(audioEntity);
		return fp;
	}

	private FingerprintDataObj save(Session s, String entityId,
			List<Segment> segments, List<Repeat> repeats, IUser user) {
		FingerprintDataObj fingerprint;
		AudioEntityDataObj entity = (AudioEntityDataObj) s.createCriteria(
				AudioEntityDataObj.class).add(
				Restrictions.eq("identifier", entityId)).uniqueResult();
		if (entity != null) {
			fingerprint = append(s, entity, segments, repeats);
		} else {
			fingerprint = insert(s, entityId, segments, repeats, user);
		}
		log(s, FingerprintDataObj.class.getSimpleName(), fingerprint.getId(),
				user == null ? null : user.getId(), "CREATE", fingerprint
						.getIdentifier());
		return fingerprint;
	}

	public FingerprintDataObj saveFingerprint(final String entityId,
			final List<Segment> segments, final List<Repeat> repeats,
			final IUser user) {
		if (segments.isEmpty() && repeats.isEmpty()) {
			throw new IllegalStateException(
					"Tried to save entityId with zero fingerprint data. entityId="
							+ entityId);
		}
		Transactional<Void, FingerprintDataObj> transactional = new Transactional<Void, FingerprintDataObj>() {

			@Override
			public FingerprintDataObj with(Session s, Void arg) {
				return save(s, entityId, segments, repeats, user);
			}
		};
		return dao.execute(transactional, null);
	}

}
