/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.ElementCollection;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.ManyToOne;
import javax.persistence.OrderColumn;
import javax.persistence.Transient;

import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.Segment;

/**
 * Represents persisted portion of user submitted fingerprint.
 * <p>
 * Fingerprint consist from set of {@link Segment}s. Each fingerprint is
 * identified uniquely by {@link #getIdentifier()}. In addition it contains some
 * meta data such as a creation time. Users of this class are responsible of
 * {@link Segment} data correctness.
 * <p>
 * Alternatively the fingerprint can consist from run-length-encoded symbols (
 * {@link Repeat}s). These are needed when the fingerprint contains only long
 * stretches of symbols.
 * <p>
 * Both the {@link Segment}s and {@link Repeat}s can be stored and used at the
 * same time since they both describe the submitted fingerprint sequence.
 * <p>
 * Internally this class stores {@link Segment#getPosition()} and
 * {@link Segment#getData()} separately. {@link Segment}s are assembled and
 * disassembled internally by {@link #getSegments()} and
 * {@link #setSegments(List)}, respectively.
 * <p>
 * Similarly the {@link Repeat}s are disassembled when set via
 * {@link #setRepeats(List)} and assembled when get via {@link #getRepeats()}.
 */
@Entity
public class FingerprintDataObj implements Serializable {

	private static final long serialVersionUID = 6629560386280530712L;

	private Long approvals;

	private AudioEntityDataObj audioEntity;

	private Long id;

	private String identifier;

	private List<Character> repeatCharList;

	private List<Integer> repeatLengthList;

	private List<Integer> repeatPositionList;

	@Transient
	private List<Repeat> repeats;

	private List<String> segmentDataList;

	private List<Integer> segmentPositionList;

	@Transient
	private List<Segment> segments;

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		FingerprintDataObj other = (FingerprintDataObj) obj;
		if (identifier == null) {
			if (other.identifier != null)
				return false;
		} else if (!identifier.equals(other.identifier))
			return false;
		return true;
	}

	/**
	 * The number of times this fingerprint being approved to be correct.
	 * 
	 * @return
	 */
	@Column(nullable = false)
	public Long getApprovals() {
		return approvals;
	}

	/**
	 * Returns entity in which this fingerprint is associated with.
	 * 
	 * @return
	 */
	@ManyToOne(fetch = FetchType.LAZY, cascade = { CascadeType.DETACH,
			CascadeType.MERGE, CascadeType.PERSIST, CascadeType.REFRESH })
	public AudioEntityDataObj getAudioEntity() {
		return audioEntity;
	}

	/**
	 * Internal identifier. Should not be exposed to clients.
	 * 
	 * @return
	 */
	@Id
	@GeneratedValue
	public Long getId() {
		return id;
	}

	/**
	 * Allows clients to refer this fingerprint..
	 * 
	 * @return
	 */
	@Column(nullable = false, unique = true, updatable = false)
	public String getIdentifier() {
		return identifier;
	}

	@ElementCollection(fetch = FetchType.LAZY)
	@OrderColumn
	@Column(name = "repeatCharList")
	private List<Character> getRepeatCharList() {
		return repeatCharList;
	}

	@ElementCollection(fetch = FetchType.LAZY)
	@OrderColumn
	@Column(name = "repeatLengthList")
	private List<Integer> getRepeatLengthList() {
		return repeatLengthList;
	}

	@ElementCollection(fetch = FetchType.LAZY)
	@OrderColumn
	@Column(name = "repeatPositionList")
	private List<Integer> getRepeatPositionList() {
		return repeatPositionList;
	}

	@Transient
	public List<Repeat> getRepeats() {
		if (this.repeats != null) {
			return this.repeats;
		}
		final List<Integer> positions = getRepeatPositionList();
		final List<Integer> lengths = getRepeatLengthList();
		final List<Character> chars = getRepeatCharList();
		final int size = positions.size();
		final List<Repeat> repeats = new ArrayList<Repeat>(size);
		for (int i = 0; i < size; i++) {
			repeats.add(new Repeat(chars.get(i).charValue(), positions.get(i)
					.intValue(), lengths.get(i).intValue()));
		}
		this.repeats = repeats;
		return repeats;
	}

	@ElementCollection
	@OrderColumn
	@Column(name = "segmentDataList")
	private List<String> getSegmentDataList() {
		return segmentDataList;
	}

	@ElementCollection
	@OrderColumn
	@Column(name = "segmentPositionList")
	private List<Integer> getSegmentPositionList() {
		return segmentPositionList;
	}

	/**
	 * Return list of {@link Segment}s from which this fingerprint consist of.
	 * 
	 * @return
	 */
	@Transient
	public List<Segment> getSegments() {
		if (this.segments != null) {
			return this.segments;
		}
		final List<Integer> segPosList = getSegmentPositionList();
		final List<String> slist = getSegmentDataList();
		final int size = segPosList.size();
		final List<Segment> segList = new ArrayList<Segment>(size);
		for (int i = 0; i < size; i++) {
			segList.add(new Segment(slist.get(i), segPosList.get(i)));
		}
		this.segments = segList;
		return segList;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((identifier == null) ? 0 : identifier.hashCode());
		return result;
	}

	public void setApprovals(Long approvals) {
		this.approvals = approvals;
	}

	public void setAudioEntity(AudioEntityDataObj audioEntity) {
		this.audioEntity = audioEntity;
	}

	@SuppressWarnings("unused")
	private void setId(Long id) {
		this.id = id;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	private void setRepeatCharList(List<Character> repeatCharList) {
		this.repeatCharList = repeatCharList;
	}

	private void setRepeatLengthList(List<Integer> repeatLengthList) {
		this.repeatLengthList = repeatLengthList;
	}

	private void setRepeatPositionList(List<Integer> repeatPositionList) {
		this.repeatPositionList = repeatPositionList;
	}

	public void setRepeats(List<Repeat> repeats) {
		final int size = repeats.size();
		final List<Integer> positions = new ArrayList<Integer>(size);
		final List<Integer> lengths = new ArrayList<Integer>(size);
		final List<Character> chars = new ArrayList<Character>(size);
		for (Repeat r : repeats) {
			positions.add(Integer.valueOf(r.getPosition()));
			lengths.add(Integer.valueOf(r.getLength()));
			chars.add(Character.valueOf(r.getCh()));
		}
		setRepeatPositionList(positions);
		setRepeatLengthList(lengths);
		setRepeatCharList(chars);
		this.repeats = repeats;
	}

	private void setSegmentDataList(List<String> segmentDataList) {
		this.segmentDataList = segmentDataList;
	}

	private void setSegmentPositionList(List<Integer> segmentPositionList) {
		this.segmentPositionList = segmentPositionList;
	}

	public void setSegments(List<Segment> segments) {
		final int size = segments.size();
		final List<Integer> positions = new ArrayList<Integer>(size);
		final List<String> datas = new ArrayList<String>(size);
		for (Segment s : segments) {
			positions.add(Integer.valueOf(s.getPosition()));
			datas.add(s.getData());
		}
		setSegmentPositionList(positions);
		setSegmentDataList(datas);
		this.segments = segments;
	}

	@Override
	public String toString() {
		return FingerprintDataObj.class.getSimpleName() + "[identifier="
				+ identifier + "]";
		// + " [audioEntity.identifier=" + audioEntity.getIdentifier()
		// + ", creationTime=" + creationTime + ", id=" + id
		// + ", identifier=" + identifier + ", repeatCharList="
		// + repeatCharList + ", repeatLengthList=" + repeatLengthList
		// + ", repeatPositionList=" + repeatPositionList
		// + ", segmentDataList=" + segmentDataList
		// + ", segmentPositionList=" + segmentPositionList + "]";
	}

}
