/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.Iterator;
import java.util.Map.Entry;

import org.jhears.data.DatabaseMap;
import org.jhears.filter.bloom.IBloomFilter;
import org.jhears.util.IterationObserver;
import org.jhears.util.AbstractIterationObserver;
import org.jhears.util.ObservableIterator;

import com.google.common.base.Predicate;
import com.google.common.collect.Iterators;

public class BloomFilteredDBMap<K, V> implements DatabaseMap<K, V> {

	private long failCnt;

	private final IBloomFilter<K> filter;

	private final Predicate<K> itrPredicate;

	private final DatabaseMap<K, V> map;

	public BloomFilteredDBMap(DatabaseMap<K, V> map, IBloomFilter<K> filter) {
		this.map = map;
		this.filter = filter;
		itrPredicate = new Predicate<K>() {
			@Override
			public boolean apply(K input) {
				return BloomFilteredDBMap.this.filter.contains(input);
			}
		};
		this.failCnt = 0L;
	}

	@Override
	public void clear() {
		map.clear();
		filter.clear();
		failCnt = 0L;
	}

	public boolean filterSupportsRemove() {
		return filter.supportsRemove();
	}

	@Override
	public V get(K key) {
		if (filter.contains(key)) {
			return map.get(key);
		}
		return null;
	}

	@Override
	public void getAll(final Iterator<K> iterator, IterationObserver<V> observer) {
		map.getAll(Iterators.filter(iterator, itrPredicate), observer);
	}

	/**
	 * 
	 * @return the number of times the {@link IBloomFilter#remove(Object)}
	 *         returned false when element(s) where removed from
	 *         {@link DatabaseMap}.
	 */
	public long getFailedFilterRemoves() {
		return failCnt;
	}

	@Override
	public void iterateKeys(IterationObserver<K> observer) {
		map.iterateKeys(observer);
	}

	@Override
	public void iterateMapEntries(IterationObserver<Entry<K, V>> observer) {
		map.iterateMapEntries(observer);
	}

	@Override
	public void put(K key, V value) {
		map.put(key, value);
		filter.add(key);
	}

	@Override
	public void putAll(Iterator<Entry<K, V>> iterator) {
		ObservableIterator<Entry<K, V>> itr = new ObservableIterator<Entry<K, V>>(
				iterator, new AbstractIterationObserver<Entry<K, V>>() {
					@Override
					public void observe(Entry<K, V> elt) {
						filter.add(elt.getKey());
					}
				});
		map.putAll(itr);
	}

	@Override
	public boolean remove(K key) {
		boolean removed = false;
		if (filter.contains(key)) {
			removed = map.remove(key);
			if (removed && filter.supportsRemove()) {
				if (!filter.remove(key)) {
					failCnt++;
				}
			}
		}
		return removed;
	}

	@Override
	public void removeAll(Iterator<K> iterator) {
		Iterator<K> itr;
		if (filter.supportsRemove()) {
			itr = new ObservableIterator<K>(iterator,
					new AbstractIterationObserver<K>() {
						@Override
						public void observe(K elt) {
							if (!filter.remove(elt)) {
								failCnt++;
							}
						}
					});
		} else {
			itr = iterator;
		}
		map.removeAll(itr);
	}

	@Override
	public long size() {
		return map.size();
	}

}
