/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Random;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.jhears.cache.DBMapRepeatCache;
import org.jhears.cache.DBMapSegmentCache;
import org.jhears.cluster.ClusterDetector;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.CnfManager;
import org.jhears.cnf.CnfUtil;
import org.jhears.cnf.Cnf.SearchKey;
import org.jhears.cnf.Cnf.ServerKey;
import org.jhears.data.BerkeleyMap;
import org.jhears.data.BerkeleyMapManager;
import org.jhears.data.CountingDatabaseSBFState;
import org.jhears.data.DAO;
import org.jhears.data.FPDAO;
import org.jhears.data.GeneralDao;
import org.jhears.data.HibernateSessionProvider;
import org.jhears.data.SessionProvider;
import org.jhears.filter.bloom.Hasher;
import org.jhears.filter.bloom.scalable.ScalableBloomFilter;
import org.jhears.filter.bloom.scalable.HVScalableBloomFilter;
import org.jhears.filter.bloom.util.BloomMath;
import org.jhears.filter.bloom.util.Growth;
import org.jhears.seq.CombinedSmoother;
import org.jhears.seq.FSeqFactory;
import org.jhears.seq.ISmoother;
import org.jhears.seq.IntQuantizer;
import org.jhears.seq.RLEnc;
import org.jhears.seq.RepeatGrouper;
import org.jhears.seq.SeqMatch;
import org.jhears.seq.SequenceSmoother;
import org.jhears.server.tool.RepeatProducer;
import org.jhears.util.DefaultHashFunction;
import org.jhears.util.HashFunction;
import org.jhears.util.LCG;
import org.jhears.util.RuntimeIOException;
import org.jhears.util.Serialization;
import org.jhears.util.Serializer;
import org.picocontainer.MutablePicoContainer;
import org.picocontainer.Parameter;
import org.picocontainer.parameters.ComponentParameter;
import org.picocontainer.parameters.ConstantParameter;
import org.picocontainer.parameters.NullParameter;

public class CatalogEngineBuilder {

	private final CnfManager cnfManager;
	private final MutablePicoContainer container;
	private final Random random;

	// private final Logger logger;
	// private final MutablePicoContainer container;

	public CatalogEngineBuilder(MutablePicoContainer container, Random random) throws IOException {
		CnfManager cnf = container.getComponent(CnfManager.class);
		if (cnf == null) {
			cnf = new CnfManager();
			container.addComponent(cnf);
		}
		this.cnfManager = cnf;
		this.container = container;
		this.random = random;

	}

	public FPCatalogEngine build() throws ConfigurationException {
		BerkeleyMapManager bdManager = new BerkeleyMapManager(
				getBerkelyHome(cnfManager));
		SessionProvider provider = getSessionProvider(cnfManager);
		Configuration serverConf = cnfManager.getConfiguration(Cnf.server);
		Configuration searchConf = cnfManager.getConfiguration(Cnf.search);
		Configuration fpConf = cnfManager.getConfiguration(Cnf.fingerprint);
		container.addComponent(bdManager);
		container.addComponent(provider);
		container.addComponent(DAO.class);
		container.addComponent(GeneralDao.class);
		container.addComponent(FPDAO.class);
		container.addComponent(LCG.class);
		container.addComponent(HashFunction.class, DefaultHashFunction.class);
		container.addComponent(Serialization.class, Serializer.class);
		container.addComponent(createRLEnc(searchConf));
		container.addComponent(createSmoother(searchConf));
		container.addComponent(RepeatProducer.class);
		container.addComponent(new FSeqFactory(fpConf));

		container.addComponent(SegmentSearchFilter.class,
				SegmentSearchFilter.class, new ConstantParameter(
						createSegmentClusterDetector(searchConf)));
		container.addComponent(RepeatSearchFilter.class,
				RepeatSearchFilter.class, new ConstantParameter(
						createRepeatClusterDetector(searchConf)));
		container.addComponent(Hasher.class);
		container.addComponent(createRepeatGrouper(searchConf));
		container.addComponent(createIntQuantizer(searchConf));
		container.addComponent(RepeatGroupKeyMaker.class);
		container.addComponent(createSeqMatcher(searchConf));
		container.addComponent(createRepeatInsertFilter(searchConf));
		container.addComponent(SegmentInsertFilter.class,
				new SegmentInsertFilter(fpConf, searchConf, random));
		container.start();

		MutablePicoContainer repeatContainer = container.makeChildContainer();
		buildRepeatContainer(serverConf, repeatContainer);
		DBMapRepeatCache rc = repeatContainer
				.getComponent(DBMapRepeatCache.class);
		container.addComponent(rc);

		MutablePicoContainer segmentContainer = container.makeChildContainer();
		buildSegmentContainer(serverConf, segmentContainer);
		DBMapSegmentCache sc = segmentContainer
				.getComponent(DBMapSegmentCache.class);
		container.addComponent(sc);

		container.addComponent(new SegmentCacheSearch(sc, fpConf, searchConf));
		container.addComponent(RepeatCacheFacade.class);

		container.addComponent(InsertableExtractor.class);
		container.addComponent(SearchFacade.class);

		container.addComponent(FingerprintStorageImpl.class);
		container.addComponent(FPCatalogEngine.class, FPCatalogEngine.class,
				new ComponentParameter(FingerprintStorageImpl.class),
				new ComponentParameter());
		return container.getComponent(FPCatalogEngine.class);
	}

	private void buildRepeatContainer(Configuration serverConf,
			MutablePicoContainer repeatContainer) {
		Parameter[] berkeleyPar = new Parameter[5];
		berkeleyPar[0] = new ComponentParameter();
		berkeleyPar[1] = new ConstantParameter("repeatGroupCacheDb");
		berkeleyPar[2] = NullParameter.INSTANCE;
		berkeleyPar[3] = new ComponentParameter();
		berkeleyPar[4] = new ComponentParameter();
		repeatContainer.addComponent(BerkeleyMap.class, BerkeleyMap.class,
				berkeleyPar);

		Growth growth = Growth.MAG_3;
		double p = serverConf
				.getDouble(Cnf.ServerKey.REPEAT_GROUP_CACHE_SBF_ERROR_PROB);
		int initialCapacity = serverConf
				.getInt(Cnf.ServerKey.REPEAT_GROUP_CACHE_SBF_INITIAL_SIZE);
		int m0 = BloomMath.optimalSliceSize(initialCapacity);
		double p0 = BloomMath.optimalSBFP0(p, growth.getR());
		Parameter[] statePar = new Parameter[8];
		statePar[0] = new ConstantParameter(growth.getS());
		statePar[1] = new ConstantParameter(growth.getR());
		statePar[2] = new ConstantParameter(m0);
		statePar[3] = new ConstantParameter(p0);
		statePar[4] = new ConstantParameter("repeatGroupSBFState");
		statePar[5] = new ComponentParameter();
		statePar[6] = new ComponentParameter();
		statePar[7] = new ComponentParameter();

		repeatContainer.addComponent(CountingDatabaseSBFState.class,
				CountingDatabaseSBFState.class, statePar);
		repeatContainer.addComponent(HVScalableBloomFilter.class);
		repeatContainer.addComponent(ScalableBloomFilter.class);
		repeatContainer.addComponent(BloomFilteredDBMap.class);
		repeatContainer.addComponent(DBMapRepeatCache.class,
				DBMapRepeatCache.class, new ComponentParameter(
						BloomFilteredDBMap.class));
	}

	private void buildSegmentContainer(Configuration serverConf,
			MutablePicoContainer segmentContainer) {
		Parameter[] berkeleyPar = new Parameter[5];
		berkeleyPar[0] = new ComponentParameter();
		berkeleyPar[1] = new ConstantParameter("segmentCacheDb");
		berkeleyPar[2] = NullParameter.INSTANCE;
		berkeleyPar[3] = new ComponentParameter();
		berkeleyPar[4] = new ComponentParameter();
		segmentContainer.addComponent(BerkeleyMap.class, BerkeleyMap.class,
				berkeleyPar);

		Growth growth = Growth.MAG_3;
		double p = serverConf
				.getDouble(Cnf.ServerKey.SEGMENT_CACHE_SBF_ERROR_PROB);
		int initialCapacity = serverConf
				.getInt(Cnf.ServerKey.SEGMENT_CACHE_SBF_INITIAL_SIZE);
		int m0 = BloomMath.optimalSliceSize(initialCapacity);
		double p0 = BloomMath.optimalSBFP0(p, growth.getR());
		Parameter[] statePar = new Parameter[8];
		statePar[0] = new ConstantParameter(growth.getS());
		statePar[1] = new ConstantParameter(growth.getR());
		statePar[2] = new ConstantParameter(m0);
		statePar[3] = new ConstantParameter(p0);
		statePar[4] = new ConstantParameter("segmentSBFState");
		statePar[5] = new ComponentParameter();
		statePar[6] = new ComponentParameter();
		statePar[7] = new ComponentParameter();

		segmentContainer.addComponent(CountingDatabaseSBFState.class,
				CountingDatabaseSBFState.class, statePar);
		segmentContainer.addComponent(HVScalableBloomFilter.class);
		segmentContainer.addComponent(ScalableBloomFilter.class);
		segmentContainer.addComponent(BloomFilteredDBMap.class);
		segmentContainer.addComponent(DBMapSegmentCache.class,
				DBMapSegmentCache.class, new ComponentParameter(
						BloomFilteredDBMap.class));

	}

	private IntQuantizer createIntQuantizer(Configuration searchConf) {
		String[] codb = searchConf
				.getStringArray(SearchKey.REPEAT_QUANTIZER_CODEBOOK);
		int[] codebook = CnfUtil.toIntArray(codb);
		int offset = searchConf
				.getInt(SearchKey.REPEAT_QUANTIZER_CONSTANT_MARGIN);
		double minMultiplier = searchConf
				.getDouble(SearchKey.REPEAT_QUANTIZER_MIN_MULTIPLIER);
		double maxMultiplier = searchConf
				.getDouble(SearchKey.REPEAT_QUANTIZER_MAX_MULTIPLIER);
		return new IntQuantizer(codebook, offset, minMultiplier, maxMultiplier);
	}

	private ClusterDetector createRepeatClusterDetector(Configuration searchConf) {
		int repeatMaxDist = searchConf
				.getInt(SearchKey.REPEAT_MAX_CLUSTER_DISTANCE);
		int repeatMinSize = searchConf
				.getInt(SearchKey.REPEAT_MIN_CLUSTER_SIZE);
		return new ClusterDetector(repeatMaxDist, repeatMinSize);
	}

	private RepeatGrouper createRepeatGrouper(Configuration searchConf) {
		int repeatsPerGroup = searchConf.getInt(SearchKey.REPEATS_PER_GROUP);
		return new RepeatGrouper(repeatsPerGroup);
	}

	private RepeatInsertFilter createRepeatInsertFilter(Configuration searchConf) {
		int minInsertSize = searchConf.getInt(SearchKey.REPEATS_FOR_INSERT);
		return new RepeatInsertFilter(minInsertSize);
	}

	private RLEnc createRLEnc(Configuration searchConf) {
		int minRleLen = searchConf.getInt(SearchKey.REPEAT_MIN_LENGHT);
		char ignoreChar = searchConf.getString(SearchKey.REPEAT_IGNORE_CHAR)
				.charAt(0);
		return new RLEnc(minRleLen, ignoreChar);
	}

	private ClusterDetector createSegmentClusterDetector(
			Configuration searchConf) {
		int maxDist = searchConf.getInt(SearchKey.MAX_CLUSTER_DISTANCE);
		int minSize = searchConf.getInt(SearchKey.MIN_SEARCH_SCORE);
		return new ClusterDetector(maxDist, minSize);
	}

	private SeqMatch createSeqMatcher(Configuration searchConf) {
		float sequenceMatchThreshold = searchConf
				.getFloat(SearchKey.SEQUENCE_MATCH_THRESHOLD);
		return new SeqMatch(sequenceMatchThreshold);
	}

	private ISmoother createSmoother(Configuration searchConf) {
		Collection<ISmoother> smoothers = new ArrayList<ISmoother>();

		String[] smoothersCnf = searchConf
				.getStringArray(SearchKey.SEQUENCE_SMOOTHERS);
		for (int i = 0; i < smoothersCnf.length / 2; i++) {
			int smootherMinSame = Integer.parseInt(smoothersCnf[i * 2]);
			int smootherMaxGap = Integer.parseInt(smoothersCnf[i * 2 + 1]);
			smoothers
					.add(new SequenceSmoother(smootherMinSame, smootherMaxGap));
		}
		return new CombinedSmoother(smoothers);
	}

	private File getBerkelyHome(CnfManager cnfMngr)
			throws ConfigurationException {
		Configuration serverConfiguration = cnfMngr
				.getConfiguration(Cnf.server);
		File defaultDataDir = new File(cnfMngr.getHomeDir(),
				serverConfiguration.getString(ServerKey.DATADIR));
		File envHome;
		String bdb = "bdb";
		String bdbProperty = Cnf.PREFIX + Cnf.SEPARATOR + bdb;
		if (System.getProperty(bdbProperty) != null) {
			envHome = new File(System.getProperty(bdbProperty));
		} else {
			envHome = new File(defaultDataDir, bdb);
		}

		if (!envHome.exists()) {
			if (!envHome.mkdirs()) {
				throw new RuntimeIOException(new IOException(
						"Failed to create directory: "
								+ envHome.getAbsolutePath()));
			}
		}
		return envHome;
	}

	private SessionProvider getSessionProvider(CnfManager cnfMngr) {
		File cnfDir = cnfMngr.getConfigDir();
		String hibernate = "hibernate";
		String hibernateProperty = Cnf.PREFIX + Cnf.SEPARATOR + hibernate;
		File hibernateCnfFile;
		if (System.getProperty(hibernateProperty) != null) {
			hibernateCnfFile = new File(System.getProperty(hibernateProperty));
		} else {
			hibernateCnfFile = new File(cnfDir, "hibernate.cfg.xml");
		}
		SessionProvider provider;
		if (hibernateCnfFile.exists()) {
			provider = new HibernateSessionProvider(hibernateCnfFile);
		} else {
			provider = new HibernateSessionProvider();
		}
		return provider;
	}

	// public CacheManager getCacheManager() {
	// return cacheManager;
	// }

	// public FPCatalogEngine getEngine() {
	// return engine;
	// }
	//
	// public GeneralDao getGeneralDao() {
	// return generalDao;
	// }

	// @SuppressWarnings("unchecked")
	// private Iterator<String> getTypeSafeKeyIterator(Configuration conf) {
	// return conf.getKeys();
	// }

	// @SuppressWarnings("unchecked")
	// private IRepeatGroupCache makeRepeatGroupCache(
	// Configuration serverConfiguration) {
	// // DatabaseMap<IRepeatGroupKey, Set<IRepeatGroupValue>> repeatGroupMap =
	// // new BerkeleyMap<IRepeatGroupKey, Set<IRepeatGroupValue>>(
	// // bdManager, "repeatGroupCacheDb", null, serialization,
	// // (HashFunction<IRepeatGroupKey>) hashFunction);
	// Growth growth = Growth.MAG_3;
	// double p = serverConfiguration
	// .getDouble(Cnf.ServerKey.REPEAT_GROUP_CACHE_SBF_ERROR_PROB);
	// int initialCapacity = serverConfiguration
	// .getInt(Cnf.ServerKey.REPEAT_GROUP_CACHE_SBF_INITIAL_SIZE);
	// int m0 = BloomMath.optimalSliceSize(initialCapacity);
	// double p0 = BloomMath.optimalSBFP0(p, growth.getR());
	// logger.info("Loading repeat group cache bloomfilter");
	// IScalableBloomFilterState state = new CountingDatabaseSBFState(growth
	// .getS(), growth.getR(), m0, p0, "repeatGroupSBFState",
	// bdManager, hashFunction, serialization);
	// logger.info("Repeat group cache bloomfilter loaded");
	// ScalableBloomFilterAlg segmentBfAlg = new ScalableBloomFilterAlg(state);
	// IBloomFilter<IRepeatGroupKey> filter = new
	// ScalableBloomFilter<IRepeatGroupKey>(
	// segmentBfAlg, hasher);
	// BloomFilteredDBMap<IRepeatGroupKey, Set<IRepeatGroupValue>> map = new
	// BloomFilteredDBMap<IRepeatGroupKey, Set<IRepeatGroupValue>>(
	// repeatGroupMap, filter);
	// return new DBMapRepeatCache(map);
	// }

	// @SuppressWarnings("unchecked")
	// private ISegmentCache makeSegmentCache(Configuration serverConfiguration)
	// {
	// DatabaseMap<String, Set<ISegmentCacheValue>> segmentMap = new
	// BerkeleyMap<String, Set<ISegmentCacheValue>>(
	// bdManager, "segmentCacheDb", null, serialization,
	// (HashFunction<String>) hashFunction);
	// Growth growth = Growth.MAG_3;
	// double p = serverConfiguration
	// .getDouble(Cnf.ServerKey.SEGMENT_CACHE_SBF_ERROR_PROB);
	// int initialCapacity = serverConfiguration
	// .getInt(Cnf.ServerKey.SEGMENT_CACHE_SBF_INITIAL_SIZE);
	// int m0 = BloomMath.optimalSliceSize(initialCapacity);
	// double p0 = BloomMath.optimalSBFP0(p, growth.getR());
	// logger.info("Loading segment cache bloomfilter");
	// IScalableBloomFilterState state = new CountingDatabaseSBFState(growth
	// .getS(), growth.getR(), m0, p0, "segmentSBFState", bdManager,
	// hashFunction, serialization);
	// logger.info("Segment cache bloomfilter loaded");
	// ScalableBloomFilterAlg segmentBfAlg = new ScalableBloomFilterAlg(state);
	// IBloomFilter<String> segmentFilter = new ScalableBloomFilter<String>(
	// segmentBfAlg, hasher);
	// BloomFilteredDBMap<String, Set<ISegmentCacheValue>> seMap = new
	// BloomFilteredDBMap<String, Set<ISegmentCacheValue>>(
	// segmentMap, segmentFilter);
	// return new DBMapSegmentCache(seMap);
	// }

	// private Map<String, Object> convert() {
	// HashMap<String, Object> fingerprintConfiguration = new HashMap<String,
	// Object>();
	// Iterator<String> ki = getTypeSafeKeyIterator(frameworkConfiguration);
	// while (ki.hasNext()) {
	// String key = ki.next();
	// fingerprintConfiguration.put(key, frameworkConfiguration
	// .getProperty(key));
	// }
	// return fingerprintConfiguration;
	// }
}
