/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.lang.StringUtils;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.CnfManager;
import org.jhears.cnf.CnfUtil;
import org.jhears.web.rs.AudioEntity;
import org.jhears.web.rs.Fingerprint;
import org.jhears.web.rs.SearchResult;
import org.jhears.web.rs.Version;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.base.Throwables;

public class FPCatalogEngine implements FingerprintStorage {

	private final ReentrantReadWriteLock lock;

	private final Logger logger;

	private final FingerprintStorage storageDelegate;

	private final CnfManager cnfManager;

	private final Map<String, Map<String, Object>> cMaps;

	private final Version version;

	public FPCatalogEngine(FingerprintStorage storageDelegate,
			CnfManager cnfManager) {
		this.cnfManager = cnfManager;
		logger = LoggerFactory.getLogger(FPCatalogEngine.class);
		lock = new ReentrantReadWriteLock(true);
		this.storageDelegate = storageDelegate;
		cMaps = new HashMap<String, Map<String, Object>>();
		String ver = CnfManager.getVersion("jhears-server");
		this.version = new Version(ver == null ? "Unknown" : ver, "jhears-server");
		try {
			populateCMaps();
		} catch (ConfigurationException e) {
			throw Throwables.propagate(e);
		}
	}

	public void populateCMaps() throws ConfigurationException {
		Cnf[] cArr = new Cnf[] { Cnf.audioPrePros, Cnf.fingerprint,
				Cnf.framework, Cnf.search };
		for (Cnf cnf : cArr) {
			cMaps.put(cnf.toString(),
					CnfUtil.toMap(cnfManager.getConfiguration(cnf)));
		}
	}

	@Override
	public boolean deleteAudioEntity(String audioEntityId, ICtx ctx) {
		WriteLock wlock = lock.writeLock();
		wlock.lock();
		try {
			boolean success = storageDelegate.deleteAudioEntity(audioEntityId,
					ctx);
			if (logger.isDebugEnabled()) {
				if (success) {
					logger.debug("Deleted audio entity with id="
							+ audioEntityId);
				} else {
					logger.debug("Failed to delete audio entity with id="
							+ audioEntityId);
				}
			}
			return success;
		} finally {
			wlock.unlock();
		}
	}

	public Map<String, Object> getConfMap(String name) {
		return cMaps.get(name);
	}

	@Override
	public boolean deleteFingerprint(String fingerprintId, ICtx ctx) {
		WriteLock wlock = lock.writeLock();
		wlock.lock();
		try {
			boolean success = storageDelegate.deleteFingerprint(fingerprintId,
					ctx);
			if (logger.isDebugEnabled()) {
				if (success) {
					logger.debug("Deleted fingerprint with id=" + fingerprintId);
				} else {
					logger.debug("Failed to delete fingerprint with id="
							+ fingerprintId);
				}
			}
			return success;
		} finally {
			wlock.unlock();
		}
	}

	@Override
	public AudioEntity getAudioEntity(String audioEntityId, ICtx ctx) {
		ReadLock rlock = lock.readLock();
		rlock.lock();
		try {
			AudioEntity entity = storageDelegate.getAudioEntity(audioEntityId,
					ctx);
			if (logger.isDebugEnabled()) {
				if (entity != null) {
					logger.debug("Audio entity found. id=" + audioEntityId);
				} else {
					logger.debug("Audio entity not found. id=" + audioEntityId);
				}
			}
			return entity;
		} finally {
			rlock.unlock();
		}
	}

	@Override
	public Fingerprint getFingerprint(String fingerprintId, ICtx ctx) {
		ReadLock rlock = lock.readLock();
		rlock.lock();
		try {
			Fingerprint fingerprint = storageDelegate.getFingerprint(
					fingerprintId, ctx);
			if (logger.isDebugEnabled()) {
				if (fingerprint != null) {
					logger.debug("Fingerprint found. id=" + fingerprintId);
				} else {
					logger.debug("Fingerprint not found. id=" + fingerprintId);
				}
			}
			return fingerprint;
		} finally {
			rlock.unlock();
		}
	}

	@Override
	public Fingerprint insertFingerprint(String audioEntityId,
			String fingerprintData, ICtx ctx) {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(fingerprintData),
				"Tried to insert an empty fingerprint");
		Preconditions.checkArgument(!Strings.isNullOrEmpty(audioEntityId),
				"Tried to insert a fingerprint with empty audioEntityId");
		WriteLock wlock = lock.writeLock();
		wlock.lock();
		try {
			Fingerprint fingerprint = storageDelegate.insertFingerprint(
					audioEntityId, fingerprintData, ctx);
			if (logger.isDebugEnabled()) {
				if (fingerprint != null) {
					logger.debug("Fingerprint inserted. fingerprintId="
							+ fingerprint.getIdentifier() + ",  audioEntityId="
							+ audioEntityId);
				} else {
					logger.debug("Fingerprint insert failed. audioEntityId="
							+ audioEntityId);
				}
			}
			return fingerprint;
		} finally {
			wlock.unlock();
		}
	}

	@Override
	public SearchResult[] search(String fingerprintData, ICtx ctx) {
		ReadLock rlock = lock.readLock();
		rlock.lock();
		try {
			SearchResult[] results = storageDelegate.search(fingerprintData,
					ctx);
			if (logger.isDebugEnabled()) {
				if (results != null && results.length > 0) {
					String[] ids = new String[results.length];
					for (int i = 0; i < results.length; i++) {
						StringBuilder sb = new StringBuilder(128);
						sb.append("[audioEntityId=")
								.append(results[i].getAudioEntityId())
								.append(", matchCount=")
								.append(results[i].getMatches().size())
								.append(']');
						ids[i] = sb.toString();
					}
					logger.debug("Search found: " + StringUtils.join(ids, ','));
				} else {
					logger.debug("No results for search.");
				}
			}
			return results;
		} finally {
			rlock.unlock();
		}
	}

	public Version getVersion() {
		return version;
	}

	@Override
	public Boolean approveFingerprint(String identifer, ICtx ctx) {
		return storageDelegate.approveFingerprint(identifer, ctx);
	}

}
