/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.jhears.data.AudioEntityDataObj;
import org.jhears.data.FPDAO;
import org.jhears.data.FingerprintDataObj;
import org.jhears.seq.FSeqFactory;
import org.jhears.seq.MutableFSeq;
import org.jhears.server.tool.RepeatProducer;
import org.jhears.tuple.Tuple2;
import org.jhears.web.rs.AudioEntity;
import org.jhears.web.rs.Fingerprint;
import org.jhears.web.rs.Match;
import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.RepeatCluster;
import org.jhears.web.rs.SearchResult;
import org.jhears.web.rs.Segment;
import org.jhears.web.rs.SegmentCluster;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class FingerprintStorageImpl implements FingerprintStorage {

	private final FSeqFactory fSeqFactory;

	private final InsertableExtractor insertableExtractor;

	private final Logger logger;

	private final FPDAO objAccess;

	private final RepeatProducer repeatProducer;

	private final SearchFacade searchFacade;

	public FingerprintStorageImpl(SearchFacade searchFacade,
			RepeatProducer repeateProducer, FPDAO objAccess,
			FSeqFactory fSeqFactory, InsertableExtractor insertHelper) {
		logger = LoggerFactory.getLogger(FingerprintStorageImpl.class);
		this.repeatProducer = repeateProducer;
		this.insertableExtractor = insertHelper;
		this.objAccess = objAccess;
		this.fSeqFactory = fSeqFactory;
		this.searchFacade = searchFacade;
	}

	/**
	 * 
	 * @param segmentResults
	 *            Map which key set represents
	 *            {@link FingerprintDataObj#getId()}.
	 * @param repeatResults
	 *            Map which key set represents
	 *            {@link FingerprintDataObj#getId()}.
	 * @return
	 */
	private SearchResult[] constructResults(
			Map<Long, List<SegmentCluster>> segmentResults,
			Map<Long, List<RepeatCluster>> repeatResults) {
		HashSet<Long> fpIds = new HashSet<Long>(segmentResults.size()
				+ repeatResults.size());
		fpIds.addAll(segmentResults.keySet());
		fpIds.addAll(repeatResults.keySet());
		List<AudioEntityDataObj> audioEntitys = objAccess
				.getAudioEntitiesByFingerprintId(fpIds);
		ArrayList<SearchResult> results = new ArrayList<SearchResult>(
				segmentResults.size());
		for (AudioEntityDataObj entity : audioEntitys) {
			HashSet<Match> matches = new HashSet<Match>();
			for (FingerprintDataObj fp : entity.getFingerprints()) {
				List<SegmentCluster> segClusters = segmentResults.get(fp
						.getId());
				List<RepeatCluster> repClusters = repeatResults.get(fp.getId());
				Match m = new Match(fp.getIdentifier(), segClusters, fp
						.getSegments(), repClusters, fp.getRepeats());
				matches.add(m);
			}
			String name = entity.getIdentifier();
			results.add(new SearchResult(name, matches));
		}
		return results.toArray(new SearchResult[0]);
	}

	@Override
	public boolean deleteAudioEntity(String audioEntityId, ICtx ctx) {
		boolean retval = false;
		AudioEntityDataObj entity = objAccess
				.getEntityWithIdentifier(audioEntityId);
		if (entity != null) {
			for (FingerprintDataObj fp : entity.getFingerprints()) {
				searchFacade.deleteFromCache(fp.getId().longValue(), fp
						.getSegments(), fp.getRepeats());
			}
			if (objAccess.deleteEntityWithIdentifier(audioEntityId,
					ctx == null ? null : ctx.getUser())) {
				retval = true;
				logger.info("Deleted successful. audioEntityId="
						+ audioEntityId);
			} else {
				logger.error("Failed to delete audio entity. audioEntityId="
						+ audioEntityId);
			}
		}
		return retval;
	}

	@Override
	public boolean deleteFingerprint(String fingerprintId, ICtx ctx) {
		boolean retval = false;
		FingerprintDataObj dataObj = objAccess
				.getFingerprintWithIdentifier(fingerprintId);
		if (dataObj != null) {
			searchFacade.deleteFromCache(dataObj.getId().longValue(), dataObj
					.getSegments(), dataObj.getRepeats());
			if (objAccess.deleteFingerprintWithIdentifier(fingerprintId,
					ctx == null ? null : ctx.getUser())) {
				logger.info("Deleted successful. fingerprintId="
						+ fingerprintId);
				retval = true;
			} else {
				logger.error("Failed to delete fingerprint. fingerprintId="
						+ fingerprintId);
			}
		}
		return retval;
	}

	@Override
	public AudioEntity getAudioEntity(String audioEntityId, ICtx ctx) {
		AudioEntityDataObj entity = objAccess
				.getEntityWithIdentifier(audioEntityId);
		AudioEntity retval = null;
		if (entity != null) {
			List<Fingerprint> list = new ArrayList<Fingerprint>(entity
					.getFingerprints().size());
			for (FingerprintDataObj fp : entity.getFingerprints()) {
				list.add(new Fingerprint(fp.getIdentifier(), entity
						.getIdentifier(), fp.getSegments(), fp.getRepeats(), fp
						.getApprovals().longValue()));
			}
			retval = new AudioEntity(entity.getIdentifier(), list);
		}
		return retval;
	}

	@Override
	public Fingerprint getFingerprint(String fingerprintId, ICtx ctx) {
		FingerprintDataObj fp = objAccess
				.getFingerprintWithIdentifierPlusEntity(fingerprintId);
		Fingerprint retval = null;
		if (fp != null) {
			retval = new Fingerprint(fp.getIdentifier(), fp.getAudioEntity()
					.getIdentifier(), fp.getSegments(), fp.getRepeats(), fp
					.getApprovals().longValue());
		}
		return retval;
	}

	private Fingerprint insert(String audioEntityId,
			Tuple2<List<Segment>, List<Repeat>> data, ICtx ctx) {
		FingerprintDataObj fp = null;
		if (!data.get1().isEmpty() || !data.get2().isEmpty()) {
			fp = objAccess.saveFingerprint(audioEntityId, data.get1(), data
					.get2(), ctx == null ? null : ctx.getUser());
			List<Segment> segs = fp.getSegments();
			List<Repeat> reps = fp.getRepeats();
			searchFacade.addToCache(fp.getId(), segs, reps);
			logger.info("Associated fingerprintId=" + fp.getIdentifier()
					+ " to entityId=" + audioEntityId);
			return new Fingerprint(fp.getIdentifier(), fp.getAudioEntity()
					.getIdentifier(), segs, reps, fp.getApprovals().longValue());
		}
		return null;
	}

	@Override
	public Fingerprint insertFingerprint(String audioEntityId,
			String fingerprintData, ICtx ctx) {
		MutableFSeq builder = fSeqFactory.getNewInstance();
		builder.append(fingerprintData);
		if (logger.isDebugEnabled()) {
			logger.debug("Attempting to insert fingerprint for "
					+ audioEntityId);
		}
		Tuple2<List<Segment>, List<Repeat>> insertable = insertableExtractor
				.extract(audioEntityId, builder);
		return insert(audioEntityId, insertable, ctx);
	}

	@Override
	public SearchResult[] search(String fingerprintData, ICtx ctx) {
		MutableFSeq seq = fSeqFactory.getNewInstance();
		seq.append(fingerprintData);
		List<Segment> slist = seq.toSegments();
		Map<Long, List<SegmentCluster>> segmentResults = searchFacade
				.searchSegments(slist);
		if (segmentResults.isEmpty()) {
			seq.flatten();
			List<Repeat> repeats = repeatProducer.toRepeats(seq.asString());
			Map<Long, List<RepeatCluster>> repeatResults = searchFacade
					.searchRepeats(repeats);
			if (repeatResults.isEmpty()) {
				return new SearchResult[0];
			}
			return constructResults(segmentResults, repeatResults);
		}
		return constructResults(segmentResults,
				new HashMap<Long, List<RepeatCluster>>(0));
	}

	@Override
	public Boolean approveFingerprint(String fingerprintId, ICtx ctx) {
		Boolean retval = objAccess.inApp(fingerprintId, ctx == null ? null
				: ctx.getUser());
		return retval;
	}

}
