/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.jhears.data.FPDAO;
import org.jhears.data.FingerprintDataObj;
import org.jhears.data.GeneralDao;
import org.jhears.seq.MutableFSeq;
import org.jhears.seq.SeqMatch;
import org.jhears.server.tool.RepeatProducer;
import org.jhears.tuple.Tuple2;
import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.RepeatCluster;
import org.jhears.web.rs.Segment;
import org.jhears.web.rs.SegmentCluster;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class InsertableExtractor {

	private final GeneralDao generalDao;

	private final Logger logger;

	private final SeqMatch matcher;

	private final RepeatProducer repeateProducer;

	private final RepeatInsertFilter repeatInsertFilter;

	private final SearchFacade searchFacade;

	private final SegmentInsertFilter segmentInsertFilter;

	public InsertableExtractor(SegmentInsertFilter segmentInsertFilter,
			SeqMatch matcher, RepeatProducer repeateProducer,
			RepeatInsertFilter repeatInsertFilter, SearchFacade searchFacade,
			GeneralDao generalDao) {
		this.generalDao = generalDao;
		this.logger = LoggerFactory.getLogger(InsertableExtractor.class);
		this.segmentInsertFilter = segmentInsertFilter;
		this.matcher = matcher;
		this.repeateProducer = repeateProducer;
		this.repeatInsertFilter = repeatInsertFilter;
		this.searchFacade = searchFacade;
	}

	public Tuple2<List<Segment>, List<Repeat>> extract(String audioEntityId,
			MutableFSeq builder) {
		List<Segment> segments = builder.toSegments();
		builder.flatten();
		List<Segment> existingSegments = locateExistingMatchingSegments(
				audioEntityId, builder, segments);
		List<Segment> insertableSegments = segmentInsertFilter
				.extractInsertableSegments(builder, existingSegments);
		List<Repeat> seqRepeats = repeateProducer.toRepeats(builder.asString());
		List<Repeat> existingRepeats = locateExistingMatchingRepeats(
				audioEntityId, builder, seqRepeats);
		List<Repeat> insertableRepeats = repeatInsertFilter
				.extractInsertableRepeats(seqRepeats, existingRepeats);
		return new Tuple2<List<Segment>, List<Repeat>>(insertableSegments,
				insertableRepeats);
	}

	private List<Repeat> locateExistingMatchingRepeats(String audioEntityId,
			MutableFSeq builder, List<Repeat> seqRepeats) {
		List<Repeat> retval = new ArrayList<Repeat>();
		Map<Long, List<RepeatCluster>> searchResults = searchFacade
				.searchRepeats(seqRepeats);

		for (Map.Entry<Long, List<RepeatCluster>> entry : searchResults
				.entrySet()) {
			Long fpId = entry.getKey();
			List<RepeatCluster> v = entry.getValue();
			FPDAO dd = new FPDAO(generalDao);
			FingerprintDataObj fpDataObj = dd.getFingerpritWithId(fpId);
			// TODO when delete is implemented, check if null
			List<Repeat> databaseRepeats = fpDataObj.getRepeats();
			for (RepeatCluster result : v) {
				if (matcher.repeatsInSeq(databaseRepeats, result
						.getRelativePosition(), builder)) {
					// user is trying to submit already existing
					// fingerprint
					if (logger.isDebugEnabled()) {
						logger.debug("Fingerprint already exists fpId=" + fpId
								+ " entityId=" + audioEntityId);
					}
					// record already existing segments
					List<Repeat> querSegments = result.getRepeatSearchEntries();
					retval.addAll(querSegments);
				} else {
					// better log this for later analysis
					logger
							.warn("Possible false positive during pre-insert search. fpId="
									+ fpId + " entityId=" + audioEntityId);
				}
			}
		}
		return retval;
	}

	/**
	 * Performs fingerprint search and returns segments which matched the query
	 * sequence.
	 * 
	 * @param audioEntityId
	 * @param builder
	 * @return
	 */
	private List<Segment> locateExistingMatchingSegments(String audioEntityId,
			MutableFSeq builder, List<Segment> allSegments) {
		List<Segment> retval = new ArrayList<Segment>();
		Map<Long, List<SegmentCluster>> searchResults = searchFacade
				.searchSegments(allSegments);
		for (Map.Entry<Long, List<SegmentCluster>> entry : searchResults
				.entrySet()) {
			Long fpId = entry.getKey();
			List<SegmentCluster> v = entry.getValue();
			FPDAO oo = new FPDAO(generalDao);
			FingerprintDataObj fpDataObj = oo.getFingerpritWithId(fpId);
			// TODO when delete is implemented, check if null
			List<Segment> databaseSegments = fpDataObj.getSegments();
			for (SegmentCluster result : v) {
				if (matcher.segmentsInSeq(databaseSegments, result
						.getRelativePosition(), builder)) {
					// user is trying to submit already existing
					// fingerprint
					if (logger.isDebugEnabled()) {
						logger.debug("Fingerprint already exists fpId=" + fpId
								+ " entityId=" + audioEntityId);
					}
					// record already existing segments
					List<Segment> querSegments = result.getQuerySegments();
					retval.addAll(querSegments);
				} else {
					// better log this for later analysis
					logger
							.warn("Possible false positive during pre-insert search. fpId="
									+ fpId + " entityId=" + audioEntityId);
				}
			}
		}
		return retval;
	}

}
