/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jhears.cluster.Cluster;
import org.jhears.cluster.ClusterDetector;
import org.jhears.util.IntList;
import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.RepeatCluster;

/**
 * Detects closely packed {@link RepeatSearchEntry}s.
 * <p>
 * This class is thread safe.
 */
public class RepeatSearchFilter {

	/**
	 * Compare {@link RepeatSearchEntry}s based on
	 * {@link RepeatSearchEntry#getRelativePosition()} values.
	 */
	private static final Comparator<RepeatSearchEntry> REL_POS_CMP = new Comparator<RepeatSearchEntry>() {

		@Override
		public int compare(RepeatSearchEntry o1, RepeatSearchEntry o2) {
			int a1 = o1.getRelativePosition();
			int a2 = o2.getRelativePosition();
			return (a1 < a2 ? -1 : (a1 == a2 ? 0 : 1));
		}

	};

	private final ClusterDetector clusterDetector;

	public RepeatSearchFilter(ClusterDetector clusterDetector) {
		this.clusterDetector = clusterDetector;
	}

	/**
	 * Detects clusters in data and returns all found clusters.
	 * 
	 * @param list
	 * @return
	 */
	private List<RepeatCluster> clusterAndConvert(List<RepeatSearchEntry> list) {
		Collections.sort(list, REL_POS_CMP);
		int[] positions = relativePositions(list);
		List<Cluster> clusters = clusterDetector.detectClusters(positions);
		return convert(clusters, list);
	}

	/**
	 * Group {@link RepeatSearchEntry}s into {@link RepeatCluster}s based on
	 * information available in <code>clusters</code> argument.
	 * 
	 * @param clusters
	 * @param list
	 * @return
	 */
	private List<RepeatCluster> convert(List<Cluster> clusters,
			List<RepeatSearchEntry> list) {
		List<RepeatCluster> retval = new ArrayList<RepeatCluster>();
		for (Cluster cluster : clusters) {
			int[] indexes = cluster.getIndexes();
			if (indexes.length > 0) {
				List<Repeat> rseList = new ArrayList<Repeat>(indexes.length);
				IntList relativePositions = new IntList(
						indexes.length);
				for (int index : indexes) {
					RepeatSearchEntry rse = list.get(index);
					rseList.add(rse.getRepeat());
					relativePositions.add(rse.getRelativePosition());
				}
				int position = getAvg(relativePositions);
				retval.add(new RepeatCluster(rseList, position));
			}
		}
		return retval;
	}

	/**
	 * Detects argument contains clustered {@link RepeatSearchEntry}s and
	 * returns them.
	 * 
	 * @param map
	 * @return Values with very high confidence being correct.
	 */
	public Map<Long, List<RepeatCluster>> filter(
			Map<Long, Set<RepeatSearchEntry>> map) {
		Map<Long, List<RepeatCluster>> retval = new HashMap<Long, List<RepeatCluster>>();
		for (Map.Entry<Long, Set<RepeatSearchEntry>> entry : map.entrySet()) {
			Long fpId = entry.getKey();
			Set<RepeatSearchEntry> v = entry.getValue();
			List<RepeatSearchEntry> list = new ArrayList<RepeatSearchEntry>(v);
			List<RepeatCluster> cluster = clusterAndConvert(list);
			if (!cluster.isEmpty()) {
				retval.put(fpId, cluster);
			}
		}
		return retval;
	}

	/**
	 * Calculate average.
	 * 
	 * @param repos
	 * @return
	 */
	private int getAvg(IntList repos) {
		int size = repos.size();
		int sum = 0;
		for (int i = 0; i < size; i++) {
			sum += repos.get(i);
		}
		return (int) Math.round((double) sum / (double) size);
	}

	/**
	 * Extract relative positions from list entries.
	 * 
	 * @param list
	 * @return
	 */
	private int[] relativePositions(List<RepeatSearchEntry> list) {
		int[] arr = new int[list.size()];
		for (int i = 0; i < arr.length; i++) {
			arr[i] = list.get(i).getRelativePosition();
		}
		return arr;
	}

}
