/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.apache.commons.configuration.Configuration;
import org.jhears.cnf.Cnf.FingerprintKey;
import org.jhears.cnf.Cnf.SearchKey;
import org.jhears.seq.MutableFSeq;
import org.jhears.server.tool.SeqUtil;
import org.jhears.util.Mth;
import org.jhears.web.rs.Segment;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SegmentInsertFilter {

	private final char censorChar;

	private final int complexityMargin;

	private final int complexityTreshold;

	private final Logger logger;

	private final int maxDuplicateSegments;

	private final int minSearchHits;

	private final int minUniqueSeg;

	private final int nonComplexMargin;

	private final Random random;

	private final int seglen;

	public SegmentInsertFilter(Configuration fpConf, Configuration searchConf,
			Random random) {
		logger = LoggerFactory.getLogger(getClass());
		seglen = fpConf.getInt(FingerprintKey.SEGMENT_LENGTH);
		minUniqueSeg = searchConf.getInt(SearchKey.MIN_UNIQ_SEGMENTS);
		maxDuplicateSegments = searchConf
				.getInt(SearchKey.MAX_DUPLICATE_SEGMENTS);
		complexityTreshold = searchConf.getInt(SearchKey.COMPLEXITY_TRESHOLD);
		complexityMargin = searchConf.getInt(SearchKey.COMPLEXITY_MARGIN);
		censorChar = fpConf.getString(FingerprintKey.SYMBOLS).charAt(
				fpConf.getInt(FingerprintKey.ZERO_SYMBOL_IDX));
		minSearchHits = searchConf.getInt(SearchKey.MINIMUM_HITS);
		nonComplexMargin = searchConf.getInt(SearchKey.NONCOMPLEX_MARGIN);
		this.random = random == null ? new Random() : random;
	}

	public List<Segment> extractInsertableSegments(MutableFSeq origSeq,
			List<Segment> existingSegments) {
		MutableFSeq workSeq = origSeq.copy();
		workSeq.replaceLowComplexity(seglen, complexityTreshold,
				complexityMargin, censorChar);
		List<Segment> hiComplexSegs = workSeq.getNonLowComplexitySegments(
				nonComplexMargin, censorChar);
		Collections.sort(hiComplexSegs, Segment.POSITION_COMPARATOR);
		SeqUtil.removeOrphans(hiComplexSegs, seglen, minSearchHits);
		Map<String, List<Segment>> segmentsByData = groupBySegmentData(hiComplexSegs);
		List<Integer> remainingPositions = removeExcessiveDuplicates(
				hiComplexSegs, segmentsByData);
		List<Segment> fromOriginal = origSeq
				.getSegmentsFromPositions(remainingPositions);
		fromOriginal.removeAll(existingSegments);
		Collections.sort(fromOriginal, Segment.POSITION_COMPARATOR);
		SeqUtil.removeOrphans(fromOriginal, seglen, minSearchHits);
		if (fromOriginal.size() < minUniqueSeg) {
			return new ArrayList<Segment>(0);
		} else {
			return fromOriginal;
		}
	}

	public char getCensorChar() {
		return censorChar;
	}

	public int getMinUniqueSegments() {
		return minUniqueSeg;
	}

	private Map<String, List<Segment>> groupBySegmentData(List<Segment> slist) {
		HashMap<String, List<Segment>> segmentsByData = new HashMap<String, List<Segment>>(
				slist.size());
		for (Segment s : slist) {
			List<Segment> list = segmentsByData.get(s.getData());
			if (list == null) {
				list = new ArrayList<Segment>(2);
				segmentsByData.put(s.getData(), list);
			}
			list.add(s);
		}
		return segmentsByData;
	}

	private List<Integer> removeExcessiveDuplicates(List<Segment> slist,
			Map<String, List<Segment>> duplicateSegmentMap) {
		List<Integer> save = new ArrayList<Integer>(slist.size());

		for (Map.Entry<String, List<Segment>> entry : duplicateSegmentMap
				.entrySet()) {
			String key = entry.getKey();
			removeExtraDuplicates(key, duplicateSegmentMap);
			for (Segment seg : entry.getValue()) {
				save.add(Integer.valueOf(seg.getPosition()));
			}
		}
		return save;
	}

	/**
	 * If map has too many segments matching the key then the extra entries are
	 * filtered out by selecting only {@link #maxDuplicateSegments}.
	 * 
	 * @param key
	 * @param duplicates
	 */
	private void removeExtraDuplicates(String key,
			Map<String, List<Segment>> duplicates) {
		if (duplicates.get(key).size() > maxDuplicateSegments) {
			if (logger.isTraceEnabled()) {
				logger.trace("Too many hits " + duplicates.get(key).size()
						+ "for segment " + key + ". Selecting randomly "
						+ maxDuplicateSegments + " entries.");
			}
			List<Segment> newList = Mth.sampleRandom(maxDuplicateSegments,
					duplicates.get(key), random);
			duplicates.put(key, newList);
		}
	}

}
