/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jhears.cluster.Cluster;
import org.jhears.cluster.ClusterDetector;
import org.jhears.util.IntList;
import org.jhears.web.rs.Segment;
import org.jhears.web.rs.SegmentCluster;

/**
 * Detects closely packed {@link PotentialMatch}s.
 * <p>
 * This class is thread safe.
 */
public class SegmentSearchFilter {

	/**
	 * Compares {@link PotentialMatch} objects based on
	 * {@link PotentialMatch#getRelativePosition()}.
	 */
	private static final Comparator<PotentialMatch> POT_MATCH_POS_CMP = new Comparator<PotentialMatch>() {

		@Override
		public int compare(PotentialMatch o1, PotentialMatch o2) {
			int a1 = o1.getRelativePosition();
			int a2 = o2.getRelativePosition();
			return (a1 < a2 ? -1 : (a1 == a2 ? 0 : 1));
		}

	};

	private final ClusterDetector clusterDetector;

	public SegmentSearchFilter(ClusterDetector clusterDetector) {
		this.clusterDetector = clusterDetector;
	}

	/**
	 * Perform cluster detection and by using results from cluster detection
	 * convert data into {@link SegmentCluster}s.
	 * 
	 * @param list
	 * @return
	 */
	private List<SegmentCluster> clusterAndConvert(List<PotentialMatch> list) {
		Collections.sort(list, POT_MATCH_POS_CMP);
		int[] positions = relativePositions(list);
		List<Cluster> clusters = clusterDetector.detectClusters(positions);
		return convertClustersToHighScoringResults(clusters, list);
	}

	private List<SegmentCluster> convertClustersToHighScoringResults(
			List<Cluster> clusters, List<PotentialMatch> list) {
		List<SegmentCluster> retval = new ArrayList<SegmentCluster>();
		for (Cluster cluster : clusters) {
			int[] indexes = cluster.getIndexes();
			if (indexes.length > 0) {
				List<Segment> querySegments = new ArrayList<Segment>(
						indexes.length);
				IntList repos = new IntList(indexes.length);
				for (int index : indexes) {
					PotentialMatch pmatch = list.get(index);
					querySegments.add(pmatch.getQuerySegment());
					repos.add(pmatch.getRelativePosition());
				}
				int position = getAvg(repos);
				retval.add(new SegmentCluster(position, querySegments));
			}
		}
		return retval;
	}

	/**
	 * Detects high scoring clusters from given argument.
	 * 
	 * @param map
	 * @return Values with very high confidence being correct.
	 */
	public Map<Long, List<SegmentCluster>> filter(Map<Long, SearchEntry> map) {
		Map<Long, List<SegmentCluster>> retval = new HashMap<Long, List<SegmentCluster>>();
		for (Map.Entry<Long, SearchEntry> entry : map.entrySet()) {
			Long fpId = entry.getKey();
			List<PotentialMatch> list = merge(entry.getValue().getMap()
					.values());
			List<SegmentCluster> cluster = clusterAndConvert(list);
			if (!cluster.isEmpty()) {
				retval.put(fpId, cluster);
			}
		}
		return retval;
	}

	private int getAvg(IntList repos) {
		int size = repos.size();
		int sum = 0;
		for (int i = 0; i < size; i++) {
			sum += repos.get(i);
		}
		return (int) Math.round((double) sum / (double) size);
	}

	/**
	 * Puts contents of all given lists into to new list.
	 * 
	 * @param potentialMatches
	 * @return
	 */
	private List<PotentialMatch> merge(
			Collection<List<PotentialMatch>> potentialMatches) {
		List<PotentialMatch> retval = new ArrayList<PotentialMatch>();
		for (List<PotentialMatch> list : potentialMatches) {
			retval.addAll(list);
		}
		return retval;
	}

	/**
	 * Extracts {@link PotentialMatch#getRelativePosition()}s from given list.
	 * 
	 * @param list
	 * @return
	 */
	private int[] relativePositions(List<PotentialMatch> list) {
		int[] arr = new int[list.size()];
		for (int i = 0; i < arr.length; i++) {
			arr[i] = list.get(i).getRelativePosition();
		}
		return arr;
	}

}
