/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.server.tool;

import java.util.ArrayList;
import java.util.List;

import org.jhears.web.rs.Segment;

public class SeqUtil {

	/**
	 * Removes orphaned {@link Segment}s from the <code>list</code>. Orphaned
	 * {@link Segment}s are ones which have less than
	 * <code>requredAdjacent</code> adjacent siblings. The check is based on
	 * value of the {@link Segment#getPosition()}.
	 * <p>
	 * List must be sorted. {@link Segment#getPosition()} must be multiple of
	 * <code>segLen</code>. Each {@link Segment#getPosition()} must return a
	 * unique value.
	 * 
	 * @param list
	 *            List sorted to ascending order by
	 *            {@link Segment#getPosition()}. Elements which return same
	 *            value for {@link Segment#getPosition()} are removed.
	 * @param segLen
	 *            length between two adjacent {@link Segment}s.
	 * @param requredAdjacent
	 *            the number of adjacent {@link Segment}s which the
	 *            {@link Segment} must have. Otherwise the {@link Segment} is
	 *            removed from the list.
	 * @throws IllegalArgumentException
	 *             if segLen is less than 1 or if requredAdjacent is less than
	 *             two.
	 */
	public static void removeOrphans(List<Segment> list, int segLen,
			int requredAdjacent) throws IllegalArgumentException {
		if (segLen < 1) {
			throw new IllegalArgumentException(
					"Segment length must be greater than zero.");
		}
		if (requredAdjacent < 2) {
			throw new IllegalArgumentException(
					"Required adjacent value must be greater than one.");
		}
		int len = list.size();
		if (len < requredAdjacent) {
			list.clear();
			return;
		}
		int okCnt = 0; // the number of adjacent segments so far
		int i = 0; // current list index
		len--; // prevent list.get(i + 1) from exploding
		requredAdjacent--; // allow to check with okCnt < requredAdjacent
		for (; i < len; i++) {
			if (list.get(i).getPosition() == list.get(i + 1).getPosition()
					- segLen) {
				okCnt++;
			} else if (okCnt < requredAdjacent) {
				// remove okCnt + 1 elements
				for (int j = 0; j < okCnt + 1; j++) {
					list.remove(i - okCnt);
				}
				i -= okCnt + 1;
				len -= okCnt + 1;
				okCnt = 0;
			} else {
				// had at least requredAdjacent elements.
				// start checking the next one.
				okCnt = 0;
			}
		}
		if (okCnt < requredAdjacent) {
			// remove okCnt + 1 elements
			for (int j = 0; j < okCnt + 1; j++) {
				list.remove(i - okCnt);
			}
		}
	}

	/**
	 * Selects <code>selectCnt</code> first entries from list which have at
	 * least <code>requredAdjacent</code> adjacent entries.
	 * 
	 * @param list
	 *            list from which to select segments
	 * @param segLen
	 *            Segment length.
	 * @param requredAdjacent
	 *            The number of segments which must be adjacent before they are
	 *            considered.
	 * @param selectCnt
	 *            The number of segments to include at minimum.
	 * @return at least <code>selectCnt</code> segments or an empty list if
	 *         selection failed.
	 */
	public static List<Segment> select(List<Segment> list, int segLen,
			int requredAdjacent, int selectCnt) {
		if (segLen < 1) {
			throw new IllegalArgumentException(
					"Segment length must be greater than zero.");
		}
		if (requredAdjacent < 2) {
			throw new IllegalArgumentException(
					"Required adjacent value must be greater than one.");
		}
		if (selectCnt < requredAdjacent) {
			throw new IllegalArgumentException(
					"Select count must be greater than or equal to require adjacent.");
		}
		List<Segment> selected = new ArrayList<Segment>();
		int len = list.size();
		if (len < requredAdjacent) {
			return selected;
		}
		int okCnt = 0; // the number of adjacent segments so far
		int i = 0; // current list index
		len--; // prevent list.get(i + 1) from exploding
		requredAdjacent--; // allow to check with okCnt < requredAdjacent
		for (; i < len; i++) {
			if (list.get(i).getPosition() == list.get(i + 1).getPosition()
					- segLen) {
				okCnt++;
			} else if (okCnt < requredAdjacent) {
				okCnt = 0;
			} else {
				for (int j = i - okCnt; j < i + 1; j++) {
					selected.add(list.get(j));
					if (selected.size() >= selectCnt
							&& j >= i - okCnt + requredAdjacent) {
						return selected;
					}
				}
				okCnt = 0;
			}
		}
		if (okCnt > 0) {
			for (int j = i - okCnt; j < i + 1; j++) {
				selected.add(list.get(j));
				if (selected.size() >= selectCnt
						&& j >= i - okCnt + requredAdjacent) {
					return selected;
				}
			}
		}
		// fail
		selected.clear();
		return selected;
	}

}
