/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.catalog;

import java.beans.PropertyVetoException;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Random;
import java.util.Set;

import javax.naming.NamingException;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.FileUtils;
import org.jhears.cnf.CnfManager;
import org.jhears.server.CatalogEngineBuilder;
import org.jhears.server.FPCatalogEngine;
import org.jhears.web.rs.Fingerprint;
import org.jhears.web.rs.InsertRequest;
import org.jhears.web.rs.Match;
import org.jhears.web.rs.SearchResult;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.picocontainer.DefaultPicoContainer;

import com.google.common.io.Files;

public class FPCatalogEngineIntegrationTest {

	private static File tmpDir;

	@BeforeClass
	public static void setupClass() throws IOException, PropertyVetoException,
			NamingException {

		tmpDir = Files.createTempDir();
		System.setProperty("database.home", tmpDir.getAbsolutePath());
		System.setProperty(CnfManager.PROP_NAME_HOME, tmpDir.getAbsolutePath());
	}

	@AfterClass
	public static void tearDownClass() throws IOException {
		FileUtils.deleteDirectory(tmpDir);
		// Files.deleteRecursively(tmpDir); // guava bug 359
	}

	private Random r;

	private char[] fakeInput(int len, int maxLen) {
		StringBuilder sb = new StringBuilder(len * maxLen);
		for (char c : getRnd(len).toCharArray()) {
			int x = r.nextInt(maxLen);
			char[] arr = new char[x];
			Arrays.fill(arr, c);
			sb.append(arr);
		}
		return sb.toString().toCharArray();
	}

	private String getRnd(int len) {
		byte[] data = new byte[len];
		r.nextBytes(data);
		return new String(Base64.encodeBase64(data)).substring(0, len);
	}

	@Before
	public void setup() {
		r = new Random(0);
	}

	@Test
	public void testInsertSearch() throws ConfigurationException, IOException {

		CatalogEngineBuilder builder = new CatalogEngineBuilder(
				new DefaultPicoContainer(), r);
		FPCatalogEngine engine = builder.build();

		/*
		 * Test Segment inserting
		 */
		InsertRequest insertReq = new InsertRequest(getRnd(6), getRnd(256));
		// not inserted yet
		SearchResult[] res = engine.search(insertReq.getFingerprint(), null);
		Assert.assertEquals(0, res.length);

		Fingerprint insertResult = engine.insertFingerprint(insertReq
				.getAudioEntityId(), insertReq.getFingerprint(), null);
		Assert.assertEquals(insertReq.getAudioEntityId(), insertResult
				.getAudioEntityId());

		// search for it
		res = engine.search(insertReq.getFingerprint(), null);
		Assert.assertEquals(1, res.length);
		Assert.assertEquals(insertReq.getAudioEntityId(), res[0]
				.getAudioEntityId());
		Set<Match> matches = res[0].getMatches();
		Assert.assertEquals(1, matches.size());
		Match match = matches.iterator().next();
		Assert.assertEquals(insertResult.getIdentifier(), match
				.getFingerprintId());
		Assert.assertTrue(match.getRepeats().isEmpty());
		Assert.assertFalse(match.getSegments().isEmpty());

		// check loading with fingerprint id
		engine.getFingerprint(res[0].getMatches().iterator().next()
				.getFingerprintId(), null);

		// reload caches and see if the data is still available
		// TODO: fixme? pro.getCacheManager().reloadCaches();
		res = engine.search(insertReq.getFingerprint(), null);
		Assert.assertEquals(1, res.length);
		Assert.assertEquals(insertReq.getAudioEntityId(), res[0]
				.getAudioEntityId());
		matches = res[0].getMatches();
		Assert.assertEquals(1, matches.size());
		match = matches.iterator().next();
		Assert.assertTrue(match.getRepeats().isEmpty());
		Assert.assertFalse(match.getSegments().isEmpty());
		Assert.assertEquals(insertResult.getIdentifier(), matches.iterator()
				.next().getFingerprintId());

		// /*
		// * try to insert fingerprint again. This succeeds since the
		// fingerprint
		// * contains non-indexed data.
		// */
		// insertResult = engine.insertFingerprint(insertReq.getAudioEntityId(),
		// insertReq.getFingerprint());
		//
		// res = engine.search(insertReq.getFingerprint());
		// Assert.assertEquals(1, res.length);
		// Assert.assertEquals(insertReq.getAudioEntityId(), res[0]
		// .getAudioEntityId());
		// matches = res[0].getMatches();
		// Assert.assertEquals(2, matches.size());

		// do some basic checks. validating fingerprint identifiers is tricky
		// since the set ordering is not consistent.

		for (Match m : matches) {
			Assert.assertTrue(m.getRepeats().isEmpty());
			Assert.assertFalse(m.getSegments().isEmpty());
		}
		Assert.assertEquals(insertReq.getAudioEntityId(), insertResult
				.getAudioEntityId());

		/*
		 * Insert new entity using repeats.
		 */
		insertReq = new InsertRequest(getRnd(6), new String(fakeInput(80, 512)));
		// not inserted yet
		res = engine.search(insertReq.getFingerprint(), null);
		Assert.assertEquals(0, res.length);
		insertResult = engine.insertFingerprint(insertReq.getAudioEntityId(),
				insertReq.getFingerprint(), null);

		res = engine.search(insertReq.getFingerprint(), null);
		Assert.assertEquals(1, res.length);
		Assert.assertEquals(insertReq.getAudioEntityId(), res[0]
				.getAudioEntityId());
		matches = res[0].getMatches();
		Assert.assertEquals(1, matches.size());
		match = matches.iterator().next();
		Assert.assertFalse(match.getRepeats().isEmpty());
		Assert.assertTrue(match.getSegments().isEmpty());
		Assert.assertEquals(insertResult.getIdentifier(), matches.iterator()
				.next().getFingerprintId());

		// reload caches and see if the data is still available
		// TODO: fixme? pro.getCacheManager().reloadCaches();
		res = engine.search(insertReq.getFingerprint(), null);
		Assert.assertEquals(1, res.length);
		Assert.assertEquals(insertReq.getAudioEntityId(), res[0]
				.getAudioEntityId());
		matches = res[0].getMatches();
		Assert.assertEquals(1, matches.size());
		match = matches.iterator().next();
		Assert.assertFalse(match.getRepeats().isEmpty());
		Assert.assertTrue(match.getSegments().isEmpty());
		Assert.assertEquals(insertResult.getIdentifier(), matches.iterator()
				.next().getFingerprintId());

		System.out.println(Arrays.toString(res));

	}
}
