/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.catalog;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.FileUtils;
import org.jhears.cache.MemRepeatGroupCache;
import org.jhears.cnf.CnfManager;
import org.jhears.seq.IntQuantizer;
import org.jhears.seq.RLEnc;
import org.jhears.seq.RepeatGrouper;
import org.jhears.server.RepeatCacheFacade;
import org.jhears.server.RepeatGroupKeyMaker;
import org.jhears.server.RepeatSearchEntry;
import org.jhears.util.Mth;
import org.jhears.web.rs.Repeat;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.google.common.io.Files;

public class RepeatCacheFacadeTest {

	private static final Comparator<Repeat> CMP = new Comparator<Repeat>() {

		@Override
		public int compare(Repeat o1, Repeat o2) {
			int v1 = o1.getPosition();
			int v2 = o2.getPosition();
			return (v1 < v2 ? -1 : (v1 == v2 ? 0 : 1));
		}

	};

	private static File tmpDir;

	@BeforeClass
	public static void setupClass() throws IOException {
		tmpDir = Files.createTempDir();
		System.setProperty("database.home", tmpDir.getAbsolutePath());
		System.setProperty(CnfManager.PROP_NAME_HOME, tmpDir.getAbsolutePath());
	}

	@AfterClass
	public static void tearDownClass() throws IOException {
		FileUtils.deleteDirectory(tmpDir);
		// Files.deleteRecursively(tmpDir); // guava bug 359
	}

	private RLEnc enc;

	private Random r;

	private RepeatCacheFacade search;

	private char[] fakeInput(int len, int maxLen) {
		StringBuilder sb = new StringBuilder(len * maxLen);
		for (char c : getRnd(len).toCharArray()) {
			int x = r.nextInt(maxLen);
			char[] arr = new char[x];
			Arrays.fill(arr, c);
			sb.append(arr);
		}
		return sb.toString().toCharArray();
	}

	private String getRnd(int len) {
		byte[] data = new byte[len];
		r.nextBytes(data);
		return new String(Base64.encodeBase64(data)).substring(0, len);
	}

	@Before
	public void setup() {
		r = new Random(0);
		enc = new RLEnc(7, '/');
		// RepeatGroupCache cache = new InMemoryRepeatGroupCache();
		RepeatGrouper grouper = new RepeatGrouper(4);
		Collection<Integer> codebook = new ArrayList<Integer>();
		codebook.add(Integer.valueOf(5));
		codebook.add(Integer.valueOf(7));
		codebook.add(Integer.valueOf(10));
		codebook.add(Integer.valueOf(13));
		codebook.add(Integer.valueOf(18));
		codebook.add(Integer.valueOf(25));
		codebook.add(Integer.valueOf(35));
		codebook.add(Integer.valueOf(49));
		codebook.add(Integer.valueOf(67));
		codebook.add(Integer.valueOf(93));
		codebook.add(Integer.valueOf(128));
		codebook.add(Integer.valueOf(177));
		codebook.add(Integer.valueOf(245));
		codebook.add(Integer.valueOf(338));
		codebook.add(Integer.valueOf(467));
		codebook.add(Integer.valueOf(646));
		int offset = 1;
		double minMultiplier = 0.9;
		double maxMultiplier = 1.1;
		IntQuantizer quantizer = new IntQuantizer(codebook, offset,
				minMultiplier, maxMultiplier);
		RepeatGroupKeyMaker keyMaker = new RepeatGroupKeyMaker(grouper,
				quantizer);
		search = new RepeatCacheFacade(new MemRepeatGroupCache(), keyMaker);
	}

	@Test
	public void testEmptyInput() {
		char[] ca = new char[0];
		List<Repeat> repeats = enc.encode(ca);
		Map<Long, Set<RepeatSearchEntry>> results = search.query(repeats);
		Assert.assertTrue(results.isEmpty());
	}

	@Test
	public void testMatchCorrupted() {
		char[] ca = fakeInput(80, 512);
		List<Repeat> repeats = enc.encode(ca);
		Random rnd = new Random(0);
		// corrupt by adding gaps
		repeats = Mth.sampleRandom(repeats.size() / 2, repeats, rnd);
		Collections.sort(repeats, CMP);
		search.add(123L, repeats);
		// corrupt by adding more gaps
		repeats = Mth.sampleRandom(repeats.size() / 2, repeats, rnd);
		Collections.sort(repeats, CMP);
		Map<Long, Set<RepeatSearchEntry>> results = search.query(repeats);
		Assert.assertEquals(1, results.size());
		Set<RepeatSearchEntry> rseSet = results.get(Long.valueOf(123L));
		// Only single RepeatGroup was unaffected
		Assert.assertEquals(4, rseSet.size());
		for (RepeatSearchEntry rse : rseSet) {
			Assert.assertEquals(0, rse.getRelativePosition());
		}
	}

	@Test
	public void testMatchSimple() {
		char[] ca = fakeInput(80, 512);
		List<Repeat> repeats = enc.encode(ca);
		search.add(123L, repeats);
		repeats = enc.encode(ca);
		Map<Long, Set<RepeatSearchEntry>> results = search.query(repeats);
		Assert.assertEquals(1, results.size());
		Set<RepeatSearchEntry> rseSet = results.get(Long.valueOf(123L));
		Assert.assertEquals(72, rseSet.size());
		for (RepeatSearchEntry rse : rseSet) {
			Assert.assertEquals(0, rse.getRelativePosition());
		}
	}

	@Test
	public void testOffset_MiddlePartCached() {
		char[] ca = fakeInput(120, 512);
		char[] caInsert = Arrays.copyOfRange(ca, ca.length / 3,
				ca.length * 2 / 3);
		List<Repeat> repeats = enc.encode(caInsert);
		search.add(123L, repeats);
		repeats = enc.encode(ca);
		Map<Long, Set<RepeatSearchEntry>> results = search.query(repeats);
		Assert.assertEquals(1, results.size());
		Set<RepeatSearchEntry> rseSet = results.get(Long.valueOf(123L));
		Assert.assertEquals(32, rseSet.size());
		int offset = 10589;
		for (RepeatSearchEntry rse : rseSet) {
			Assert.assertEquals(offset, rse.getRelativePosition());
		}
	}

	@Test
	public void testOffset_MoreDataInCacheThanInQuery() {
		char[] ca = fakeInput(120, 512);
		List<Repeat> repeats = enc.encode(ca);
		search.add(123L, repeats);
		ca = Arrays.copyOfRange(ca, ca.length / 3, ca.length * 2 / 3);
		repeats = enc.encode(ca);
		Map<Long, Set<RepeatSearchEntry>> results = search.query(repeats);
		Assert.assertEquals(1, results.size());
		Set<RepeatSearchEntry> rseSet = results.get(Long.valueOf(123L));
		Assert.assertEquals(36, rseSet.size());
		int offset = -10589;
		for (RepeatSearchEntry rse : rseSet) {
			Assert.assertEquals(offset, rse.getRelativePosition());
		}
	}

	@Test
	public void testSearchEmpty() {
		char[] ca = fakeInput(64, 512);
		List<Repeat> repeats = enc.encode(ca);
		Map<Long, Set<RepeatSearchEntry>> results = search.query(repeats);
		Assert.assertTrue(results.isEmpty());
	}

}
