/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.catalog;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.io.FileUtils;
import org.jhears.cache.IRepeatGroupCache;
import org.jhears.cache.ISegmentCache;
import org.jhears.cache.MemRepeatGroupCache;
import org.jhears.cache.MemSegmentCache;
import org.jhears.cache.SegmentCacheValue;
import org.jhears.cluster.ClusterDetector;
import org.jhears.cnf.Cnf;
import org.jhears.cnf.CnfManager;
import org.jhears.cnf.CnfUtil;
import org.jhears.cnf.Cnf.SearchKey;
import org.jhears.seq.FSeq;
import org.jhears.seq.FSeqFactory;
import org.jhears.seq.IntQuantizer;
import org.jhears.seq.MutableFSeq;
import org.jhears.seq.RLEnc;
import org.jhears.seq.RepeatGrouper;
import org.jhears.seq.SeqMatch;
import org.jhears.server.RepeatCacheFacade;
import org.jhears.server.RepeatGroupKeyMaker;
import org.jhears.server.RepeatSearchFilter;
import org.jhears.server.SearchFacade;
import org.jhears.server.SegmentCacheSearch;
import org.jhears.server.SegmentSearchFilter;
import org.jhears.util.Mth;
import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.RepeatCluster;
import org.jhears.web.rs.Segment;
import org.jhears.web.rs.SegmentCluster;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.google.common.io.Files;

public class SearchFacadeTest {

	private static final Comparator<Repeat> CMP = new Comparator<Repeat>() {

		@Override
		public int compare(Repeat o1, Repeat o2) {
			int v1 = o1.getPosition();
			int v2 = o2.getPosition();
			return (v1 < v2 ? -1 : (v1 == v2 ? 0 : 1));
		}

	};

	private static CnfManager factory;

	private static File tmpDir;

	@BeforeClass
	public static void setupClass() throws IOException {
		tmpDir = Files.createTempDir();
		System.setProperty("database.home", tmpDir.getAbsolutePath());
		factory = new CnfManager();
		System.setProperty(CnfManager.PROP_NAME_HOME, tmpDir.getAbsolutePath());
	}

	@AfterClass
	public static void tearDownClass() throws IOException {
		FileUtils.deleteDirectory(tmpDir);
		// Files.deleteRecursively(tmpDir); // guava bug 359
	}

	private Configuration configuration;
	private RLEnc enc;
	private Random r;

	private RepeatCacheFacade repeatSearch;

	private ISegmentCache segmentCache;

	private FSeqFactory seqFactory;

	private SearchFacade sf;

	private String fakeInput(int len, int maxLen) {
		StringBuilder sb = new StringBuilder(len * maxLen);
		for (char c : getRnd(len).toCharArray()) {
			int x = r.nextInt(maxLen);
			char[] arr = new char[x];
			Arrays.fill(arr, c);
			sb.append(arr);
		}
		return sb.toString();
	}

	private List<Segment> fullSegs(FSeq seq) {
		List<Segment> retval = new ArrayList<Segment>();
		for (Segment s : seq) {
			if (s.getPosition() % seq.getSegmentLength() == 0) {
				retval.add(s);
			}
		}
		return retval;
	}

	private String getRnd(int len) {
		byte[] data = new byte[len];
		r.nextBytes(data);
		return new String(Base64.encodeBase64(data)).substring(0, len);
	}

	@Before
	public void setup() throws ConfigurationException {
		r = new Random(0);
		configuration = factory.getConfiguration(Cnf.framework,
				Cnf.fingerprint, Cnf.search);
		Configuration fpConf = factory.getConfiguration(Cnf.fingerprint);
		Configuration searchCnf = factory.getConfiguration(Cnf.search);
		seqFactory = new FSeqFactory(configuration);
		segmentCache = new MemSegmentCache();

		int minRleLen = configuration.getInt(SearchKey.REPEAT_MIN_LENGHT);
		char ignoreChar = configuration.getString(SearchKey.REPEAT_IGNORE_CHAR)
				.charAt(0);
		enc = new RLEnc(minRleLen, ignoreChar);

		int maxDist = configuration.getInt(SearchKey.MAX_CLUSTER_DISTANCE);
		int minSize = configuration.getInt(SearchKey.MIN_SEARCH_SCORE);
		ClusterDetector segmentClusterDetector = new ClusterDetector(maxDist,
				minSize);
		SegmentSearchFilter segmentSearchFilter = new SegmentSearchFilter(
				segmentClusterDetector);
		SegmentCacheSearch cacheSearch = new SegmentCacheSearch(segmentCache,
				fpConf, searchCnf);

		int repeatMaxDist = configuration
				.getInt(SearchKey.REPEAT_MAX_CLUSTER_DISTANCE);
		int repeatMinSize = configuration
				.getInt(SearchKey.REPEAT_MIN_CLUSTER_SIZE);
		ClusterDetector repeatClusterDetector = new ClusterDetector(
				repeatMaxDist, repeatMinSize);
		RepeatSearchFilter repeatSearchFilter = new RepeatSearchFilter(
				repeatClusterDetector);
		IRepeatGroupCache repeatGroupCache = new MemRepeatGroupCache();
		// RepeatGroupCache repeatGroupCache = new InMemoryRepeatGroupCache();
		int repeatsPerGroup = configuration.getInt(SearchKey.REPEATS_PER_GROUP);
		RepeatGrouper grouper = new RepeatGrouper(repeatsPerGroup);
		String[] codb = configuration
				.getStringArray(SearchKey.REPEAT_QUANTIZER_CODEBOOK);
		int[] codebook = CnfUtil.toIntArray(codb);
		int offset = configuration
				.getInt(SearchKey.REPEAT_QUANTIZER_CONSTANT_MARGIN);
		double minMultiplier = configuration
				.getDouble(SearchKey.REPEAT_QUANTIZER_MIN_MULTIPLIER);
		double maxMultiplier = configuration
				.getDouble(SearchKey.REPEAT_QUANTIZER_MAX_MULTIPLIER);
		IntQuantizer quantizer = new IntQuantizer(codebook, offset,
				minMultiplier, maxMultiplier);
		RepeatGroupKeyMaker keyMaker = new RepeatGroupKeyMaker(grouper,
				quantizer);
		repeatSearch = new RepeatCacheFacade(repeatGroupCache, keyMaker);
		sf = new SearchFacade(cacheSearch, segmentSearchFilter,
				repeatSearchFilter, repeatSearch);
	}

	@Test
	public void testSearchRepeatsEmpty() {
		MutableFSeq fseq = seqFactory.getNewInstance();
		fseq.append(fakeInput(64, 512));
		List<Repeat> repeats = enc.encode(fseq.asString().toCharArray());
		Map<Long, List<RepeatCluster>> results = sf.searchRepeats(repeats);
		Assert.assertTrue(results.isEmpty());
	}

	@Test
	public void testSearchRepeatsExact() {
		MutableFSeq fseq = seqFactory.getNewInstance();
		fseq.append(fakeInput(64, 512));
		List<Repeat> repeats = enc.encode(fseq.asString().toCharArray());
		repeatSearch.add(123L, repeats);
		Map<Long, List<RepeatCluster>> results = sf.searchRepeats(repeats);
		Assert.assertEquals(1, results.size());
	}

	@Test
	public void testSearchRepeatsNegativeOffset() {

		MutableFSeq query = seqFactory.getNewInstance();
		MutableFSeq cached = seqFactory.getNewInstance();
		query.append(fakeInput(64, 512));
		cached.append(fakeInput(64, 512));
		cached.append(query.asString());
		cached.append(fakeInput(64, 512));
		repeatSearch.add(123L, enc.encode(cached.asString().toCharArray()));

		// corrupt mildly
		List<Repeat> repeats = enc.encode(query.asString().toCharArray());
		repeats = Mth.sampleRandom(repeats.size() * 3 / 4, repeats, r);
		Collections.sort(repeats, CMP);
		Map<Long, List<RepeatCluster>> results = sf.searchRepeats(repeats);
		Assert.assertEquals(1, results.size());
		RepeatCluster cluster = results.get(Long.valueOf(123L)).get(0);
		int pos = cluster.getRelativePosition();
		Assert.assertEquals(-17345, pos);

		// verify
		SeqMatch sm = new SeqMatch(0.3f); // we have only ~1/3 of data
		Assert.assertTrue(sm.repeatsInSeq(
				enc.encode(cached.asString().toCharArray()), pos, query));
	}

	@Test
	public void testSearchRepeatsPartialWithHeadAndTail() {

		MutableFSeq cached = seqFactory.getNewInstance();
		cached.append(fakeInput(64, 512));

		repeatSearch.add(123L, enc.encode(cached.asString().toCharArray()));

		MutableFSeq query = seqFactory.getNewInstance();
		query.append(fakeInput(64, 512));
		query.append(cached.asString());
		query.append(fakeInput(64, 512));
		List<Repeat> queryRepeats = enc.encode(query.asString().toCharArray());

		// corrupt mildly
		queryRepeats = Mth.sampleRandom(queryRepeats.size() * 5 / 6,
				queryRepeats, r);
		Collections.sort(queryRepeats, CMP);

		Map<Long, List<RepeatCluster>> results = sf.searchRepeats(queryRepeats);
		Assert.assertEquals(1, results.size());

		RepeatCluster cluster = results.get(Long.valueOf(123L)).get(0);
		int pos = cluster.getRelativePosition();
		Assert.assertEquals(17345, pos);

		// verify
		SeqMatch sm = new SeqMatch(0.8f);
		Assert.assertTrue(sm.repeatsInSeq(
				enc.encode(cached.asString().toCharArray()), pos, query));
	}

	@Test
	public void testSearchRepeatsVoid() {
		List<Repeat> repeats = new ArrayList<Repeat>();
		Map<Long, List<RepeatCluster>> results = sf.searchRepeats(repeats);
		Assert.assertTrue(results.isEmpty());
	}

	@Test
	public void testSearchSegmentsEmpty() {
		MutableFSeq fseq = seqFactory.getNewInstance();
		fseq.append(getRnd(120));

		List<Segment> segments = fullSegs(fseq);
		Map<Long, List<SegmentCluster>> results = sf.searchSegments(segments);
		Assert.assertTrue(results.isEmpty());
	}

	@Test
	public void testSearchSegmentsExact() {

		MutableFSeq fseq = seqFactory.getNewInstance();
		fseq.append(getRnd(120));
		for (Segment s : fullSegs(fseq)) {
			segmentCache.add(s.getData(),
					new SegmentCacheValue(123L, s.getPosition()));
		}
		List<Segment> segments = fullSegs(fseq);
		Map<Long, List<SegmentCluster>> results = sf.searchSegments(segments);
		Assert.assertEquals(1, results.size());
	}

	@Test
	public void testSearchSegmentsNegativeOffset() {

		MutableFSeq query = seqFactory.getNewInstance();
		MutableFSeq cached = seqFactory.getNewInstance();
		int offset = 80;
		query.append(getRnd(120));
		cached.append(getRnd(offset));
		cached.append(query.asString());
		cached.append(getRnd(120));
		for (Segment s : fullSegs(cached)) {
			segmentCache.add(s.getData(),
					new SegmentCacheValue(123L, s.getPosition()));
		}

		// corrupt mildly
		List<Segment> querySegs = fullSegs(query);
		querySegs = Mth.sampleRandom(querySegs.size() * 2 / 3, querySegs, r);
		Collections.sort(querySegs, Segment.POSITION_COMPARATOR);

		Map<Long, List<SegmentCluster>> results = sf.searchSegments(querySegs);
		Assert.assertEquals(1, results.size());
		SegmentCluster cluster = results.get(Long.valueOf(123L)).get(0);
		int pos = cluster.getRelativePosition();
		Assert.assertEquals(-offset, pos);

		// verify
		SeqMatch sm = new SeqMatch(0.3f); // we have only ~1/3 of data
		Assert.assertTrue(sm.segmentsInSeq(fullSegs(cached), pos, query));
	}

	@Test
	public void testSearchSegmentsPartialWithHeadAndTail() {

		MutableFSeq cached = seqFactory.getNewInstance();
		cached.append(getRnd(120));
		for (Segment s : fullSegs(cached)) {
			segmentCache.add(s.getData(),
					new SegmentCacheValue(123L, s.getPosition()));
		}
		MutableFSeq query = seqFactory.getNewInstance();
		query.append(getRnd(80));
		query.append(cached.asString());
		query.append(getRnd(120));
		List<Segment> querySegs = fullSegs(query);

		// corrupt mildly
		querySegs = Mth.sampleRandom(querySegs.size() * 2 / 3, querySegs, r);
		Collections.sort(querySegs, Segment.POSITION_COMPARATOR);

		Map<Long, List<SegmentCluster>> results = sf.searchSegments(querySegs);
		Assert.assertEquals(1, results.size());
		SegmentCluster cluster = results.get(Long.valueOf(123L)).get(0);
		int pos = cluster.getRelativePosition();
		Assert.assertEquals(80, pos);

		// verify
		SeqMatch sm = new SeqMatch(0.8f);
		Assert.assertTrue(sm.segmentsInSeq(fullSegs(cached), pos, query));
	}

	@Test
	public void testSearchSegmentsVoid() {
		List<Segment> segments = new ArrayList<Segment>();
		Map<Long, List<SegmentCluster>> results = sf.searchSegments(segments);
		Assert.assertTrue(results.isEmpty());
	}

}
