/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.jhears.bitarray.BitArray;
import org.jhears.bitarray.BitArrayUtil;
import org.jhears.data.DatabaseMapTest.IdGn;
import org.jhears.util.DefaultHashFunction;
import org.jhears.util.HashFunction;
import org.jhears.util.Serialization;
import org.jhears.util.Serializer;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.google.common.io.Files;
import com.sleepycat.je.DatabaseConfig;
import com.sleepycat.je.Durability;
import com.sleepycat.je.EnvironmentConfig;

public class DBBitArrayUtilTest {

	private static final int[] capacities = new int[] { 32, 64, 128, 256, 293 };

	private static HashFunction<Long> hashFunction;

	private static BerkeleyMapManager manager;

	private static Serialization serialization;

	@AfterClass
	public static void afterClass() throws IOException {
		manager.shutDown();
		FileUtils.deleteDirectory(envHome);
		// Files.deleteRecursively(envHome); // guava bug 359
	}

	@BeforeClass
	public static void beforeClass() {

		envHome = Files.createTempDir();
		EnvironmentConfig cnf = BerkeleyMapManager
				.getDefaultEnvironmentConfig();
		cnf.setDurability(Durability.COMMIT_NO_SYNC);
		cnf.setLocking(false);
		manager = new BerkeleyMapManager(envHome);

		serialization = new Serializer();
		hashFunction = new DefaultHashFunction<Long>();
	}

	private String databaseName;

	private static File envHome;

	@Before
	public void beforeTest() {
		databaseName = DBBitArrayUtilTest.class.getSimpleName() + "-"
				+ IdGn.get();
	}

	protected DatabaseBitArray getBitArray(int capacity) {
		DatabaseConfig databaseConfig = BerkeleyMapManager
				.getDefaultDatabaseConfig();
		DatabaseMap<Long, Long> map = new BerkeleyMap<Long, Long>(manager,
				databaseName, databaseConfig, serialization, hashFunction);
		DatabaseBitArray arr = new DatabaseBitArray(map, capacity);
		arr.clear();
		return arr;
	}

	protected DatabaseCountingBitArray getCountingBitArray(int capacity) {
		DatabaseConfig databaseConfig = BerkeleyMapManager
				.getDefaultDatabaseConfig();
		DatabaseMap<Long, Byte> map = new BerkeleyMap<Long, Byte>(manager,
				databaseName, databaseConfig, serialization, hashFunction);
		DatabaseCountingBitArray arr = new DatabaseCountingBitArray(map,
				capacity);
		arr.clear();
		return arr;
	}

	@Test
	public void testBitArrayCopy() {
		Random r = new Random(0);

		for (int capacity : capacities) {
			DatabaseBitArray dbBitArr = getBitArray(capacity);
			Set<Integer> iset = new HashSet<Integer>(capacity / 2);
			while (iset.size() < capacity / 2) {
				iset.add(r.nextInt(capacity));
			}
			BitArray expected = new BitArray(capacity);
			for (Integer i : iset) {
				dbBitArr.set(i.intValue());
				expected.set(i.intValue());
			}
			BitArray actual = DBBitArrayUtil.bitArrayCopy(dbBitArr);
			Assert.assertTrue("Bit arrays not equal",
					BitArrayUtil.equals(expected, actual));
		}
	}

	@Test
	public void testCountingBitArrayCopy() {
		Random r = new Random(0);

		for (int capacity : capacities) {
			DatabaseCountingBitArray dbBitArr = getCountingBitArray(capacity);
			Set<Integer> iset = new HashSet<Integer>(capacity / 2);
			while (iset.size() < capacity / 2) {
				iset.add(r.nextInt(capacity));
			}
			BitArray expected = new BitArray(capacity);
			for (Integer i : iset) {
				dbBitArr.set(i.intValue());
				expected.set(i.intValue());
			}
			BitArray actual = DBBitArrayUtil.bitArrayCopy(dbBitArr);
			Assert.assertTrue("Bit arrays not equal",
					BitArrayUtil.equals(expected, actual));
		}
	}

	@Test
	public void testCopy() {
		Random r = new Random(0);
		for (int capacity : capacities) {
			Set<Integer> iset = new HashSet<Integer>(capacity / 2);
			while (iset.size() < capacity / 2) {
				iset.add(r.nextInt(capacity));
			}
			BitArray expected = new BitArray(capacity);
			for (Integer i : iset) {
				expected.set(i.intValue());
			}

			DatabaseBitArray actual = getBitArray(capacity);
			DBBitArrayUtil.copy(expected, actual);

			Assert.assertTrue("Bit arrays not equal",
					BitArrayUtil.equals(expected, actual));
		}
	}

}
