/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.jhears.cnf.CnfManager;
import org.jhears.web.rs.Repeat;
import org.jhears.web.rs.Segment;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.google.common.io.Files;

public class DataObjAccessTest {

//	private static CnfManager factory;
	private static GeneralDao generalDao;
	private static File tmpDir;

	@AfterClass
	public static void afterClass() throws IOException {
		FileUtils.deleteDirectory(tmpDir);
		// Files.deleteRecursively(tmpDir); // guava bug 359
	}

	@BeforeClass
	public static void beforeClass() throws IOException {

		tmpDir = Files.createTempDir();
		System.setProperty("database.home", tmpDir.getAbsolutePath());
//		factory = new CnfManager();
		System.setProperty(CnfManager.PROP_NAME_HOME, tmpDir.getAbsolutePath());
		HibernateSessionProvider provider = new HibernateSessionProvider();
		DAO dao = new DAO(provider);
		generalDao = new GeneralDao(dao);
	}

	@Before
	public void before() {

	}

	private void check(AudioEntityDataObj entity, Repeat repeat, Segment segment) {
		Assert.assertNotNull(entity);
		Set<FingerprintDataObj> fps = entity.getFingerprints();
		Assert.assertEquals(1, fps.size());
		FingerprintDataObj fp = fps.iterator().next();
		Assert.assertEquals(1, fp.getRepeats().size());
		Assert.assertEquals(repeat, fp.getRepeats().get(0));
		Assert.assertEquals(1, fp.getSegments().size());
		Assert.assertEquals(segment, fp.getSegments().get(0));
	}

	private void check(String entityId, FingerprintDataObj fp, Repeat repeat,
			Segment segment) {
		Assert.assertNotNull(fp);
		Assert.assertEquals(entityId, fp.getAudioEntity().getIdentifier());
		Assert.assertEquals(1, fp.getRepeats().size());
		Assert.assertEquals(repeat, fp.getRepeats().get(0));
		Assert.assertEquals(1, fp.getSegments().size());
		Assert.assertEquals(segment, fp.getSegments().get(0));
	}

	private void deleteAll(FPDAO doa) {
		doa.deleteAll();
		Assert.assertEquals(0, generalDao.count(AudioEntityDataObj.class));
		Assert.assertEquals(0, generalDao.count(FingerprintDataObj.class));
	}

	@Test
	public void saveAndDeleteSingleEntity() {
		FPDAO doa = new FPDAO(generalDao);

		deleteAll(doa);

		List<Segment> segments = new ArrayList<Segment>();
		List<Repeat> repeats = new ArrayList<Repeat>();
		segments.add(new Segment("foo", 0));
		repeats.add(new Repeat('x', 1, 2));
		String entityId = "123";

		// test saving
		FingerprintDataObj fp = doa.saveFingerprint(entityId, segments,
				repeats, null);
		String fpId = fp.getIdentifier();

		fp = doa.getFingerprintWithIdentifier(fpId);
		Assert.assertNotNull(fp);
		doa.deleteEntityWithIdentifier(entityId, null);

		Assert.assertNull(doa.getFingerprintWithIdentifier(fpId));
		Assert.assertNull(doa.getEntityWithIdentifier(entityId));

		fp = doa.saveFingerprint(entityId, segments, repeats, null);
		Assert.assertNotNull(fp);
		doa.deleteEntityWithId(fp.getAudioEntity().getId());

		Assert.assertNull(doa.getFingerprintWithIdentifier(fpId));
		Assert.assertNull(doa.getEntityWithIdentifier(entityId));

	}

	@Test
	public void saveAndDeleteSingleFingerprint() {
		FPDAO doa = new FPDAO(generalDao);

		deleteAll(doa);

		List<Segment> segments = new ArrayList<Segment>();
		List<Repeat> repeats = new ArrayList<Repeat>();
		segments.add(new Segment("foo", 0));
		repeats.add(new Repeat('x', 1, 2));
		String entityId = "123";

		// test saving
		FingerprintDataObj fp = doa.saveFingerprint(entityId, segments,
				repeats, null);
		String fpId = fp.getIdentifier();

		fp = doa.getFingerprintWithIdentifier(fpId);
		Assert.assertNotNull(fp);
		doa.deleteFingerprintWithIdentifier(fpId, null);

		Assert.assertNull(doa.getFingerprintWithIdentifier(fpId));
		Assert.assertEquals(0, doa.getEntityWithIdentifier(entityId)
				.getFingerprints().size());

		fp = doa.saveFingerprint(entityId, segments, repeats, null);
		Assert.assertNotNull(fp);
		fpId = fp.getIdentifier();
		Assert.assertEquals(1, doa.getEntityWithIdentifier(entityId)
				.getFingerprints().size());
		doa.deleteFingerprintWithId(fp.getId());
		Assert.assertNull(doa.getFingerprintWithIdentifier(fpId));
		Assert.assertEquals(0, doa.getEntityWithIdentifier(entityId)
				.getFingerprints().size());

	}

	@Test
	public void saveAndFetchSingleFingerprint() {
		FPDAO doa = new FPDAO(generalDao);

		deleteAll(doa);

		List<Segment> segments = new ArrayList<Segment>();
		List<Repeat> repeats = new ArrayList<Repeat>();
		segments.add(new Segment("foo", 0));
		repeats.add(new Repeat('x', 1, 2));
		String entityId = "123";

		// test saving
		FingerprintDataObj fp = doa.saveFingerprint(entityId, segments,
				repeats, null);
		check(entityId, fp, new Repeat('x', 1, 2), new Segment("foo", 0));
		String fpId = fp.getIdentifier();

		// test fetching with identifier
		fp = doa.getFingerprintWithIdentifierPlusEntity(fpId);
		check(entityId, fp, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching with id
		fp = doa.getFingerpritWithIdPlusEntity(fp.getId());
		check(entityId, fp, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching entity with indentifier
		AudioEntityDataObj entity = doa.getEntityWithIdentifier(entityId);
		check(entity, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching entity with id
		entity = doa.getEntityWithId(entity.getId());
		check(entity, new Repeat('x', 1, 2), new Segment("foo", 0));

	}

	@Test
	public void saveAndFetchTwoFingerprints1() {
		FPDAO doa = new FPDAO(generalDao);

		deleteAll(doa);

		List<Segment> segments = new ArrayList<Segment>();
		List<Repeat> repeats = new ArrayList<Repeat>();
		segments.add(new Segment("foo", 0));
		repeats.add(new Repeat('x', 1, 2));
		String entityId = "123";

		List<AudioEntityDataObj> l = generalDao
				.getAll(AudioEntityDataObj.class);
		Assert.assertEquals(0, l.size());

		FingerprintDataObj fp1 = doa.saveFingerprint(entityId, segments,
				repeats, null);

		l = generalDao.getAll(AudioEntityDataObj.class);
		Assert.assertEquals(1, l.size());

		segments = new ArrayList<Segment>();
		repeats = new ArrayList<Repeat>();
		segments.add(new Segment("bar", 1));
		repeats.add(new Repeat('y', 3, 4));
		FingerprintDataObj fp2 = doa.saveFingerprint(entityId, segments,
				repeats, null);

		l = generalDao.getAll(AudioEntityDataObj.class);
		Assert.assertEquals(2, generalDao.count(FingerprintDataObj.class));
		Assert.assertEquals(1, generalDao.count(AudioEntityDataObj.class));
		Assert.assertEquals(1, l.size());

		String fpId1 = fp1.getIdentifier();
		check(entityId, fp1, new Repeat('x', 1, 2), new Segment("foo", 0));

		Assert.assertNotNull(fp2);
		check(entityId, fp2, new Repeat('y', 3, 4), new Segment("bar", 1));
		String fpId2 = fp2.getIdentifier();

		Assert.assertNotSame(fpId1, fpId2);

		// test fetching with identifier1
		fp1 = doa.getFingerprintWithIdentifierPlusEntity(fpId1);
		check(entityId, fp1, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching with identifier2
		fp2 = doa.getFingerprintWithIdentifierPlusEntity(fpId2);
		// Assert.assertEquals(1,
		// fp2.getAudioEntity().getFingerprints().size());
		check(entityId, fp2, new Repeat('y', 3, 4), new Segment("bar", 1));

		// test fetching with id1
		fp1 = doa.getFingerpritWithIdPlusEntity(fp1.getId());
		check(entityId, fp1, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching with id2
		fp2 = doa.getFingerpritWithIdPlusEntity(fp2.getId());
		check(entityId, fp2, new Repeat('y', 3, 4), new Segment("bar", 1));

		AudioEntityDataObj entity = doa.getEntityWithIdentifier(entityId);
		Assert.assertEquals(2, entity.getFingerprints().size());

		HashSet<Object> expected = new HashSet<Object>();
		expected.add(new Repeat('x', 1, 2));
		expected.add(new Repeat('y', 3, 4));
		expected.add(new Segment("foo", 0));
		expected.add(new Segment("bar", 1));
		HashSet<Object> actual = new HashSet<Object>();
		for (FingerprintDataObj fp : entity.getFingerprints()) {
			actual.addAll(fp.getRepeats());
			actual.addAll(fp.getSegments());
		}
		Assert.assertEquals(expected, actual);

		entity = doa.getEntityWithId(entity.getId());
		Assert.assertEquals(2, entity.getFingerprints().size());
		actual = new HashSet<Object>();
		for (FingerprintDataObj fp : entity.getFingerprints()) {
			actual.addAll(fp.getRepeats());
			actual.addAll(fp.getSegments());
		}
		Assert.assertEquals(expected, actual);

		l = generalDao.getAll(AudioEntityDataObj.class);
		Assert.assertEquals(1, l.size());

		Assert.assertEquals(entity.getIdentifier(), l.get(0).getIdentifier());

		entity = doa.getEntityWithId(entity.getId());
		Assert.assertEquals(2, entity.getFingerprints().size());
		actual = new HashSet<Object>();
		for (FingerprintDataObj fp : entity.getFingerprints()) {
			actual.addAll(fp.getRepeats());
			actual.addAll(fp.getSegments());
		}
		Assert.assertEquals(expected, actual);

	}

	@Test
	public void saveAndFetchTwoFingerprints2() {
		FPDAO doa = new FPDAO(generalDao);

		deleteAll(doa);

		List<Segment> segments = new ArrayList<Segment>();
		List<Repeat> repeats = new ArrayList<Repeat>();
		segments.add(new Segment("foo", 0));
		repeats.add(new Repeat('x', 1, 2));
		String entityId1 = "123";
		String entityId2 = "987";

		FingerprintDataObj fp1 = doa.saveFingerprint(entityId1, segments,
				repeats, null);

		segments = new ArrayList<Segment>();
		repeats = new ArrayList<Repeat>();
		segments.add(new Segment("bar", 1));
		repeats.add(new Repeat('y', 3, 4));
		FingerprintDataObj fp2 = doa.saveFingerprint(entityId2, segments,
				repeats, null);

		String fpId1 = fp1.getIdentifier();
		check(entityId1, fp1, new Repeat('x', 1, 2), new Segment("foo", 0));

		Assert.assertNotNull(fp2);
		check(entityId2, fp2, new Repeat('y', 3, 4), new Segment("bar", 1));
		String fpId2 = fp2.getIdentifier();

		Assert.assertNotSame(fpId1, fpId2);

		// test fetching with identifier1
		fp1 = doa.getFingerprintWithIdentifierPlusEntity(fpId1);
		check(entityId1, fp1, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching with identifier2
		fp2 = doa.getFingerprintWithIdentifierPlusEntity(fpId2);
		check(entityId2, fp2, new Repeat('y', 3, 4), new Segment("bar", 1));

		// test fetching with id1
		fp1 = doa.getFingerpritWithIdPlusEntity(fp1.getId());
		check(entityId1, fp1, new Repeat('x', 1, 2), new Segment("foo", 0));

		// test fetching with id2
		fp2 = doa.getFingerpritWithIdPlusEntity(fp2.getId());
		check(entityId2, fp2, new Repeat('y', 3, 4), new Segment("bar", 1));

	}
}
