/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import junit.framework.Assert;

import org.apache.commons.io.FileUtils;
import org.jhears.data.DatabaseMapTest.IdGn;
import org.jhears.filter.bloom.AbstractCountingBloomFilterTest;
import org.jhears.filter.bloom.BloomFilter;
import org.jhears.filter.bloom.HVBloomFilter;
import org.jhears.filter.bloom.Hasher;
import org.jhears.filter.bloom.IBloomFilter;
import org.jhears.filter.bloom.IBloomFilterState;
import org.jhears.filter.bloom.IHasher;
import org.jhears.filter.bloom.util.BloomMath;
import org.jhears.util.DefaultHashFunction;
import org.jhears.util.HashFunction;
import org.jhears.util.LCG;
import org.jhears.util.Serialization;
import org.jhears.util.Serializer;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import com.google.common.io.Files;
import com.sleepycat.je.Durability;
import com.sleepycat.je.EnvironmentConfig;

public class DatabaseCountingBloomFilterTest extends
		AbstractCountingBloomFilterTest {

	private static HashFunction<?> hashFunction;
	private static BerkeleyMapManager manager;

	private static Serialization serialization;
	private static File envHome;

	@AfterClass
	public static void afterClass() throws IOException {
		manager.shutDown();
		FileUtils.deleteDirectory(envHome);
		// Files.deleteRecursively(envHome); // guava bug 359
	}

	@BeforeClass
	public static void before() {
		envHome = Files.createTempDir();
		EnvironmentConfig cnf = BerkeleyMapManager
				.getDefaultEnvironmentConfig();
		cnf.setDurability(Durability.COMMIT_NO_SYNC);
		cnf.setLocking(false);
		manager = new BerkeleyMapManager(envHome);
		serialization = new Serializer();
		hashFunction = new DefaultHashFunction<Object>();
	}

	private void add(IBloomFilter<Long> filter, Set<Long> reference, Random rnd) {
		for (int i = 0; i < getRandomTestSize(); i++) {
			Long value = Long.valueOf(rnd.nextLong());
			reference.add(value);
			filter.add(value);
		}
	}

	@Override
	protected IBloomFilter<Long> createFilter(int size, double p0) {
		int m0 = BloomMath.optimalSliceSize(size);
		int k0 = BloomMath.optimalSliceCount(p0);
		DatabaseBFState state = new DatabaseBFState(m0, k0, manager,
				DatabaseCountingBloomFilterTest.class.getSimpleName() + "-"
						+ IdGn.get(), serialization, hashFunction);
		state.clear();
		HVBloomFilter alg = new HVBloomFilter(state);
		HashFunction<Object> genericHash = new DefaultHashFunction<Object>();
		IHasher hasher = new Hasher(genericHash, new LCG());
		return new BloomFilter<Long>(alg, hasher);
	}

	@SuppressWarnings("unchecked")
	private IBloomFilter<Long> cro(IBloomFilterState state) {
		HVBloomFilter alg = new HVBloomFilter(state);
		IHasher hasher = new Hasher((HashFunction<Object>) hashFunction,
				new LCG());
		return new BloomFilter<Long>(alg, hasher);
	}

	@Override
	protected int getRandomTestSize() {
		return 10;
	}

	@Override
	public void testP0OneBasispoint() {
	}

	@Override
	public void testP0OnePermil() {
	}

	@Test
	public void testStateSaving() throws IOException {
		Random rnd = new Random(0);
		int isize = getRandomTestSize();
		double r = 0.85;
		double p0 = BloomMath.optimalSBFP0(0.01, r);
		String name = DatabaseCountingBloomFilterTest.class.getSimpleName()
				+ "-" + IdGn.get();

		int m0 = BloomMath.optimalSliceSize(isize);
		int k0 = BloomMath.optimalSliceCount(p0);
		DatabaseBFState state = new DatabaseBFState(m0, k0, manager, name,
				serialization, hashFunction);
		state.clear();
		IBloomFilter<Long> filter = cro(state);
		HashSet<Long> reference = new HashSet<Long>();
		add(filter, reference, rnd);
		state.close();
		state = new DatabaseBFState(manager, name, serialization, hashFunction);
		filter = cro(state);
		add(filter, reference, rnd);
		state.close();
		state = new DatabaseBFState(manager, name, serialization, hashFunction);
		filter = cro(state);
		verify(filter, reference);
		state.close();
	}

	private void verify(IBloomFilter<Long> filter, Set<Long> reference) {
		for (Long l : reference) {
			Assert.assertTrue(filter.contains(l));
		}
	}
}
