/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.data;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import junit.framework.Assert;

import org.jhears.filter.bloom.Hasher;
import org.jhears.filter.bloom.IBloomFilter;
import org.jhears.filter.bloom.IHasher;
import org.jhears.filter.bloom.scalable.AbstractCountingScalableBloomFilterTest;
import org.jhears.filter.bloom.scalable.HVScalableBloomFilter;
import org.jhears.filter.bloom.scalable.IScalableBloomFilter;
import org.jhears.filter.bloom.scalable.IScalableBloomFilterState;
import org.jhears.filter.bloom.scalable.ScalableBloomFilter;
import org.jhears.filter.bloom.util.BloomMath;
import org.jhears.util.DefaultHashFunction;
import org.jhears.util.HashFunction;
import org.jhears.util.LCG;
import org.jhears.util.Serialization;
import org.jhears.util.Serializer;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;

import com.google.common.io.Files;
import com.sleepycat.je.Durability;
import com.sleepycat.je.EnvironmentConfig;

public class DatabaseScalableCountingBloomFilterTest extends
		AbstractCountingScalableBloomFilterTest {

	private static HashFunction<?> hashFunction;

	private static BerkeleyMapManager manager;

	private static Serialization serialization;

	@AfterClass
	public static void afterClass() throws IOException {
		manager.shutDown();
		Files.deleteDirectoryContents(envHome);
	}

	@BeforeClass
	public static void beforeClass() {
		envHome = Files.createTempDir();
		EnvironmentConfig cnf = BerkeleyMapManager
				.getDefaultEnvironmentConfig();
		cnf.setDurability(Durability.COMMIT_NO_SYNC);
		cnf.setLocking(false);
		manager = new BerkeleyMapManager(envHome);
		serialization = new Serializer();
		hashFunction = new DefaultHashFunction<Object>();
	}

	@Rule
	public TemporaryFolder folder = new TemporaryFolder();

	int i = 0;

	private static File envHome;

	private void add(IBloomFilter<Long> filter, Set<Long> reference, Random rnd) {
		for (int i = 0; i < getRandomTestSize(); i++) {
			Long value = Long.valueOf(rnd.nextLong());
			reference.add(value);
			filter.add(value);
		}
	}

	@Override
	protected IScalableBloomFilter<Long> createFilter(int size, double p0,
			double r, int s, int m0) {
		IScalableBloomFilterState state = new CountingDatabaseSBFState(s, r,
				m0, p0, Integer.toString(i++), manager, hashFunction,
				serialization);
		IScalableBloomFilter<Long> f = cro(state);
		f.clear();
		return f;
	}

	@SuppressWarnings("unchecked")
	private IScalableBloomFilter<Long> cro(IScalableBloomFilterState state) {
		HVScalableBloomFilter alg = new HVScalableBloomFilter(state);
		IHasher hasher = new Hasher((HashFunction<Object>) hashFunction,
				new LCG());
		return new ScalableBloomFilter<Long>(alg, hasher);
	}

	@Override
	protected int getRandomTestSize() {
		return 200;
	}

	@Override
	@Test
	public void test() {
		int isize = 10; // initial set size
		List<Double> PList = new ArrayList<Double>();
		PList.add(0.01);
		List<Double> rList = new ArrayList<Double>();
		rList.add(0.85);
		// rList.add(0.975);
		List<Integer> sList = new ArrayList<Integer>();
		sList.add(2);
		// sList.add(16);
		List<Integer> magList = new ArrayList<Integer>();
		magList.add(2);
		for (double P : PList) {
			for (double r : rList) {
				for (int s : sList) {
					for (int mag : magList) {
						int size = (int) Math.round(Math.pow(isize, mag));
						int m0 = BloomMath.optimalSliceSize(isize);
						test(size, P, r, s, m0, mag);
					}
				}
			}
		}
	}

	@Test
	public void testStateSaving() throws IOException {
		Random rnd = new Random(0);
		int isize = 10;
		double r = 0.85;
		double p0 = BloomMath.optimalSBFP0(0.01, r);
		int m0 = BloomMath.optimalSliceSize(isize);
		String name = "foo";
		int s = 2;
		CountingDatabaseSBFState state = new CountingDatabaseSBFState(s, r, m0,
				p0, name, manager, hashFunction, serialization);
		IBloomFilter<Long> filter = cro(state);
		HashSet<Long> reference = new HashSet<Long>();
		add(filter, reference, rnd);
		state.close();
		state = new CountingDatabaseSBFState(name, manager, hashFunction,
				serialization);
		filter = cro(state);
		add(filter, reference, rnd);
		state.close();
		state = new CountingDatabaseSBFState(name, manager, hashFunction,
				serialization);
		filter = cro(state);
		verify(filter, reference);
	}

	private void verify(IBloomFilter<Long> filter, Set<Long> reference) {
		for (Long l : reference) {
			Assert.assertTrue(filter.contains(l));
		}
	}

}
