/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.web.rs;

import java.io.IOException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;

import org.codehaus.jackson.JsonParseException;
import org.codehaus.jackson.map.JsonMappingException;
import org.codehaus.jackson.map.ObjectMapper;
import org.jhears.data.GeneralDao;
import org.jhears.data.UserDataObj;
import org.jhears.server.Ctx;
import org.jhears.server.FPCatalogEngine;
import org.jhears.server.ICtx;
import org.jhears.user.UserInformation;

public class BaseResource {

	public static class BaseResourceParam {

		private final GeneralDao dao;

		private final FPCatalogEngine engine;

		private final ObjectMapper mapper;

		private final boolean securityEnabled;

		public BaseResourceParam(FPCatalogEngine engine, ObjectMapper mapper,
				boolean securityEnabled, GeneralDao dao) {
			this.engine = engine;
			this.mapper = mapper;
			this.securityEnabled = securityEnabled;
			this.dao = dao;
		}

		public GeneralDao getDao() {
			return dao;
		}

		public FPCatalogEngine getEngine() {
			return engine;
		}

		public ObjectMapper getMapper() {
			return mapper;
		}

		public boolean isSecurityEnabled() {
			return securityEnabled;
		}

	}

	public static enum Format {
		json, plain;
	}

	private final GeneralDao dao;

	private final FPCatalogEngine engine;

	private final ObjectMapper mapper;

	@Context
	private HttpServletRequest req;

	private final boolean securityEnabled;

	public BaseResource(BaseResourceParam param) {
		this.engine = param.getEngine();
		this.mapper = param.getMapper();
		this.securityEnabled = param.isSecurityEnabled();
		this.dao = param.getDao();
	}

	public FPCatalogEngine getCatalog() {
		return engine;
	}

	protected ICtx getCtx() {
		HttpSession session = req.getSession(false);
		if (securityEnabled && session != null
				&& session.getAttribute("jhears-user-information") != null) {
			UserInformation info = (UserInformation) session
					.getAttribute("jhears-user-information");
			UserDataObj user = dao.get(UserDataObj.class, "name", info
					.getUsername());
			return new Ctx(user);
		}
		return null;
	}

	protected boolean isAuthorized(String role) {
		if (!securityEnabled) {
			return true;
		}
		HttpSession session = req.getSession(false);
		if (session != null
				&& session.getAttribute("jhears-user-information") != null) {
			UserInformation info = (UserInformation) session
					.getAttribute("jhears-user-information");
			for (String userRole : info.getRoles()) {
				if (role.equals(userRole)) {
					return true;
				}
			}
		}
		return false;
	}

	protected <T> T readValue(String content, Class<T> valueType)
			throws JsonParseException, JsonMappingException, IOException {
		return mapper.readValue(content, valueType);
	}

	protected String toText(Object o) throws WebApplicationException {
		try {
			return mapper.writeValueAsString(o);
		} catch (JsonMappingException e) {
			throw new WebApplicationException(e,
					HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
		} catch (IOException e) {
			throw new WebApplicationException(e,
					HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
		}
	}

	protected Response unauthorized() {
		return Response.status(Status.UNAUTHORIZED).build();
	}

}
