/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.concurrent;

import java.util.Iterator;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;

import org.jhears.util.IterableObserver;
import org.jhears.util.IterationNotifier;
import org.jhears.util.IterationObserver;
import org.jhears.util.Observer;

import com.google.common.base.Function;
import com.google.common.base.Throwables;
import com.google.common.collect.Iterators;

public final class Concurrent {

	private static boolean shutdownExecutor(ExecutorService executor) {
		executor.shutdown();
		try {
			if (!executor.awaitTermination(100, TimeUnit.MILLISECONDS)) {
				Thread.sleep(15);
				return executor.shutdownNow().isEmpty();
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw Throwables.propagate(e);
		}
		return true;
	}

	/**
	 * Returns an iterator that applies {@code function} to each element of
	 * {@code fromIterator} concurrently while maintaining the element iteration
	 * order.
	 * <p>
	 * Practically the same as {@link Iterators#transform(Iterator, Function)}
	 * except the iterator {@code remove()} is not supported.
	 * <p>
	 * Note that the function should check
	 * {@code Thread.currentThread().isInterrupted()} regularly if running long.
	 * Also the function should exit immediately if the thread is interrupted.
	 * <p>
	 * Returned iterator is not thread safe.
	 * 
	 * @param <F>
	 * @param <T>
	 * @param fromIterator
	 * @param function
	 * @param executor
	 * @return
	 */
	public static <F, T> Iterator<T> transform(Iterator<F> fromIterator,
			Function<? super F, ? extends T> function, Executor executor) {
		return new TransformingConcurrentIterator<F, T>(function, fromIterator,
				executor);
	}

	/** See {@link #transform(Iterator, Function, Observer, int)}. */
	public static <F, T> void transform(Iterator<F> fromIterator,
			Function<? super F, ? extends T> function,
			Observer<? super T> observer) {
		transform(fromIterator, function, observer, -1);
	}

	/**
	 * Apply a function to iterator elements concurrently. Method blocks until
	 * all elements are processed.
	 * 
	 * @param <F>
	 * @param <T>
	 * @param fromIterator
	 *            Iterator which elements are transformed.
	 * @param function
	 *            Transforming function.
	 * @param observer
	 *            Will observe transformed element in exactly the same order as
	 *            returned by the iterator.
	 * @param threads
	 *            The number of executor threads. If less than 2 then
	 *            {@code (availableProcessors * 2 + 1)} is used.
	 * @return {@code true} if executor was shutdown properly. {@code false} if
	 *         some of the task refused to terminate.
	 */
	public static <F, T> boolean transform(Iterator<F> fromIterator,
			Function<? super F, ? extends T> function,
			Observer<? super T> observer, int threads) {
		if (threads < 2) {
			threads = Runtime.getRuntime().availableProcessors() * 2 + 1;
		}
		ExecutorService executor = Executors.newFixedBoundedThreadPool(threads);
		Iterator<T> itre = Concurrent.transform(fromIterator, function,
				executor);
		while (itre.hasNext()) {
			observer.observe(itre.next());
		}
		return shutdownExecutor(executor);
	}

	private Concurrent() {
	}

	/**
	 * See {@link #notifyObserver(IterationNotifier)}
	 * 
	 * @param <E>
	 * @param notifier
	 * @param executor
	 *            Executor used to perform
	 *            {@link IterationNotifier#notifyObserver(IterationObserver)}.
	 * @return
	 */
	public static <E> Iterator<E> notifyIterator(
			final IterationNotifier<E> notifier, Executor executor) {
		final IterableObserver<E> observer = new IterableObserver<E>();
		final Iterator<E> retval = observer.iterator();
		executor.execute(new Runnable() {
			@Override
			public void run() {
				try {
					notifier.notifyObserver(observer);
				} catch (Throwable t) {
					observer.notifyException(t);
				} finally {
					observer.completed();
				}
			}
		});
		return retval;
	}

	/**
	 * Invokes {@link IterationNotifier#notifyObserver(IterationObserver)} and
	 * returns results as an iterator. Note that the returned iterator
	 * <b>must</b> be fully consumed or resources are leaked.
	 * <p>
	 * {@link IterationNotifier#notifyObserver(IterationObserver)} is performed
	 * in different thread.
	 * 
	 * @param <E>
	 * @param notifier
	 * @return
	 */
	public static <E> Iterator<E> notifyObserver(
			final IterationNotifier<E> notifier) {
		return notifyIterator(notifier, Executors.getExecutor());
	}
}
