/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.concurrent;

import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.LinkedBlockingQueue;

import com.google.common.base.Function;

/**
 * Utility class to transform values using a {@link Function} concurrently while
 * maintaining the order in which the value where submitted.
 * 
 * @author Juha Heljoranta
 * 
 * @param <F>
 * @param <T>
 */
public class ConcurrentTransform<F, T> {

	/**
	 * Custom ExecutionException class to allow T to be instance of
	 * ExecutionException.
	 */
	private static final class WrappedException extends
			ExecutionException {

		private static final long serialVersionUID = 1L;

		public WrappedException(Throwable cause) {
			super(cause);
		}

	}

	private static enum HasNext {
		END, NO, YES;
	}

	/** holds the completed tasks */
	private final ConcurrentMap<CountDownLatch, Object> completed;

	/** maintains the order */
	private final BlockingQueue<CountDownLatch> execOrder;

	private final Function<? super F, ? extends T> function;

	private HasNext hasNext;

	/** Workaround: ConcurrentHashMap cannot have null values. */
	private final Object nullObject;

	private Object o;

	/** holds the pending tasks */
	private final BlockingQueue<Entry<CountDownLatch, F>> pending;

	/** Used to detect when {@link #hasNext} should return false. */
	private final CountDownLatch poison;

	/**
	 * Creates a new instance.
	 * 
	 * @param function
	 *            A thread safe function used to transform elements.
	 */
	public ConcurrentTransform(Function<? super F, ? extends T> function) {
		this.function = function;
		completed = new ConcurrentHashMap<CountDownLatch, Object>();
		pending = new LinkedBlockingQueue<Map.Entry<CountDownLatch, F>>();
		execOrder = new LinkedBlockingQueue<CountDownLatch>(Runtime
				.getRuntime().availableProcessors() * 2 + 1);
		poison = new CountDownLatch(0);
		nullObject = new Object();
		hasNext = HasNext.NO;
		o = null;
	}

	/**
	 * Takes a next available element from queue, applies the function and makes
	 * the function results available to {@link #next()}. This method is thread
	 * safe. No that this method will block indefinitely unless the
	 * {@link #put(Object)} is called.
	 * 
	 * @throws InterruptedException
	 */
	public void apply() throws InterruptedException {
		Entry<CountDownLatch, F> entry = pending.take();
		CountDownLatch latch = entry.getKey();
		try {
			T t = latch != poison ? function.apply(entry.getValue()) : null;
			completed.put(latch, t == null ? nullObject : t);
		} catch (RuntimeException e) {
			completed.put(latch, new WrappedException(e));
		} finally {
			latch.countDown();
		}
	}

	/**
	 * Causes {@link #hasNext()} to eventually return false. This method is
	 * thread safe. This method will block if {@link #apply()} was not called
	 * after the {@link #put(Object)}.
	 */
	public void close() throws InterruptedException {
		enque(poison, null);
	}

	private void enque(CountDownLatch latch, F f) throws InterruptedException {
		pending.put(new SimpleImmutableEntry<CountDownLatch, F>(latch, f));
		execOrder.put(latch);
	}

	/**
	 * Non-thread safe method to check if there is a next element available.
	 * Similar to {@link Iterator#hasNext()}.
	 * 
	 * @return
	 * @throws InterruptedException
	 */
	public boolean hasNext() throws InterruptedException {
		if (HasNext.NO.equals(hasNext)) {
			CountDownLatch latch = execOrder.take();
			latch.await();
			o = completed.remove(latch);
			hasNext = latch == poison ? HasNext.END : HasNext.YES;
		}
		return HasNext.YES.equals(hasNext);
	}

	/**
	 * Non-thread safe method to retrieve next element. Similar to
	 * {@link Iterator#next()}.
	 * 
	 * @return
	 * @throws ExecutionException
	 * @throws InterruptedException
	 * @throws NoSuchElementException
	 */
	@SuppressWarnings("unchecked")
	public T next() throws ExecutionException, InterruptedException,
			NoSuchElementException {
		if (hasNext()) {
			hasNext = HasNext.NO;
			if (o instanceof WrappedException) {
				throw (WrappedException) o;
			}
			return o == nullObject ? null : (T) o;
		} else {
			throw new NoSuchElementException();
		}
	}

	/**
	 * Thread safe method to queue object for later execution. Note that each
	 * time this method is used the {@link #apply()} must also be used. This
	 * method might block unless {@link #apply()} is called.
	 * 
	 * @param f
	 * @throws InterruptedException
	 */
	public void put(F f) throws InterruptedException {
		enque(new CountDownLatch(1), f);
	}

}
