/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.concurrent;

import java.util.concurrent.Executor;
import java.util.concurrent.Semaphore;

import com.google.common.base.Preconditions;

/**
 * Convenience class to execute small set of tasks. Spawns a new thread for each
 * task. If maximum concurrent threads is reached the caller executes the task.
 * 
 * @author Juha Heljoranta
 * 
 */
public class SimpleExecutor implements Executor {

	private final int maxThreads;

	private final Semaphore semaphore;

	/**
	 * Instance with maximum threads =
	 * {@code Runtime.getRuntime().availableProcessors() * 2 + 1}.
	 */
	public SimpleExecutor() {
		this(Runtime.getRuntime().availableProcessors() * 2 + 1);
	}

	public SimpleExecutor(int maxThreads) {
		Preconditions.checkArgument(maxThreads > 0, "Value must be positive");
		semaphore = new Semaphore(maxThreads);
		this.maxThreads = maxThreads;
	}

	private Runnable delegate(final Runnable command) {
		return new Runnable() {
			@Override
			public void run() {
				try {
					command.run();
				} finally {
					semaphore.release();
				}
			}
		};
	}

	@Override
	public void execute(Runnable command) {
		if (semaphore.tryAcquire()) {
			new Thread(delegate(command)).start();
		} else {
			command.run();
		}
	}

	public int getMaxThreads() {
		return maxThreads;
	}

}
