/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.concurrent;

import java.util.Iterator;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;

import com.google.common.base.Function;
import com.google.common.base.Preconditions;
import com.google.common.base.Throwables;

/**
 * Iterator which applies a function to Iterator elements. Function is applied
 * concurrently. The order of elements returned by this iterator is not affected
 * by the concurrency.
 * <p>
 * This class is not thread safe.
 * 
 * @author Juha Heljoranta
 * 
 * @param <F>
 * @param <T>
 */
public class TransformingConcurrentIterator<F, T> implements Iterator<T> {

	private final Runnable applyTransform;

	private final Executor executor;

	private final Iterator<F> fromIterator;

	private final ConcurrentTransform<F, T> transform;

	public TransformingConcurrentIterator(
			Function<? super F, ? extends T> function,
			Iterator<F> fromIterator, Executor executor) {
		Preconditions.checkNotNull(function, "Function was null");
		Preconditions.checkNotNull(fromIterator, "Iterator was null");
		Preconditions.checkNotNull(executor, "Executor was null");
		transform = new ConcurrentTransform<F, T>(function);
		this.executor = executor;
		this.fromIterator = fromIterator;
		this.applyTransform = new Runnable() {
			@Override
			public void run() {
				try {
					transform.apply();
				} catch (InterruptedException e) {
					Thread.currentThread().interrupt();
					throw Throwables.propagate(e);
				}
			}
		};
		init();
	}

	@Override
	public boolean hasNext() {
		try {
			return transform.hasNext();
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw Throwables.propagate(e);
		}
	}

	private void init() {
		executor.execute(new Runnable() {
			@Override
			public void run() {
				try {
					iterate();
				} catch (InterruptedException e) {
					Thread.currentThread().interrupt();
					throw Throwables.propagate(e);
				}
			}
		});
	}

	private void iterate() throws InterruptedException {
		try {
			while (fromIterator.hasNext()) {
				transform.put(fromIterator.next());
				executor.execute(applyTransform);
			}
		} finally {
			transform.close();
		}
	}

	@Override
	public T next() {
		try {
			return transform.next();
		} catch (ExecutionException e) {
			throw Throwables.propagate(e);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw Throwables.propagate(e);
		}
	}

	@Override
	public void remove() {
		throw new UnsupportedOperationException();
	}

}
