/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;

import com.google.common.base.Throwables;

/**
 * Retrieves queue elements by using iterator.
 * 
 * @author Juha Heljoranta
 * 
 * @param <E>
 */
public class BlockingQueueIteratorAdapter<E> implements Iterable<E> {

	private final E poison;

	private final BlockingQueue<E> queue;

	/**
	 * 
	 * @param queue
	 * @param poison
	 *            the instance used for stopping the iterator.
	 */
	public BlockingQueueIteratorAdapter(BlockingQueue<E> queue, E poison) {
		this.queue = queue;
		this.poison = poison;
	}

	/**
	 * Iterator returns elements until poison element is encountered. When
	 * poison object is encountered the {@link Iterator#hasNext()} returns false
	 * and {@link Iterator#next()} throws an {@link NoSuchElementException}.
	 * <p>
	 * {@inheritDoc}
	 */
	@Override
	public Iterator<E> iterator() {
		return new Iterator<E>() {

			private E e = null;

			private boolean hasNext = false;

			@Override
			public boolean hasNext() {
				if (!hasNext) {
					try {
						if (e == null) {
							e = queue.take();
							hasNext = e != poison;
						}
					} catch (InterruptedException e) {
						// restore interrupted status
						Thread.currentThread().interrupt();
						throw Throwables.propagate(e);
					}
				}
				return hasNext;
			}

			@Override
			public E next() {
				if (hasNext()) {
					E retval = e;
					e = null;
					hasNext = false;
					return retval;
				} else {
					throw new NoSuchElementException();
				}
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}

		};
	}

}
