/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutionException;

import com.google.common.base.Throwables;

/**
 * 
 * 
 * @author Juha Heljoranta
 * 
 * @param <E>
 */
public class IterableObserver<E> implements IterationObserver<E>, Iterable<E> {

	private static final class WrappedException extends ExecutionException {

		private static final long serialVersionUID = 1L;

		private WrappedException(Throwable cause) {
			super(cause);
		}

	}

	private volatile boolean closed;

	private final E poison;

	private final BlockingQueue<E> q;

	public IterableObserver() {
		this(10);
	}

	@SuppressWarnings("unchecked")
	protected IterableObserver(BlockingQueue<E> q) {
		this.q = q;
		this.poison = (E) new Object();
		closed = false;
	}

	public IterableObserver(int capacity) {
		this(new ArrayBlockingQueue<E>(capacity));
	}

	private E get() {
		try {
			return q.take();
		} catch (InterruptedException ex) {
			Thread.currentThread().interrupt();
			throw Throwables.propagate(ex);
		}
	}

	@Override
	public Iterator<E> iterator() {
		return new Iterator<E>() {

			private E e = null;

			private void close() {
				e = poison;
				closed = true;
				q.clear();
			}

			private void closeIfError() {
				if (e instanceof WrappedException) {
					Throwable ex = ((WrappedException) e);
					close();
					throw Throwables.propagate(ex);
				}
			}

			@Override
			public boolean hasNext() {
				if (e == poison) {
					return false;
				}
				if (e == null) {
					e = get();
					closeIfError();
				}
				return e != poison;
			}

			@Override
			public E next() {
				if (hasNext()) {
					E retval = e;
					e = null;
					return retval;
				}
				throw new NoSuchElementException();
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
	}

	@SuppressWarnings("unchecked")
	public void notifyException(Throwable t) {
		put((E) new WrappedException(t));
	}

	@Override
	public void observe(E elt) {
		put(elt);
	}

	@Override
	public boolean observeNext() {
		return true;
	}

	private void put(E e) {
		try {
			if (closed) {
				if (e != poison) {
					throw new IllegalArgumentException(
							IterableObserver.class.getSimpleName()
									+ " is closed. Refusing to observe element="
									+ e);
				}
			} else {
				q.put(e);
			}
		} catch (InterruptedException ex) {
			Thread.currentThread().interrupt();
			throw Throwables.propagate(ex);
		}
	}

	@Override
	public void completed() {
		put(poison);
	}

}
