/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Random;

/**
 * Math utilities.
 */
public class Mth {

	/**
	 * Absolute of double values {@code a}.
	 * 
	 * @param a
	 */
	public static void abs(final double[] a) {
		for (int i = 0; i < a.length; i++) {
			a[i] = Math.abs(a[i]);
		}
	}

	/**
	 * Add values from b to values of the a.
	 * 
	 * @param a
	 * @param b
	 */
	public static void add(final double[] a, final double b) {
		for (int i = 0; i < a.length; i++) {
			a[i] += b;
		}
	}

	/**
	 * Add values from b to values of the a.
	 * 
	 * @param a
	 * @param b
	 */
	public static void add(final double[] a, final double[] b) {
		for (int i = 0; i < a.length; i++) {
			a[i] += b[i];
		}
	}

	/**
	 * return squared distance between two vectors: <code>(a-b)^2</code>.
	 * 
	 * @param a
	 * @param b
	 * @return
	 */
	public static double distance(final double[] a, final double[] b) {
		double d = 0.0;
		for (int i = 0; i < a.length; i++) {
			d += ((a[i] - b[i]) * (a[i] - b[i]));
		}
		return d;
	}

	/**
	 * Divide values of the a with value b.
	 * 
	 * @param a
	 * @param b
	 */
	public static void div(final double a, final double[] b) {
		for (int i = 0; i < b.length; i++) {
			b[i] = a / b[i];
		}
	}

	/**
	 * Divide values of the a with value b.
	 * 
	 * @param a
	 * @param b
	 */
	public static void div(final double[] a, final double b) {
		for (int i = 0; i < a.length; i++) {
			a[i] /= b;
		}
	}

	/**
	 * Divide values of the a with value b.
	 * 
	 * @param a
	 * @param b
	 */
	public static void div(final double[] a, final double[] b) {
		for (int i = 0; i < a.length; i++) {
			a[i] /= b[i];
		}
	}

	/**
	 * Returns a matrix row number which is closest to given reference vector.
	 * In case where there are several rows with the smallest possible distance
	 * to the reference vector the index of the first found row is returned.
	 * 
	 * @param matrix
	 * @param reference
	 * @return index of a row which is closest to the given reference row.
	 */
	public static int getClosestRow(final double[][] matrix,
			final double[] reference) {
		double dist = Double.POSITIVE_INFINITY;
		int row = -1;
		for (int r = 0; r < matrix.length; r++) {
			double d = distance(matrix[r], reference);
			if (d < dist) {
				dist = d;
				row = r;
			}
		}
		return row;
	}

	public static double max(final double[] a) {
		double max = Double.NEGATIVE_INFINITY;
		for (double d : a) {
			if (d > max) {
				max = d;
			}
		}
		return max;
	}

	/**
	 * Calculates a mean: sum(a)/a.length.
	 * 
	 * @param a
	 * @return
	 */
	public static double mean(final double[] a) {
		return sum(a) / a.length;
	}

	/**
	 * 
	 * @param a
	 * @param from inclusive
	 * @param to exclusive
	 * @return
	 */
	public static double mean(final double[] a, final int from, final int to) {
		return sum(a, from, to) / (to - from);
	}
	
	/**
	 * Multiplies values of the a with values of the b.
	 * 
	 * @param a
	 * @param b
	 */
	public static void mult(final double[] a, final double[] b) {
		for (int i = 0; i < a.length; i++) {
			a[i] *= b[i];
		}
	}

	public static int[] seq(final int length) {
		final int[] arr = new int[length];
		for (int i = 0; i < length; i++) {
			arr[i] = i;
		}
		return arr;
	}

	/**
	 * Shuffle integer array in place by using Durstenfeld's version of the
	 * Fisher–Yates shuffle.
	 * 
	 * @param array
	 *            Array to be shuffled.
	 * @param rng
	 *            Random source.
	 */
	public static void shuffle(final int[] array, final Random rng) {
		// i is the number of items remaining to be shuffled.
		for (int i = array.length; i > 1; i--) {
			// Pick a random element to swap with the i-th element.
			int j = rng.nextInt(i); // 0 <= j <= i-1 (0-based array)
			// Swap array elements.
			int tmp = array[j];
			array[j] = array[i - 1];
			array[i - 1] = tmp;
		}
	}

	/** Double variant of {@link #shuffle(int[], Random)}. */
	public static void shuffle(final double[] array, final Random rng) {
		for (int i = array.length; i > 1; i--) {
			int j = rng.nextInt(i);
			double tmp = array[j];
			array[j] = array[i - 1];
			array[i - 1] = tmp;
		}
	}

	/** Long variant of {@link #shuffle(int[], Random)}. */
	public static void shuffle(final long[] array, final Random rng) {
		for (int i = array.length; i > 1; i--) {
			int j = rng.nextInt(i);
			long tmp = array[j];
			array[j] = array[i - 1];
			array[i - 1] = tmp;
		}
	}

	/**
	 * Multiplies values of the a with constant value b.
	 * 
	 * @param a
	 * @param b
	 */
	public static void mult(final double[] a, final double b) {
		for (int i = 0; i < a.length; i++) {
			a[i] *= b;
		}
	}

	/**
	 * Select n random samples from list. Delegates to
	 * {@link #sampleRandom(int, List, Random)} with new random instance.
	 * 
	 * @see #sampleRandom(int, List, Random)
	 */
	public static <T> List<T> sampleRandom(int n, List<T> list) {
		return sampleRandom(n, list, new Random());
	}

	/**
	 * Select n random samples from list.
	 * 
	 * @param n
	 *            The number of samples to get from the list. Must be less than
	 *            or equal to list.size().
	 * @param list
	 *            List of elements from which the samples are taken.
	 * @param random
	 *            Random used to generate random numbers during the samping.
	 * @return Randomly selected samples from the list.
	 * @throws NullPointerException
	 *             If list is null or the random is null.
	 * @throws IllegalArgumentException
	 *             If list size is less than n.
	 */
	public static <T> List<T> sampleRandom(int n, List<T> list, Random random) {
		if (list == null) {
			throw new NullPointerException("List is null.");
		}
		int size = list.size();
		if (size < n) {
			throw new IllegalArgumentException(
					"List size must be at least equal to number of samples");
		}
		HashSet<T> newList = new HashSet<T>(n);
		if (size > n * 2) {
			HashSet<Integer> iset = new HashSet<Integer>(n);
			// low collision probability on random selection
			// relatively fast if size is large and samples is small
			while (iset.size() < n) {
				iset.add(Integer.valueOf(random.nextInt(size)));
			}
			for (Integer i : iset) {
				newList.add(list.get(i.intValue()));
			}
		} else {
			// high collision probability on random selection
			// Collections.suffle() might do better
			ArrayList<Integer> ilist = new ArrayList<Integer>(size);
			for (int i = 0; i < size; i++) {
				ilist.add(Integer.valueOf(i));
			}
			for (int i = 0; i < n; i++) {
				int idx = ilist.remove(random.nextInt(ilist.size())).intValue();
				newList.add(list.get(idx));
			}
		}
		return new ArrayList<T>(newList);
	}

	/**
	 * Calculates a moving average using variant of original SMA algorithm.
	 * <p>
	 * The idea is to slide a window over array elements and take an average of
	 * window elements.
	 * <p>
	 * Example:
	 * 
	 * <pre>
	 * data = new double[] { 7, 11, 13, 17, 19 };
	 * n = 6;
	 * sma = simpleMovingAverage(data, n);
	 * // sma contains values:
	 * sma[0] = 7;
	 * sma[1] = (7 + 11 + 13) / 3.0;
	 * sma[2] = (7 + 11 + 13 + 17 + 19) / 5.0;
	 * sma[3] = (13 + 17 + 19) / 3.0;
	 * sma[4] = 19;
	 * </pre>
	 * 
	 * @param data
	 * @param n
	 *            the number of elements around each array element to be
	 *            averaged.
	 * @return array with SMA applied. null if array is null. Copy of array if
	 *         data.length < 3 or n is zero.
	 * @throws IllegalArgumentException
	 *             if n is negative.
	 */
	public static final double[] simpleMovingAverage(final double[] data,
			final int n) throws IllegalArgumentException {
		if (data == null) {
			return null;
		}
		if (n == 0 || data.length < 3) {
			return Arrays.copyOf(data, data.length);
		}
		if (n < 0) {
			throw new IllegalArgumentException("n must be positive. n=" + n);
		}
		final double[] retval = new double[data.length];
		final int last = data.length - 1;

		// first and last value are always unaffected
		retval[0] = data[0];
		retval[last] = data[last];

		// "average" remaining values

		// average first n values
		int i = 1;
		for (; i < n; i++) {
			final int len = i * 2 + 1;
			if (len > data.length) {
				break;
			}
			retval[i] = sum_(data, 0, len) / len;
		}

		final int fulLen = n * 2 + 1;
		for (; i < data.length - n; i++) {
			retval[i] = sum_(data, i - n, fulLen) / fulLen;
		}

		// average last n values
		for (; i < last; i++) {
			final int j = (last - i);
			final int len = j * 2 + 1;
			retval[i] = sum_(data, i - j, len) / len;
		}
		return retval;
	}

	/**
	 * Subtract value b from values of the a.
	 * 
	 * @param a
	 * @param b
	 */
	public static void subtract(final double[] a, final double b) {
		for (int i = 0; i < a.length; i++) {
			a[i] -= b;
		}
	}

	/**
	 * Subtract values b from values of the a.
	 * 
	 * @param a
	 * @param b
	 */
	public static void subtract(final double[] a, final double[] b) {
		for (int i = 0; i < a.length; i++) {
			a[i] -= b[i];
		}
	}

	/**
	 * Calculates sum.
	 * 
	 * @param a
	 * @return
	 */
	public static double sum(final double[] a) {
		double sum = 0.0;
		for (int i = 0; i < a.length; i++) {
			sum += a[i];
		}
		return sum;
	}

	public static double[] cumulativeSum(final double[] a) {
		final double[] b = new double[a.length]; 
		double sum = 0.0;
		for (int i = 0; i < a.length; i++) {
			sum += a[i];
			b[i] = sum;
		}
		return b;
	}
	
	public static double[] cumulativeSum(final double[] a, final int from, final int to) {
		final double[] b = new double[to - from]; 
		double sum = 0.0;
		for (int i = 0; i < b.length; i++) {
			sum += a[i + from];
			b[i] = sum;
		}
		return b;
	}
	
	public static double sum(final double[] values, final int from, final int to) {
		double sum = 0.0;
		for (int j = from; j < to; j++) {
			sum += values[j];
		}
		return sum;
	}

	public static double sum(final double[] values, final int[] indexes) {
		double sum = 0.0;
		for (final int i : indexes) {
			sum += values[i];
		}
		return sum;
	}

	/**
	 * Sum array elements in given range.
	 * 
	 * @param data
	 * @param from
	 * @param count
	 * @return
	 */
	private static double sum_(final double[] data, final int from,
			final int count) {
		double sum = 0.0;
		for (int j = 0; j < count; j++) {
			sum += data[from + j];
		}
		return sum;
	}


}
