/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * 
 */
package org.jhears.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import com.google.common.base.Throwables;

/**
 * Convenience class for executing tasks in parallel.
 * <p>
 * This class is not thread safe.
 * 
 * @param <T>
 */
public class ParallelExecutor<T> {

	private final ExecutorCompletionService<T> ecs;

	private final boolean mayInterruptIfRunning;

	private final long timeout;

	private final TimeUnit unit;

	/**
	 * 
	 * @param executor
	 *            Executor used for executing the tasks.
	 * @param timeout
	 *            Minimum amount of time to wait for each task.
	 * @param unit
	 *            Determining how to interpret the timeout parameter
	 * @param mayInterruptIfRunning
	 *            If task executing fails then should the cancellation of
	 *            remaining task allow interrupt.
	 */
	public ParallelExecutor(Executor executor, long timeout, TimeUnit unit,
			boolean mayInterruptIfRunning) {
		this.timeout = timeout;
		this.unit = unit;
		this.mayInterruptIfRunning = mayInterruptIfRunning;
		ecs = new ExecutorCompletionService<T>(executor);
	}

	/**
	 * Executes tasks in parallel, waits until all tasks are executed and
	 * finally returns. Results are in the task iterator order. That is fist
	 * result corresponds to the first task returned by the tasks iterator.
	 * 
	 * @param tasks
	 *            Task which will be executed in parallel.
	 * @return List of results in the same order as the tasks where executed.
	 */
	public List<T> execute(Collection<Callable<T>> tasks) {
		final int size = tasks.size();
		List<T> retval = new ArrayList<T>(size);
		if (size == 0) {
			return retval;
		} else if (size == 1) {
			T t;
			try {
				t = tasks.iterator().next().call();
			} catch (Exception e) {
				throw new RuntimeException("Execution failure.", e);
			}
			retval.add(t);
			return retval;
		}

		Map<Future<T>, Reference<T>> futures = new LinkedHashMap<Future<T>, Reference<T>>(
				size);
		try {
			for (Callable<T> task : tasks) {
				futures.put(ecs.submit(task), new Reference<T>());
			}
			for (int i = 0; i < size; i++) {
				poll(futures);
			}
		} catch (RuntimeException e) {
			for (Future<T> future : futures.keySet()) {
				future.cancel(mayInterruptIfRunning);
			}
			throw Throwables.propagate(e);
		}

		for (Reference<T> r : futures.values()) {
			retval.add(r.getValue());
		}

		return retval;
	}

	private void poll(Map<Future<T>, Reference<T>> futures) {
		Future<T> future = null;
		try {
			future = ecs.poll(timeout, unit);
			if (future == null) {
				throw new TimeoutException(
						"ExecutorCompletionService poll() timed out");
			}
			futures.get(future).setValue(future.get(timeout, unit));
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		} catch (ExecutionException e) {
			throw new RuntimeException("Execution failure. Future=" + future, e);
		} catch (TimeoutException e) {
			throw new RuntimeException("Execution timed out. Future=" + future,
					e);
		}
	}

}
