/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 * This class is thread safe.
 * 
 * @author Juha Heljoranta
 * 
 */
public class Serializer implements Serialization {

	private static final class BAIS extends ByteArrayInputStream {

		public BAIS(byte[] buf) {
			super(buf);
		}

		public void setBuffer(byte[] buf, int pos) {
			this.buf = buf;
			this.pos = pos;
			this.count = buf.length;
		}

	}

	private final BAIS bais;

	private final ByteArrayOutputStream baos;

	private final Object inMutex;

	private final Object outMutex;

	private final ObjectInputStream ois;

	private final ObjectOutputStream oos;

	private final byte[] header;

	public Serializer() {
		this(1024);
	}

	public Serializer(int bufferSize) {
		if (bufferSize < 32) {
			bufferSize = 32;
		}
		try {
			this.baos = new ByteArrayOutputStream(bufferSize);
			this.oos = new ObjectOutputStream(baos);
			// oos will write headers and ois will require them
			header = baos.toByteArray();
			this.bais = new BAIS(header);
			this.ois = new ObjectInputStream(bais);
			this.inMutex = new Object();
			this.outMutex = new Object();
		} catch (IOException e) {
			throw new SerializationException(e);
		}
	}

	@SuppressWarnings("unchecked")
	public <E> E deserialize(final byte[] arr) throws SerializationException {
		try {
			synchronized (inMutex) {
				bais.setBuffer(arr, 0);
				try {
					return (E) ois.readUnshared();
				} finally {
					ois.close();
				}
			}
		} catch (IOException e) {
			throw new SerializationException(e);
		} catch (ClassNotFoundException e) {
			throw new SerializationException(e);
		}
	}

	public <E> byte[] serialize(final E obj) throws SerializationException {
		try {
			synchronized (outMutex) {
				baos.reset();
				oos.writeUnshared(obj);
				oos.close();
				return baos.toByteArray();
			}
		} catch (IOException e) {
			throw new SerializationException(e);
		}
	}

}
