/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.util.Arrays;
import java.util.Random;

import org.apache.commons.lang.ArrayUtils;

import com.google.common.base.Preconditions;

/**
 * 
 * Provides a weighted random shuffle.
 * 
 * @author Juha Heljoranta
 * 
 */
public class WeightedRandom {

	private static boolean ok(final double value) {
		return !(Double.isNaN(value) || Double.isInfinite(value) || value < 0.0);
	}

	private static boolean ok(final double[] values) {
		for (double value : values) {
			if (!ok(value)) {
				return false;
			}
		}
		return true;
	}

	private final Random random;

	public WeightedRandom() {
		this(new Random());
	}

	public WeightedRandom(Random random) {
		this.random = random;
	}

	int select(final double[] weights) {
		return select(weights, 0, weights.length);
	}

	/**
	 * Select random index from the list presented by the weights. The higher
	 * the weight values the higher probability it has to get selected. If all
	 * values are the same then a random element is selected.
	 * 
	 * @param weights
	 *            weights to choose from
	 * @param from
	 *            start index from which the weights are included
	 * @param to
	 *            the index from which the weights are excluded.
	 * @return index of the weight array chosen randomly. -1 if array was empty.
	 */
	int select(final double[] weights, final int from, final int to) {
		final double total = Mth.sum(weights, from, to);
		double limit = random.nextDouble() * total;
		for (int i = from; i < to; i++) {
			limit -= weights[i];
			if (limit < 0.0) {
				return i;
			}
		}
		if ((to - from) > 0) {
			// all the elements are zeroes, select random
			return random.nextInt(to - from) + from;
		}
		// nothing to select
		return -1;
	}

	/**
	 * Perform weighted shuffle. The method uses the original Fisher-Yates
	 * algorithm.
	 * <p>
	 * For example for a weight array {@code 1.0, 2.0, 4.0} the probability to
	 * get an index array {@code 2, 1, 0} is {@code 4 / (1 + 2 + 4) * 2 / (1 +
	 * 2) = 0.381} . Using the same weights the probability to get an index
	 * array {@code 0, 1, 2} is {@code 1 / (1 + 2 + 4) * 2 / (2 + 4) = 0.048}.
	 * <p>
	 * If all the weights are the same then the probabilities of resulting index
	 * arrays are equal. That is, the method behaves as non-weighted shuffle.
	 * 
	 * @param weights
	 *            Array of non-negative weight values.
	 * @return shuffled indexes of the weight array.
	 * @throws NullPointerException
	 *             if weight array is null.
	 * @throws IllegalArgumentException
	 *             if any of the arguments is negative.
	 */
	public int[] shuffle(final double[] weights) throws NullPointerException,
			IllegalArgumentException {
		/*
		 * Note: Durstenfeld's version is not suitable since the last
		 * non-swapped element in an array has probably the lowest weight and
		 * therefore it ought to be the last.
		 */
		Preconditions
				.checkArgument(ok(weights), "Weights must be non-negative");
		// weight array indexes
		final IntList scratch = new IntList(Mth.seq(weights.length));
		final IntList result = new IntList(weights.length);
		double[] scratchWeights = Arrays.copyOf(weights, weights.length);
		while (!scratch.isEmpty()) {
			// roll a random number k from 0 to scratch.size - 1
			int k = select(scratchWeights);
			// strike out kth number on the scratch pad
			scratchWeights = ArrayUtils.remove(scratchWeights, k);
			int number = scratch.remove(k);
			// and write it down as the result
			result.add(number);
		}
		return result.toArray();
	}

}
