/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.concurrent;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

import org.jhears.util.IterationNotifier;
import org.jhears.util.IterationObserver;
import org.jhears.util.Observer;
import org.junit.Assert;
import org.junit.Test;

import com.google.common.base.Function;
import com.google.common.base.Functions;
import com.google.common.base.Throwables;

public class ConcurrentTest {

	private static final double SLEEP = 256.0;

	private static final long TIMEOUT = 3000;

	@Test(timeout = TIMEOUT)
	public void testElements0() {
		testNElements(0);
	}

	@Test(timeout = TIMEOUT)
	public void testElements1() {
		testNElements(1);
	}

	@Test(timeout = TIMEOUT)
	public void testElements2() {
		testNElements(2);
	}

	@Test(timeout = TIMEOUT)
	public void testElements3() {
		testNElements(3);
	}

	@Test(timeout = TIMEOUT)
	public void testExecutor0() throws InterruptedException {
		testExecutorService(0);
	}

	@Test(timeout = TIMEOUT)
	public void testExecutor1() throws InterruptedException {
		testExecutorService(1);
	}

	@Test(timeout = TIMEOUT)
	public void testExecutor16() throws InterruptedException {
		testExecutorService(16);
	}

	@Test(timeout = TIMEOUT)
	public void testExecutor2() throws InterruptedException {
		testExecutorService(2);
	}

	@Test(timeout = TIMEOUT)
	public void testExecutor8() throws InterruptedException {
		testExecutorService(8);
	}

	private void testExecutorService(int count) throws InterruptedException {
		List<String> fromList = new ArrayList<String>();
		List<String> toList = new ArrayList<String>();
		for (int i = 0; i < count; i++) {
			fromList.add(Integer.toString(i));
		}

		Iterator<String> from = fromList.iterator();
		Function<String, String> function = Functions.identity();
		ExecutorService executor = Executors.newFixedThreadPool(8);
		Iterator<String> to = Concurrent.transform(from, function, executor);
		for (; to.hasNext();) {
			toList.add(to.next());
		}
		Assert.assertEquals(fromList, toList);
		executor.shutdown();
		Assert.assertTrue(executor.awaitTermination(TIMEOUT,
				TimeUnit.MILLISECONDS));
	}

	@Test(timeout = TIMEOUT)
	public void testIteratorOrder() {
		List<String> fromList = new ArrayList<String>();
		List<String> toList = new ArrayList<String>();
		for (int i = 0; i < 10; i++) {
			fromList.add(Integer.toString(i));
		}
		Iterator<String> from = fromList.iterator();
		Function<String, String> function = new Function<String, String>() {

			@Override
			public String apply(String from) {
				try {
					Thread.sleep((int) Math.round(Math.random() * SLEEP));
				} catch (InterruptedException e) {
					Thread.currentThread().interrupt();
					throw Throwables.propagate(e);
				}
				return from;
			}
		};
		Iterator<String> to = Concurrent.transform(from, function,
				new SimpleExecutor());
		for (; to.hasNext();) {
			toList.add(to.next());
		}
		Assert.assertEquals(fromList, toList);
	}

	private void testNElements(int count) {
		List<String> fromList = new ArrayList<String>();
		List<String> toList = new ArrayList<String>();
		for (int i = 0; i < count; i++) {
			fromList.add(Integer.toString(i));
		}

		Iterator<String> from = fromList.iterator();
		Function<String, String> function = Functions.identity();
		Iterator<String> to = Concurrent.transform(from, function,
				new SimpleExecutor());
		for (; to.hasNext();) {
			toList.add(to.next());
		}
		Assert.assertEquals(fromList, toList);
	}

	@Test(timeout = TIMEOUT)
	public void testNull() {
		List<String> fromList = new ArrayList<String>();
		List<String> toList = new ArrayList<String>();
		for (int i = 0; i < 3; i++) {
			fromList.add(null);
		}

		Iterator<String> from = fromList.iterator();
		Function<String, String> function = Functions.identity();
		Iterator<String> to = Concurrent.transform(from, function,
				new SimpleExecutor());
		for (; to.hasNext();) {
			toList.add(to.next());
		}
		Assert.assertEquals(fromList, toList);
	}

	@Test(timeout = TIMEOUT)
	public void testTransformObserver() {
		List<String> fromList = new ArrayList<String>();
		for (int i = 0; i < 4; i++) {
			fromList.add(Integer.toString(i));
		}

		Iterator<String> from = fromList.iterator();
		Function<String, String> function = Functions.identity();
		final List<String> toList = new ArrayList<String>();
		Observer<String> observer = new Observer<String>() {
			@Override
			public void observe(String elt) {
				toList.add(elt);
			}
		};
		Concurrent.transform(from, function, observer);
		Assert.assertEquals(fromList, toList);
	}

	@Test
	public void testNotifyObserver() {
		final int cnt = 3;
		IterationNotifier<String> notifier = new IterationNotifier<String>() {

			@Override
			public void notifyObserver(IterationObserver<String> observer) {
				for (int i = 0; i < cnt; i++) {
					observer.observe(Integer.toString(i));
				}
				observer.completed();
			}
		};
		Iterator<String> si = Concurrent.notifyObserver(notifier);
		Assert.assertTrue(si.hasNext());
		Assert.assertEquals("0", si.next());
		Assert.assertTrue(si.hasNext());
		Assert.assertEquals("1", si.next());
		Assert.assertTrue(si.hasNext());
		Assert.assertEquals("2", si.next());
		Assert.assertFalse(si.hasNext());
	}

	@Test
	public void testNotifyObserverExceptions() {
		final AtomicReference<IterationObserver<String>> observerRef = new AtomicReference<IterationObserver<String>>();
		IterationNotifier<String> notifier = new IterationNotifier<String>() {

			@Override
			public void notifyObserver(IterationObserver<String> observer) {
				observerRef.set(observer);
				throw new NullPointerException();
			}
		};
		Iterator<String> si = Concurrent.notifyObserver(notifier);
		Exception ex = null;
		try {
			si.next();
			Assert.fail();
		} catch (Exception e) {
			ex = e;
			// null pointer was thrown, further observe not possible
			try {
				observerRef.get().observe("foo");
				Assert.fail();
			} catch (IllegalArgumentException e1) {
				Assert.assertNotNull(e1);
			}
		}
		Assert.assertNotNull(ex);
	}

}
