/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.io.Serializable;
import java.util.Random;

import org.junit.Assert;
import org.junit.Test;

public class SerializerTest {

	private static final class SerializerCrashTestDummy implements Serializable {

		private static final long serialVersionUID = 5497733429384934830L;

		private final int age;

		private final String name;

		public SerializerCrashTestDummy(String name, int age) {
			this.name = name;
			this.age = age;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			SerializerCrashTestDummy other = (SerializerCrashTestDummy) obj;
			if (age != other.age)
				return false;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + age;
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			return result;
		}

		@Override
		public String toString() {
			return SerializerCrashTestDummy.class.getSimpleName() + " [age="
					+ age + ", name=" + name + "]";
		}

	}

	@Test
	public void testSerialization() throws Exception {
		Serializer s = new Serializer();
		for (int j = 0; j < 10; j++) {
			byte[] data = s.serialize(new SerializerCrashTestDummy(Integer
					.toString(j + 1), j));
			SerializerCrashTestDummy actual = (SerializerCrashTestDummy) s
					.deserialize(data);
			Assert.assertEquals(new SerializerCrashTestDummy(Integer
					.toString(j + 1), j), actual);
		}
	}

	@Test
	public void testIndepended() throws Exception {
		Serializer ser = new Serializer();
		Serializer deser = new Serializer();
		for (int j = 0; j < 10; j++) {
			byte[] data = ser.serialize(new SerializerCrashTestDummy(Integer
					.toString(j + 1), j));
			SerializerCrashTestDummy actual = (SerializerCrashTestDummy) deser
					.deserialize(data);
			Assert.assertEquals(new SerializerCrashTestDummy(Integer
					.toString(j + 1), j), actual);
		}
	}

	@Test
	public void testDynamic() throws Exception {
		for (int j = 0; j < 10; j++) {
			byte[] data = new Serializer()
					.serialize(new SerializerCrashTestDummy(Integer
							.toString(j + 1), j));
			SerializerCrashTestDummy actual = (SerializerCrashTestDummy) new Serializer()
					.deserialize(data);
			Assert.assertEquals(new SerializerCrashTestDummy(Integer
					.toString(j + 1), j), actual);
		}
	}

	@Test
	public void testRandom() throws Exception {
		Serializer s = new Serializer();
		Random rnd = new Random(0);
		for (int j = 0; j < 1000; j++) {
			switch (rnd.nextInt(8)) {
			case 0:
				Assert.assertEquals(null, s.deserialize(s.serialize(null)));
				break;
			case 1:
				Assert.assertEquals("String", s.deserialize(s
						.serialize("String")));
				break;
			case 2:
				Integer ival = Integer.valueOf(rnd.nextInt());
				Assert.assertEquals(ival, s.deserialize(s.serialize(ival)));
				break;
			case 3:
				Long lval = Long.valueOf(rnd.nextLong());
				Assert.assertEquals(lval, s.deserialize(s.serialize(lval)));

				break;
			case 4:
				byte[] baval = new byte[rnd.nextInt(128)];
				rnd.nextBytes(baval);
				Assert.assertArrayEquals(baval, (byte[]) s.deserialize(s
						.serialize(baval)));
				break;
			case 5: {
				String name = Long.toString(rnd.nextLong());
				int age = rnd.nextInt();
				SerializerCrashTestDummy actual = (SerializerCrashTestDummy) s
						.deserialize(s.serialize(new SerializerCrashTestDummy(
								name, age)));
				Assert.assertEquals(new SerializerCrashTestDummy(name, age),
						actual);
				break;
			}
			case 6:
				long[] laval = new long[rnd.nextInt(128)];
				for (int i = 0; i < laval.length; i++) {
					laval[i] = rnd.nextLong();
				}
				Assert.assertArrayEquals(laval, (long[]) s.deserialize(s
						.serialize(laval)));
				break;
			case 7:
				SerializerCrashTestDummy[] oaval = new SerializerCrashTestDummy[rnd
						.nextInt(128)];
				for (int i = 0; i < oaval.length; i++) {
					String name = Long.toString(rnd.nextLong());
					int age = rnd.nextInt();
					oaval[i] = new SerializerCrashTestDummy(name, age);
				}
				Assert.assertArrayEquals(oaval, (SerializerCrashTestDummy[]) s
						.deserialize(s.serialize(oaval)));
				break;
			default:
				break;
			}
		}
	}
}
