/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.jhears.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class WeightedRandomTest {

	private static final int TESTCNT = 128;

	private WeightedRandom wr;

	@Before
	public void before() {
		wr = new WeightedRandom(new Random(16289973));
	}

	private int count(List<int[]> suffled, int[] expected) {
		int cnt = 0;
		for (int[] arr : suffled) {
			if (Arrays.equals(expected, arr)) {
				cnt++;
			}
		}
		return cnt;
	}

	private void neverSelectZero(double[] weights, int zeroIdx) {
		for (int i = 0; i < TESTCNT; i++) {
			int actual = wr.select(weights);
			if (zeroIdx == actual) {
				Assert.fail("Index with zero weight selected");
			}
		}
	}

	@Test
	public void selectEmpty() {
		Assert.assertEquals(-1, wr.select(new double[0]));
	}

	@Test
	public void selectOne() {
		Assert.assertEquals(0, wr.select(new double[] { 1.0 }));
	}

	@Test
	public void selectOneZero() {
		Assert.assertEquals(0, wr.select(new double[] { 1.0, 0.0 }));
	}

	private void selectTwoWithEqualWeight(double[] weights) {
		int aCnt = 0;
		int bCnt = 0;
		for (int i = 0; i < TESTCNT; i++) {
			int actual = wr.select(weights);
			if (0 == actual) {
				aCnt++;
			} else if (1 == actual) {
				bCnt++;
			} else {
				throw new IllegalStateException(Integer.toString(actual));
			}
		}
		// equal probability with +-20% error margin
		Assert.assertEquals(1.0, (double) aCnt / (double) bCnt, 0.20);
	}

	@Test
	public void selectZero() {
		Assert.assertEquals(0, wr.select(new double[] { 0.0 }));
	}

	@Test
	public void selectZeroOne() {
		Assert.assertEquals(1, wr.select(new double[] { 0.0, 1.0 }));
	}

	private void suffleTwoNonZeroEqualsAndOneZero(double[] weights, int zeroIdx) {
		for (int i = 0; i < TESTCNT; i++) {
			int[] suffled = wr.shuffle(weights);
			Assert.assertEquals(
					"Element with zero probability should the last", zeroIdx,
					suffled[2]);
		}
	}

	private void suffleTwoWithEqualWeight(double[] weights) {
		int[] expectedA = new int[] { 0, 1 };
		int[] expectedB = new int[] { 1, 0 };
		int aCnt = 0;
		int bCnt = 0;
		for (int i = 0; i < TESTCNT; i++) {
			int[] suffled = wr.shuffle(weights);
			if (Arrays.equals(expectedA, suffled)) {
				aCnt++;
			} else if (Arrays.equals(expectedB, suffled)) {
				bCnt++;
			} else {
				throw new IllegalStateException();
			}
		}
		// equal probability with +-15% error margin
		Assert.assertEquals(1.0, (double) aCnt / (double) bCnt, 0.15);
	}

	@Test
	public void testNeverSelectFirstZero() {
		neverSelectZero(new double[] { 0.0, 1.0, 1.0 }, 0);
	}

	@Test
	public void testNeverSelectLastZero() {
		neverSelectZero(new double[] { 1.0, 1.0, 0.0 }, 2);
	}

	@Test
	public void testNeverSelectMiddleZero() {
		neverSelectZero(new double[] { 1.0, 0.0, 1.0 }, 1);
	}

	@Test
	public void testSelectWithOnes() {
		selectTwoWithEqualWeight(new double[] { 1.0, 1.0 });
	}

	@Test
	public void testSelectWithZeroes() {
		selectTwoWithEqualWeight(new double[] { 0.0, 0.0 });
	}

	@Test
	public void testSuffle() {
		int shuffleCnt = TESTCNT << 3;

		double[] weights = new double[] { 1.0, 2.0, 4.0 };
		ArrayList<int[]> list = new ArrayList<int[]>(shuffleCnt);
		for (int i = 0; i < shuffleCnt; i++) {
			list.add(wr.shuffle(weights));
		}

		int actual, expected;
		// 10% error margin
		double delta = 0.15;

		// 4 / 7 * 2 / 3
		actual = count(list, new int[] { 2, 1, 0 });
		expected = shuffleCnt * 4 * 2 / 3 / 7;
		Assert.assertEquals(expected, actual, expected * delta);

		// 2 / 7 * 4 / 5
		actual = count(list, new int[] { 1, 2, 0 });
		expected = shuffleCnt * 2 * 4 / 5 / 7;
		Assert.assertEquals(expected, actual, expected * delta);

		// 4 / 7 * 1 / 3
		actual = count(list, new int[] { 2, 0, 1 });
		expected = shuffleCnt * 4 * 1 / 3 / 7;
		Assert.assertEquals(expected, actual, expected * delta);

		// 1 / 7 * 4 / 6
		actual = count(list, new int[] { 0, 2, 1 });
		expected = shuffleCnt * 1 * 4 / 6 / 7;
		Assert.assertEquals(expected, actual, expected * delta);

		// 2 / 7 * 1 / 5
		actual = count(list, new int[] { 1, 0, 2 });
		expected = shuffleCnt * 2 * 1 / 5 / 7;
		Assert.assertEquals(expected, actual, expected * delta);

		// 1 / 7 * 2 / 6
		actual = count(list, new int[] { 0, 1, 2 });
		expected = shuffleCnt * 1 * 2 / 6 / 7;
		Assert.assertEquals(expected, actual, expected * delta);
	}

	@Test
	public void testSuffleEmpty() {
		Assert.assertArrayEquals(new int[0], wr.shuffle(new double[0]));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testSuffleNegative() {
		wr.shuffle(new double[] { -1.0 });
	}

	@Test(expected = NullPointerException.class)
	public void testSuffleNull() {
		wr.shuffle((double[]) null);
	}
	
	@Test
	public void testSuffleOne() {
		Assert.assertArrayEquals(new int[] { 0 }, wr
				.shuffle(new double[] { 1.0 }));
	}

	@Test
	public void testSuffleWithOnes() {
		suffleTwoWithEqualWeight(new double[] { 1.0, 1.0 });
	}

	@Test
	public void testSuffleWithZeroes() {
		suffleTwoWithEqualWeight(new double[] { 0.0, 0.0 });
	}

	@Test
	public void testSuffleZero() {
		Assert.assertArrayEquals(new int[] { 0 }, wr
				.shuffle(new double[] { 0.0 }));
	}

	@Test
	public void testSuffleZeroFirst() {
		suffleTwoNonZeroEqualsAndOneZero(new double[] { 0.0, 1.0, 1.0 }, 0);
	}

	@Test
	public void testSuffleZeroLast() {
		suffleTwoNonZeroEqualsAndOneZero(new double[] { 1.0, 1.0, 0.0 }, 2);
	}

	@Test
	public void testSuffleZeroMiddle() {
		suffleTwoNonZeroEqualsAndOneZero(new double[] { 1.0, 0.0, 1.0 }, 1);
	}

}
