/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.apache.wicket.contrib.recaptcha;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.conn.ssl.AllowAllHostnameVerifier;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.apache.wicket.behavior.SimpleAttributeModifier;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.protocol.http.WebRequest;

import com.google.common.io.CharStreams;

public class ReCaptchaPanel extends Panel {

	private static final long serialVersionUID = 1L;

	private final ReCaptchaConfig config;

	public ReCaptchaPanel(String id, IModel<?> model, ReCaptchaConfig config) {
		super(id, model);
		this.config = config;
		addPanelComponents();
	}

	public ReCaptchaPanel(String id, ReCaptchaConfig config) {
		super(id);
		this.config = config;
		addPanelComponents();
	}

	private void addPanelComponents() {

		String reCaptchaScriptSrc = config.getScriptUrlFragment()
				+ config.getPublicKey();

		String reCaptchaNoScriptSrc = config.getNoscriptUrlFragment()
				+ config.getPublicKey();

		add(new WebMarkupContainer("reCaptchaScript")
				.add(new SimpleAttributeModifier("src", reCaptchaScriptSrc)));

		add(new WebMarkupContainer("reCaptchaNoScript")
				.add(new SimpleAttributeModifier("src", reCaptchaNoScriptSrc)));

	}

	private List<String> executeHttpRequest(HttpParams httpParams,
			HttpRequestBase httpReq) throws IOException {
		DefaultHttpClient httpclient = null;
		try {
			httpclient = new DefaultHttpClient(httpParams);

			// TODO: fix this when google fixes its api
			// https://api-verify.recaptcha.net cert refers to www.google.com
			SSLSocketFactory sslsf = (SSLSocketFactory) httpclient
					.getConnectionManager().getSchemeRegistry().get("https")
					.getSocketFactory();
			sslsf.setHostnameVerifier(new AllowAllHostnameVerifier());

			HttpResponse httpResponse = httpclient.execute(httpReq);
			if (httpResponse.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
				HttpEntity entity = httpResponse.getEntity();
				if (entity != null) {
					InputStream is = entity.getContent();
					Charset cs = Charset.forName("UTF-8");
					List<String> retval = CharStreams
							.readLines(new InputStreamReader(is, cs));
					is.close();
					return retval;
				}
			} else {
				throw new IOException(
						"HTTP status not ok. Server returned code="
								+ httpResponse.getStatusLine().getStatusCode()
								+ ", reason="
								+ httpResponse.getStatusLine()
										.getReasonPhrase());
			}
		} finally {
			if (httpclient != null) {
				httpclient.getConnectionManager().shutdown();
			}
		}
		return null;
	}

	/**
	 * Simple check to see if we should invoke {@link #verifyCaptcha()}. If this
	 * method returns false then it is guaranteed that the captcha verify fails.
	 * 
	 * @return true if both the "recaptcha_challenge_field" and
	 *         "recaptcha_response_field" are set.
	 */
	public boolean isReCaptchaFieldsFilled() {
		String challenge = getRequest().getParameter(
				"recaptcha_challenge_field");
		String response = getRequest().getParameter("recaptcha_response_field");
		return !(challenge == null || response == null
				|| challenge.trim().isEmpty() || response.trim().isEmpty());
	}

	/**
	 * Verify captcha.
	 * 
	 * @return true only if user solved the captcha correctly. false if server
	 *         returned "incorrect-captcha-sol".
	 * @throws IOException
	 *             If communication with the recaptcha server failed.
	 * @throws ReCaptchaException
	 *             If recaptcha server returned some other error than
	 *             "incorrect-captcha-sol".
	 */
	public boolean verifyCaptcha() throws IOException, ReCaptchaException {

		HttpParams params = new BasicHttpParams();
		params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, config
				.getVerifyConnectionTimeoutInMillis());
		params.setParameter(CoreConnectionPNames.SO_TIMEOUT, config
				.getVerifySocketTimeoutInMillis());

		WebRequest wrq = (WebRequest) getRequest();
		String remoteip = wrq.getHttpServletRequest().getRemoteAddr();
		String challenge = wrq.getParameter("recaptcha_challenge_field");
		String response = wrq.getParameter("recaptcha_response_field");

		List<NameValuePair> formparams = new ArrayList<NameValuePair>();
		formparams.add(new BasicNameValuePair("privatekey", config
				.getPrivateKey()));
		formparams.add(new BasicNameValuePair("remoteip", remoteip));
		formparams.add(new BasicNameValuePair("challenge", challenge));
		formparams.add(new BasicNameValuePair("response", response));

		UrlEncodedFormEntity postentity = new UrlEncodedFormEntity(formparams,
				"UTF-8");

		HttpPost httppost = new HttpPost(config.getVerifyUrl());
		httppost.setEntity(postentity);
		List<String> res = executeHttpRequest(params, httppost);

		if (res != null) {
			if (!res.isEmpty() && "true".equals(res.get(0))) {
				return true;
			} else {
				if (res.size() > 1 && "false".equals(res.get(0))) {
					if ("incorrect-captcha-sol".equals(res.get(1))) {
						return false;
					}
				}
			}
			throw new ReCaptchaException("Server returned an error", res);
		}
		throw new IOException("Failed to get content from server");
	}

}
