/*
 * jHears, acoustic fingerprinting framework.
 * Copyright (C) 2009-2010  Juha Heljoranta.
 *
 * This file is part of jHears.
 *
 * jHears is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * jHears is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with jHears.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * 
 */
package org.jhears.webgui;

import java.security.SecureRandom;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import org.apache.commons.codec.binary.Hex;
import org.hibernate.Session;
import org.hibernate.criterion.Restrictions;
import org.jhears.crypto.AuthCode;
import org.jhears.data.GeneralDao;
import org.jhears.data.IGenericDao;
import org.jhears.data.RoleDataObj;
import org.jhears.data.Transactional;
import org.jhears.data.UserDataObj;
import org.joda.time.DateTime;
import org.joda.time.format.ISODateTimeFormat;
import org.slf4j.LoggerFactory;

import com.google.common.base.Charsets;

public class AccountManager implements IAccountManager {

	private final GeneralDao dao;

	private final Random random;

	private AuthCode authCode;

	public AccountManager(GeneralDao dao) {
		this.dao = dao;
		this.random = new Random(new SecureRandom().nextLong());
		this.authCode = new AuthCode();
	}

	private void add(Session s, long accontId, long roleId) {
		LoggerFactory.getLogger(WebguiModule.class).info(
				"accontId=" + accontId + " roleId=" + roleId);
		s.createSQLQuery(
				"insert into " + UserDataObj.class.getSimpleName() + "_"
						+ RoleDataObj.class.getSimpleName() + " values (?,?)")
				.setLong(0, accontId).setLong(1, roleId).executeUpdate();
	}

	private void delete(Session s, long accontId, long roleId) {
		s
				.createSQLQuery(
						"delete from " + UserDataObj.class.getSimpleName() + "_"
						+ RoleDataObj.class.getSimpleName() + " where users_id=? and roles_id=?")
				.setLong(0, accontId).setLong(1, roleId).executeUpdate();
	}

	@Override
	public UserDataObj createAccount(String name, String pass,
			final String[] roles) {
		final UserDataObj userDataObj = new UserDataObj();
		byte[] saltBytes = new byte[64];
		random.nextBytes(saltBytes);
		// be a paranoid and re-seed
		random.setSeed(random.nextLong() ^ System.nanoTime());
		byte[] authCodeBytes = authCode.calculate(saltBytes, pass
				.getBytes(Charsets.UTF_8));
		String saltString = Hex.encodeHexString(saltBytes);
		String authCodeString = Hex.encodeHexString(authCodeBytes);
		Map<String, String> properties = new HashMap<String, String>();
		properties.put(UserDataObj.PASSWORD_SALT, saltString);
		properties.put(UserDataObj.PASSWORD, authCodeString);
		properties.put(UserDataObj.CREATED_TIMESTAMP, ISODateTimeFormat
				.dateTime().print(new DateTime()));
		properties.put("locked", Boolean.FALSE.toString());
		userDataObj.setName(name);
		userDataObj.setProperties(properties);
		// save user and add role mappings
		Transactional<Void, Void> transactional = new Transactional<Void, Void>() {

			@Override
			public Void with(Session s, Void arg) {
				s.save(userDataObj);
				s.flush();
				for (String roleName : roles) {
					RoleDataObj r = (RoleDataObj) s.createCriteria(
							RoleDataObj.class).add(
							Restrictions.eq("name", roleName)).uniqueResult();
					add(s, userDataObj.getId().longValue(), r.getId()
							.longValue());
				}
				return null;
			}

		};
		dao.getDao().execute(transactional, null);
		return userDataObj;
	}

	@Override
	public IGenericDao getGenericDao() {
		return dao;
	}

	// @Override
	// public List<String> getRoleNames(final String username) {
	// Transactional<Void, List<String>> transactional = new Transactional<Void,
	// List<String>>() {
	//
	// @Override
	// public List<String> with(Session s, Void arg) {
	// UserDataObj account = (UserDataObj) s.createCriteria(
	// UserDataObj.class).add(
	// Restrictions.eq("name", username)).uniqueResult();
	// List<String> retval = null;
	// if (account != null) {
	// retval = new ArrayList<String>();
	// for (RoleDataObj role : account.getRoles()) {
	// retval.add(role.getName());
	// }
	// }
	// return retval;
	// }
	//
	// };
	// return dao.getDao().execute(transactional, null);
	// }

}
